//************************************************************
//              Simple conversion functions
//************************************************************
// Reference:   SOFTWARE SOLUTIONS
// Workspace  : Appendix A\Test Un32_1
// Entry point: Test Un32_1
// Other files: math32.lib
//              Un32_1.h
// Development: Visual C++ 6.0 Console Application
// Date:        January 12/2006
//              March 31/2006
// Description:
//              C++ interface functions for the Un32_1 module
//              of the MATH32 library
//************************************************************
// Library functions
extern "C" BIN_TO_ASC10(); // 16-bit binary into 5-digit ASCII decimal
extern "C" BIN_TO_ASC16(); // 16-bit binary into 4-digit ASCII hex
extern "C" ASCII_TO_EDX(); // 4-digit ASCII decimal to binary in EDX
extern "C" BINF_TO_DECF(); // 8-bit binary to 8-digit decimal fraction
extern "C" DECODE_SINGLE();// Isolate elements of single precision
extern "C" BIN_TO_BCD();   // 16-bit binary to 5 BCD digits
extern "C" BCD_TO_ASCII(); // 5 unpacked BCD to 5 ASCII digits
extern "C" BCD_INTO_REG(); // 5-digit BCD into a binary in DX
extern "C" BCD12_TO_ASCSTR();// BCD12 number in ASCII string
extern "C" ASCII_TO_BCD12(); // ASCII decimal string into BCD12 format
extern "C" ASCII_TO_EDX(); // ASCII to binary in EDX register

//**************************************************************
//                       Prototypes
//**************************************************************
void IntToAsc10(int, char[]);
void IntToAsc16(int, char[]);
int AscToInt(char[]);

//**************************************************************
//            C++ interface functions to Un32_1 module
//**************************************************************
void IntToAsc10(int value, char *buffer)
{
// Function to convert a 16-bit binary number into 5 ASCII
// decimal digits
// On entry:
//         parameter value holds binary number
//         parameter buffer[] is and array of char for result
// Uses the BIN_TO_ASC10 procedure in the Un32_1 module of
// the MATH32 library
_asm
    {
        MOV      EAX,value
        MOV      EDI,buffer
		CALL     BIN_TO_ASC10
	}
  return;
}


void IntToAsc16(int value, char *buffer)
{
// Function to convert a 16-bit binary number into 4 ASCII
// hexadecimal digits
// On entry:
//         parameter value holds binary number
//         parameter buffer[] is and array of char for result
// Uses the BIN_TO_ASC16 procedure in the Un32_1 module of
// the MATH32 library
_asm
    {
        MOV      EAX,value
        MOV      EDI,buffer
    	CALL     BIN_TO_ASC16
	}
return;
}

int AscToInt(char *buffer)
// Convert an ASCII string to binary
// On entry:
//         parameter buffer[] is an array of char holding
//         10-digit ASCII decimal string in the range
//         0 to 4,294,967,295
// On exit:
//         returns binary value in an int type
// Uses the ASCII_TO_EDX procedure in the Un32_1 module of
// the MATH32 library
{
int aVal = 0;
_asm
    {
        MOV      ESI,buffer
		CALL     ASCII_TO_EDX
		MOV      aVal,EDX     
	}
return aVal;
}

