//************************************************************
//               Integer conversions module
//************************************************************
// Reference:   SOFTWARE SOLUTIONS - Chapter 1
// Project:     Chapter1\Integer Conversions
// Entry point: Integer Conversions.cpp
// This file:   math1.cpp
// Header:      math1.h
// Development: Visual C++ 6.0 Console Application
// Date:        December 2000
//
// Description:
//              Function module with integer conversion
//              functions
// Note:
//              Functions are coded in C++
//***********************************************************

#include "stdafx.h"
#include <iostream.h>
#include "math1.h"


unsigned long AscBin2Bin( char digits[])
// Function to convert string of binary digits to a binary
// value
// Pre:
//    digits[] is a char array of binary digits.
//    The array can contain spaces and dashes as
//    separators.
// Post:
//    Returns the binary value of the digits
//    in a variable of type unsigned long
//
{

     int size = 0;                  // Array size
     unsigned long binWeight = 1;   // Binary digit weight
     unsigned long accumulator = 0; // Temp value
     unsigned int bitOffset;        // Offset in bit array
     unsigned int digitPos = 0;     // Relative digit position

     // Calculate array size
     while(digits[size])
		size++;

     // Routine starts at the least significant digit, which
     // has a weight of 1. The weight of other digits to the
     // left of the first one is 2 * the weight of the
     // previous bit
     for(int counter = 0; counter < size; counter++)
     {
	bitOffset = (size - 1) - digitPos;

	if(digits[bitOffset] == ' ' || digits[bitOffset] == '-')
	{
	   digitPos++;
	   continue;
	}

	// Test for 1-digit and add weight to accumulator
	if(digits[bitOffset] == '1')
	  accumulator = accumulator + binWeight;
	// Update digit weight
	binWeight *= 2;
	// Update digit position in string
	digitPos++;
     }
     return accumulator;

}


unsigned long AscDec2Bin( char digits[])
// Function to convert string of decimal digits to a binary
// value
// Pre:
//    digits[] is a char array of decimal digits.
//    The array can contain commas as separators.
// Post:
//    Returns the binary value of the digits
//    in a variable of type unsigned long
//
{

     int size = 0;                  // Array size
     unsigned long decWeight = 1;   // Decimal digit weight
     unsigned long accumulator = 0; // Temp value
     unsigned int bitOffset;        // Offset in bit array
     unsigned int digitPos = 0;     // Relative digit position
     char thisDigit;                // Storage for digit

     // Calculate array size
     while(digits[size])
		size++;

     // Routine starts at the least significant digit, which
     // has a weight of 1. The weight of other digits to the
     // left of the first one is 10 * the weight of the
     // previous one
     for(int counter = 0; counter < size; counter++)
     {
	 bitOffset = (size - 1) - digitPos;

	 if(digits[bitOffset] == ',')
	 {
	   digitPos++;
	   continue;
	 }
	 // ASSERT:
	 //       Valid decimal digit in the range 0 to 9
	 // Convert ASCII ditit to binary
	 thisDigit = digits[bitOffset] - 0x30;
	 // Accumulate digit weight
	 accumulator = accumulator + (thisDigit * decWeight);
	 // Update digit weight
	 decWeight *= 10;
	 // Update digit position in string
	 digitPos++;
     }
     return accumulator;
}


unsigned long AscHex2Bin( char digits[])
// Function to convert string of hex digits to a binary
// value
// Pre:
//    digits[] is a char array of hex digits.
// Post:
//    Returns the binary value of the digits
//    in a variable of type unsigned long
// Note:
//    The function allows the standard C/C++ hex
//    format characters (0x) in the digit string.
//    Hex alpha characters (A through F) can be
//    upper- or lower-case.
//
{
     int size = 0;                  // Array size
     unsigned long hexWeight = 1;   // Digit weight
     unsigned long accumulator = 0; // Temp value
     unsigned int offset = 0;       // Offset in array
     unsigned int digitPos = 0;     // Relative digit position
     unsigned int digitValue;       // Binary digit value

     // Obtain the size of the array
     while(digits[size])
	 size++;

     // Routine starts at the least significant digit, which
     // has a weight of 1. The weight of other digits to the
     // left of the first one is 16 * the weight of the
     // previous bit
     for(int counter = 0; counter < size; counter++)
     {
	offset = (size - 1) - digitPos;

	if(digits[offset] == 'x')
	{
	   digitPos++;
	   continue;
	}
	// ASSERT:
	//       Valid hex digit in the range 0 to F
	// Convert ASCII ditit to binary

	if(digits[offset] < 'A')
	   digitValue = digits[offset] - 0x30;
	else
	{
	   // Test for upper case hex digit
	   if(digits[offset] < 'G')
	      digitValue = digits[offset] - 0x37;
	   else
	      digitValue = digits[offset] - 0x57;
	}

	// Accumulate digit weight
	accumulator = accumulator + (digitValue * hexWeight);
	// Update digit weight
	hexWeight *= 16;
	// Update digit position in string
	digitPos++;
     }
     return accumulator;

}


void Bin2AscHex( char digits[], unsigned long value)
// Function to convert a binary value into a string
// of ASCII hexadecimal digits
// Pre:
//    1. digits[] is a char array for storing the ASCII
//       hex digits
//    2. value is the binary number stored in a variable
//       of type unsigned long
//    3. Array is of sufficient size for storing all
//       ASCII hex digits. If not, the returned result is
//       invalid.
// Post:
//    Returns a string of ASCII hexadecimal digits in the
//    argument array

{
   int size = 0;                 // Elements in array
   unsigned int hexDigit;
   unsigned int factor = 0x37;   // Change to 0x57 to obtain
				 // lower-case digits in the
				 // range a to f

   // Calculate array size
   while(digits[size])
      size++;

   while(value > 0 && size > 0)
   {

    hexDigit = value % 16;
    // Convert and store digits in the range 0 to 9
    if(hexDigit < 10)
      digits[size-1] = hexDigit + 0x30;   // Obtain and store
					   // low-order digit
    else
      digits[size-1] = hexDigit + factor;

    // Note:
    //      If size is the number of elements in the digits
    //      array, then the storage offset is (size - 1)

    value = value / 16;                   // New operand
    size --;                              // Update digit position
   }
    return;
}


void Bin2AscDec( char digits[], unsigned long value)
// Function to convert a binary value into a string
// of ASCII decimal digits
// Pre:
//    1. digits[] is a char array for storing the ASCII
//       decimal digits
//    2. value is the binary number stored in a variable
//       of type unsigned long
//    3. Array is of sufficient size for storing all
//       ASCII digits. If not, the returned result is
//       invalid.
// Post:
//    Returns a string of ASCII decimal digits in the
//    argument array

{
   int size = 0;

   // Calculate array size
   while(digits[size])
      size++;

   while(value > 0 && size > 0)
   {

    digits[size-1] = (char)((value % 10) + 0x30);// Obtain and store
                                         // low-order digit
    // Note:
    //      aSize is the number of elements in the digits
    //      array. The storage offset is aSize - 1
    value = value / 10;                   // New operand
    size --;                             // Update digit position
   }
    return;
}

