//************************************************************
//           Probability and statistical functions
//************************************************************
// Reference:   NSOFTWARE SOLUTIONS
// Workspace  : Chapter11\Test Un32_9
// Entry point: Test Un32_9
// Other files: Un32_9.asm
//              Un32_9.h
//              Un32_9.obj      
// Development: Visual C++ 6.0 Console Application
// Date:        January 25/2001
//
// Description:
//              Header file for testing calculator primtive
//              functions in the Un32_9 module of math32.lib
//************************************************************

// Library functions
extern "C" PERMUTATIONS();       //.. Number of permutations
extern "C" COMBINATIONS();       //.. Number of combinations
extern "C" BINOMIAL_PROB();      //.. Binomial probability
extern "C" NORMAL_CURVE_Y();     //.. f(x) for normal curve
extern "C" LOADE();              //.. Load constant e to ST(0)
//**************************************************************
//                       Prototypes
//**************************************************************
int Permutations(int, int);
int Combinations(int, int);
double BinomialProb(int, int);
double ArraySum(class, int, int, int);
void Sort(class, int);
void SortCopy(class, class, int);
void ArrayShow(class, int);
double MeanOf(class, int, int, int);
double Mean(class, int);
double Median(class, int);
double Mode(class, int *, int);
double WeightedMean(class, int *, int);
double WeightedMedian(class, int *, int);
double WeightedMode(class, int *, int);
double Range(class, int);
double VarianceS(class, int);
double VarianceP(class, int);
double StdDevS(class, int);
double StdDevP(class, int);
void NormalDist4Pts(double *, double, double);
double NormalDistY(double, double, double);
double NormalProbTrpz(double, double, double, double);
double NormalProbSimp(double, double, double, double);
double NormalProbError(double, double, double, double);
  
//**************************************************************
//            C++ interface functions to Un32_9 module
//**************************************************************
int Permutations(int n, int r)
{
// Ordered permutation without repetitions of n elements, in set P,
// taken r at a time, according to the formula:
//
//              n!
//     nPr = ---------
//            (n - r)!
//
// where:

//          n is the number of elements in set P
//          r is number of elements taken at a time
//          nPr = number of permutations in a set of n elements
//                taken r at a time
// On entry:
//          parameter n is n in formula
//          parameter r is r in formula
// On exit:
//          returns nPr
int perms;
_asm
{
	FILD       r
	FILD       n
	CALL       PERMUTATIONS
    FISTP      perms
}
return perms;
}

int Combinations(int n, int r)
{
// Combination of r elements, in a set P that contains 
// n elements, according to the formula:
//
//                n!
//     nCr =  ----------
//            r!(n - r)!
//
// where:
//          n = number of elements in set P
//          r = number of combinations of r elements
//          nCr is the number of combinations of r elements in
//              a set of n elements
//
// Note: code assumes that n and r are integers
// On entry:
//          parameter n is variable n in formula
//          parameter r is variable r in formula
// On exit:
//          Returns the number of possible combinations

int combs;
_asm
{
	FILD       r
	FILD       n
	CALL       COMBINATIONS
    FISTP      combs
}
return combs;
}

double BinomialProb(int n, int r)
{
// In an event that can ocurr in two equally likely ways,
// calculate the probability of obtaining r repettitions out
// of n trials, according to the formula:
//
//              nCr
//     P =  ----------
//              2^n
//
// where:
//          nCr = number of combinations of r elements in a set of
//                n elements
//          n = number of trials
//          r = number of repetitions
//          P = binomial probability of event r
//
// Note: code assumes that n and r are integers
// On entry:
//          parameter n is variable n in formula
//          parameter r is variable r in formula
// On exit:
//         returns binomial probability

double prob2;
_asm
{
	FILD       r
	FILD       n
	CALL       BINOMIAL_PROB
    FSTP       prob2
}
return prob2;
}
//***************************************
//       standard summations
//       and array operations
//***************************************
template <class A>
double SIGx(A *arr, int size)
{
// Function to find the sum of values in an array of any
// numeric type,
// On entry:
//         arr[] is a numeric array
//         size if the number of elements in the array
//         calculated by the caller, as follows:
//             size = sizeof dArray / sizeof dArray[0];
// On exit:
//         Returns the sum of the array elements
// First test for valid input data
double aSum = 0;

for(int x = 0; x < size; x++)
    aSum += arr[x];
return aSum;
}

template <class A>
double SIGxSq(A *arr, int size)
{
// Function to find the sum of the squares of the values in
// an array of any numeric type,
// On entry:
//         arr[] is a numeric array
//         size if the number of elements in the array
//         calculated by the caller, as follows:
//             size = sizeof dArray / sizeof dArray[0];
// On exit:
//         Returns the sum of square the array elements
double aSum = 0;

for(int x = 0; x < size; x++)
    aSum += (arr[x] * arr[x]);
return aSum;
}

template <class A>
double SIGxToSq(A *arr, int size)
{
// Function to find the square of the sum of the values in
// an array of any numeric type,
// On entry:
//         arr[] is a numeric array
//         size if the number of elements in the array
//         calculated by the caller, as follows:
//             size = sizeof dArray / sizeof dArray[0];
// On exit:
//         Returns the square of the sum of the array
//         elements
double aSum = 0;

for(int x = 0; x < size; x++)
    aSum += arr[x];
return aSum * aSum;
}

template <class A>
double SIGxy(A *arrX, A *arrY, int size)
{
// Function to find the square of the sum of the values in
// an array of any numeric type,
// On entry:
//         arrX[] is a numeric array
//         arrY[] is a second numeric array
//         size if the number of elements in the array
//         calculated by the caller, as follows:
//             size = sizeof dArray / sizeof dArray[0];
// On exit:
//         Returns the sum of the products of the
//         corresponding elements in both arrays
double aSum = 0;

for(int x = 0; x < size; x++)
    aSum += (arrX[x] * arrY[x]);
return aSum;
}

template <class A, class B>
double SIGxyByn(A *arrX, A *arrY, B n, int size)
{
// Function to find the square of the sum of the values in
// an array of any numeric type, multiplied by a constant
// of any numeric type
// On entry:
//         arrX[] is a numeric array
//         arrY[] is a second numeric array
//         n is the array sum multiplier
//         size if the number of elements in the array
//         calculated by the caller, as follows:
//             size = sizeof dArray / sizeof dArray[0];
// On exit:
//         Returns the sum of the products of the
//         corresponding elements in both arrays
//         multiplied by n
double aSum = 0;

for(int x = 0; x < size; x++)
    aSum += (arrX[x] * arrY[x]);
return ((double) aSum * n);
}

template <class A, class B>
double SIGxBySIGy(A *arrX, B *arrY, int size)
{
// Function to find the product of the sum of the values in
// two arrays of any numeric type
// On entry:
//         arrX[] is a numeric array
//         arrY[] is a second numeric array
//         size if the number of elements in the array
//         calculated by the caller, as follows:
//             size = sizeof dArray / sizeof dArray[0];
// On exit:
//         Returns the product of the sum of the
//         corresponding elements in both arrays
double aSum = 0;
double bSum = 0;

for(int x = 0; x < size; x++)
{
    aSum += arrX[x];
    bSum += arrY[x];
}
return (double(aSum * bSum));
}

template <class A>
double SIGxSkipCnt(A *arr, int skip, int count, int size)
{
// Function to find the sum of values in an array of any
// numeric type, starting at a given offset from the first
// element, for a number of elements
// On entry:
//         arr[] is a numeric array
//         skip is the start offset
//         count is the number of elements to use
//         size if the number of elements in the array
//         calculated by the caller, as follows:
//             size = sizeof dArray / sizeof dArray[0];
// On exit:
//         Returns the sum of the array elements
//         Returns 0 if invalid data
// First test for valid input data
if((skip + count) > size)
   return 0.0; 
	
double aSum = 0;

for(int x = skip; x < (count + skip); x++)
    aSum = arr[x] + aSum;
return aSum;
}


template <class A>
void Sort(A *arrS, int size )
{
// Sort an array of any numeric type using the selection
// sort algorithm, as follows:
// 
//         [0] [1] [2] [3]    <== offset
//	      | 9 | 7 | 4 | 5 |   <== original array
//                      
//	      | 5 | 7 | 4 | 9 |   <== pass 1          
//                      ^
//                      |--------- bottom
//
//	      | 5 | 4 | 7 | 9 |   <== pass 2          
//                  ^
//                  |------------- bottom
//
//	      | 5 | 4 | 7 | 9 |   <== pass 3          
//              ^
//              |----------------- bottom
//
//	      | 4 | 5 | 7 | 9 |   <== pass 4
//          ^
//          |--------------------- bottom
// DONE

    int maxInd;   // Index of largest value in each pass
    int bottom;   // False bottom index for each pass
    int i;        // Local counter
    A temp;       // Temporary storage 

    for (bottom = size-1; bottom >= 1; bottom--)
	{
        maxInd = 0;
        for (i = 1; i <= bottom; i++) 
            if (arrS[i] > arrS[maxInd])
                maxInd = i;
        temp = arrS[bottom];
        arrS[bottom] = arrS[maxInd];
        arrS[maxInd] = temp; 
    }
}

template <class A>
void SortCopy(A *arrS, A *arrC, int size )
{
// Sort an array of any numeric type using the selection
// sort algorithm into a second array passed by the caller
    

    int maxInd;   // Index of largest value in each pass
    int bottom;   // False bottom index for each pass
    int i;        // Local counter
    A temp;       // Temporary storage 

    // Copy data from original array
	for(int j = 0; j < size; j++)
		arrC[j] = arrS[j];

    for (bottom = size-1; bottom >= 1; bottom--)
	{
        maxInd = 0;
        for (i = 1; i <= bottom; i++) 
            if (arrC[i] > arrC[maxInd])
                maxInd = i;
        temp = arrC[bottom];
        arrC[bottom] = arrC[maxInd];
        arrC[maxInd] = temp; 
    }
}


template <class A>
void ArrayShow(A *arr, int size)
{
// Function to display the elements in an array of any
// numeric type.
// On entry:
//         arr[] is a numeric array
//         size if the number of elements in the array
//         calculated by the caller, as follows:
//             size = sizeof dArray / sizeof dArray[0];
// On exit:
//         array elements are displayed on the console
//         returns nothing
	
	for(int x = 0; x < size; x++)
		cout << arr[x] << " ";
    return;
}

//**********************************
//   MEASURES OF CENTRAL TENDENCY
//**********************************
template <class A>
double MeanOf(A *arr, int skip, int count, int size)
{
// Function to find the mean in an array of any type
// starting at a given offset from the first element,
// for a given number of elements
// On entry:
//         arr[] is a numeric array
//         skip is the start offset
//         count is the number of elements to use
//         size is the number of elements in the array
//         calculated by the caller, as follows:
//             size = sizeof dArray / sizeof dArray[0];
// On exit:
//         Returns the mean value in the array
//         Returns 0 if invalid data
// First test for valid input data
if((skip + count) > size)
   return 0.0; 

double aSum = 0;

for(int x = skip; x < (count + skip); x++)
    aSum = arr[x] + aSum;
return aSum/count;
}

template <class A>
double Mean(A *arr, int size)
{
// Function to find the mean in an array of any type
// On entry:
//         arr[] is a numeric array
//         size is the number of elements in the array
//         calculated by the caller, as follows:
//             size = sizeof dArray / sizeof dArray[0];
// On exit:
//         Returns the mean value in the array
double aSum = 0;

for(int x = 0; x < size; x++)
    aSum = arr[x] + aSum;
return aSum/size;
}

template <class A>
double Median(A *arr, int size)
{
// Function to find the median in an array of any type
// On entry:
//         arr[] is a numeric array
//         size is the number of elements in the array
//         calculated by the caller, as follows:
//             size = sizeof dArray / sizeof dArray[0];
// Terms:
//         case 1:
//           If arr[] has an odd number of elements,
//           then the median is the value of the middle
//           element
//         case 2:
//           If arr[] has an even number of elements,
//           then the median is the mean of the two middle
//           elements
// On exit:
//         Returns the median value in the array
double median = 0;

int midE = size / 2;
// Sort a copy of the array passed as an argument

A *arrCopy = new A[size];

// Copy data from original array
for(int j = 0; j < size; j++)
	arrCopy[j] = arr[j];

Sort(arrCopy, size);


if(size % 2 != 0)        // if odd number of elements
   median = (double) arrCopy[midE];
else
   median = (double) ((arrCopy[midE] + arrCopy[midE - 1]) /2.0);

return median;
}

template <class A>
double Midrange(A *arr, int size)
{
// Function to find the midrange in an array of any type.
// The midrange is the value midway between the highest
// and the lowest entry.
// On entry:
//         arr[] is a numeric array
//         size is the number of elements in the array
//         calculated by the caller, as follows:
//             size = sizeof dArray / sizeof dArray[0];
// On exit:
//         Returns the midrange value in the array
double midrange = 0;

// Sort a copy of the array passed as an argument
A *arrCopy = new A[size];

// Copy data from original array
for(int j = 0; j < size; j++)
	arrCopy[j] = arr[j];

Sort(arrCopy, size);

midrange = (arrCopy[size - 1] - arrCopy[0]) / 2.0;
return midrange;
}

template <class A>
double Mode(A *arr, int *modeCode, int size)
{
// Function to find the mode, if one exists, in an array
// of any type
// On entry:
//         arr[] is a numeric array
//         modeCode is a user variable to hold a code
//         returned by the function
//         size is the number of elements in the array
//         calculated by the caller, as follows:
//             size = sizeof dArray / sizeof dArray[0];
// Terms:
//         The mode is the value that occurs most
//         frequently in the array.
//         case 1:
//       	 If no value occurs more than once, then
//           the array has no mode. Function retunrs
//           mode = zero, modeCode = 0.
//         case 2:
//           It is possible for an array to have more than
//           one mode. In this case the function returns:
//           mode = mode with highest index
//           modeCode = number of modes.
// On exit:
//         Returns the mode in the array, if there is one
//         Zero if no mode found.
//         Mode is always returned in a variable of type
//         double
//         Modecode holds the number of modes.

double mode = 0;

int numModes = 0;       // Number of modes found
int oldM = 0, newM = 0; // Multiplicity counters
int sPtr = 0, ePtr = 0;       // Offset pointers
int loc;                // Offset location of mode
// A match;

// Create a copy of the array passed as an argument
A *arc = new A[size];
// Copy data from original array
for(int j = 0; j < size; j++)
	arc[j] = arr[j];
// Sort the copy
Sort(arc, size);

// Find mode

  while(sPtr < (size - 1))
  {
	 ePtr = sPtr + 1; 
     if(arc[sPtr] != arc[ePtr]) // No match condition
	 {                          // s = e, e = e + 1
        sPtr = ePtr;            // Search for new match
		newM = 0;               // Reset multiplier
    	continue;         
	 }
	 else                     // Match 	 
     {
		 newM++;              // Bump multiplier
		   if(newM == oldM)
			   numModes++;
      	   else if(newM > oldM)    // New multiplier is larger
		   {
               oldM = newM;   // Reset old multiplier
		       loc = sPtr;    // Store location of match
	    	   numModes = 1;  // bump modes counter
		   }
     }
    sPtr++; 
  } // end of loop

//    Case 1: No mode in set
//            numModes = 0;
//    Case 2: One mode in set
//            numModes = 1;
//    Case 3: More than one mode   
//            numModes > 1
mode = (double) arc[loc];
*modeCode = numModes;

return mode;
}

template <class A>
double WeightedMean(A *dat, int dist[], int size)
{
// Function to find the wieghted mean according to the
// frequency distribution of the data
// On entry:
//         dat[] is a numeric array of the data
//         dis[] is an integer array of the frequency
//         distribution of the data 
//         size is the number of elements in the array
//         calculated by the caller, as follows:
//             size = sizeof dArray / sizeof dArray[0];
// According to the formula:
//
//              SIG xw 
//       Wx = ------------
//               SIG w
// where:
//       x is the data elements
//       w is the frequency
//       Wx is the weighted mean
// On exit:
//         Returns the wieghted mean value in the data
//         array according to the frequency distribution

double wMean = 0;
A sumwx = 0;
int sumw = 0;

// calculate SIG w * x and SIG W
for(int x = 0; x < size; x++)
{
    sumwx = sumwx + (dat[x] * dist[x]);
    sumw = sumw + dist[x];
}
wMean = (double)sumwx / (double)sumw;

return wMean;
}

template <class A>
double WeightedMedian(A *dat, int dist[], int size)
{
// Function to find the wieghted median according to the
// frequency distribution of the data
// On entry:
//         dat[] is a numeric array of the data
//         dis[] is an integer array of the frequency
//         distribution of the data 
//         size is the number of elements in the array
//         calculated by the caller, as follows:
//             size = sizeof dArray / sizeof dArray[0];
//
// Weighted median is calculated by expanding the data
// array using the frequency distribution and then
// obtaining the median in the resulting array.
// On exit:
//         Returns the wieghted median value in the data
//         array according to the frequency distribution

double wMedian = 0;
int aoff = 0;           // Offset in array
int wMsize = 0;

// Calculate the size of the expanded array by counting
// the data's frequency distribution
for(int x = 0; x < size; x++)
    wMsize = wMsize + dist[x];

// Create the expanded array
A *exp = new A[wMsize];

// Fill expanded array
for(x = 0; x < size; x++)
{
    for(int y = 0; y < dist[x]; y++)
    { 
		exp[aoff] = dat[x];
		aoff++;
	}
}
// Use the Median() function on the expanded array
wMedian = Median(exp, wMsize);

return wMedian;
}

template <class A>
double WeightedMode(A *dat, int dist[], int size)
{
// Function to find the wieghted mode according to the
// frequency distribution of the data
// On entry:
//         dat[] is a numeric array of the data
//         dis[] is an integer array of the frequency
//         distribution of the data 
//         size is the number of elements in the array
//         calculated by the caller, as follows:
//             size = sizeof dArray / sizeof dArray[0];
//
// Weighted mode is the data value that occurs most
// frequently in the set
// On exit:
//         Returns the wieghted mode in the data array
//         according to the frequency distribution

double wMode = 0;
int largest = 0;    // Largest value 
int offL = 0;       // Offset of largest element

// Obtain the largest value in the dist[] array
for(int x = 0; x < size; x++)
{
    if(dist[x] > largest)
    {
		largest = dist[x];
	    offL = x;
	}
}
wMode = dat[offL]; 
return wMode;
}


//**********************************
//     MEASURES OF DISPERSION
//**********************************
template <class A>
double Range(A *arr, int size)
{
// Function to find the range of the data set, defined
// as the difference between the largest and the smallest
// element in the set
// On entry:
//         arr[] is a numeric array
//         size is the number of elements in the array
//         calculated by the caller, as follows:
//             size = sizeof dArray / sizeof dArray[0];
// On exit:
//         Returns the data range

double range = 0;

A largest = 0;    // Largest value 
A smallest;

// Obtain the largest value in the array
for(int x = 0; x < size; x++)
{
    if(arr[x] > largest)
		largest = arr[x];
}

smallest = largest;

// Obtain the smallest value in the array
for(x = 0; x < size; x++)
{
    if(arr[x] < smallest)
		smallest = arr[x];
}

range = largest - smallest;

return range;
}

template <class A>
double VarianceS(A *dat, int size)
{
// Function to find the variance of a sample set, defined
// by the following formula:
//
//            SIG(x - m)^2 
//      vS = ----------------
//               n - 1
// where:
//       x are the population data
//       m is the mean of the population data
//       n is the number of data entries
//       vS is the variance
// On entry:
//         dat[] is a numeric array
//         size is the number of elements in the array
//         calculated by the caller, as follows:
//             size = sizeof dArray / sizeof dArray[0];
// On exit:
//         Returns the variance

double varS = 0;
double dMean;
double sumXm = 0;

// Obtain and store mean
dMean = Mean(dat, size);

// Calculate SIG(x - m)^2 
for(int x = 0; x < size;x++)
   sumXm = sumXm + (dat[x] - dMean)*(dat[x] - dMean);

varS = sumXm / (size - 1);

return varS;
}

template <class A>
double VarianceP(A *dat, int size)
{
// Function to find the variance of a population, defined
// by the following formula:
//
//            SIG(x - m)^2 
//     vP = ----------------
//                 n
// where:
//       x are the population data
//       m is the mean of the population data
//       n is the number of data entries
//       vP is the variance
// On entry:
//         dat[] is a numeric array
//         size is the number of elements in the array
//         calculated by the caller, as follows:
//             size = sizeof dArray / sizeof dArray[0];
// On exit:
//         Returns the variance

double varP = 0;
double dMean;
double sumXm = 0;

// Obtain and store mean
dMean = Mean(dat, size);

// Calculate SIG(x - m)^2 
for(int x = 0; x < size;x++)
   sumXm = sumXm + (dat[x] - dMean)*(dat[x] - dMean);

varP = sumXm / size;

return varP;
}

template <class A>
double StdDevS(A *dat, int size)
{
// Function to find the standard deviation of a sample
// set, defined by the following formula:
//
//      sdS = SQR vS
//
// where:
//       vS is the variance of the set
//       sdS is the standard deviation
// On entry:
//         dat[] is a numeric array
//         size is the number of elements in the array
//         calculated by the caller, as follows:
//             size = sizeof dArray / sizeof dArray[0];
// On exit:
//         Returns the standard deviation for the sample

double sdS;
double varS;

// Obtain the variance of the sample set
varS = VarianceS(dat, size);

// Calculate square root of variance 
_asm
{
	  FLD     varS
	  FSQRT
	  FSTP    sdS
}
return sdS;
}

template <class A>
double StdDevP(A *dat, int size)
{
// Function to find the standard deviation of a
// population, defined by the following formula:
//
//      sdP = SQR vP
//
// where:
//       vP is the variance of the population
//       sdS is the standard deviation
// On entry:
//         dat[] is a numeric array
//         size is the number of elements in the array
//         calculated by the caller, as follows:
//             size = sizeof dArray / sizeof dArray[0];
// On exit:
//         Returns the standard deviation for the
//         population

double sdP;
double varP;

// Obtain the variance of the sample set
varP = VarianceP(dat, size);
// Calculate square root of variance 
_asm
{
	  FLD     varP
	  FSQRT
	  FSTP    sdP
}
return sdP;
}

//***********************************************
//      Normal Distribution and Probabiliaty
//***********************************************
void NormalDist4Pts(double yArray[], double mean, double sd)
{
// Function to calculate normal distribution given the mean
// and the standard deviation, using the formula:
// formula:
//             e^-((x-u)^2/2s^2)
//     y = ---------------------
//                s * SQR 2Pi  
// where:
//       x is the cartesian coordinate in terms of the
//         mean. 
//       y is the value of the dependent variable
//       u is the mean
//       s is the standard deviation
// On entry:
//        yArray[] is an array of 4 entries for results
//                 at x = mean, x = mean + sd,
//                 x = mean + 2*sd, x = mean + 3*sd
//        mean is formula variable u
//        sd is formula variable s
// On exit:
//        yArray[] contains the following results:
//           yArray[0] = y for x = mean
//           yArray[1] = y for x = mean + sd
//           yArray[2] = y for x = mean + 2*sd
//           yArray[3] = y for x = mean + 3*sd

double x = mean;
double fOfx;

  for(int y = 0; y < 4; y++)
  {
	// Calculate x = mean
	_asm
	{
	; Function expects:         |  ST0  |  ST1  |  ST2  |
								;   x   |  mean |   sd  |
	   FLD      sd
       FLD      mean
	   FLD      x
       CALL     NORMAL_CURVE_Y
	   FSTP     fOfx
	}
    // Store results in array
    yArray[y] = fOfx;
	x = x + sd;
  }
return;
}

double NormalDistY(double x, double mean, double sd)
{
// Function to calculate normal distribution given the mean
// and the standard deviation, using the formula:
// formula:
//             e^-((x-u)^2/2s^2)
//     y = ---------------------
//                s * SQR 2Pi  
// where:
//       x is the cartesian coordinate in terms of the
//         mean. 
//       y is the value of the dependent variable
//       u is the mean
//       s is the standard deviation
// On entry:
//        x is the independent variable
//        mean is formula variable u
//        sd is formula variable s
// On exit:
//        return f(x) = y for normal distribution

double fOfx;

	// Calculate y = f(x)
	_asm
	{
	; Function expects:         |  ST0  |  ST1  |  ST2  |
								;   x   |  mean |   sd  |
	   FLD      sd
       FLD      mean
	   FLD      x
       CALL     NORMAL_CURVE_Y
	   FSTP     fOfx
	}
return fOfx;
}


double NormalProbTrpz(double a, double b, double mean, double sd)
{
// Calculate normal probability by estimating the area under
// the normal distribution curve, as follows:
//
//      Np = INTEGRAL(from a to b) f(x)dx	
//
// where:
//       a is the x coordinate of the first point in the curve
//       b is the x coordinate of the second point
//       f(x) is the normal distribution function:
//                    e^-((x-u)^2/2s^2)
//            f(x) = --------------------
//                       s * SQR 2Pi
//	          where:
//            x is the cartesian coordinate
//            f(x) is the dependent variable
//            u is the mean
//            s is the standard deviation
//       Np is the area under the curve from point a to
//       point b.
//
// Integration is performed using the trapezoidal rule,
// as follows:
//           
//       Np = (step / 2) * (y1 + 2y2 + 2y3 + 2y4 ... yTRPZ+1)
//
// where:
//       TRPZ is a constant defining the number of trapezoids
//            used in the computation
//       y1, y2 .. yTRPZ+1 are values of f(x)
//       a is x coordinate of the lower limit of the range
//       b is x coordinate of the upper limit of the range     
//       step = b - a / TRPZ
//    
// Variables
const int TRPZ = 100;          // Number of trapezoids

// Local variables
double area = 0;
double xVal;
double fOfx;
double step = (b - a) / TRPZ;      // x increment  
double tAreas[TRPZ + 1];           // Array for f(x)
  
// Calculate and store y1
	_asm
	{
	   ; Function expects:         |  ST0  |  ST1  |  ST2  |
				    			   ;   x   |  mean |   sd  |
	   FLD      sd
       FLD      mean
	   FLD      a
       CALL     NORMAL_CURVE_Y
	   FSTP     fOfx
	}
    tAreas[0] = fOfx;

// Calculate and store yTRPZ+1
	_asm
	{
	; Function expects:         |  ST0  |  ST1  |  ST2  |
								;   x   |  mean |   sd  |
	   FLD      sd
       FLD      mean
	   FLD      b
       CALL     NORMAL_CURVE_Y
	   FSTP     fOfx
	}
    tAreas[1] = fOfx;

// Calculate and store 2y2 to 2TRPZ+1
xVal = a + step;          // x1 coordinate

for(int j = 2; j < TRPZ + 1; j++)
{
 	_asm
	{
	; Function expects:         |  ST0  |  ST1  |  ST2  |
								;   x   |  mean |   sd  |
	   FLD      sd
       FLD      mean
	   FLD      xVal
       CALL     NORMAL_CURVE_Y
	   FSTP     fOfx
	}
	// Calculate area for this entry
    tAreas[j] = 2 * fOfx;
    xVal = xVal + step;	
}

// At this point there are TRPZ + 1 entries in the local array
// Add values to calculate area
for(int z = 0; z < (TRPZ + 1) ; z++)
  area = area + tAreas[z];  

return (area * (step / 2) * 100);
}

double NormalProb(double a, double b, double mean, double sd)
{
// Calculate normal probability by estimating the area under
// the normal distribution curve, as follows:
//
//      Np = INTEGRAL(from a to b) f(x)dx	
//
// where:
//       a is the x coordinate of the first point in the curve
//       b is the x coordinate of the second point
//       f(x) is the normal distribution function:
//                    e^-((x-u)^2/2s^2)
//            f(x) = --------------------
//                       s * SQR 2Pi
//	          where:
//            x is the cartesian coordinate
//            f(x) is the dependent variable
//            u is the mean
//            s is the standard deviation
//       Np is the area under the curve from point a to
//       point b.
//
// Integration is performed using Simpson's rule, as follows:
//           
//    Np =  step/3 * (y1 + 4y2 + 2y3 + 4y4 + 2y5 + 4y6 .. yCNT+1 
//
// where:
//       N is a constant defining the number of area elements
//           used in the computation
//       y1, y2 .. yN+1 are values of f(x)
//       a is x coordinate of the lower limit of the range
//       b is x coordinate of the upper limit of the range     
//       step = b - a / N
// The sum consists of 4 elements
//    1. y1
//    2. 4 * (sums of all yj, j even from 2 to N)
//    3. 2 * (sums of all yk, k odd from 3 to N - 1)
//    4. yN+1	
//  
// Variables
const int N = 20;          // Number of areas

// Local variables
double area = 0;
double xVal;
double fOfx;
double step = (b - a) / N;       // x increment  
double tAreas[N + 1];            // Array for f(x)

// Clear data array
for(int c = 0; c < N + 1; c++)
   tAreas[c] =  0;
  
// Calculate and sum element 1
// y1
	_asm
	{
	   ; Function expects:         |  ST0  |  ST1  |  ST2  |
				    			   ;   x   |  mean |   sd  |
	   FLD      sd
       FLD      mean
	   FLD      a
       CALL     NORMAL_CURVE_Y
	   FSTP     fOfx
	}
    tAreas[0] = fOfx;

// Calculate and store sum element 4
// yN+1
	_asm
	{
	; Function expects:         |  ST0  |  ST1  |  ST2  |
								;   x   |  mean |   sd  |
	   FLD      sd
       FLD      mean
	   FLD      b
       CALL     NORMAL_CURVE_Y
	   FSTP     fOfx
	}
    tAreas[1] = fOfx;

// Calculate and store sum element 2
// 4 * (sums of all yj, j even from 2 to CNT)
//
xVal = a + step;          // first j

for(int j = 2; j < (N + 1); j += 2)
{
 	_asm
	{
	; Function expects:         |  ST0  |  ST1  |  ST2  |
								;   x   |  mean |   sd  |
	   FLD      sd
       FLD      mean
	   FLD      xVal
       CALL     NORMAL_CURVE_Y
	   FSTP     fOfx
	}
	// Calculate area for this entry
    tAreas[j] = 4 * fOfx;
    xVal = xVal + (2 * step);	
}

// Calculate and store sum element 3
// 2 * (sums of all yk, k odd from 3 to CNT-1)
//
xVal = a + (2 * step);          // first k

for(int k = 3; k < N; k += 2)
{
 	_asm
	{
	; Function expects:         |  ST0  |  ST1  |  ST2  |
								;   x   |  mean |   sd  |
	   FLD      sd
       FLD      mean
	   FLD      xVal
       CALL     NORMAL_CURVE_Y
	   FSTP     fOfx
	}
	// Calculate area for this entry
    tAreas[k] = 2 * fOfx;
    xVal = xVal + (2 * step);	
}

// At this point there are CNT + 1 entries in the local array
// Add values to calculate area
for(int z = 0; z < (N + 1) ; z++)
  area = area + tAreas[z];  

return (area * (step / 3) * 100); 
}



double NormalProbError(double a, double b, double mean, double sd)
{
// Use interval arithmetic to determine rounding error in normal
// distribution calculation using the NormalProb() function

double errorVal = 0;
double probDown;
double probUp;

const unsigned short RND_DOWN = 0x14ff;
const unsigned short RND_UP = 0x18ff;
unsigned short status;

// First set math unit to round down
_asm
{ 
     FSTSW    status     ; Store status word
     FLDCW    RND_DOWN
}
// Calculate normal probability
probDown = NormalProbTrpz(a, b, mean, sd);

// Now set to round up
_asm
{ 
     FLDCW    RND_UP
}
// Re-calculate normal probability
probUp = NormalProbTrpz(a, b, mean, sd);

// Restore original status
_asm
{
	FLDCW    status
}

// Compare results
if(probUp > probDown)
   errorVal = probUp - probDown;
else
   errorVal = probDown - probUp;

return errorVal;
}

//**********************************************
//          Correlation and Regression
//**********************************************
template <class A>
double LinCorCoef(A *datX, A *datY, int size)
{
// Function to find the linear correlation coefficient
// between two data sets.
// On entry:
//         datX[] is a numeric data array
//         datY[] is a second numeric data array
//         size is the number of elements in both arrays
//         calculated by the caller, as follows:
//             size = sizeof datX / sizeof datY[0];
// Linear correlation between the data in both arrays is
// calculated according to the formula:
//
//              n(SIGxy) - (SIGx)(SIGy)    
//  r = ------------------------------------------------
//      SQR(n(SIGx^2)-(SIGx)^2 * SQR(n(SIGy^2)-(SIGy)^2
//
// where:
//       x are the data elements in the array datX
//       y are the data elements in the array dayY
//       n is the number of elements
//       r is the linear correlation coefficient
// On exit:
//       Returns the linear correlation coefficient

double lcc = 0;
double nume;         // Storage for numerator
double d1, d2;       // Storage for denominator

A sumXY = 0;       // SIGxy
A sumX = 0;        // SIGx
A sumY = 0;        // SIGy
A sumXSq = 0;      // SIGx^2
A sumYSq = 0;      // SIGy^2



// Calcuate sums required in formulat
for(int i = 0; i < size; i++)
{
    sumXY = sumXY + (datX[i] * datY[i]);
	sumX = sumX + datX[i];
	sumY = sumY + datY[i];
	sumXSq = sumXSq + (datX[i] * datX[i]);
    sumYSq = sumYSq + (datY[i] * datY[i]); 
}

// Calculate numerator: n(SIGxy) - (SIGx)(SIGy)
nume = (double) size * sumXY - (sumX * sumY);

// Calculate denominator left and right elements
// d1 = SQR(n(SIGx^2)-(SIGx)^2
// d2 = SQR(n(SIGy^2)-(SIGy)^2
d1 = (size * sumXSq) - (sumX * sumX);
d2 = (size * sumYSq) - (sumY * sumY);
d1 = sqrt(d1);
d2 = sqrt(d2);
lcc = nume / (d1 * d2);

return lcc;

}

template <class A>
double RegrYInt(A *datX, A *datY, int size)
{
// Function to find the y-intercept of a linear regression
// equation related to two data sets.
// On entry:
//         datX[] is a numeric data array
//         datY[] is a second numeric data array
//         size is the number of elements in both arrays
//         calculated by the caller, as follows:
//             size = sizeof datX / sizeof datY[0];
//
// Calculates y-intercept (m) in the formula:
//
//      y' = mx + b
//
//            n(SIGxy) - (SIGx)(SIGy)
//      m = -----------------------------------
//               n(SIGx^2) - (SIGx)^2
// where:
//       x are the data elements of the array datX
//       y are the data elements of the array dayY
//       n is the number of elements
//       m is the y-intercept of y' = mx + b
// On exit:
//       Returns the y-intercept

double m = 0;
double nume;         // Storage for numerator
double d1, d2;       // Storage for denominator

A sumXY = 0;       // SIGxy
A sumX = 0;        // SIGx
A sumY = 0;        // SIGy
A sumXSq = 0;      // SIGx^2


// Calcuate sums required in formulat
for(int i = 0; i < size; i++)
{
      sumXY = sumXY + (datX[i] * datY[i]);
	sumX = sumX + datX[i];
	sumY = sumY + datY[i];
	sumXSq = sumXSq + (datX[i] * datX[i]);
}

// Calculate numerator: nSIGxy - (SIGx)(SIGy)
nume = (double) size * sumXY - (sumX * sumY);

// Calculate denominator left and right elements
// d1 = n(SIGx^2)
// d2 = (SIGx)^2
d1 = size * sumXSq;
d2 = sumX * sumX;
m = nume / (d1 - d2);

return m;

}

template <class A>
double RegrSlope(A *datX, A *datY, int size)
{
// Function to find the slope of a linear regression
// equation between two data sets.
// On entry:
//         datX[] is a numeric data array
//         datY[] is a second numeric data array
//         size is the number of elements in both arrays
//         calculated by the caller, as follows:
//             size = sizeof datX / sizeof datY[0];
//
// Calculates slope (b) in the formula:
//
//      y' = mx + b
//
//            SIGy - m(SIGx)
//      b = -----------------
//                 n
// where:
//       x are the data elements of the array datX
//       y are the data elements of the array dayY
//       n is the number of elements
//       m is the y-intercept of y' = mx + b
// Uses:
//       the y-intercept returned by the RegrYInt()
//       function
// On exit:
//       Returns the slope

double b = 0;
double nume;         // Storage for numerator
double m;            // Storage for y-intercept

A sumX = 0;        // SIGx
A sumY = 0;        // SIGy

// Calcuate sums required in formulat
for(int i = 0; i < size; i++)
{
	sumX = sumX + datX[i];
	sumY = sumY + datY[i];
}

// Obtain yIntercept
m = RegrYInt(datX, datY, size);

// Calculate numerator: SIGy - m(SIGx)
nume = (double) sumY - (m * sumX);

b = nume / size;

return b;

}


double LoadE()
{

double eConst;
_asm
{
       CALL  LOADE 
	   FSTP     eConst
       NOP
}
return eConst;
}