//************************************************************
//                    Linear Systems
//************************************************************
// Reference:   SOFTWARE SOLUTIONS
// Workspace  : Chapter13\Test Un32_12
// Entry point: Test Un32_12
// Other files:
// Development: Visual C++ 6.0 Console Application
// Date:        March 19/2001
//
// Description:
//              Header file for primtive functions used in
//              linear systems programming.
//************************************************************

// MATH32 library functions
extern "C" void ROW_TIMES_SCALAR(); 
extern "C" void ROW_PLUS_SCALAR(); 
extern "C" void ROW_DIV_SCALAR(); 
extern "C" void ROW_MINUS_SCALAR(); 
extern "C" void MAT_TIMES_SCALAR(); 
extern "C" void MAT_PLUS_SCALAR(); 
extern "C" void MAT_DIV_SCALAR(); 
extern "C" void MAT_MINUS_SCALAR(); 
extern "C" void ADD_MATRICES();
extern "C" void MUL_MATRICES();
extern "C" void GAUSS_SOLVE();

//**************************************************************
//                      Support Functions
//**************************************************************

template <class A>
void ShowMatrix(A *matx, int rows, int cols)
{
// Display a matrix of any numeric type
// On entry:
//       *mat is caller's matrix 
//       rows is nomber of rows in matrix
//       cols is number of columns in matrix

for(int i = 0;i < rows; i++)
{
   cout << "\n";
   for(int j = 0; j < cols; j++)
   {
       cout <<  matx[(i * cols) + j];
	   cout << "\t";
   }
}
}

//**************************************************************
//            C++ interface functions to Un32_13 module
//**************************************************************
template <class A>
void RowTimesScalarLL(A *matx, int i, int cols, A scalar)
{
// Multiply a matrix row times a scalar using low-level code
// in the Un32_13 module
// On entry:
//       *mat is caller's matrix (type double) 
//       i is number of the row
//       cols is number of columns in the matrix
//       scalar is the value to multiply (type double)
// Routine expects:
//       ST(0) holds scalar
//       ESI --> matrix
//       AL = row vector number
//       CL = number of columns in matrix
// On exit:
//       elements in matrix row i are multipliced by scalar
int eSize = sizeof(matx[0]);

_asm
{

         MOV     ECX,cols          // Columns to ECX 
		 MOV     EAX,i             // Row number to EAX 
		 MOV     ESI,matx          // Address to ESI
		 FLD     scalar            // Scalar to ST(0)
         MOV     EDX,eSize         // Horizontal skip
         CALL    ROW_TIMES_SCALAR
}
return;
}

template <class A>
void RowPlusScalarLL(A *matx, int i, int cols, A scalar)
{
// Multiply a matrix row times a scalar using low-level code
// in the Un32_13 module
// On entry:
//       *mat is caller's matrix (type double) 
//       i is number of the row
//       cols is number of columns in the matrix
//       scalar is the value to add (type double)
// Routine expects:
//       ST(0) holds scalar
//       ESI --> matrix
//       EAX = row vector number
//       ECX = number of columns in matrix
// On exit:
//       elements in matrix row i are multipliced by scalar
int eSize = sizeof(matx[0]);
	
	_asm
{

         MOV     ECX,cols          // Columns to ECX 
		 MOV     EAX,i             // Row number to EAX 
		 MOV     ESI,matx          // Address to ESI
		 FLD     scalar            // Scalar to ST(0)
		 MOV     EDX,eSize         // Horizontal skip
		 CALL    ROW_PLUS_SCALAR
}
return;
}

template <class A>
void RowDivScalarLL(A *matx, int i, int cols, A scalar)
{
// Divide a matrix row by a scalar using low-level code
// in the Un32_13 module
// On entry:
//       *mat is caller's matrix (type double) 
//       i is number of the row
//       cols is number of columns in the matrix
//       scalar is the value to add (type double)
// Routine expects:
//       ST(0) holds scalar
//       ESI --> matrix
//       EAX = row vector number
//       ECX = number of columns in matrix
// On exit:
//       elements in matrix row i are multipliced by scalar
int eSize = sizeof(matx[0]);
_asm
{

         MOV     ECX,cols          // Columns to ECX 
		 MOV     EAX,i             // Row number to EAX 
		 MOV     ESI,matx          // Address to ESI
		 FLD     scalar            // Scalar to ST(0)
		 MOV     EDX,eSize         // Horizontal skip
         CALL    ROW_DIV_SCALAR
}
return;
}

template <class A>
void RowMinusScalarLL(A *matx, int i, int cols, A scalar)
{
// Subtract a scalar from a matrix row using low-level code
// in the Un32_13 module
// On entry:
//       *mat is caller's matrix (type double) 
//       i is number of the row
//       cols is number of columns in the matrix
//       scalar is the value to add (type double)
// Routine expects:
//       ST(0) holds scalar
//       ESI --> matrix
//       EAX = row vector number
//       ECX = number of columns in matrix
// On exit:
//       elements in matrix row i are multipliced by scalar
int eSize = sizeof(matx[0]);
_asm
{

         MOV     ECX,cols          // Columns to ECX 
		 MOV     EAX,i             // Row number to EAX 
		 MOV     ESI,matx          // Address to ESI
		 FLD     scalar            // Scalar to ST(0)
		 MOV     EDX,eSize         // Horizontal skip
         CALL    ROW_MINUS_SCALAR
}
return;
}

//*******************************************
//     matrix-by-scalar operations
//*******************************************
template <class A>
void MatTimesScalarLL(A *matx, int rows, int cols, A scalar)
{
// Multiply a matrix times a scalar using low-level code
// in the Un32_13 module
// On entry:
//       *mat is caller's matrix (type double) 
//       rows is number of the rows in matrix
//       cols is number of columns in the matrix
//       scalar is the value to multiply (floating point type)
// Routine expects:
//       ST(0) holds scalar
//       ESI --> matrix
//       AL = row vector number
//       CL = number of columns in matrix
// On exit:
//       elements in matrix are multipliced by scalar
int eSize = sizeof(matx[0]);

_asm
{

         MOV     ECX,cols          // Columns to ECX 
		 MOV     EAX,rows          // Rows to EAX 
		 MOV     ESI,matx          // Address to ESI
		 FLD     scalar            // Scalar to ST(0)
		 MOV     EDX,eSize         // Horizontal skip
         CALL    MAT_TIMES_SCALAR
}
return;
}

template <class A>
void MatPlusScalarLL(A *matx, int rows, int cols, A scalar)
{
// Add a scalar to a matrix using low-level code in the
// Un32_13 module
// On entry:
//       *mat is caller's matrix (type double) 
//       rows is number of the rows in matrix
//       cols is number of columns in the matrix
//       scalar is the value to add (floating-point type)
// Routine expects:
//       ST(0) holds scalar
//       ESI --> matrix
//       AL = row vector number
//       CL = number of columns in matrix
// On exit:
//       scalar is added to all elements in matrix
int eSize = sizeof(matx[0]);

_asm
{

         MOV     ECX,cols          // Columns to ECX 
		 MOV     EAX,rows          // Rows to EAX 
		 MOV     ESI,matx          // Address to ESI
		 FLD     scalar            // Scalar to ST(0)
		 MOV     EDX,eSize         // Horizontal skip
         CALL    MAT_PLUS_SCALAR
}
return;
}


template <class A>
void MatDivScalarLL(A *matx, int rows, int cols, A scalar)
{
// Divide a matrix by a scalar using low-level code in the
// Un32_13 module
// On entry:
//       *mat is caller's matrix (type double) 
//       rows is number of the rows in matrix
//       cols is number of columns in the matrix
//       scalar is the value to add (floating-point type)
// Routine expects:
//       ST(0) holds scalar
//       ESI --> matrix
//       AL = row vector number
//       CL = number of columns in matrix
// On exit:
//       elements in matrix row i are divided by scalar
int eSize = sizeof(matx[0]);

_asm
{

         MOV     ECX,cols          // Columns to ECX 
		 MOV     EAX,rows          // Rows to EAX 
		 MOV     ESI,matx          // Address to ESI
		 FLD     scalar            // Scalar to ST(0)
		 MOV     EDX,eSize         // Horizontal skip
         CALL    MAT_DIV_SCALAR
}
return;
}


template <class A>
void MatMinusScalarLL(A *matx, int rows, int cols, A scalar)
{
// Subtract a scalar to a matrix using low-level code in the
// Un32_13 module
// On entry:
//       *mat is caller's matrix (type double) 
//       rows is number of the rows in matrix
//       cols is number of columns in the matrix
//       scalar is the value to add (floating-point type)
// Routine expects:
//       ST(0) holds scalar
//       ESI --> matrix
//       AL = row vector number
//       CL = number of columns in matrix
// On exit:
//       scalar is subratcted from all elements in matrix
int eSize = sizeof(matx[0]);

_asm
{

         MOV     ECX,cols          // Columns to ECX 
		 MOV     EAX,rows          // Rows to EAX 
		 MOV     ESI,matx          // Address to ESI
		 FLD     scalar            // Scalar to ST(0)
		 MOV     EDX,eSize         // Horizontal skip
         CALL    MAT_MINUS_SCALAR
}
return;
}

template <class A>
void AddMatrices(A *matA, A *matB, A *matC, int rows, int cols)
{
// Perform matrix addition: C = A + B using low-level code in the
// Un32_13 module
// On entry:
//       *matA and *matB are matrices to be added
//       *matC is matrix for sums 
//       rows is number of the rows in matrix
//       cols is number of columns in the matrix
// Requires:
//       All three matrices must be of the same dimensions
//       All three matrices must be of the same float
//       data type
// Routine expects:
//      ESI --> first matrix (A)
//      EDI --> second matrix (B)
//      EBX --> storage area for addition matrix (C)
//      EAX = number of rows in matrices
//      ECX = number of columns in matrices
//      EDX = horizontal skip factor
// On exit:
//      returns matC[] = matA[] + matB[]

int eSize = sizeof(matA[0]);

_asm
{

         MOV     ECX,cols          // Columns to ECX 
		 MOV     EAX,rows          // Rows to EAX 
		 MOV     ESI,matA          // Address to ESI
		 MOV     EDI,matB
		 MOV     EBX,matC
		 MOV     EDX,eSize         // Horizontal skip
         CALL    ADD_MATRICES
}
return;
}


template <class A>
bool MulMatrices(A *matA, A *matB, A *matC,
				 int rowsA, int colsA,
				 int rowsB, int colsB)
{
// Perform matrix addition: C = A + B using low-level code in the
// Un32_13 module
// On entry:
//       *matA and *matB are matrices to be added
//       *matC is matrix for sums 
//       rowsA is number of the rows in matrix A
//       colsA is number of columns in the matrix A
//       rowsB is number of the rows in matrix B
//       colsB is number of columns in the matrix B
// Requires:
//       All three matrices must be of the same dimensions
//       All three matrices must be of the same float
//       data type
// Asumes:
//       Matrix C dimensions are the product of the
//       columns of matrix B times the rows or matrix A
// Routine expects:
//      ESI --> first matrix (A)
//      EDI --> second matrix (B)
//      EBX --> storage area for addition matrix (C)
//      AH = number of rows in matrix A
//      AL = number of columns in matrix A
//      CH = number of rows in matrix B
//      CL = number of columns in matrix B
//      EDX = horizontal skil factor
// On exit:
//      returns true if matC[] = matA[] * matB[]
//      returns false if columns of matA[] not = rows 
//      of matB[]. If so, matC[] is undefined

int eSize = sizeof(matA[0]);

// Test for valid matrix sizes:
//     columns of matA[] = rows of matB[]
if(colsA != rowsB)
   return false;


_asm
{
         MOV     AH,BYTE PTR rowsA
         MOV     AL,BYTE PTR colsA
		 MOV     CH,BYTE PTR rowsB
		 MOV     CL,BYTE PTR colsB
		 MOV     ESI,matA          // Address to ESI
		 MOV     EDI,matB
		 MOV     EBX,matC
		 MOV     EDX,eSize         // Horizontal skip
         CALL    MUL_MATRICES
}
return true;
}


void GaussSolve(double *matS,
				 double *matR,
				 int rowsS, int colsS)
{
// Use Guass-Jordan elimination with pivoting to solve a system
// of linear equations stored in an augmented matrix.
// Uses GAUSS_SOLVE procedure in Un32_13.asm
// On entry:
//       *matS is the source matrix
//       *matR is the results matrix
//       rowsS is number of the rows in matrix S
//       colsS is number of columns in the matrix S
// Requires:
//       Both matrices must be of the same dimensions
//       Both matrices must be of double precision type
// Routine expects:
//      ESI --> first matrix (S)
//      EDI --> second matrix (R)
//      AH = number of rows in matrix S
//      AL = number of columns in matrix S
// On exit:
//      returns the solution to the linear system in a
//      matrix in row-echelon form

_asm
{
         MOV     AH,BYTE PTR rowsS
         MOV     AL,BYTE PTR colsS
		 MOV     ESI,matS          // Address to ESI
		 MOV     EDI,matR
         CALL    GAUSS_SOLVE
}
return;
}
