//************************************************************
//              Simple Neurons and Neural Units
//************************************************************
// Reference:   SOFTWARE SOLUTIONS
// Workspace  : Chapter15\Neural Units
// Entry point: Neural Units
// Other files: 
//
// Development: Visual C++ 6.0 Console Application
// Date:        December 23/2001
//
// Description:
//             A binary output neuron based on the
//             McCulloch-Pitts model.
//             Neural units for AND, OR, NOT, and XOR
//             operations 
// ***********************************************************
// ***********************************************************

#include <iostream.h>

// Prototypes
int NeuronB(int, int[], int[]);
int LogicAnd(int, int);
int LogicOr(int, int);
int LogicNot(int);
int LogicXor(int, int);

main()
{
    // Variables
    int xVal = 0;
	int yVal = 0;
	int output;
	int option = 1;
     

  while(option > 0)
  {

    // Menu for neuron testing
	cout << "TESTING NEURAL UNITS\n";
	cout << "   1 for AND\n";
	cout << "   2 for OR\n";
	cout << "   3 for NOT\n";
	cout << "   4 for XOR\n";
	cout << "   0 to end\n";
    cout << "Enter option number: ";
	cin >> option;
	cout << "\n\n";

      switch(option)
      {
	  case 1:
		// Testing AND
		cout << "Testing AND operation\n";
		cout << "Enter 1 or 0 for x: ";
		cin >> xVal;
		cout << "Enter 1 or 0 for y: ";
		cin >> yVal;
		// Call LogicAnd()
		output = LogicAnd(xVal, yVal);
		cout << "Result of AND is: " << output << "\n\n";
		break;
	   
	  case 2:
		// Testing OR
		cout << "Testing OR operation\n";
		cout << "Enter 1 or 0 for x: ";
		cin >> xVal;
		cout << "Enter 1 or 0 for y: ";
		cin >> yVal;
		// Call LogicOr()
		output = LogicOr(xVal, yVal);
		cout << "Result of OR is: " << output << "\n\n";
		break;

	  case 3:
		// Testing NOT
		cout << "Testing NOT operation\n";
		cout << "Enter 1 or 0 for x: ";
		cin >> xVal;
		// Call LogicNot()
		output = LogicNot(xVal);
		cout << "Result of NOT is: " << output << "\n\n";
		break;

      case 4:
        // Testing XOR
        cout << "Testing XOR operation\n";
		cout << "Enter 1 or 0 for x: ";
		cin >> xVal;
		cout << "Enter 1 or 0 for y: ";
		cin >> yVal;
        // Call LogicXor()
		output = LogicXor(xVal, yVal);
		cout << "Result of XOR is: " << output << "\n\n";
		break;        
	  }
  }
  return 0;
}

//**************************************************************
//                      LogicAnd()
//**************************************************************
int LogicAnd(int x, int y)
{
// Pre conditions:
// 1. Receives values of x and y in integer variables. Caller
//    is resposible for ensuring binary values only (0 or 1)
// 2. Function sets weights and bias line for performing logical
//    AND operation, using a neuron that follows the
//    McCulloch-Pitts model
// Post conditions:
//    Returns logical 0 or 1

  int weights[3];        // Array for weights
  int inputs[3];         // Array for inputs

  // Set weights
  weights[0] = -2;        // Bias line weight
  weights[1] = 1;
  weights[2] = 1;

  inputs[0] = 1;          // Bias line input
  inputs[1] = x;
  inputs[2] = y;
  // Call NeuronB()
  return (NeuronB(3, weights, inputs));
}

//**************************************************************
//                      LogicOr()
//**************************************************************
int LogicOr(int x, int y)
{
// Pre conditions:
// 1. Receives values of x and y in integer variables. Caller
//    is resposible for ensuring binary values only (0 or 1)
// 2. Function sets weights and bias line for performing logical
//    OR operation, using a neuron that follows the
//    McCulloch-Pitts model
// Post conditions:
//    Returns logical 0 or 1

  int weights[3];        // Array for weights
  int inputs[3];         // Array for inputs

  // Set weights
  weights[0] = -1;          // Bias line weight
  weights[1] = 1;
  weights[2] = 1;
  // Inputs
  inputs[0] = 1;            // Bias line input
  inputs[1] = x;
  inputs[2] = y;
  // Call NeuronB()
  return (NeuronB(3, weights, inputs));
}

//**************************************************************
//                      LogicNot()
//**************************************************************
int LogicNot(int x)
{
// Pre conditions:
// 1. Receives values of x in an integer variable. Caller
//    is resposible for ensuring binary value only (0 or 1)
// 2. Function sets weights and bias line for performing logical
//    NOT operation, using a neuron that follows the
//    McCulloch-Pitts model
// Post conditions:
//    Returns logical 0 or 1
	
	// Pre conditions:
// 1. Receives value of x in array values[0]
// 2. Set weights for logical NOT operation using the
//    McCulloch-Pitts neuron model
// 3. Calls Neuron() 
// Post conditions:
//    Returns logical 0 or 1

  int weights[3];        // Array for weights
  int inputs[3];         // Array for inputs

  // Set weights
  weights[0] = 0;        // Bias line weight
  weights[1] = -2;
  weights[2] = 1;

  inputs[0] = 1;         // Value for bias line 
  inputs[1] = x;
  inputs[2] = 0;
  // Call NeuronB()
  return (NeuronB(3, weights, inputs));
}


//**************************************************************
//                       LogicXor()
//**************************************************************

int LogicXor(int x, int y)
{
// This function performs a logical XOR operation using a
// 2-stage network based on the LogicAnd(), LogicOr() and
// LogicNot() functions, as follows:
//
//    x XOR y = [NOT(x) AND y] OR [NOT(y) AND x] 
//
// Stage 1:
//    calculate outputA = NOT(x) AND y
//              outputB = NOT(y) AND x
// Stage 2:
//    calculate result = outputA OR outputB

	int inx= x;           // Save inputs
	int iny = y;
	int tempx, tempy;     // Temporary storage
	int outputA, outputB;

	// Save data in temp[] array
	// Call LogicNot() and LogicAnd()operations, as follow:
	//   outputA = NOT(x) AND y
	//   outputB = NOT(y) AND x
	tempx = LogicNot(x);
	outputA = LogicAnd(tempx, y);   // NOTx AND y
    //
	tempy = LogicNot(y);
	outputB = LogicAnd(x, tempy);   // x AND NOTy
    // Perform OR
	return (LogicOr(outputA, outputB));

}	
//**************************************************************
//                 primtive function Neuron()
//**************************************************************
int NeuronB(int count,            // number of data elements
		   int weights[],        // array of weights
		   int values[])         // array of inputs
{
// Computes sum-of-products of = values[x] * weight[x]
// for numValues number of array elements
// the applies the threshold function:
//        if(sum < 0)
//           output 0
//        else
//           output 1

int result = 0;
// Routine logic as per McCulloch-Pitts model
for(int x = 0; x < count; x++)
	result += weights[x] * values[x];
if(result < 0)
	return 0;
else
	return 1;
}


