/**********************************************************************
  PROJECT: Keyboard Demo
  Source: kbddemo.cpp
  Chapter reference: 21
***********************************************************************
  Description: 
    Keyboard data processing
  Topics:
    1. Keystroke processing
	2. Determining key states
	3. Character codes processing
***********************************************************************/


#include <windows.h>  // standard Windows header

// Text strings 
char TextStr0[] = {"00000000 00000000 00000000 00000000 <-- lParam\n"
				  "-------- -------- -----------------\n"
				  "|||    |     |            |        \n"
				  "|||    |     |            |____ repeat count\n"
				  "|||    |     |_________________ OEM code\n"
				  "|||    |_______________ extended key flag\n"
				  "|||____________________ context code\n"
				  "||_____________________ previous key state\n"
				  "|______________________ transition state"};

char TextStr1[] = {"-------- -------- 00000000 00000000 <-- wParam\n"};

char TextStr2[] = {"CHARACTER TYPED:   "};

char TextStr3[] = {"Backspace key:   "};

char TextStr4[] = {"Shift state:      "};

char StrON[] = {"ON "};
char StrOFF[] = {"OFF"};


//************************
//       WinMain()
//************************
LRESULT CALLBACK WndProc (HWND, UINT, WPARAM, LPARAM) ;

int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance,
                    PSTR szCmdLine, int iCmdShow)
     {
     static char szAppName[] = "Demos" ;
     HWND        hwnd ;
     MSG         msg ;
     WNDCLASSEX  wndclass ;

     wndclass.cbSize        = sizeof (wndclass) ;
     wndclass.style         = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
     wndclass.lpfnWndProc   = WndProc ;
     wndclass.cbClsExtra    = 0 ;
     wndclass.cbWndExtra    = 0 ;
     wndclass.hInstance     = hInstance ;
     wndclass.hIcon         = LoadIcon (NULL, IDI_APPLICATION) ;
     wndclass.hCursor       = LoadCursor (NULL, IDC_ARROW) ;
     wndclass.hbrBackground = (HBRUSH) GetStockObject (WHITE_BRUSH) ;
     wndclass.lpszMenuName  = NULL ;
     wndclass.lpszClassName = szAppName ;
     wndclass.hIconSm       = LoadIcon (NULL, IDI_APPLICATION) ;

// Registering the wmdclass structure
     RegisterClassEx (&wndclass) ;

	 hwnd = CreateWindow (szAppName,
		  "Keyboard Demo",             // Window caption
           WS_OVERLAPPEDWINDOW,				// style
		   CW_USEDEFAULT,					// initial x
		   CW_USEDEFAULT,					// initial y
		   CW_USEDEFAULT,					// initial x size
		   CW_USEDEFAULT,					// initial y size
		   NULL,							// parent window handle
		   NULL,							// window menu handle
		   hInstance,						// program handle
		   NULL) ;							// creation parameters

     ShowWindow (hwnd, iCmdShow) ;
     UpdateWindow (hwnd) ;

//************************
//    Message loop
//************************
     while (GetMessage (&msg, NULL, 0, 0))
          {
          TranslateMessage (&msg) ;
          DispatchMessage (&msg) ;
          }
     return msg.wParam ;
     }


//************************
//    Windows Procedure
//************************
LRESULT CALLBACK WndProc (HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam)
{

 	 static int  cxChar, cyChar ;        // Character dimensions
	 static int  cxClient, cyClient;     // Client area parameters
     static HDC         hdc ;            // handle to private DC
	 unsigned long    keycode;           // storage for keystroke
	 unsigned long    keymask;           // bit mask
	 unsigned int     virtkey;           // virtual-key
	 int         i, j;                   // counters
	 char        aChar;                  // character code

	 // Structures
     PAINTSTRUCT ps;
	 TEXTMETRIC  tm;
	 RECT		 textRect;                // RECT-type
	 HFONT       hFont;

	 switch (iMsg)
          {

		 case WM_CREATE :
               hdc = GetDC (hwnd) ;


			   // Create a logical font
			   hFont = CreateFont (
					   30 * GetDeviceCaps (hdc, LOGPIXELSY) / 72, // height
					   0,                  // width
					   0,                  // escapement angle
					   0,                  // orientation angle
					   FW_NORMAL,          // weight
					   0,                  // not italics
					   0,                  // not underlined
					   0,                  // not strikedout
					   DEFAULT_CHARSET,    // character set
					   OUT_DEFAULT_PRECIS, // precision
					   CLIP_DEFAULT_PRECIS,// clipping precision
					   PROOF_QUALITY,      // quality
					   DEFAULT_PITCH | FF_DONTCARE, // pitch and family
					   "Courier");         // typeface name

			   SelectObject( hdc, hFont);  // Select font into DC

               // Text metrics are obtained after font is selected 
               GetTextMetrics (hdc, &tm) ;

               // Calculate and store character dimensions
               cxChar = tm.tmAveCharWidth ;
               cyChar = tm.tmHeight + tm.tmExternalLeading ;
			   // Store size of client area
			   cxClient = LOWORD(lParam);
			   cyClient = HIWORD(lParam);
               
			   return 0 ;

		 case WM_SIZE:
			   // Determine and store size of client area
			   cxClient = LOWORD(lParam);
			   cyClient = HIWORD(lParam);
			   return 0;
         
		  case WM_PAINT :
               BeginPaint (hwnd, &ps) ;
               // Initialize rectangle structure
			   SetRect (&textRect,                // address of structure
				        2 * cxChar,               // x for start
						cyChar,                   // y for start
						cxClient -(2 * cxChar),   // x for end
						cyClient);                // y for end 
               // Display multi-line text string 
               DrawText( hdc, TextStr0, -1, &textRect,
				         DT_LEFT | DT_WORDBREAK);

			   // Display second text string
			   SetRect (&textRect,                // address of structure
				        2 * cxChar,               // x for start
						13 * cyChar,              // y for start
						cxClient -(2 * cxChar),   // x for end
						cyClient);                // y for end 
               // Display text string 
               DrawText( hdc, TextStr1, -1, &textRect,
				         DT_LEFT | DT_WORDBREAK);

			   // Display third text string
			   SetRect (&textRect,                // address of structure
				        2 * cxChar,               // x for start
						15 * cyChar,              // y for start
						cxClient -(2 * cxChar),   // x for end
						cyClient);                // y for end 

               // Display text string 
               DrawText( hdc, TextStr2, -1, &textRect,
				         DT_LEFT | DT_WORDBREAK);

			   // Display text string
			   SetRect (&textRect,                // address of structure
				        2 * cxChar,               // x for start
						17 * cyChar,              // y for start
						cxClient -(2 * cxChar),   // x for end
						cyClient);                // y for end 

               // Display text string 
               DrawText( hdc, TextStr3, -1, &textRect,
				         DT_LEFT | DT_WORDBREAK);
			   
			   // Display last text string
			   SetRect (&textRect,                // address of structure
				        2 * cxChar,               // x for start
						19 * cyChar,              // y for start
						cxClient -(2 * cxChar),   // x for end
						cyClient);                // y for end 

               // Display text string 
               DrawText( hdc, TextStr4, -1, &textRect,
				         DT_LEFT | DT_WORDBREAK);

               EndPaint (hwnd, &ps);

               return 0 ;
 
          //****************************
          //     keystroke procssing
          //****************************
		  // Character code processing
		  case WM_CHAR:

			  aChar = (char) wParam;

			  // Test for control codes and replace with space
			  if (aChar < 0x30)
				  aChar = 0x20;

			  // Test for shift key pressed
			  if(GetKeyState (VK_SHIFT) < 0) {
                  i = 0;        // counter
				  j = 13;       // string offset
				  for(i = 0; i < 3; i++){
				      TextStr4[j] = StrON[i];
					  j++;
				  }
			  }
				  else {
                  i = 0;        // counter
				  j = 13;       // string offset
				  for(i = 0; i < 3; i++){
				      TextStr4[j] = StrOFF[i];
					  j++;
				  }
			  }
			  		  

			  TextStr2[17] = aChar;
			  return 0;

          // Scan code and keystroke data processing 
          // Display space if a system key
 		  case WM_SYSKEYDOWN:
			  TextStr2[17] = 0x20;

		  case WM_KEYDOWN:
			   // Store bits for lParam in TextStr0[]
			   keycode = lParam;    // get 32-bit keycode value
			   i = 0;               // counter for keystroke bits
			   j = 0;               // offset into string
			   keymask = 0x80000000;// bitmask

			   for (i = 0; i < 32; i++) {
				   // Test for separators and skip
				   if(i == 8 || i == 16 || i == 24) {
				       TextStr0[j] = 0x20;
					   j++;
				   }
				   // Test for 1 and 0 bits and display digits
				   if(keycode & keymask)
					   TextStr0[j] = '1';
				   else
					   TextStr0[j] = '0';
               keymask = keymask >> 1; 
               j++;
			   }
			   
			   // Store bits for wParam in TextStr1[]
			   // Since the virtual-key code is a 16-bit value only
			   // the 16-low order bits are examined in the code
			   keycode = wParam;    // get 32-bit keycode value
			   i = 0;               // counter for keystroke bits
			   j = 18;              // initial offset into string
			   keymask = 0x8000;    // bitmask

               // 16-bit loop
			   for (i = 0; i < 16; i++) {
				   // Test for separators and skip
				   if(i == 8) {
				       TextStr1[j] = 0x20;
					   j++;
				   }
				   // Test for 1 and 0 bits and display digits
				   if(keycode & keymask)
					   TextStr1[j] = '1';
				   else
					   TextStr1[j] = '0';
               keymask = keymask >> 1; 
               j++;
			   }

			   // Test for Backspace key pressed
               virtkey = (unsigned int) wParam;
			   if (virtkey == VK_BACK)
				   TextStr3[15] = 'Y';
			   else
				   TextStr3[15] = 'N';


			   // Force WM_PAINT message
			   InvalidateRect(NULL, NULL, TRUE);
               
               return 0;


          case WM_DESTROY :
               PostQuitMessage (0) ;
               return 0 ;
          }

     return DefWindowProc (hwnd, iMsg, wParam, lParam) ;
     }
