/***************************************************
  MEN_DEMO.CPP -- Menu demonstration program
  Topic: demonstration of a program main menu, of
         shortcut keys(accelerators), and of a
		 popup menu.
  Chapter reference: 22
****************************************************/

#include <windows.h>
#include "resource.h"

// Global variable for handle to program instance
HINSTANCE   pInstance;

// Predeclaration of the Windows Procedure
LRESULT CALLBACK WndProc (HWND, UINT, WPARAM, LPARAM) ;

//****************************
//         WinMain
//****************************
int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance,
                    PSTR szCmdLine, int iCmdShow) {

     static char szAppName[] = "Demo" ; // Application name
     HWND        hwnd ;
     MSG         msg ;
	 HACCEL      hAccel;                // Handle to accelerator

     // Defining a structure of type WNDCLASSEX
     WNDCLASSEX  wndclass ;
     wndclass.cbSize        = sizeof (wndclass) ;
     wndclass.style         = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
     wndclass.lpfnWndProc   = WndProc ;
     wndclass.cbClsExtra    = 0 ;
     wndclass.cbWndExtra    = 0 ;
     wndclass.hInstance     = hInstance ;
     wndclass.hIcon         = LoadIcon (NULL, IDI_APPLICATION) ;
     wndclass.hCursor       = LoadCursor (NULL, IDC_ARROW) ;
     wndclass.hbrBackground = (HBRUSH) GetStockObject
                              (WHITE_BRUSH) ;
	 // The next structure member defines the menu resource
     wndclass.lpszMenuName  = MAKEINTRESOURCE(IDR_MENU1);
     wndclass.lpszClassName = szAppName ; 
     wndclass.hIconSm       = LoadIcon (NULL, IDI_APPLICATION) ;

     // Store handle to instance
	 pInstance = hInstance;
     
	 // Registering the structure wmdclass
     RegisterClassEx (&wndclass) ;

     // CreateWindow()
     hwnd = CreateWindow (szAppName,
           "Menu Demo",         // window caption
            WS_OVERLAPPEDWINDOW,     // window style
            CW_USEDEFAULT,           // initial x position
            CW_USEDEFAULT,           // initial y position
            CW_USEDEFAULT,           // initial x size
            CW_USEDEFAULT,           // initial y size
            NULL,                    // parent window handle
            NULL,                    // window menu handle
            hInstance,               // program instance handle
            NULL) ;                  // creation parameters

     ShowWindow (hwnd, iCmdShow) ;
     UpdateWindow (hwnd) ;

	 // Load accelerators
	 hAccel = LoadAccelerators (hInstance,
		                        MAKEINTRESOURCE (IDR_ACCELERATOR1));

     // Message loop
     while (GetMessage (&msg, NULL, 0, 0))
          {
		 if (!TranslateAccelerator (hwnd, hAccel, &msg)) {
          TranslateMessage (&msg) ;
          DispatchMessage (&msg) ;
          }
	 }
     return msg.wParam ;
}
//****************************
//    Windows Procedure
//****************************
LRESULT CALLBACK WndProc (HWND hwnd, UINT iMsg, WPARAM wParam,
LPARAM lParam)
{
     static HDC  hdc ;
     PAINTSTRUCT ps ;
     RECT        rect ;
	 static  HMENU   pMenu;    // Handle to popup menu
	 POINT   aPoint;

     switch (iMsg)
         {
     // Windows message processing
	      case WM_CREATE:
			  hdc = GetDC(hwnd);    // Obtain handle to private DC

			  // Get handle to popup menu
			  pMenu = LoadMenu(pInstance,
				               (MAKEINTRESOURCE(IDR_MENU2)));
			  pMenu = GetSubMenu(pMenu, 0);
              
			  return 0;

          case WM_PAINT :
               BeginPaint (hwnd, &ps) ;
               GetClientRect (hwnd, &rect) ;
          // Initial display operations here
               return 0 ;

          //********************************
          //     menu command processing
          //********************************
		  case WM_COMMAND:
			  switch (LOWORD (wParam)) {

			  case ID_MYFILE_OPEN:
				  MessageBox(hwnd,
					         "File Open Requested",
							 "Menu Action",
							 MB_ICONEXCLAMATION | MB_OK);
				  break;

			  case ID_MYFILE_CLOSE:
				  MessageBox(hwnd,
					         "File Close Requested",
							 "Menu Action",
							 MB_ICONEXCLAMATION | MB_OK);
				  break;

			  case ID_MYFILE_SAVEAS:
				  MessageBox(hwnd,
					         "File Save As Requested",
							 "Menu Action",
							 MB_ICONEXCLAMATION | MB_OK);
				  break;

			  case ID_MYFILE_PRINT:
				  break;

			  case ID_MYEDIT_UNDO:
				  MessageBox(hwnd,
					         "Edit Undo Requested",
							 "Menu Action",
							 MB_ICONEXCLAMATION | MB_OK);
				  break;

			  case ID_MYEDIT_REPEAT:
				  MessageBox(hwnd,
					         "Edit Repeat Requested",
							 "Menu Action",
							 MB_ICONEXCLAMATION | MB_OK);
				  break;

			  case ID_MYEDIT_DELETE:
				  MessageBox(hwnd,
					         "Edit Delete Requested",
							 "Menu Action",
							 MB_ICONEXCLAMATION | MB_OK);
				  break;

			  case ID_MYHELP_TOPICS:
				  MessageBox(hwnd,
					         "Help Topics Requested",
							 "Menu Action",
							 MB_ICONEXCLAMATION | MB_OK);
				  break;

			  case ID_MYHELP_CONTEXT:
				  MessageBox(hwnd,
					         "Help Context Requested",
							 "Menu Action",
							 MB_ICONEXCLAMATION | MB_OK);
				  break;

              // Popup menu processing
     		  case ID_POPUP1_UNDO:
				  MessageBox(hwnd,
					         "Undo Requested",
							 "Popup Menu Action",
							 MB_ICONEXCLAMATION | MB_OK);
				  break;
     		  case ID_POPUP1_CUT:
				  MessageBox(hwnd,
					         "Cut Requested",
							 "Popup Menu Action",
							 MB_ICONEXCLAMATION | MB_OK);
				  break;
     		  case ID_POPUP1_COPY:
				  MessageBox(hwnd,
					         "Copy Requested",
							 "Popup Menu Action",
							 MB_ICONEXCLAMATION | MB_OK);
				  break;
     		  case ID_POPUP1_PASTE:
				  MessageBox(hwnd,
					         "Paste Requested",
							 "Popup Menu Action",
							 MB_ICONEXCLAMATION | MB_OK);
				  break;
     		  case ID_POPUP1_SELECTALL:
				  break;

		  }
		  return 0;

		  //****************************
		  //  popup menu processing
		  //****************************
		  case WM_RBUTTONDOWN:
			  // Get mouse coordinates
			  aPoint.x = LOWORD(lParam);
			  aPoint.y = HIWORD(lParam);
			  ClientToScreen(hwnd, &aPoint);
			  TrackPopupMenu(pMenu,
				             TPM_LEFTALIGN | TPM_TOPALIGN |\
							 TPM_LEFTBUTTON,
							 aPoint.x, aPoint.y,
							 0,
							 hwnd,
							 NULL);
			  return 0;

          // End of program execution
          case WM_DESTROY :
               PostQuitMessage (0) ;
               return 0 ;
          }
     return DefWindowProc (hwnd, iMsg, wParam, lParam) ;
}
