/**********************************************************************
  PROJECT: Filled Figure Demo
  Source: Fil_demo.cpp
  Chapter reference: 24
***********************************************************************
  Description:
    GDI filled figures operations

  Topics:
    1. Offsetting the hatch orgin
	2. Drawing geometrical figures
	3. Operations on rectangles
	4. Operations on regions
	5. Clipping
	6. Rendering open and closed paths
	7. Setting the miter limit
***********************************************************************/

#include <windows.h>
#include "resource.h"
#include <math.h>
#include <windowsx.h>


// Global variable for handle to program instance
HINSTANCE   pInstance;

// Function prototypes
LRESULT CALLBACK WndProc (HWND, UINT, WPARAM, LPARAM) ;
BOOL DrawRect (HDC, LPRECT);


//****************************
//         WinMain
//****************************
int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance,
                    PSTR szCmdLine, int iCmdShow)
{
     static char szAppName[] = "Demo" ; // Application name
     HWND        hwnd ;
     MSG         msg ;

     // Defining a structure of type WNDCLASSEX
     WNDCLASSEX  wndclass ;
     wndclass.cbSize        = sizeof (wndclass) ;
     wndclass.style         = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
     wndclass.lpfnWndProc   = WndProc ;
     wndclass.cbClsExtra    = 0 ;
     wndclass.cbWndExtra    = 0 ;
     wndclass.hInstance     = hInstance ;
     wndclass.hIcon         = LoadIcon (NULL, IDI_APPLICATION) ;
     wndclass.hCursor       = LoadCursor (NULL, IDC_ARROW) ;
     wndclass.hbrBackground = (HBRUSH) GetStockObject
                              (WHITE_BRUSH) ;
     wndclass.lpszMenuName  = MAKEINTRESOURCE(IDR_MENU1);
     wndclass.lpszClassName = szAppName ; 
     wndclass.hIconSm       = LoadIcon (NULL, IDI_APPLICATION) ;

     // Store handle to instance
	 pInstance = hInstance;
     
	 // Registering the structure wmdclass
     RegisterClassEx (&wndclass) ;

     // CreateWindow()
     hwnd = CreateWindow (szAppName,
           "Filled Figure Demo",         // window caption
            WS_OVERLAPPEDWINDOW,     // window style
            CW_USEDEFAULT,           // initial x position
            CW_USEDEFAULT,           // initial y position
            CW_USEDEFAULT,           // initial x size
            CW_USEDEFAULT,           // initial y size
            NULL,                    // parent window handle
            NULL,                    // window menu handle
            hInstance,               // program instance handle
            NULL) ;                  // creation parameters

     ShowWindow (hwnd, iCmdShow) ;
     UpdateWindow (hwnd) ;

     // Message loop
     while (GetMessage (&msg, NULL, 0, 0))
          {
          TranslateMessage (&msg) ;
          DispatchMessage (&msg) ;
          }
     return msg.wParam ;
}


//****************************
//    Windows Procedure
//****************************
LRESULT CALLBACK WndProc (HWND hwnd, UINT iMsg, WPARAM wParam,
LPARAM lParam)
{
     static int  cxChar, cxCaps, cyChar, cxClient, cyClient ;
	 
	 static int drawMode = 0;  // Display mode control

     static HDC     hdc ;
     PAINTSTRUCT	ps ;
     TEXTMETRIC		tm ;
	 static HPEN    blackPenSol, blackPenDot, redPen2, fatPen;
	 static HBRUSH  vertBrush1, diagBrush1, vertBrush2, stockBrush;
	 static HBRUSH  redSolBrush, bluSolBrush, grnSolBrush;
	 LOGBRUSH		brush1;
	 RECT			Rect1, Rect2, Rect3;
	 DWORD			aColor;
	 static	HRGN    rectRgn, ellipRgn, resultRgn, clipRgn;
	 int            combined;
	 static int		clipState = 0;  // Clipping ON/OFF control
	 static HMENU   hMenu;
	 static LOGBRUSH fatBrush;

	 POINT	pointsArray[10];   // Array for coordinates
	 static FLOAT	oldMiter;  // Storage for miter limit

     // Arrays of POINT structures for polygon vertices
	 POINT  polyPoints1[] = {
		 { 100, 100 }, // 1
         { 150, 100 }, // 2
		 { 150, 150 }, // 3
		 { 300, 150 }, // 4
		 { 300, 300 }, // 5
		 { 150, 300 }, // 6
		 { 150, 150 }, // 7
		 { 200, 150 }, // 8
		 { 200, 200 }, // 9
		 { 250, 200 }, // 10
		 { 250, 250 }, // 11
		 { 200, 250 }, // 12
		 { 200, 200 }, // 13
		 { 150, 200 }, // 14
		 { 150, 150 }, // 15
		 { 100, 150 }  // 16
	 } ;

	 POINT  polyPoints2[] = {
		 { 400, 100 }, // 1
         { 450, 100 }, // 2
		 { 450, 150 }, // 3
		 { 600, 150 }, // 4
		 { 600, 300 }, // 5
		 { 450, 300 }, // 6
		 { 450, 150 }, // 7
		 { 500, 150 }, // 8
		 { 500, 200 }, // 9
		 { 550, 200 }, // 10
		 { 550, 250 }, // 11
		 { 500, 250 }, // 12
		 { 500, 200 }, // 13
		 { 450, 200 }, // 14
		 { 450, 150 }, // 15
		 { 400, 150 }  // 16
	 } ;

	 POINT  polyPoints3[] = {
		 { 400, 150 }, // 1
         { 550, 450 }, // 2
		 { 250, 450 }  // 3
	 } ;


	 POINT  polyPoints4[] = {
		 { 150, 300 }, // 1
         { 100, 200 }, // 2
		 { 200, 100 }, // 3
		 { 350, 250 }, // 4
		 { 125, 350 }, // 5
		 { 150, 250 }, // 6
		 { 250, 400 }, // 7
		 { 250, 250 }, // 8
		 { 200, 175 }  // 9
	 } ;


	 POINT  polyPoints5[] = {
		 { 450, 300 }, // 1
         { 400, 200 }, // 2
		 { 500, 100 }, // 3
		 { 650, 250 }, // 4
		 { 425, 350 }, // 5
		 { 450, 250 }, // 6
		 { 550, 400 }, // 7
		 { 550, 250 }, // 8
		 { 500, 175 }  // 9
	 } ;

	 POINT  polyPoly1[] = {
		 { 150, 150 }, // 1  |
         { 300, 150 }, // 2  |
		 { 300, 300 }, // 3  |-- first polygon
		 { 150, 300 }, // 4  |
		 { 150, 150 }, // 5  |

		 { 200, 200 }, // 6  |
		 { 250, 200 }, // 7  |
		 { 250, 250 }, // 8  |-- second polygon
		 { 200, 250 }, // 9  |
		 { 200, 200 }, // 10 |

		 { 150, 150 }, // 11 |
         { 200, 150 }, // 12 |
		 { 200, 200 }, // 13 |-- third polygon
		 { 150, 200 }, // 14 |
		 { 150, 150 }, // 15 |

		 { 100, 100 }, // 11 |
         { 150, 100 }, // 12 |
		 { 150, 150 }, // 13 |-- fourth polygon
		 { 100, 150 }, // 14 |
		 { 100, 100 }, // 15 |

	 } ;

	 int  vertexArray[] = {
		 { 5 },
		 { 5 },
		 { 5 },
		 { 5 }
	 };

     switch (iMsg)
         {
     // Windows message processing
	      case WM_CREATE:
			  hdc = GetDC(hwnd);    // Obtain handle to private DC
              // Text metrics are obtained after font is selected 
              GetTextMetrics (hdc, &tm) ;
              cxChar = tm.tmAveCharWidth ;
              cxCaps = (tm.tmPitchAndFamily & 1 ? 3 : 2) * cxChar / 2 ;
              cyChar = tm.tmHeight + tm.tmExternalLeading ;
              // Create pens
			  blackPenSol = CreatePen (PS_SOLID, 2, 0);
			  blackPenDot = CreatePen (PS_DOT, 1, 0);
			  redPen2 = CreatePen(PS_SOLID, 2, RGB(0xff, 0x00, 0x00));
			  // Create brushes
			  brush1.lbStyle = BS_HATCHED;
              brush1.lbColor = RGB(0,0,0);
			  brush1.lbHatch = HS_VERTICAL;
			  vertBrush1 = CreateBrushIndirect (&brush1);
			  brush1.lbStyle = BS_HATCHED;
              brush1.lbColor = RGB(0xFF,0,0);
			  brush1.lbHatch = HS_BDIAGONAL;
			  diagBrush1 = CreateBrushIndirect (&brush1);			  
			  // Create a solid, red brush
			  brush1.lbStyle = BS_SOLID;
              brush1.lbColor = RGB(0xff,0x0,0x0);
			  brush1.lbHatch = NULL;
			  redSolBrush = CreateBrushIndirect (&brush1);
			  // Create a solid blue and green brushes
			  bluSolBrush = CreateSolidBrush (RGB(0x0, 0x0, 0xff));
			  grnSolBrush = CreateSolidBrush (RGB(0x0, 0xff, 0x0));
			  // Get menu handle
			  hMenu = GetMenu (hwnd);

              return 0;

          case WM_PAINT :
               BeginPaint (hwnd, &ps) ;

			   switch (drawMode)
			   {
			   case 1:
				  SelectPen (hdc, blackPenSol);
				  SelectBrush (hdc, vertBrush1);
                  Rectangle (hdc, 150, 150, 302, 300);

			  // Create a new hatched brush with offset origin
			      brush1.lbStyle = BS_HATCHED;
                  brush1.lbColor = RGB(0,0,0);
			      brush1.lbHatch = HS_VERTICAL;
                  SetBrushOrgEx (hdc, 5, 0, NULL);
		          vertBrush2 = CreateBrushIndirect (&brush1);
				  SelectBrush (hdc, vertBrush2); 
                  Rectangle (hdc, 350, 150, 502, 300);
			   break;

			   case 2:
				  SelectPen (hdc, blackPenSol);
				  SelectBrush (hdc, redSolBrush);
                  Rectangle (hdc, 100, 100, 180, 180);
                  SelectBrush (hdc, bluSolBrush);
				  Rectangle (hdc, 100, 200, 180, 280);
                  // Draw a rectangle with a stock brush
				  SelectBrush (hdc, GetStockObject (GRAY_BRUSH));
				  Rectangle (hdc, 100, 300, 180, 380);
			   break;

			   case 3:
                  // Clear the client area
				  InvalidateRect (hwnd, NULL, TRUE);
				  UpdateWindow (hwnd);
				  SelectPen (hdc, blackPenSol);
				  SelectBrush (hdc, bluSolBrush);
				  Ellipse (hdc, 100, 100, 300, 200);
			   break;

			   case 4:
				  SelectPen (hdc, blackPenSol);
				  SelectBrush (hdc, redSolBrush);
                  RoundRect (hdc, 100, 100, 350, 200, 40, 40);
 			   break;
				   
			   case 5:
				  // Select the solid black pen
				  // and draw bounding box
				  SelectPen(hdc, blackPenSol);
				  MoveToEx (hdc, 150, 150, NULL);
				  LineTo (hdc, 350, 150);
				  LineTo (hdc, 350, 250);
				  LineTo (hdc, 150, 250);
				  LineTo (hdc, 150, 150);

                  // Select the dot black pen and draw
				  // arc start/end lines
				  SelectPen (hdc, blackPenDot);
				  MoveToEx (hdc, 250, 260, NULL);
				  LineTo (hdc, 200, 140);

				  SelectPen (hdc, redPen2);
				  SelectBrush (hdc, GetStockObject (LTGRAY_BRUSH));

				  // Draw the arc with the red pen
				  Chord (hdc, 150, 150,    // Upper-left corner
					        350, 250,    // lower-right corner
							250, 260,    // start point
							200, 140);   // end point
			   break;

			   case 6:
				  // Select the solid black pen
				  // and draw bounding box
				  SelectPen(hdc, blackPenSol);
				  MoveToEx (hdc, 150, 150, NULL);
				  LineTo (hdc, 350, 150);
				  LineTo (hdc, 350, 250);
				  LineTo (hdc, 150, 250);
				  LineTo (hdc, 150, 150);

                  // Select the dot black pen and draw
				  // arc start/end lines
				  SelectObject (hdc, blackPenDot);
				  MoveToEx (hdc, 250, 200, NULL);
				  LineTo (hdc, 250, 260);
				  MoveToEx (hdc, 250, 200, NULL);
				  LineTo (hdc, 200, 140);

    			  SelectPen (hdc, redPen2);
				  SelectBrush (hdc, GetStockObject (LTGRAY_BRUSH));

				  // Draw the pie with the red pen
				  Pie (hdc, 150, 150,    // Upper-left corner
					        350, 250,    // lower-right corner
							250, 260,    // start point
							200, 140);   // end point
			   break;

			   case 7:
    			  SelectPen (hdc, redPen2);
				  SelectBrush (hdc, GetStockObject (LTGRAY_BRUSH));
				  // Draw the polygon using array data
				  SetPolyFillMode (hdc, ALTERNATE);
				  Polygon (hdc, polyPoints1, 16);
				  SetPolyFillMode (hdc, WINDING);
				  Polygon (hdc, polyPoints2, 16);
			   break;

			   case 8:
                  // Select pen and brush
    			  SelectPen (hdc, redPen2);
				  SelectBrush (hdc, GetStockObject (LTGRAY_BRUSH));
				  // Draw the polygon using array data
				  SetPolyFillMode (hdc, ALTERNATE);
				  PolyPolygon (hdc, polyPoly1, vertexArray, 4);
     		   break;

			   case 9:
    			  SelectPen (hdc, redPen2);
				  SelectBrush (hdc, GetStockObject (LTGRAY_BRUSH));
				  // Draw the polygon using array data
				  SetPolyFillMode (hdc, ALTERNATE);
				  Polygon (hdc, polyPoints4, 9);
				  SetPolyFillMode (hdc, WINDING);
				  Polygon (hdc, polyPoints5, 9);
			   break;

			   case 10:
                  // Clear the client area
				  InvalidateRect (hwnd, NULL, TRUE);
				  UpdateWindow (hwnd);
				  // Fill RECT structure variables
				  Rect1.left = 150;
				  Rect1.top = 150;
				  Rect1.right = 300;
				  Rect1.bottom = 300;
				  // Get a system color
				  aColor = GetSysColor (COLOR_MENU);
				  // Fill the rectangle
				  FillRect (hdc, &Rect1, (HBRUSH) (COLOR_ACTIVEBORDER+1));
     		   break;

			   case 11:
				  // Fill RECT structure variables
				  Rect1.left = 150;
				  Rect1.top = 150;
				  Rect1.right = 305;
				  Rect1.bottom = 305;
				  // Frame the rectangle
				  FrameRect (hdc, &Rect1, vertBrush1);
     		   break;

			   case 12:
				  // Fill RECT structure variables
				  Rect1.left = 150;
				  Rect1.top = 150;
				  Rect1.right = 300;
				  Rect1.bottom = 300;
				  // Fill the rectangle
				  FillRect (hdc, &Rect1, bluSolBrush);
				  // Determine rectangle for inversion
				  Rect1.left = 175;
				  Rect1.top = 175;
				  Rect1.right = 275;
				  Rect1.bottom = 275;
				  InvertRect (hdc, &Rect1);
     		   break;

			   case 13:
				  SelectPen (hdc, blackPenSol);
				  SelectBrush (hdc, GetStockObject (BLACK_BRUSH));
				  Rectangle (hdc, 100, 100, 200, 200);

				  // Fill RECT structure variables
				  Rect1.left = 150;
				  Rect1.top = 150;
				  Rect1.right = 300;
				  Rect1.bottom = 300;
				  // Fill the rectangle
				  DrawFocusRect (hdc, &Rect1);
    		   break;

			   case 14:
				  SelectPen (hdc, blackPenSol);
				  SelectBrush (hdc, GetStockObject (GRAY_BRUSH));
				  // Fill RECT structure variables
				  Rect1.left = 150;
				  Rect1.top = 150;
				  Rect1.right = 300;
				  Rect1.bottom = 300;
				  // Fill the rectangle
				  DrawRect (hdc, &Rect1);
				  // Translate rectangle
				  OffsetRect (&Rect1, 75, -75);
				  SelectBrush (hdc, GetStockObject (LTGRAY_BRUSH));
				  // Draw the new rectangle
				  DrawRect (hdc, &Rect1);
			   break;

			   case 15:
				  SelectPen (hdc, blackPenSol);
				  SelectBrush (hdc, GetStockObject (GRAY_BRUSH));
				  // Fill RECT structure variables
				  Rect1.left = 150;
				  Rect1.top = 150;
				  Rect1.right = 300;
				  Rect1.bottom = 300;
				  // Fill the rectangle
				  DrawRect (hdc, &Rect1);
				  // Inflate the rectangle
				  InflateRect (&Rect1, 100, -55);
				  SelectBrush (hdc, GetStockObject (LTGRAY_BRUSH));
				  // Draw the new rectangle
				  DrawRect (hdc, &Rect1);
			   break;
			   
			   case 16:
				  SelectPen (hdc, blackPenSol);
				  SelectBrush (hdc, GetStockObject (GRAY_BRUSH));
				  // Fill RECT structure variables
				  Rect1.left = 100;
				  Rect1.top = 100;
				  Rect1.right = 300;
				  Rect1.bottom = 300;
				  // Draw the Rectangle
				  DrawRect (hdc, &Rect1);
				  SelectBrush (hdc, GetStockObject (DKGRAY_BRUSH));
				  // Fill RECT structure variables
				  Rect2.left = 200;
				  Rect2.top = 200;
				  Rect2.right = 400;
				  Rect2.bottom = 400;
				  // Draw the Rectangle
				  DrawRect (hdc, &Rect2);
                  // Calculate the intersection rectangle
				  IntersectRect (&Rect3, &Rect1, &Rect2);
				  SelectBrush (hdc, GetStockObject (LTGRAY_BRUSH));
				  DrawRect (hdc, &Rect3);
			   break;

			   case 17:
				  SelectPen (hdc, blackPenSol);
				  SelectBrush (hdc, GetStockObject (GRAY_BRUSH));
				  // Fill RECT structure variables
				  Rect1.left = 100;
				  Rect1.top = 100;
				  Rect1.right = 300;
				  Rect1.bottom = 300;
				  // Draw the Rectangle
				  DrawRect (hdc, &Rect1);
				  SelectBrush (hdc, GetStockObject (DKGRAY_BRUSH));
				  // Fill RECT structure variables
				  Rect2.left = 200;
				  Rect2.top = 200;
				  Rect2.right = 400;
				  Rect2.bottom = 400;
				  // Draw the Rectangle
				  DrawRect (hdc, &Rect2);
                  // Calculate the intersection rectangle
				  UnionRect (&Rect3, &Rect1, &Rect2);
				  SelectBrush (hdc, GetStockObject (LTGRAY_BRUSH));
				  DrawFocusRect (hdc, &Rect3);
			   break;

			   case 18:
				  SelectPen (hdc, blackPenSol);
				  SelectBrush (hdc, GetStockObject (DKGRAY_BRUSH));
				  // Fill RECT structure variables
				  Rect1.left = 100;
				  Rect1.top = 100;
				  Rect1.right = 300;
				  Rect1.bottom = 300;
				  // Draw the Rectangle
				  DrawRect (hdc, &Rect1);
				  SelectBrush (hdc, GetStockObject (GRAY_BRUSH));
				  // Fill RECT structure variables
				  Rect2.left = 150;
				  Rect2.top = 100;
				  Rect2.right = 350;
				  Rect2.bottom = 350;
				  // Draw the Rectangle
				  DrawRect (hdc, &Rect2);
                  // Calculate the difference rectangle
				  SubtractRect (&Rect3, &Rect1, &Rect2);
				  // Move the rectangle
				  OffsetRect (&Rect3, 400, 0);
				  SelectBrush (hdc, GetStockObject (LTGRAY_BRUSH));
				  DrawRect (hdc, &Rect3);
			   break;

			   case 19:
                 // Create a rectangular region
				  rectRgn = CreateRectRgn (100, 100, 300, 200);
				  // Create an elliptical region
				  ellipRgn = CreateEllipticRgn (200, 100, 400, 200);
				  // Create a region for results
				  resultRgn = CreateRectRgn (0, 0, 0, 0);
				  // Combine regions and fill
				  combined = CombineRgn (resultRgn, rectRgn, ellipRgn, RGN_COPY);
				  FillRgn (hdc, resultRgn, redSolBrush);

				  // Draw dotted lines arround original regions
				  SelectBrush (hdc, GetStockObject (NULL_BRUSH));
				  SelectPen (hdc, blackPenDot);
				  Rectangle (hdc, 100, 100, 300, 200);
				  Ellipse (hdc, 200, 100, 400, 200);
			   break;

			   case 20:
                  // Create a rectangular region
				  rectRgn = CreateRectRgn (100, 100, 300, 200);
				  // Create an elliptical region
				  ellipRgn = CreateEllipticRgn (200, 100, 400, 200);
				  // Create a region for results
				  resultRgn = CreateRectRgn (0, 0, 0, 0);
				  // Combine regions and fill
				  combined = CombineRgn (resultRgn, rectRgn, ellipRgn, RGN_AND);
				  FillRgn (hdc, resultRgn, redSolBrush);

				  // Draw dotted lines arround original regions
				  SelectBrush (hdc, GetStockObject (NULL_BRUSH));
				  SelectPen (hdc, blackPenDot);
				  Rectangle (hdc, 100, 100, 300, 200);
				  Ellipse (hdc, 200, 100, 400, 200);
			   break;

			   case 21:
                  // Create a rectangular region
				  rectRgn = CreateRectRgn (100, 100, 300, 200);
				  // Create an elliptical region
				  ellipRgn = CreateEllipticRgn (200, 100, 400, 200);
				  // Create a region for results
				  resultRgn = CreateRectRgn (0, 0, 0, 0);
				  // Combine regions and fill
				  combined = CombineRgn (resultRgn, rectRgn, ellipRgn, RGN_DIFF);
				  FillRgn (hdc, resultRgn, redSolBrush);

				  // Draw dotted lines arround original regions
				  SelectBrush (hdc, GetStockObject (NULL_BRUSH));
				  SelectPen (hdc, blackPenDot);
				  Rectangle (hdc, 100, 100, 300, 200);
				  Ellipse (hdc, 200, 100, 400, 200);
			   break;

			   case 22:
                  // Create a rectangular region
				  rectRgn = CreateRectRgn (100, 100, 300, 200);
				  // Create an elliptical region
				  ellipRgn = CreateEllipticRgn (200, 100, 400, 200);
				  // Create a region for results
				  resultRgn = CreateRectRgn (0, 0, 0, 0);
				  // Combine regions and fill
				  combined = CombineRgn (resultRgn, rectRgn, ellipRgn, RGN_OR);
				  FillRgn (hdc, resultRgn, redSolBrush);

				  // Draw dotted lines arround original regions
				  SelectBrush (hdc, GetStockObject (NULL_BRUSH));
				  SelectPen (hdc, blackPenDot);
				  Rectangle (hdc, 100, 100, 300, 200);
				  Ellipse (hdc, 200, 100, 400, 200);
			   break;

			   case 23:
                  // Create a rectangular region
				  rectRgn = CreateRectRgn (100, 100, 300, 200);
				  // Create an elliptical region
				  ellipRgn = CreateEllipticRgn (200, 100, 400, 200);
				  // Create a region for results
				  resultRgn = CreateRectRgn (0, 0, 0, 0);
				  // Combine regions and fill
				  combined = CombineRgn (resultRgn, rectRgn, ellipRgn, RGN_XOR);
				  FillRgn (hdc, resultRgn, redSolBrush);

				  // Draw dotted lines arround original regions
				  SelectBrush (hdc, GetStockObject (NULL_BRUSH));
				  SelectPen (hdc, blackPenDot);
				  Rectangle (hdc, 100, 100, 300, 200);
				  Ellipse (hdc, 200, 100, 400, 200);
			   break;

			   case 24:
                  // Create a rectangular region
				  rectRgn = CreateRectRgn (100, 100, 300, 200);
				  // Create an elliptical region
				  ellipRgn = CreateEllipticRgn (200, 100, 400, 200);
				  // Create a region for results
				  resultRgn = CreateRectRgn (0, 0, 0, 0);
				  // Combine regions and fill
				  stockBrush = GetStockObject (BLACK_BRUSH);
				  combined = CombineRgn (resultRgn, rectRgn, ellipRgn, RGN_OR);
				  FrameRgn (hdc, resultRgn, stockBrush, 4, 2);
			   break;

			   case 25:
                  // Create a rectangular region
				  rectRgn = CreateRectRgn (100, 100, 300, 200);
				  // Create an elliptical region
				  ellipRgn = CreateEllipticRgn (200, 100, 400, 200);
				  // Create a region for results
				  resultRgn = CreateRectRgn (0, 0, 0, 0);
				  // Combine regions and fill
				  stockBrush = GetStockObject (BLACK_BRUSH);
				  combined = CombineRgn (resultRgn, rectRgn, ellipRgn, RGN_OR);
				  FillRgn (hdc, resultRgn, redSolBrush);
				  InvertRgn (hdc, resultRgn);
			   break;

			   case 26:
				  // Create an elliptical region
				  rectRgn = CreateRectRgn (0, 0, 300, cyClient + 100);
				  SetWindowRgn (hwnd, rectRgn, TRUE);
			   break;

			   case 27:
                  // Create a rectangular region
				  rectRgn = CreateRectRgn (100, 100, 300, 200);
				  FillRgn (hdc, rectRgn, redSolBrush);
				  //  Move the region
				  OffsetRgn (rectRgn, 50, 50);
				  InvertRgn (hdc, rectRgn);
			   break;

			   case 28:
                  // Display geometrical figures
				  SelectPen (hdc, blackPenSol);
				  SelectBrush (hdc, redSolBrush);
                  Rectangle (hdc, 60, 60, 300, 250);
				  SelectBrush (hdc, GetStockObject (LTGRAY_BRUSH));
				  SetPolyFillMode (hdc, ALTERNATE);
				  Polygon (hdc, polyPoints3, 3);
				  SelectBrush (hdc, bluSolBrush);
				  Ellipse (hdc, 180, 180, 380, 380);
				  Rect1.left = 90;
				  Rect1.top = 90;
				  Rect1.right = 430;
				  Rect1.bottom = 400;
				  // Fill the rectangle
				  DrawFocusRect (hdc, &Rect1);
     		   break;

			   case 29:
				  if (clipState == 0)
				  {
                    clipState = 1;
					CheckMenuItem (hMenu,
						           ID_OPERATIONS_CLIPON,
								   MF_CHECKED);
					// Create a region
					clipRgn = CreateRectRgn (90, 90, 430, 400);
					SelectClipRgn (hdc, clipRgn);
				  }
				  else
				  {
                    clipState = 0;
				    // Remove the clipping region
				    SelectClipRgn (hdc, NULL);
					CheckMenuItem (hMenu,
						           ID_OPERATIONS_CLIPON,
								   MF_UNCHECKED);
				  }
			   break;

			   case 30:
                  // Create a simple path
				  BeginPath (hdc);
				  MoveToEx (hdc, 100, 100, NULL);
				  LineTo (hdc, 300, 100);
				  LineTo (hdc, 300, 250);
				  LineTo (hdc, 100, 250);
				  CloseFigure (hdc);
				  MoveToEx (hdc, 120, 120, NULL);
				  LineTo (hdc, 280, 120);
				  LineTo (hdc, 280, 230);
				  LineTo (hdc, 120, 230);
				  CloseFigure (hdc);
				  EndPath (hdc);
				  SelectPen (hdc, blackPenSol);
				  SelectBrush (hdc, diagBrush1);
				  StrokeAndFillPath (hdc);
			   break;

			   case 31:
				  // Create a brush
				  fatBrush.lbStyle = BS_SOLID;
				  fatBrush.lbColor = RGB(0,0,0);
				  fatBrush.lbHatch = 0;
				  // Create a special pen
				  fatPen = ExtCreatePen (PS_GEOMETRIC | PS_SOLID |
					         PS_ENDCAP_ROUND | PS_JOIN_MITER,
							 15,
							 &fatBrush, 0, NULL);
				  SelectBrush (hdc, GetStockObject (LTGRAY_BRUSH));
                  // Create a simple path
				  BeginPath (hdc);
				  MoveToEx (hdc, 200, 100, NULL);
				  LineTo (hdc, 300, 180);
				  LineTo (hdc, 100, 180);
				  CloseFigure (hdc);
				  EndPath (hdc);
				  SelectPen (hdc, fatPen);
				  StrokePath (hdc);
                  // Create a simple path
				  BeginPath (hdc);
				  MoveToEx (hdc, 450, 100, NULL);
				  LineTo (hdc, 550, 180);
				  LineTo (hdc, 350, 180);
				  LineTo (hdc, 450, 100);
				  EndPath (hdc);
				  SelectPen (hdc, fatPen);
				  StrokePath (hdc);
			   break;

			   case 32:
    			  // Fill array of points for Bezier spline
				  pointsArray[0].x = 250; pointsArray[0].y = 200;
				  pointsArray[1].x = 150; pointsArray[1].y = 300;
				  pointsArray[2].x = 100; pointsArray[2].y = 250;
                  // Create a simple path
				  BeginPath (hdc);
				  MoveToEx (hdc, 100, 100, NULL);
				  LineTo (hdc, 300, 100);
				  LineTo (hdc, 300, 250);
				  PolyBezierTo (hdc, pointsArray, 3);
				  CloseFigure (hdc);
				  EndPath (hdc);
				  SelectPen (hdc, blackPenSol);
				  SelectBrush (hdc, GetStockObject (LTGRAY_BRUSH));
				  StrokeAndFillPath (hdc);

    			  // Fill array of points for Bezier spline
				  pointsArray[0].x = 500; pointsArray[0].y = 200;
				  pointsArray[1].x = 400; pointsArray[1].y = 300;
				  pointsArray[2].x = 350; pointsArray[2].y = 250;
                  // Create a simple path
				  BeginPath (hdc);
				  MoveToEx (hdc, 350, 100, NULL);
				  LineTo (hdc, 550, 100);
				  LineTo (hdc, 550, 250);
				  PolyBezierTo (hdc, pointsArray, 3);
				  CloseFigure (hdc);
				  EndPath (hdc);
				  FlattenPath (hdc);
				  SelectPen (hdc, blackPenSol);
				  SelectBrush (hdc, GetStockObject (LTGRAY_BRUSH));
				  StrokeAndFillPath (hdc);
			   break;

			   case 33:
				  // Create a special pen
				  fatPen = ExtCreatePen (PS_GEOMETRIC | PS_SOLID |
					         PS_ENDCAP_ROUND | PS_JOIN_MITER,
							 15,
							 &fatBrush, 0, NULL);
				  SelectBrush (hdc, GetStockObject (LTGRAY_BRUSH));
				  SelectPen (hdc, fatPen);
				  // Draw first angle
				  BeginPath (hdc);
				  MoveToEx (hdc, 100, 100, NULL);
				  LineTo (hdc, 250, 100);
				  LineTo (hdc, 150, 180);
				  EndPath (hdc);
				  StrokePath (hdc);
				  // Draw second angle
				  GetMiterLimit (hdc, &oldMiter);
				  SetMiterLimit (hdc, 2, &oldMiter);
				  BeginPath (hdc);
				  MoveToEx (hdc, 300, 100, NULL);
				  LineTo (hdc, 450, 100);
				  LineTo (hdc, 350, 180);
				  EndPath (hdc);
				  StrokePath (hdc);
				  SetMiterLimit (hdc, oldMiter, NULL);
			   break;

			   default:
			   break;

			   }
			   EndPaint (hwnd, &ps);
			   return 0 ;

          case WM_SIZE:
			  // Determine and store size of client area
			  cxClient = LOWORD(lParam);
			  cyClient = HIWORD(lParam);
			  return 0;

          //********************************
          //     menu command processing
          //********************************
		  case WM_COMMAND:
			  switch (LOWORD (wParam)) {

              //****************************
              //    reset hatch origin
			  //****************************
			  case ID_OPERATIONS_HATCHORG:
				  // Set display mode variable
				  drawMode = 1;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;

			  //*****************************************************
              //              figure drawing primities
			  //*****************************************************
			  case ID_OPERATIONS_SOLID_RECT:
				  // Set display mode variable
				  drawMode = 2;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;

			  case ID_OPERATIONS_ELLIPSE:
				  // Set display mode variable
				  drawMode = 3;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;

			  case ID_OPERATIONS_ROUND_RECT:
				  // Set display mode variable
				  drawMode = 4;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;

			  case ID_OPERATIONS_CHORD:
				  // Set display mode variable
				  drawMode = 5;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;

			  case ID_OPERATIONS_PIE:
				  // Set display mode variable
				  drawMode = 6;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;

			  case ID_OPERATIONS_POLYGON:
				  // Set display mode variable
				  drawMode = 7;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;

			  case ID_OPERATIONS_POLYPOLY:
				  // Set display mode variable
				  drawMode = 8;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;

			  case ID_OPERATIONS_POLYWIN:
				  // Set display mode variable
				  drawMode = 9;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;
			  
              //************************************************************
              //                  rectangle operations
			  //************************************************************
			  case ID_OPERATIONS_FILLRECT:
				  // Set display mode variable
				  drawMode = 10;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;

			  case ID_OPERATIONS_FRAME:
				  // Set display mode variable
				  drawMode = 11;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;

			  case ID_OPERATIONS_INVERT:
				  // Set display mode variable
				  drawMode = 12;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;

			  case ID_OPERATIONS_FOCUSRECT:
				  // Set display mode variable
				  drawMode = 13;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;

			  case ID_OPERATIONS_OFFSET:
 				  // Set display mode variable
				  drawMode = 14;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;

			  case ID_OPERATIONS_INFLATE:
 				  // Set display mode variable
				  drawMode = 15;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;

			  case ID_OPERATIONS_INTERSECT:
 				  // Set display mode variable
				  drawMode = 16;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;

			  case ID_OPERATIONS_UNION:
 				  // Set display mode variable
				  drawMode = 17;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;

			  case ID_OPERATIONS_SUBTRACT:
 				  // Set display mode variable
				  drawMode = 18;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;

              //************************************************************
              //                    region operations
			  //************************************************************
			  case ID_OPERATIONS_RGNCOPY:
 				  // Set display mode variable
				  drawMode = 19;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;
 
			  case ID_OPERATIONS_RGNAND:
 				  // Set display mode variable
				  drawMode = 20;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;

			  case ID_OPERATIONS_RGNDIFF:
 				  // Set display mode variable
				  drawMode = 21;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;

			  case ID_OPERATIONS_RGNOR:
 				  // Set display mode variable
				  drawMode = 22;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;

			  case ID_OPERATIONS_RGNXOR:
  				  // Set display mode variable
				  drawMode = 23;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;
				  
			  case ID_OPERATIONS_FRAMERGN:
  				  // Set display mode variable
				  drawMode = 24;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;
			
			  case ID_OPERATIONS_INVERTRGN:
  				  // Set display mode variable
				  drawMode = 25;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;

			  case ID_OPERATIONS_WINDOWRGN:
  				  // Set display mode variable
				  drawMode = 26;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;

			  case ID_OPERATIONS_OFFSETRGN:
  				  // Set display mode variable
				  drawMode = 27;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;

			  //**********************************************************
			  //                clipping operations
			  //**********************************************************
			  case ID_OPERATIONS_CLIPDRAW:
  				  // Set display mode variable
				  drawMode = 28;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;

			  case ID_OPERATIONS_CLIPON:
  				  // Set display mode variable
				  drawMode = 29;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;

			  //**********************************************************
			  //                path operations
			  //**********************************************************

			  case ID_OPERATIONS_PATHRENDER:
  				  // Set display mode variable
				  drawMode = 30;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;
		  
			  case ID_OPERATIONS_OPENCLOSE:
  				  // Set display mode variable
				  drawMode = 31;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;

			  
			  case ID_OPERATIONS_FLATTEN:
  				  // Set display mode variable
				  drawMode = 32;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;

			  //**********************************************************
			  //                   miter limit
			  //**********************************************************
			  case ID_OPERATIONS_MITERLIMIT:
  				  // Set display mode variable
				  drawMode = 33;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;
				  
				  
			  //**********************************************************
			  //                     program exit
			  //**********************************************************
			  case ID_OPERATIONS_EXIT:
				  SendMessage(hwnd, WM_DESTROY, 0, 0);
				  return 0;
		  }
		  return 0;

          case WM_DESTROY :
			   // Delete all graphics objects

               PostQuitMessage (0) ;
               return 0 ;
          }
     return DefWindowProc (hwnd, iMsg, wParam, lParam) ;
}

//*********************************
// Rectangle() function modified to
//       read RECT data
//*********************************
BOOL DrawRect (HDC hdc,
				  LPRECT aRect) {

  return (Rectangle (hdc, aRect->left,
	              aRect->top,
				  aRect->right,
				  aRect->bottom));
}