/**********************************************************************
  PROJECT: Bitmap Demo
  Source: BMP_Demo.cpp
  Chapter reference: 25
***********************************************************************
  Description:
    GDI operations on bitmaps

  Topics:
    1. Displaying a DIB
	2. Displaying a bitmap in BMP format
	3. Color and monochrome hand-coded bitmaps
    4. Operations on a memory-resident bitmap
	5. Using a blnak bitmap
	6. Using a DIB section
	7. Pattern brushing
	8. Pattern block transfer
	9. Stretching a bitmap
***********************************************************************/


#include <windows.h>
#include "resource.h"
#include <math.h>
#include <windowsx.h>
#include <assert.h>			// for assert() macro

// Global variable for handle to program instance
HINSTANCE   pInstance;


// Function prototypes
LRESULT CALLBACK WndProc (HWND, UINT, WPARAM, LPARAM);
void ShowBitmap (HDC, HBITMAP, int, int, DWORD);

//****************************
//         WinMain
//****************************
int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance,
                    PSTR szCmdLine, int iCmdShow)
{
     static char szAppName[] = "Demo" ; // Application name
     HWND        hwnd ;
     MSG         msg ;

     // Defining a structure of type WNDCLASSEX
     WNDCLASSEX  wndclass ;
     wndclass.cbSize        = sizeof (wndclass) ;
     wndclass.style         = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
     wndclass.lpfnWndProc   = WndProc ;
     wndclass.cbClsExtra    = 0 ;
     wndclass.cbWndExtra    = 0 ;
     wndclass.hInstance     = hInstance ;
     wndclass.hIcon         = LoadIcon (NULL, IDI_APPLICATION) ;
     wndclass.hCursor       = LoadCursor (NULL, IDC_ARROW) ;
     wndclass.hbrBackground = (HBRUSH) GetStockObject
                              (WHITE_BRUSH) ;
     wndclass.lpszMenuName  = MAKEINTRESOURCE(IDR_MENU1);
     wndclass.lpszClassName = szAppName ; 
     wndclass.hIconSm       = LoadIcon (NULL, IDI_APPLICATION) ;

     // Store handle to instance
	 pInstance = hInstance;
     
	 // Registering the structure wmdclass
     RegisterClassEx (&wndclass) ;


     // CreateWindow()
     hwnd = CreateWindow (szAppName,
           "Bitmap Demo",            // window caption
            WS_OVERLAPPEDWINDOW,     // window style
            CW_USEDEFAULT,           // initial x position
            CW_USEDEFAULT,           // initial y position
            CW_USEDEFAULT,           // initial x size
            CW_USEDEFAULT,           // initial y size
            NULL,                    // parent window handle
            NULL,                    // window menu handle
            hInstance,               // program instance handle
            NULL) ;                  // creation parameters

     ShowWindow (hwnd, iCmdShow) ;
     UpdateWindow (hwnd) ;

     // Message loop
     while (GetMessage (&msg, NULL, 0, 0))
          {
          TranslateMessage (&msg) ;
          DispatchMessage (&msg) ;
          }
     return msg.wParam ;
}
//****************************************************************
//                     Windows Procedure
//****************************************************************
LRESULT CALLBACK WndProc (HWND hwnd, UINT iMsg, WPARAM wParam,
LPARAM lParam)
{
     static int  cxChar, cxCaps, cyChar, cxClient, cyClient, mapMode ;

	 static int drawMode;		// Screen mode control

    static HDC         hdc, aMemDC;
    PAINTSTRUCT ps ;
    TEXTMETRIC  tm ;

	static HBITMAP      hBitmap, bmImage1, bmImage2, bmImage3;
	static HBITMAP      nebula, hstScope, wfpc2, bmBlank, bmDib;
	static HBITMAP      brushBM1, brushBM2, patBM1;
	static HGDIOBJ	    oldObject;
	static HBRUSH       patBrush;

	static BYTE hexBits[] = {0xc6, 0x7c, 0xc1, 0x81, 0xf0, 0x0, // 0
							 0xc6, 0x7c, 0xc1, 0x83, 0xf8, 0x0, // 1
							 0xc6, 0x60, 0xc1, 0x83, 0x18, 0x0, // 2		
							 0xc6, 0x60, 0xc1, 0x83, 0x18, 0x0, // 3
							 0xfe, 0x78, 0xc1, 0x83, 0x18, 0x0, // 4		
							 0xfe, 0x78, 0xc1, 0x83, 0x18, 0x0, // 5		
							 0xc6, 0x60, 0xc1, 0x83, 0x18, 0x0, // 6
							 0xc6, 0x60, 0xc1, 0x83, 0x18, 0x0, // 7		
							 0xc6, 0x7c, 0xf9, 0xf3, 0xf8, 0x0, // 8	
							 0xc6, 0x7c, 0xf9, 0xf1, 0xf0, 0x0};// 9

	POINT rectangle[5] = { 10, 10, 210, 10, 210, 110, 10, 110, 10, 10};  
	POINT rectsmall[5] = { 2, 2, 48, 2, 48, 48, 2, 48, 2, 2}; 


	static PBITMAPINFO  pDibInfo;    // pointer to BITMAPINFO structure
	static BYTE*        pDib;        // pointer to bitmap data

	int		i, j, k;
	BYTE	shade;

	// Bitmap parameters
	static  int		BMWidth = 200;
	static  int     BMScanLines = 255;

	static HPEN    blackPenSol, redPenSol;
	static HBRUSH  bluSolBrush;


    static HBITMAP    aBitmap = NULL;
	static BYTE*      lpBits = NULL;

	BITMAPINFOHEADER	bi;
	BITMAPINFOHEADER*	lpbi;
	HANDLE				hDIB;

	// Variables for bitmap stretching
	BITMAP			bm;
	RECT			rect;


	
     switch (iMsg)
         {
     // Windows message processing
	      case WM_CREATE:
			 hdc = GetDC(hwnd);    // Obtain handle to private DC
              // Text metrics are obtained after font is selected 
              GetTextMetrics (hdc, &tm) ;
              cxChar = tm.tmAveCharWidth ;
              cxCaps = (tm.tmPitchAndFamily & 1 ? 3 : 2) * cxChar / 2 ;
              cyChar = tm.tmHeight + tm.tmExternalLeading ;

			  // Load bitmaps into memory
			  nebula = LoadBitmap (pInstance, MAKEINTRESOURCE(IDB_BITMAP1));
			  hstScope = LoadBitmap (pInstance, MAKEINTRESOURCE(IDB_BITMAP2));
			  wfpc2 = LoadBitmap (pInstance, MAKEINTRESOURCE(IDB_BITMAP3));
			  bmImage2 = LoadBitmap (pInstance, MAKEINTRESOURCE(IDB_BITMAP4));
			  brushBM1 = LoadBitmap (pInstance, MAKEINTRESOURCE(IDB_BITMAP5));
			  brushBM2 = LoadBitmap (pInstance, MAKEINTRESOURCE(IDB_BITMAP6));
			  patBM1 = LoadBitmap (pInstance, MAKEINTRESOURCE(IDB_BITMAP7));
			  // Create a monochrome bitmap from array data
			  bmImage1 = CreateBitmap (37, 10, 1, 1, hexBits);
			  //                        |   |  |  |  -------
			  //                        |   |  |  |     |____ bitmap
			  //                        |   |  |  |____ bit per pixel
			  //                        |   |  |_______ color planes
			  //                        |   |__________ scan lines
			  //                        |______________ pixel width
 
        	return 0;

          case WM_PAINT :
               BeginPaint (hwnd, &ps) ;

               switch (drawMode)
			   {
			   case 1:
				  ShowBitmap (hdc, nebula, 10, 10, NOTSRCCOPY);
				  break;

			   case 2:
				  ShowBitmap (hdc, hstScope, 10, 10, NULL);
				  break;

			   case 3:
				  ShowBitmap (hdc, wfpc2, 10, 10, SRCCOPY);
				  break;

			   case 4:
				  ShowBitmap (hdc, bmImage1, 50, 50, NOTSRCCOPY);
				  break;

			   case 5:
				  ShowBitmap (hdc, bmImage2, 50, 50, SRCCOPY);
				  break;

			   case 6:
				  // Establish bitmap size
				  BMScanLines = 255;
				  BMWidth = 200;
				  // Create a bitmap in heap memory
				  // Allocate memory for bitmap data structure
				  pDibInfo = (PBITMAPINFO) GlobalAlloc(GMEM_FIXED, 
                          sizeof(BITMAPINFOHEADER));
				  pDibInfo->bmiHeader.biSize = (LONG) sizeof(BITMAPINFOHEADER);
				  pDibInfo->bmiHeader.biWidth = (LONG) BMWidth;      // pixel width (long)
				  pDibInfo->bmiHeader.biHeight = (LONG) BMScanLines; // pixel height (long)
				  pDibInfo->bmiHeader.biPlanes = 1;      // number of planes 
				  pDibInfo->bmiHeader.biBitCount = 32;   // bits per pixel

				  // Allocate memory for bitmap data
				  pDib = (BYTE*) GlobalAlloc(GMEM_FIXED, \
				             (BMScanLines * (BMWidth * 4) + 4));
 
				  // Fill the bitmap using 32-bit color data
				  // <--------- BMWidth * 4 -------->
				  // |
				  // | ... BMScanLines

					shade = 0;

					for (k = 0; k < BMScanLines; k++){    // Counts 255 scan lines
						for (i = 0; i < BMWidth; i++){ // Counter 50 pixels per line
							for(j = 0; j < 4; j++) {  //Counts 4 bytes per pixel 
								pDib[(k*(BMWidth*4))+(i*4)+0] = shade; // blue
								pDib[(k*(BMWidth*4))+(i*4)+1] = shade; // green
								pDib[(k*(BMWidth*4))+(i*4)+2] = 0;     // red
								pDib[(k*(BMWidth*4))+(i*4)+3] = 0;     // must be zero
								};
							};
						shade++;
					};
				  // Create the device dependant bitmap
				  bmImage3 = CreateDIBitmap ( hdc,
				        (LPBITMAPINFOHEADER) &pDibInfo->bmiHeader,
					     CBM_INIT,
						(LPSTR) pDib,
						(LPBITMAPINFO) pDibInfo,
						 DIB_RGB_COLORS );
                  // Display bitmap
				  SetDIBitsToDevice (hdc, 50, 50,
			           pDibInfo->bmiHeader.biWidth,
				       pDibInfo->bmiHeader.biHeight,
					   0, 0, 0,
					   pDibInfo->bmiHeader.biHeight,
					   pDib,
					   (BITMAPINFO FAR*) pDibInfo,
					  DIB_RGB_COLORS);

				  // Alternatively, the bitmap can be displayed by
				  // calling ShowBitmap()
				  ShowBitmap (hdc, bmImage3, 300, 50, NULL);
                  // Erase bitmap and free heap memory
				  DeleteObject(bmImage3);   // Delete the object
				  GlobalFree((HANDLE) pDibInfo);  // Free memory
				  GlobalFree((HANDLE) pDib);  
				  break;

			   case 7:
				  // Create a brush and a pen
				  blackPenSol = CreatePen (PS_SOLID, 2, 0);
				  redPenSol = CreatePen (PS_SOLID, 2, (RGB (0xff, 0x0, 0x0)));
				  bluSolBrush = CreateSolidBrush (RGB(0x0, 0x0, 0xff));
				  // Bitmap operations
 				  aMemDC = CreateCompatibleDC (NULL);   // Memory device handle
				  mapMode = GetMapMode (hdc);             // Obtain mapping mode
				  SetMapMode (aMemDC, mapMode);         // Set memory DC mapping mode
                  // Create the bitmap
				  bmBlank = CreateCompatibleBitmap (hdc, 300, 300);
				  oldObject = SelectObject (aMemDC, bmBlank);
				  PatBlt (aMemDC, 0, 0, 300, 300, WHITENESS);
				  SelectPen (aMemDC, blackPenSol);
				  SelectBrush (aMemDC, bluSolBrush);
				  Ellipse (aMemDC, 10, 10, 210, 110);
				  SelectPen (aMemDC, redPenSol);
				  Polyline (aMemDC, rectangle, 5);
				  BitBlt(hdc, 50, 50, 300, 300, aMemDC, 0, 0, SRCCOPY);
				  SelectObject (aMemDC, oldObject);
				  DeleteDC (aMemDC);
				  DeleteObject (bmBlank);
				  break;
				  
			   case 8:
				  // Bitmap dimensions
				  BMScanLines = 255;
				  BMWidth = 50;
				  // Initialize structures
				  bi.biSize = sizeof(BITMAPINFOHEADER);
				  bi.biWidth = BMWidth;
				  bi.biHeight = BMScanLines;
				  bi.biPlanes = 1;
				  bi.biBitCount = 32;
				  bi.biCompression = BI_RGB;
				  bi.biSizeImage = 0;
				  bi.biXPelsPerMeter = 0;
				  bi.biYPelsPerMeter = 0;
				  bi.biClrUsed = 0;
				  bi.biClrImportant = 0;
                  
				  hDIB = GlobalAlloc (GMEM_FIXED, sizeof(BITMAPINFOHEADER));
    			  lpbi = (BITMAPINFOHEADER*) hDIB;
				  *lpbi = bi;
   			      // Create the DIB
				  aBitmap = CreateDIBSection (hdc,
							(LPBITMAPINFO)lpbi,
							DIB_RGB_COLORS,
							(VOID**) &lpBits,
							NULL,
							(DWORD)0);
				  assert (aBitmap);
				  assert (lpBits);

				  // Fill the bitmap using 32-bit color data
				  // <--------- BMWidth * 4 -------->
				  // |
				  // | ... BMScanLines

					shade = 0;

					for (k = 0; k < BMScanLines; k++){    // Counts 255 scan lines
						for (i = 0; i < BMWidth; i++){ // Counter 50 pixels per line
							for(j = 0; j < 4; j++) {  //Counts 4 bytes per pixel 
								lpBits[(k*(BMWidth*4))+(i*4)+0] = shade; // blue
								lpBits[(k*(BMWidth*4))+(i*4)+1] = shade; // green
								lpBits[(k*(BMWidth*4))+(i*4)+2] = 0x0; // red
								lpBits[(k*(BMWidth*4))+(i*4)+3] = 0;    // must be zero
								};
							};
						shade++;
					};

                  // Draw on the bitmap
				  blackPenSol = CreatePen (PS_SOLID, 2, 0);
				  redPenSol = CreatePen (PS_SOLID, 2, (RGB (0xff, 0x0, 0x0)));
				  // Bitmap operations
 				  aMemDC = CreateCompatibleDC (NULL);   // Memory device handle
				  mapMode = GetMapMode (hdc);             // Obtain mapping mode
				  SetMapMode (aMemDC, mapMode);         // Set memory DC mapping mode
                  // Select the bitmap into the momory DC
				  oldObject = SelectObject (aMemDC, aBitmap);
				  SelectPen (aMemDC, blackPenSol);
				  Polyline (aMemDC, rectsmall, 5);
				  SelectPen (aMemDC, redPenSol);
				  Ellipse (aMemDC, 4, 4, 47, 47);
                  // Erase bitmap and free heap memory
				  SelectObject (aMemDC, oldObject);
				  DeleteDC (aMemDC);
				  DeleteObject (redPenSol);
				  DeleteObject (blackPenSol);

				  // Display bitmap
				  GdiFlush();
				  ShowBitmap (hdc, aBitmap, 50, 50, SRCCOPY);

                  // Erase bitmap and free heap memory
				  // Note: deleting a DIB section bitmap also frees
				  //       the allocated memory resources
				  DeleteObject (aBitmap);   // Delete the bitmap
				  GlobalFree (hDIB);
			      break;

			   case 9:
				  // Draw using 8-by-8 pixel pattern brush
				  patBrush = CreatePatternBrush (brushBM1);
				  SetStretchBltMode (hdc, HALFTONE);
				  SetBrushOrgEx (hdc, -4, -4, NULL);
				  SelectBrush (hdc, patBrush);
				  SelectPen (hdc, blackPenSol);
				  Rectangle (hdc, 20, 20, 200, 100);
				  DeleteObject (patBrush);
				  // Draw using 16-by-16 pixel pattern brush
				  patBrush = CreatePatternBrush (brushBM2);
				  SetBrushOrgEx (hdc, 0, 0, NULL);
				  SelectBrush (hdc, patBrush);
				  Rectangle (hdc, 20, 120, 200, 230);
				  DeleteObject (patBrush);
				  DeleteObject (blackPenSol);
				  break;

			   case 10:
				  // Draw using 8-by-8 pixel pattern brush
				  patBrush = CreatePatternBrush (patBM1);
				  // Bitmap operations
 				  aMemDC = CreateCompatibleDC (NULL);   // Memory device handle
				  mapMode = GetMapMode (hdc);           // Obtain mapping mode
				  SetMapMode (aMemDC, mapMode);         // Set memory DC mapping mode
                  // Create the bitmap
				  bmBlank = CreateCompatibleBitmap (hdc, 300, 300);
				  oldObject = SelectObject (aMemDC, bmBlank);
				  SelectBrush (aMemDC, patBrush);
				  PatBlt (aMemDC, 0, 0, 300, 300, PATCOPY);
				  BitBlt(hdc, 50, 50, 300, 300, aMemDC, 0, 0, SRCCOPY);
				  SelectObject (aMemDC, oldObject);
				  DeleteDC (aMemDC);
				  DeleteObject (bmBlank);
				  break;

			   case 11:
                  // Get bitmap dimensions
				  GetObject (hstScope, sizeof(BITMAP), &bm);
				  GetClientRect (hwnd, &rect);
				  aMemDC = CreateCompatibleDC (hdc);

				  SelectObject (aMemDC, hstScope);

				  SetStretchBltMode (hdc, STRETCH_DELETESCANS);
				  StretchBlt(hdc, 0, 0, rect.right, rect.bottom,
					         aMemDC, 0, 0, bm.bmWidth, bm.bmHeight,
							 SRCCOPY);
				  DeleteDC (aMemDC);
				  break;

			   default:
			      break;
			   }
			   EndPaint (hwnd, &ps);
     		   return 0 ;

          case WM_SIZE:
			  // Determine and store size of client area
			  cxClient = LOWORD(lParam);
			  cyClient = HIWORD(lParam);
			  return 0;

          //********************************
          //     menu command processing
          //********************************
		  case WM_COMMAND:
			  switch (LOWORD (wParam)) {

              //****************************
              //    menu operations
			  //****************************
			  case ID_OPERATIONS_NEBULA:
				  drawMode = 1;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;

			  case ID_OPERATIONS_HST:
				  drawMode = 2;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;

			  case ID_OPERATIONS_BMP:
				  drawMode = 3;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;

			  case ID_OPERATIONS_MONOBMP:
 				  drawMode = 4;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;

			  case ID_OPERATIONS_COLORBMP:
				  drawMode = 5;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;

			  case ID_OPERATIONS_MEMORYBMP:
				  drawMode = 6;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;

			  case ID_OPERATIONS_BLANKBMP:
       			  drawMode = 7;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;

			  case ID_OPERATIONS_DIBSECTION:
				  drawMode = 8;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;

			  case ID_OPERATIONS_PATBRUSH:
				  drawMode = 9;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;

			  case ID_OPERATIONS_PATBLOCK:
				  drawMode = 10;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;

			  case ID_OPERATIONS_STRETCH:
				  drawMode = 11;
				  InvalidateRect (hwnd, NULL, TRUE);
				  break;


			  }

		   return 0;


          // End of program execution
          case WM_DESTROY:
			   // Delete all graphics objects
               PostQuitMessage (0) ;
               return 0 ;
          }
     return DefWindowProc (hwnd, iMsg, wParam, lParam) ;
}

BITMAP     bm;             // BITMAP structure
HDC        memoryDc;       // Handle to memory DC
POINT      ptSize;         // POINT for DC
POINT      ptOrigin;       // POINT for memory DC
int        mapMode;        // Mapping mode

/*************************************
   Function to display a bitmap
**************************************
This procedure displays a bitmap at given screen coordinates.
It requires the following parameters:
     hdc  ........... handle to the device contect
	 hBitmap ........ handle to a memory-resident bitmap
	                  (usually obtained with LoadBitmap())
	 xStart/yStart .. client area coordinates for display
	 rop3 ........... ternary ROP code for Bitblt()
*/
void ShowBitmap (HDC hdc, HBITMAP hBitmap, int xStart, int yStart,\
				 DWORD rop3)
    {


	// Test for NULL ROP3 code
	if (rop3 == NULL)
		rop3 = SRCCOPY;

	memoryDc = CreateCompatibleDC (NULL);   // Memory device handle
	mapMode = GetMapMode (hdc);             // Obtain mapping mode
	SetMapMode (memoryDc, mapMode);         // Set memory DC mapping mode
		
	// Select bitmap into memory DC
	// Note: assert statement facilitates detecting invalid
	//       bitmaps during program development
    //assert (SelectBitmap (memoryDc, hBitmap));
	SelectBitmap (memoryDc, hBitmap);
	// Obtain bitmap dimensions
    GetObject (hBitmap, sizeof(BITMAP), (LPVOID) &bm);

	// Convert device units to logical units
	ptSize.x = bm.bmWidth;
	ptSize.y = bm.bmHeight;
	DPtoLP (hdc, &ptSize, 1);

	ptOrigin.x = 0;
	ptOrigin.y = 0;
	DPtoLP (memoryDc, &ptOrigin, 1);
	
	// Bitblt bitmap onto display memory
	BitBlt( hdc, xStart, yStart, ptSize.x, ptSize.y, memoryDc,
		    ptOrigin.x, ptOrigin.y, rop3);			

	// Delete memory DC with bitmap
	DeleteDC (memoryDc);
}
