//************************************************************
//              Support functions module
//************************************************************
// Reference:   SOFTWARE SOLUTIONS - Chapters 3/5
// This file:   support.cpp
// Header:      support.h
// Development: Visual C++ 6.0 Console Application
// Date:        January 2001
//
// Description:
//              Function module with support functions
// Note:
//              Some functions uses inline assembly
//***********************************************************
#include "stdafx.h"
#include <iostream.h>


void IdCpu(int *CPUtype, int *Cid)
{
_asm
{
// Function to determine the CPU in a PC
// Post:
// Parameter CPUtype as follows:
//           1 if CPU is 8086 or 8088
//           2 if CPU is 80286
//           3 if CPU is 80386
//           4 if CPU is 486
//           5 if CPU is Pentium
//
;***********************|
;   test for 8086/8088  |
;***********************|
; Bits 12 to 15 in the flag register are always set in the 8086
; and 8088 CPU
        PUSHF                   ; Flag register to stack
        POP     AX              ; Store flags in AX
        AND     AX,0FFFH        ; Clear bits 12 to 15
        PUSH    AX              ; AX to stack
        POPF                    ; and to flags register
        PUSHF                   ; Flags to stack
        POP     AX              ; and to AX for reading
        AND     AX,0F000H       ; Preserve bits 12 to 15
        CMP     AX,0F000H       ; Test for bits set
        JNE     TEST_286        ; Go if bits not set
; At this point processor is a 8086 or 8088
        MOV     AX,1            ; Return code
		MOV     DX,0
        JMP     ID_EXIT         ; Exit
;***********************|
;     test for 80286    |
;***********************|
; Bits 12 to 15 in the flag register are always clear in the Intel
; 80286 CPU
TEST_286:
        PUSHF                   ; Flag register to stack
        POP     BX              ; Store flags in BX
        OR      BX,0F000H       ; Make sure bit field is set
        PUSH    BX              ; To stack
        POPF                    ; And to flag register
        PUSHF                   ; Flags to stack
        POP     AX              ; And to AX
        AND     AX,0F000H       ; Clear all other bits
        JNZ     TEST_386        ; Go if bits not clear
; At this point processor is an 80286
        MOV     AX,2            ; Return code
		MOV     DX,0
        JMP     ID_EXIT         ; Exit
;***********************|
;     test for 80386    |
;***********************|
; Bit 18 of the E flags register was introduced in the 486 CPU
; This bit cannot be set in the 80386
TEST_386:
        PUSHFD                  ; 32-bits E flags to stack
        POP     EAX             ; Flags to EAX
        OR      EAX,40000H      ; Make sure bit 18 is set
        PUSH    EAX             ; New flags to stack
        POPFD                   ; An to E flags register
        PUSHFD                  ; Back to stack
        POP     EAX             ; And to EAX
        AND     EAX,40000H      ; Clear all except bit 18
        JNZ     TEST_486        ; Go if bit 18 is clear
; At this point processor is a 80386
        MOV     AX,3            ; Return code
		MOV     DX,0
        JMP     ID_EXIT         ; Exit
;***********************|
;     test for 486      |
;***********************|
; Bit 21 (ID flag) of the E flags register cannot be set in the
; 486
TEST_486:
        PUSHFD                  ; 32-bits E flags to stack
        POP     EAX             ; Flags to EAX
        OR      EAX,200000H     ; Make sure bit 21 is set
        PUSH    EAX             ; New flags to stack
        POPFD                   ; An to E flags register
        PUSHFD                  ; Back to stack
        POP     EAX             ; And to EAX
        AND     EAX,200000H     ; Clear all except bit 21
        JNZ     IS_PENTIUM      ; Go if bit 21 is clear
; At this point processor is a 486
        MOV     EAX,4            ; Return code
		MOV     EDX,0
        JMP     ID_EXIT         ; Exit
;***********************|
;  processor is PENTIUM |
;***********************|
IS_PENTIUM:
;***********************|
;    use CPUID          |
;***********************|
        MOV     EAX,0
		CPUID
		CMP     EBX,'uneG'
		JE      IS_INTEL
		MOV     EDX,0
        JMP     ID_EXIT
IS_INTEL:
		MOV     EAX,1
        CPUID
; At this point:
; EAX = contains the following information
;       xxxxxxxx xxxxxxxx xxxxxxxx xxxxxxxx <= bits
;       |---------------------|--| |---|--| 
;              RESERVED        |     |  |---[3-0] stepping ID
;                              |     |------[7-4] model number
;                              |-----------[11-8] family
; 
        MOV     EDI,Cid
        MOV     [EDI],EAX
        MOV     EAX,5           ; Pentium code
ID_EXIT:
		AND     EAX,0FH         ; Clear all other bits
        MOV     EDI,CPUtype
		MOV     [EDI],EAX
}
}
