//************************************************************
//         ASCII conversions and BCD12 arithmetic
//************************************************************
// Reference:   SOFTWARE SOLUTIONS - Chapters 2/4
// Project:     BCD Tests
// Entry point: BCD Test
// This file:   bcd12math.cpp
// Header:      bcd12math.h
// Development: Visual C++ 6.0 Console Application
// Date:        December 2000
//
// Description:
//              Function module with BCD to ASCII and ASCII
//              to BCD conversion functions. BCD arithmetic.
//              Includes auxiliary funtions. 
// Note:
//              Functions use inline assembly
//***********************************************************
#include <iostream.h>

// Library functions
extern "C" NEW_DIV_BCD12(); // BCD division routine

//********************************************************
// Prototypes for user-accesible functions in math2.cpp
//********************************************************
void AsciiToBcd12(char[], char[]);
void Bcd12ToAscStr(char[], char[]);
void DecodeSingle(float, int *, int *, int *);
void SignAddBcd12(char[], char[], char[]);
void SignSubBcd12(char[], char[], char[]);
void SignMulBcd12(char[], char[], char[]);
void SignDivBcd12(char[], char[], char[]);
void NewDivBcd12(char[], char[], char[]);

// Prototypes for auxiliary functions in this module
void BIN_TO_BCD();
void ASCII_TO_DX();
void BCD_INTO_REG();
void BCD_TO_ASCII();
void RESTORE_RS();
void ROUND_SIGF();
void UNPACK_MOVE();
void EXP_2_BIAS();
void TEST_4_ZERO();
void CLEAR_SIG_BUFS();
void REVERSE_SIGN(char[]);
void RESTORE_BCD();
void COMPARE_DI_DR();
void TRANSFER_CS_CX();
void DIGIT_TO_QUOT();
void TRIAL_QUOT();
void CLEAR_BUF();


//******************************************************************
//               global data used by functions
//******************************************************************
// Storage for pointers and registers
int OFFSET_ASC = 0;
int OFFSET_BCD = 0;
int reax;
int rebx;
int recx;
int redx;
int redi;
int resi;

// Biased exponents
unsigned short BIAS_X = 0;
unsigned short BIAS_Y = 0;
unsigned short BIAS_Z = 0;

// Work areas for significand operations
char SIG_L[36];        // Larger significand area
char SIG_S[36];        // Smaller significand area
char SIG_R[37];        // Significand of result

// Other controls
char LARGER = 0;        // 'x' or 'y', whichever is larger
char CASE_12 = 0;       // Switch for CASE 1 or CASE 2
char ZERO_VAL = 0;      // Number is zero
unsigned short expDigs  = 0; // exponent digits

// Partial products buffer for multiplication operations
char PART_PROD[37];     // Buffer for partial products

// Storage for entry pointers
int ENTRY_SI = 0;
int ENTRY_DI = 0;
int ENTRY_BX = 0;
// Other controls 
unsigned short ROWS = 0;
unsigned short DIGS = 0;
char BUF_FULL = 0;

// Buffers for BCD division
char DIVISOR[18];          // Buffer for divisor
char DIVIDEND[19];         // Buffer for dividend
char NEW_DIVIDEND[19];     // Buffer for trial dividend
char TRIAL_PROD[19];       // Buffer for trial product
char QUOTIENT[37];         // Buffer for quotient


void NewDivBcd12(char bcd1[], char bcd2[], char result[])
{
_asm
{
; Store entry variables
         MOV     ESI,bcd1
         MOV     EDI,bcd2
		 MOV     EBX,result
         CALL    NEW_DIV_BCD12
}
return;
}
//*****************************************************************
//*****************************************************************
//                       conversions
//*****************************************************************
//*****************************************************************

//*******************************************
//      ASCII string to BCD12 number
//*******************************************
void AsciiToBcd12( char asc[], char bcd[] )
{
_asm
{
// Conversion of an ASCII decimal number into BCD12 format
// On entry:
// Global variable ASC_STRING
//                 an ASCII decimal string, terminated in a NULL
//                 byte or control code
//             Note: the caller's buffer may contain the $ sign
//                   and commas as formatting symbols. The string
//                   terminator (00H) may be preceded by a CR
//                   (0DH) or LF (0AH) control codes. If the
//                   characters E or e are in the string, the
//                   procedure assumes that the string is in
//                   ASCII exponential form. The code assumes that
//                   exponential strings are normalized.
//
// Global variable BCD_VALUE
//                 12-byte memory area to holds the number encoded
//                 in BCD12 format
//
// Sample input strings:
//        123.223344
//       -0000.000002345
//        1.273E-22
//       -1.44556677 E230
//
//
// On exit:
//                 the caller's buffer with the input number
//                 reformatted in BCD12 format, as follows:
//
//          S s e e e e m . m m m m m m m m m m m m m m m m m m
//                        |
//                        |____ implicit decimal point
//
//    S = sign of number (1 BCD digit)
//    s = sign of exponent (1 BCD digit)
//    e = exponent (4 BCD digits)
//    m = normalized significand (18 BCD digits)
//    . = implicit decimal point between the first and second
//        significand digits
//
// CPU compatibility chart:
//                       | 8086/8088       | NO  |
//                       | 80286           | NO  |
//                       | 80386           | YES |
//                       | 80486/PENTIUM   | YES |
//
; Set buffer pointers
		MOV		ESI,asc
		MOV		EDI,bcd
;**********************|
; save buffer addresses|
;**********************|
        MOV     OFFSET_ASC,ESI
        MOV     OFFSET_BCD,EDI
;**********************|
; save input registers |
;**********************|
; Pointer registers are saved. All others are destroyed
        PUSH    ESI
        PUSH    EDI
;**********************|
;   clear buffer for   |
;     BCD12 number     |
;**********************|
; EDI --> 12-byte buffer to hold BCD12 number
        PUSH    EDI              ; Save output pointer
        MOV     ECX,12           ; Bytes to clear
CLEAR_12:
        MOV     BYTE PTR [EDI],0           ;
        INC     EDI              ; Bump buffer pointer
        LOOP    CLEAR_12
        POP     EDI              ; Restore pointer to start
        MOV     ESI,OFFSET_ASC        ; Source pointer to start
                                        ; of ASCII
;**********************|
;  get sign of number  |
;**********************|
; Find first significant character in the caller's string
; Skip leading blanks, $ and # sign, and other invalid symbols
 GET_NB:
        MOV     AL,BYTE PTR [ESI]        ; Get character
        CMP     AL,'+'          ; Test for + sign
        JAE     FIRST_NB        ; Go if equal or larger
        INC     ESI
        JMP     GET_NB
; first non-blank is a symbol or number larger than the + sign
FIRST_NB:
        CMP     AL,'+'          ; Test for +
        JE      SIGN_IS_PLUS
        CMP     AL,'-'          ; Test for -
        JE      SIGN_IS_MINUS
; First non-blank character is not a sign. Assume a digit
; Positive number is the default setting in BCD buffer
        JMP     TEST_4_EXP      ; Test for exponential form
;**********************|
;  number is negative  |
;**********************|
SIGN_IS_MINUS:
        MOV     [EDI],00010000B ; Set first BCD to 1
;**********************|
;  number is positive  |
;**********************|
; Encoding for positive number (0000B) is already in buffer
SIGN_IS_PLUS:
        INC     ESI              ; Bump ASCII digits pointer
;**********************|
;   test for E or e    |
;**********************|
; The caller's string may be in ASCII exponential form
; Any value less than 20H terminates the ASCII string
TEST_4_EXP:
        MOV     OFFSET_ASC,ESI
                                ; Save pointer to first digit
TEST_AGAIN:
        CMP     [ESI],'E'
        JE      STRING_ISEX     ; String in FP format
        CMP     [ESI],'e'
        JE      STRING_ISEX
        CMP     [ESI],' '       ; Less than 20H is end
                                ; of string
        JB      SKIP_ZERO       ; Not exponential form
        INC     ESI             ; Bump digits pointer
        JMP     TEST_AGAIN
;**********************|
;   input is in ASCII  |
;   exponential form   |
;**********************|
STRING_ISEX:
; Sample exponential form ASCII strings:
;            1.2345 E12
;           -1.2345E-12
;           +1.2345 e+12
; ESI --> letter e or E
; First test for negative exponent
        INC     ESI             ; Bump pointer past letter E
        PUSH    ESI             ; Save pointer to possible exponent
                                ; sign
; ESI --> string of exponent digits
; Skip sign, if there is one
FIRST_0PLUS:
        CMP     [ESI],'0'       ; First valid ASCII number
        JAE     OK_SI_PTR       ; ESI pointer is OK if 0 or >
        INC     ESI             ; Bump pointer if not
        JMP     FIRST_0PLUS     ; Continue
OK_SI_PTR:
		MOV		reax,EAX        ; Save context
		MOV		rebx,EBX
		MOV		recx,ECX
		MOV     redx,EDX
        CALL    ASCII_TO_DX     ; Convert to binary in DX
; DX now holds binary value of exponent
        POP     ESI             ; Restore point to possible sign
        CMP     [ESI],'-'       ; Negative exponent
        JNE     NON_NEG_EXP     ; Go if not negative
        JMP     NEG_EXPONENT    ; Routine for negative exponent
; Positive exponent code is already in place
NON_NEG_EXP:
        JMP     CONV_EXPONENT
;**********************|
; skip leading zeros   |
;**********************|
SKIP_ZERO:
; Leading zeros are not significant. skip them
; First reset pointer to start of string
        MOV     ESI,OFFSET_ASC        ; Pointer saved in memory
SKIP_1ZERO:
        CMP     BYTE PTR[ESI],'0'       ; Test for 0
        JNE     NOT_LEAD0
        INC     ESI              ; Bump to next digit
        JMP     SKIP_1ZERO
;**********************|
;  find decimal point  |
;    (if present)      |
;**********************|
;  DX is used to count the non-zero digits preceding the decimal
;  point. This value determines the number's exponent
NOT_LEAD0:
        MOV     DX,0            ; Clear integer digit counter
; Caller's string must end in a NULL byte (00H) or other control
; code < 20H
; Test for an integer zero input in the form 0000
        CMP     BYTE PTR[ESI],' '       ; Smaller than 20H is
                                ; end of input buffer
        JAE     GET_POINT       ; Continue if not end
        JMP     NUM_ISZERO      ; Number is zero
GET_POINT:
        MOV     AL,BYTE PTR [ESI]; Caller's ASCII digit
        CMP     AL,'.'          ; Test for decimal point
        JE      POINT_FOUND
        CMP     AL,','          ; Test for formatting comma
                                ; in the caller's string
        JE      SKIP_COMMA      ; Ignore comma
        CMP     AL,20H          ; String terminator
        JNB     LOOK_4_PT       ; Continue searching if not < 20H
        JMP     NO_POINT        ; No decimal point in string
LOOK_4_PT:
        INC     DX              ; Count one integer digit
SKIP_COMMA:
        INC     ESI             ; Bump integer digits counter
        JMP     GET_POINT
;**********************|
; caller's string is a |
;   decimal fractions  |
;**********************|
; If DX = 0 then the caller's string < 1 and the EXPONENT IS
;           NEGATIVE or the number is zero
; If DX = 1 then the caller's string is already normalized
;         and the EXPONENT = +0
; If DX > 1 then the caller's string is a number =< 10 and
;         the EXPONENT IS POSITIVE
POINT_FOUND:
        CMP     DX,1            ; Test for normalized input
        JNE     TEST_4_DX0
;**********************|
;       DX = 1         |
;**********************|
; ASCII input is in the form: d.dddd...ddd
; Number is between 1 and 10. Sign of exponent = 0000B
; 4 exponent digits = 0000H (these values are already in place)
        JMP     EXP_MOVED
TEST_4_DX0:
        CMP     DX,0            ; Test for negative exponent
        JE      DX_ISZERO       ; Go to DX = 0 processing
; If DX > 1 adjust and store
        DEC     DX
        JMP     CONV_EXPONENT
;**********************|
;       DX = 0         |
;**********************|
; Possible cases:
; non-zero -->   000.002345
; non-zero -->      .2345
;     zero -->     0.00
;     zero -->      .0
; Count decimal digits to normalized position, that is, following
; the first non-zero digit
; DX is used as counter
; At this point:
;                dddd.ddd...
;                    |____ ESI points to decimal point
DX_ISZERO:
        INC     ESI              ; Skip decimal point
        INC     DX               ; Count one digit
;**********************|
;     normalize        |
;**********************|
NORMALIZE:
        CMP     BYTE PTR [ESI],'0'       ; Test for zero
        JE      ZERO_DIGIT       ; Go if zero digit
        JB      NUM_ISZERO       ; exit if less than zero
; If not equal or less than zero it must be larger
        JMP     NEG_EXPONENT     ; Normalized position found
ZERO_DIGIT:
        INC     ESI              ; Bump pointer
        INC     DX               ; and digits counter
        JMP     NORMALIZE
;**********************|
;   ASCII number = 0   |
;**********************|
NUM_ISZERO:
; Make sure zero encoding is in BCD buffer
ZERO_EXIT:
        MOV     ECX,12           ; Bytes to clear
CLEAR_12_D:
        MOV     BYTE PTR [EDI],0 ;
        INC     EDI              ; Bump buffer pointer
        LOOP    CLEAR_12_D
        JMP     A2FP_EXIT        ; exit conversion
;**********************|
;  negative exponent   |
;**********************|
; DX holds binary value of negative exponent
; Convert to BCD and store
NEG_EXPONENT:
        MOV     EDI,OFFSET_BCD   ; EDI Saved in variable
; EDI --> start of BCD buffer
        MOV     AL,BYTE PTR [EDI]; Get BCD sign digit
        MOV     AH,00000001B     ; Code for minus exponent
        OR      AL,AH            ; OR into sign digit
        MOV     BYTE PTR [EDI],AL; Store exponent sign
        JMP     CONV_EXPONENT
;**********************|
;  ASCII is an integer |
;**********************|
NO_POINT:
; Exponent of integer is positive. This encoding is already
; in place. String has no decimal point.
; DX = exponent plus 1
        DEC     DX
;**********************|
;  convert and store   |
;      exponent        |
;**********************|
CONV_EXPONENT:
		MOV		reax,EAX         ; Save context
		MOV		rebx,EBX
		MOV		recx,ECX
		MOV     redx,EDX
        CALL    BIN_TO_BCD       ; Binary to BCD conversion
        MOV     EDI,OFFSET_BCD        ; Saved in variable
; EDI --> start of BCD buffer
        INC     EDI              ; Bump pointer
; EDI --> first BCD exponent digit
        MOV     BYTE PTR [EDI],DH  ; Store two high BCD digits
        INC     EDI              ; Bump pointer again
        MOV     BYTE PTR [EDI],DL; Store two low BCD digits
;**********************|
;  convert and store   |
;     significand      |
;**********************|
; Set pointer to first non-zero and non-sign digit in ASCII number
EXP_MOVED:
        MOV     ESI,OFFSET_ASC  ; Pointer to start of number
LOOK_4_1:
        MOV     AL, BYTE PTR [ESI]       ; Get ASCII digit
        CMP     AL,'1'          ; Test for non-zero and non-control
        JAE     FIRST_NON0      ; Go if 1 or greater
        INC     ESI             ; Bump pointer
        CMP     AL,20H          ; Test for control or NULL
        JAE     LOOK_4_1        ; Continue if not control or NULL
        JMP     ZERO_EXIT       ; Force zero and exit if no valid
                                ; significand digits
FIRST_NON0:
        MOV     ECX,18          ; Counter for 18 digits maximum
        MOV     EDI,OFFSET_BCD        ; Set pointer
; EDI --> start of BCD buffer
        ADD     EDI,3            ; Offset of significand area
; EDI --> BCD significand area in buffer
; ESI --> first non-zero digits in the ASCII string
; ASCII string must have at least one valid digit at this point
; Sample significand strings:
;                1.23456(0DH)
;                1.23456(00H)
;                -1.2345 E-12
;                123.45E345
; Operation:
;    The conversion routine takes the ASCII digits from the buffer,
;    converts them to BCD, and stores the digits in packed format.
;    The DH register is used to signal the high-order BCD digit
        MOV     DH,1            ; Code for next digit is high nibble
ASCII_MOVE:
        MOV     AL,BYTE PTR [ESI]        ; ASCII digit
; Test for invalid values
        CMP     AL,'0'          ; Lower limit
        JAE     VALID_1         ; Digit is 0 or larger
; If digit is less than 0, test for period, comma, and other symbols
; and skip. Exit routine if control code or NULL
        CMP     AL,20H          ; Control codes are < 20H
        JAE     NEXT_ASCII      ; Skip if space or symbol
        JMP     A2FP_EXIT       ; Exit with first non ASCII digit
VALID_1:
        CMP     AL,'9'          ; Upper limit of digit
        JBE     VALID_2         ; Go if OK
        JMP     A2FP_EXIT       ; Exit with first non ASCII digit
                                ; greater than 9
VALID_2:
; ASCII digit in AL is ready for storing in packed BCD format
        SUB     AL,30H          ; ASCII to BCD
        CMP     DH,1            ; Test high digit indicator
        JE      HIGH_BCD        ; Go if high
; Low order BCD digit
        MOV     AH,BYTE PTR [EDI]        ; Get previous high digit
        OR      AH,AL           ; OR low digit into AH
        MOV     BYTE PTR [EDI],AH        ; Store low and high
        INC     EDI             ; Bump destination pointer
        MOV     DH,1            ; Set switch for next digit is high
        JMP     NEXT_ASCII      ; Get next ASCII digit
; Routine to store high-order ASCII digit
; Binary value (in AL) must be shifted left 4 position
HIGH_BCD:
        SHL     AL,4            ; Shift AL digits
        MOV     BYTE PTR [EDI],AL        ; Store digit in buffer
        MOV     DH,0            ; Clear high digit switch
                                ; Next digit is low
NEXT_ASCII:
        INC     ESI              ; Bump ASCII digits pointer
        LOOP    ASCII_MOVE      ; Continue moving digits
;**********************|
;  restore and exit    |
;**********************|
A2FP_EXIT:
        POP     EDI              ; Restore caller's data registers
        POP     ESI
}
}

//********************************************
//     BCD12 number to ASCII string
//********************************************
void Bcd12ToAscStr( char bcd[], char asc[] )
{
// Local data
int OFFSET_ASCBUF = 0;
char invalidBcd[] = "INVALID";

_asm
{
// Convert a number in BCD12 format into an ASCII string
// On entry:
//         array bcd[] is a 12-byte BCD12 buffer holding number
//         array asc[] is a 31-byte ASCII buffer to hold string
// On exit:
//         array asc[] holds ASCII string
//
// Operation:
//         Numbers with exponent < 10 are converted to decimal
//         format. If exponent is > 10 the number is converted to
//         exponential format
//
// BCD12 format:
//          S s e e e e m . m m m m m m m m m m m m m m m m m m
//                        |
//                        |____ implicit decimal point
//
//    S = sign of number (1 BCD digit)
//    s = sign of exponent (1 BCD digit)
//    e = exponent (4 BCD digits)
//    m = normalized significand (18 BCD digits)
//        (in positive and negative numbers the first significand
//         digit must be non-zero)
//    . = implicit decimal point between the first and second
//        significand digits
//
//    Special encodings:
//        Ss = FFH indicate invalid
//
// CPU compatibility chart:
//                       | 8086/8088       | NO  |
//                       | 80286           | NO  |
//                       | 80386           | YES |
//                       | 80486/PENTIUM   | YES |
;

;**********************|
; save buffer addresses|
;**********************|
		MOV		EDI,asc
		MOV		ESI,bcd
        MOV     OFFSET_ASCBUF,EDI
; Save pointer registers
        PUSH    ESI
        PUSH    EDI
; Clear results buffer
        PUSH    EDI              ; Save pointer
        MOV     ECX,31           ; Bytes to clear
CLEAR_ASCBUF:
        MOV     BYTE PTR [EDI],0 ; Clear to zero
        INC     EDI              ; Bump pointer
        LOOP    CLEAR_ASCBUF
        POP     EDI              ; Restore pointer to start of buffer
;**********************|
;  get sign of number  |
;**********************|
        MOV     BYTE PTR [EDI],' '      ; In case of positive
        MOV     AL,[ESI]         ; First byte of BCD12 number
; Test for encoding tagged INVALID (Ss byte = FFH)
        CMP     AL,0FFH
		JNE     VALID_BCD
; At this point an INVALID has been detected
; Move text to output string
        LEA     ESI,invalidBcd   ; Text string "INVALID" 
		MOV     EDI,OFFSET_ASCBUF
        MOV     ECX,8            ; Includes NULL terminator
MAKE_INVALID:
        MOV     AL,BYTE PTR[ESI]
        MOV     BYTE PTR[EDI],AL
		INC     ESI
		INC     EDI
		LOOP    MAKE_INVALID
        JMP     EXIT_ASCSTR
; Untagged encoding
VALID_BCD:
        TEST    AL,00010000B     ; Bit 4 is negative number
        JZ      NNOT_NEG         ; Number is not negative
; Negative number. Set sign
        MOV     BYTE PTR [EDI],'-'       ; Sign to results buffer
NNOT_NEG:
        INC     EDI              ; Bump pointer
;**********************|
; get value of exponent|
;**********************|
        INC     ESI              ; Pointer to BCD12 exponent field
        INC     ESI              ; Point to low byte of BCD exponent
; Move BCD digits to AL-BH-BL-CH-CL for conversion routine
        MOV     AL,0             ; Highest digit is zero
        MOV     DL,BYTE PTR [ESI]       ; DL holds 2 packed digits
        MOV     DH,DL            ; Copy digits to DH
        AND     DL,00001111B     ; Erase high nibble in DL
        AND     DH,11110000B     ; and low in DH
        SHR     DH,4             ; Shift digit to low position
        DEC     ESI              ; Bump pointer to previous digit
        MOV     BL,BYTE PTR [ESI]        ; BL holds 2 packed digits
        MOV     BH,BL           ; Copy digits to BH
        AND     BL,00001111B    ; Erase high nibble in BL
        AND     BH,11110000B    ; and low in BH
        SHR     BH,4            ; Shift digit to low position
        MOV     CX,DX           ; Move digits to CX
		AND     ECX,0FFFFH      ; Clear high word
        PUSH    ESI             ; Save source
        PUSH    EDI             ; and destination pointers
		MOV		reax,EAX        ; Save context 
		MOV		rebx,EBX
		MOV		recx,ECX
		MOV     redx,EDX	
        CALL    BCD_INTO_REG    ; BCD to binary conversion
        POP     EDI             ; Restore pointers
        POP     ESI
; DX holds binary value of exponent
; Numbers with exponent = or < 10 are stored in conventional form
; Number with exponent > 10 are stored in exponential form
        CMP     DX,10           ; Test for limit
        JA      EXPO_FORM       ; Exponential form routine
        JMP     NORMAL_FORM     ; Conventional form if 10 or less
;**********************|
;   exponential form   |
;**********************|
; Buffer is formatted as follows for exponential form:
;       -1.2345 ...89 E-23
;       |||          ||_________ offset 21
;       |||          |__________ offset 20
;       |||_____________________ offset 2
;       ||______________________ offset 1
;       |_______________________ offset 0 (- sign or blank)
;**********************|
;  initialize buffer   |
;**********************|
EXPO_FORM:
        MOV     EDI,OFFSET_ASCBUF       ; Start of ASCII buffer
        MOV     BYTE PTR [EDI+2],'.'    ; Set period in buffer
        MOV     BYTE PTR [EDI+20],' '   ; Blank space separator
        MOV     BYTE PTR [EDI+21],'E'   ; Letter E
; Get and encode exponent sign
        ADD     EDI,22           ; Set pointer to sign position
        DEC     ESI              ; Set pointer to first BCD digit
        MOV     AL,BYTE PTR [ESI]        ; Get BCD sign digit
        TEST    AL,00000001B     ; Low bit is negative exponent
        JZ      POS_EXP_EXP      ; Exponent is positive
        MOV     BYTE PTR [EDI],'-'      ; Place negative sign
        INC     EDI
POS_EXP_EXP:
        INC     ESI              ; BCD pointer to exponent area
; EDI --> position for first exponent digit
        MOV     EBX,OFFSET_ASCBUF     ; EBX --> start of buffer
        ADD     EBX,3            ; Buffer offset 3 to 6 is used
                                 ; as a temporary storage
        MOV     AL,[ESI]         ; Get 2 packed BCD digits
        MOV     AH,AL            ; Copy digit in AH
        AND     AL,00001111B     ; Unpack low digit
        AND     AH,11110000B     ; and high digit
        SHR     AH,4             ; Move high digit to low nibble
; Convert BCD digits to ASCII
        ADD     AL,30H           ; BCD + 30H = ASCII
        ADD     AH,30H
        MOV     BYTE PTR [EBX],AH       ; Store ASCII
        INC     EBX              ; Bump pointer to temporary area
        MOV     BYTE PTR [EBX],AL
        INC     EBX              ; Bump pointer again
        INC     ESI              ; BCD pointer to next digit pair
; Get next two BCD digits and store
        MOV     AL,[ESI]         ; Get 2 packed BCD digits
        MOV     AH,AL            ; Copy digit in AH
        AND     AL,00001111B     ; Unpack low digit
        AND     AH,11110000B     ; and high digit
        SHR     AH,4             ; Move high digit to low nibble
; Convert BCD digits to ASCII
        ADD     AL,30H           ; BCD + 30H = ASCII
        ADD     AH,30H
        MOV     BYTE PTR [EBX],AH        ; Store ASCII
        INC     EBX              ; Bump pointer to temporary area
        MOV     BYTE PTR [EBX],AL
        SUB     EBX,3            ; EBX --> first digit in temporary
                                 ; area
; Move digits to final position skipping leading zeros
        MOV     ECX,4            ; For a count of 4
SKIP_LEAD0:
        MOV     AL,BYTE PTR [EBX]        ; Digit in temporary area
        CMP     AL,'0'           ; Test for leading zero
        JNE     NOT_LEAD00       ; Go if not a leading zero
        INC     EBX              ; Bump pointer
        LOOP    SKIP_LEAD0       ; Test 4 digits
; Execution drops if all exponent digits are 0
        MOV     BYTE PTR [EDI],'0'       ; Place one zero in buffer
        INC     EDI              ; Bump buffer pointer
        JMP     EXPONENT_RDY
NOT_LEAD00:
        MOV     AL,BYTE PTR [EBX]        ; Get digit
        MOV     BYTE PTR [EDI],AL        ; and store it
        INC     EBX              ; Bump pointers
        INC     EDI
        LOOP    NOT_LEAD00       ; Continue for remaining digits
;
EXPONENT_RDY:
        MOV     BYTE PTR [EDI],0FFH ; Special code for clean up
                                    ; routine
; Exponent is in buffer. Move significand using existing routine
        SUB     ESI,2            ; Set ESI to start of BCD buffer
        MOV     EDI,OFFSET_ASCBUF; Start of ASCII buffer
        INC     EDI              ; Skip sign position
        MOV     DX,0             ; Normalized significand
        JMP     POS_EX_FORM      ; Handle by positive exponent
;**********************|
;     normal form      |
;**********************|
NORMAL_FORM:
; DX holds binary value of exponent, which must be 10 or less
; Possible storage forms:
; If sign of exponent is negative:
;       0.1234
;       0.00001234
; If sign of exponent is positive
;       12.3456
;       12345.67
;       1.2345 (exponent = 0)
;
        DEC     ESI             ; Set pointer to first BCD digit
        MOV     AL,BYTE PTR [ESI]        ; Get BCD sign digit
        TEST    AL,00000001B    ; Low bit is negative exponent
        JZ      POS_EX_FORM     ; Exponent is positive
        JMP     NEG_EX_FORM     ; Go to negative exponent routine
;**********************|
; exponent is positive |
;**********************|
; DX holds the number of digits to move to the right the decimal
; point from the normalized position
; EDI --> ASCII buffer
POS_EX_FORM:
        PUSH    EDI             ; Save start of ASCII buffer
        AND     EDX,0FFFFH      ; Clear EDX high word
        ADD     EDI,EDX         ; Add exponent to start of buffer
        INC     EDI             ; Add one for normalized format
        MOV     BYTE PTR [EDI],'.'       ; Place point in buffer
        POP     EDI             ; Restore pointer to start position
MOVE_SIG:
        ADD     ESI,3           ; Set pointer to BCD significand
; ESI --> first significand digit in packed BCD format
;**********************|
;   unpack and move    |
;     BCD digits       |
;**********************|
        MOV     DX,9            ; Counter for 9 packed BCD digits
UNPACK_2:
        MOV     AL,[ESI]        ; Get 2 packed BCD digits
        MOV     AH,AL           ; Copy digit in AH
        AND     AL,00001111B    ; Unpack low digit
        AND     AH,11110000B    ; and high digit
        MOV     CL,4            ; For a count of 4
        SHR     AH,CL           ; Move high digit to low nibble
; Convert BCD digits to ASCII
        ADD     AL,30H          ; BCD + 30H = ASCII
        ADD     AH,30H
; Place two digits in buffer
        CMP     BYTE PTR [EDI],'.'   ; Test for period in buffer
        JNE     NO_PERIOD1      ; OK if no period
        INC     EDI             ; Bump pointer if on period
NO_PERIOD1:
        MOV     BYTE PTR [EDI],AH  ; Store high digit
        INC     EDI             ; Bump pointer to next digit
        CMP     BYTE PTR [EDI],'.' ; Test for period in buffer
        JNE     NO_PERIOD2      ; OK if no period
        INC     EDI             ; Bump pointer if on period
NO_PERIOD2:
        MOV     BYTE PTR [EDI],AL  ; Store low digit
        INC     EDI
        INC     ESI             ; Bump pointer to next BCD
        DEC     DX              ; Adjust digits counter
        JNZ     UNPACK_2
;**********************|
;       clean up       |
;**********************|
; Clear trailing zeros from buffer
; Example buffer digits:  1.23450000000000000(00)
;                         1234.00000000000000(00)
; Buffer after clean-up:  1.2345(00)
;                         1234(00)
;
        MOV     EDI,OFFSET_ASCBUF ; Entry offset from storage
        ADD     EDI,30            ; Point to last digit
; Replace up to 27 trailing zeros with NULL bytes
        MOV     ECX,29            ; Counter for maximum digits
REPLACE_27:
        CMP     BYTE PTR [EDI],0FFH     ; Special code for
                                        ; exponential form
        JNE     NT_EXP_FORM       ; Go if not
        MOV     BYTE PTR [EDI],00H       ; Clear special code
        JMP     EXIT_REP27        ; Do not fix trailing zeros
NT_EXP_FORM:
        CMP     BYTE PTR [EDI],'0'       ; Test for trailing 0
        JNE     TEST_27           ; Test for other values
        MOV     BYTE PTR [EDI],0  ; Replace zero with NULL
        JMP     NEXT_27
TEST_27:
        CMP     BYTE PTR [EDI],0  ; Test for NULL at end
        JNE     EXIT_REP27        ; Exit if not 0 or NULL
NEXT_27:
        DEC     EDI               ; Adjust pointer
        LOOP    REPLACE_27        ; Continue for 27 digits
EXIT_REP27:
; If the last character found is a period, then number is in the
; form 1234.0000..00.
        CMP     BYTE PTR [EDI],'.'       ; Test for period
        JNE     NO_PERIOD_FIX     ; Go if not
        MOV     BYTE PTR [EDI],0  ; Replace period with NULL
NO_PERIOD_FIX:
        JMP     EXIT_ASCSTR
;**********************|
; exponent is negative |
;**********************|
; Display forms for negative exponent
;       0.1234 (exponent = -1)
;       0.00001234 (exponent = -5)
; Operation:
;    The decimal point is first placed in the buffer
;    DX number of zeros are placed in the buffer, skipping the
;    decimal point
NEG_EX_FORM:
        INC     EDI              ; Leave blank for leading zero
        MOV     BYTE PTR [EDI],'.'   ; Place point in buffer
        DEC     EDI
PLACE_0S:
        CMP     BYTE PTR [EDI],'.'   ; Test for period in buffer
        JNE     NO_PERIOD3      ; No period, proceed
        INC     EDI             ; Bump buffer pointer to skip
                                ; period
NO_PERIOD3:
        MOV     BYTE PTR [EDI],'0'   ; Place 0 digit in buffer
        INC     EDI              ; Bump pointer to next digit
        DEC     DX              ; Decrement counter for zeros
        JNZ     PLACE_0S        ; Continue if not zero
        JMP     MOVE_SIG        ; Use existing code to move
                                ; significand
;**********************|
;    end of routine    |
;**********************|
EXIT_ASCSTR:
        POP     EDI              ; Restore pointer registers
        POP     ESI
}
}


//*************************************************
// decode elements of number in IEEE single format
//*************************************************    
void DecodeSingle(float single, int *sig, int *exp, int *signf )
{
_asm
{
// C++ function to separate the elements of a floating point number
// stored in memory in IEEE single format
// Pre:
//         Parameter single is an IEEE floating point single
// Post:
//         Parameter sig is the number's sign field:
//                    1 = negative sign
//                    0 = positive sign
//         Parameter exp is  8-bit exponent in bias 127 form
//         Parameter signf is 24-bit significand
//                    Implicit significand bit is set if exponent
//                    is not zero
//
// CPU compatibility chart:
//                       | 8086/8088       | NO  |
//                       | 80286           | NO  |
//                       | 80386           | YES |
//                       | 80486/PENTIUM   | YES |
//

         LEA     ESI,single
;***********************|
;       decode sign     |
;***********************|
        ADD     ESI,3           ; Set pointer to last byte
        MOV     AH,BYTE PTR[ESI]         ; Get byte into AH
        AND     AH,10000000B    ; Mask off exponent bits
        MOV     CL,7            ; Counter to shift 7 bits
        SHR     AH,CL           ; Shift sign bit to right-most bit
;***********************|
;     decode exponent   |
;***********************|
        MOV     AL,[ESI]        ; Get 7 high bits of exponent
        SHL     AL,1            ; Shift left one bit
        DEC     ESI             ; Pointer to byte at offset 2
        MOV     BH,[ESI]        ; Get byte into BH
        TEST    BH,10000000B    ; Test for exponent bit 8 set
        JZ      OK_EXPONENT     ; Exponent OK if bit not set
        OR      AL,00000001B    ; Set low exponent bit
;***********************|
;  decode significand   |
;***********************|
OK_EXPONENT:
; Test for a non-zero exponent so set implicit significand bit
        CMP     AL,0            ; Is exponent zero
        JE      IMPLICIT_0      ; Implicit bit is zero with a
                                ; zero exponent. Do not set
;***********************|
;     explicit 1 bit    |
;***********************|
        OR      BH,10000000B    ; Set BH high bit
IMPLICIT_0:
        DEC     ESI             ; Point to byte at offset 1
        MOV     BL,[ESI]        ; Load into BL
        DEC     ESI             ; Point to byte at offset 0
        MOV     CH,[ESI]        ; Load into CH
;***********************|
;  Data to variables    |
;***********************|
; Move IEEE format data from registers to variables
; Start with exponent, in AL
		MOV     EDX,EAX
        AND     EDX,0FFH        ; Preserve AL
        MOV     EDI,exp
		MOV     [EDI],EDX
; Store sign, in AH
		MOV     EDX,EAX
        AND     EDX,0FF00H      ; Preserve AH
		SHR     EDX,8
		MOV     EDI,sig
		MOV     [EDI],EDX
; Store significand (in BH, BL, and CH)
        MOV     EDX,EBX
		AND     EDX,0FFFFH      ; Clear high-order bits
        SHL     EDX,8           ; Make space for low order bits
		AND     ECX,0FF00H      ; Preserve CH
		SHR     ECX,8           ; Move to low order bits
		OR      EDX,ECX
        MOV     EDI, signf
		MOV     [EDI],EDX
// End of function
}
}

//*******************************************************************
//*******************************************************************
//                     BCD arithmetic functions
//*******************************************************************
//*******************************************************************

//************************************|
// signed floating point BCD addition |
//************************************|
void SignAddBcd12(char bcd1[], char bcd2[], char result[])
{
_asm
{
// Addition of two signed BCD numbers stored in BCD12 floating point
// format
// Operation:
//           z = x + y
//           where x, y, and z are signed, floating point numbers
// On entry:
//          bcd1[] =  first addend (element x in z = x + y)
//          bcd2[] =  second addend (element y in z = x + y)
//          result[] is 12-byte storage area for result in BCD12 format
// Note: the code assumes that the BCD12 numbers are in normalized
//       form, that is, that there are no leading zeros in the
//       significand
// On exit:
//       result = Sum (element z in z = x + y)
//
// This routine operates on two numbers encoded in BCD12 format (see
// Figure 2.10) as follows:
//
//          Ss e e e e m . m m m m m m m m m m m m m m m m m m
//                       |
//                       |____ implicit decimal point
//
//    S = sign of number (1 BCD digit)
//    s = sign of exponent (1 BCD digit)
//    e = exponent (4 BCD digits)
//    m = normalized significand (18 BCD digits)
//        (first significand digit must be non-zero)
//    . = implicit decimal point between the first and second
//        significand digits
//*******************************************************************
// BCD signed addition algorithm:
// CASE 1:
//    If x and y have the same sign, the absolute values are
//    added and the result has the common sign
// CASE 2:
//    If x and y have different signs, the smaller value is
//    subtracted from the larger value and the result has the sign
//    of the larger
//*******************************************************************
// Routine operations
// CASE 1 and 2:
// A. The input elements are tested for zero values. If one element
//    is zero the result is the value of the other element
// B. The packed siginificands in BCD12 format are unpacked and moved
//    into work buffers located in the code segment
// C. The upacked significands are aligned in the work buffers SIG_L
//    (for the significand of the number with the larger absolute
//     value) and SIG_S (for the significand of the number with the
//     smaller absolute value)
// CASE 1 (x and y have the same sign)
//     SIG_R will hold the significand of the sum
//     Addition operation:
//      SIG_L =    D D D . . . D D D 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
//  +   SIG_S =    0 0 0 0 d d d . . . d d d 0 0 0 0 0 0 0 0 0 0 0
//               -------------------------------------------------
//      SIG_R =  C s s s s s s s s s s s s s 0 0 0 0 0 0 0 0 0 0 0
//          leyend:
//             D = digits in the larger significand
//             d = digits in the smaller sginificand
//             s = digits in the sum
//             C = possible carry digit in the sum significand
//                 (in this case the exponent must be adjusted)
// CASE 2 (x and y have different signs)
//     SIG_R will hold the significand of the difference
//     Subtraction operation:
//      SIG_L =    D D D . . . D D D 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
//  -   SIG_S =    0 0 0 0 d d d . . . d d d 0 0 0 0 0 0 0 0 0 0 0
//               -------------------------------------------------
//      SIG_R =  B s s s s s s s s s s s s s 0 0 0 0 0 0 0 0 0 0 0
//          leyend:
//             B = possible borrow digit in the difference
// TRIVIAL CASE:
//    If the difference between exponents is larger than the number
//    of significand digits, then the aligned significands will be
//    as follows:
//      SIG_L =    D D D . . . D D D 0 0 0 0 0 0 0 0 0 0
//      SIG_S =    0 0 0 0 0 0 0 0 0 0 d d d . . . d d d
//              ----------------------------------------
//      SIG_R =  0 D D D . . . D D D 0 ? ? ? . . . ? ? ?
//    This means that the result (rounded to the format's significand
//    size) will equal the significand of the larger number. Therefore
//    the addition or subtraction operation would be trivial
// D. The exponent of the sum is the exponent of the element with
//    the larger absolute value, adjusted according to the operations
//    performed on the significands
//*******************************************************************
// Store entry variables and clear buffers
         MOV     ESI,bcd1
         MOV     EDI,bcd2
		 MOV     EBX,result
         MOV     rebx,EBX
		 MOV     resi,ESI
		 MOV     redi,EDI
	     CALL    CLEAR_SIG_BUFS         ; Local procedure
;**********************|
;    test x or y = 0   |
;**********************|
; If x or y = 0 then the sum is the non-zero element
        MOV     resi,ESI         ; Save pointer
        CALL    TEST_4_ZERO      ; Test routine
		CMP     ZERO_VAL,1
        JNE     X_NOT_ZERO       ; Go if no carry
; At this point x is zero
        MOV     ESI,EDI          ; Set ESI to point to y
MOVE_2Z:
        MOV     ECX,12           ; Count for digits to move
MOVE_12_2Z:
        MOV     AL,BYTE PTR[ESI] ; Get byte from source
        MOV     BYTE PTR[EBX],AL ; and store by BX pointer
        INC     ESI              ; Bump pointers
        INC     EBX
        LOOP    MOVE_12_2Z
        JMP     SIGN_ADD_EXIT
; End of processing
;**********************|
;    test for y = 0    |
;**********************|
X_NOT_ZERO:
        PUSH    ESI              ; Save x pointer
        MOV     ESI,EDI          ; Pointer to ESI
		MOV     resi,ESI
        CALL    TEST_4_ZERO      ; Test routine
        POP     ESI              ; Restore x pointer
		CMP     ZERO_VAL,1
        JNE     XY_NOT_ZERO      ; Continue if no carry
        JMP     MOVE_2Z          ; Exit routine if y = 0
;**********************|
;  neither x nor y = 0 |
;**********************|
; Convert BCD signed exponent to biased binary exponent
; Binary bias = 10000
; 0, 1, 2 ............... 10000 ...............9997, 9998, 9999
;  negative exponents  <-- zero -->  positive exponents
;**********************|
;  exponent to biased  |
;      10000 form      |
;**********************|
; First convert exponent of x
; ESI points to x
XY_NOT_ZERO:
		MOV     resi,ESI
        CALL    EXP_2_BIAS      ; Conversion procedure
; DX now holds binary exponent for x in bias 10,000 form
        MOV     WORD PTR BIAS_X,DX   ; Store in scratch pad area
; Obtain biased exponent for y
        XCHG    ESI,EDI         ; Save x and set y
    	MOV     resi,ESI
        CALL    EXP_2_BIAS      ; Conversion procedure
        MOV     WORD PTR BIAS_Y,DX      ; Store in scratch pad area
        XCHG    ESI,EDI         ; Restore x and y pointers
;**********************|
;   determine which    |
;   number is larger   |
;**********************|
; DX = biased exponent for y
        CMP     BIAS_X,DX       ; Compare exponent of x and y
        JA      X_BIGGER        ; x exponent is larger
        JB      Y_BIGGER        ; y exponent is larger
;**********************|
; compare significands |
;**********************|
; At this point x and y have the same exponent
; To determine which number is larger or if they are equal it is
; necessary to look at the digits in the significands
; ESI --> x and EDI --> y
        ADD     ESI,3           ; Pointer to significand digits
        ADD     EDI,3
        MOV     ECX,9           ; 9 packed BCD digits to compare
COMPARE_SIGS:
        MOV     AL,[ESI]        ; Get two significand digits of x
        CMP     AL,BYTE PTR[EDI]        ; Compare to two y digits
        JA      X_BIGGER        ; Go if x > y
        JB      Y_BIGGER        ; or if y > x
; At this point the compared digits are equal
        INC     ESI             ; Bump pointers
        INC     EDI
        LOOP    COMPARE_SIGS    ; Continue comparing digits
;**********************|
;       x = y          |
;**********************|
; At this point x and y are identical numbers
        MOV     BYTE PTR LARGER,'e'      ; Code for x = y
        JMP     X_EQUAL_Y       ; Go to entry point
;**********************|
; y exponent is larger |
;**********************|
Y_BIGGER:
        MOV     BYTE PTR LARGER,'y'      ; y is larger than x
; Move 18 packed BCD digits from BCD12 encoding to SIG_L buffer
X_EQUAL_Y:
        MOV     ESI,bcd2        ; Offet of y value pointer
        ADD     ESI,3           ; Mantissa digits in BCD12 encoding
        LEA     EDI,SIG_L       ; Offset of CS buffer
        MOV     WORD PTR expDigs,9     ; 9 packed digits to move
		MOV     resi,ESI
		MOV     redi,EDI
        CALL    UNPACK_MOVE     ; Unpack and move digits routine
        MOV     ESI,bcd1        ; Set pointer to smaller number
; Get exponents
        MOV     DX,WORD PTR BIAS_Y      ; Load y exponent
        MOV     WORD PTR BIAS_Z,DX      ; and store as exponent of sum
        MOV     AX,BIAS_X       ; Smaller exponent to AX
        JMP     X_OR_Y          ; Go to next routine
;**********************|
; x exponent is larger |
;**********************|
X_BIGGER:
        MOV     LARGER,'x'      ; x is larger than y
; Move 18 packed BCD digits from BCD12 encoding to SIG_L buffer
        MOV     ESI,bcd1        ; Offet of x value pointer
        ADD     ESI,3           ; Mantissa digits in BCD12 encoding
        LEA     EDI,SIG_L       ; Offset of buffer
        MOV     WORD PTR expDigs,9      ; 9 packed digits to move
		MOV     resi,ESI
		MOV     redi,EDI
        CALL    UNPACK_MOVE     ; Unpack and move digits routine
        MOV     ESI,bcd2        ; Set pointer to smaller number
; Get larger exponent
        MOV     DX,BIAS_X       ; Load x exponent
        MOV     BIAS_Z,DX       ; and store as exponent of sum
        MOV     AX,BIAS_Y       ; Smaller exponent to AX
X_OR_Y:
; At this point DX holds the larger exponent and AX the smaller one
; Test for a difference that is larger than the number of digits
; in the significand (18 digits in the BCD12 format)
        SUB     DX,AX           ; Larger minus smaller
        CMP     DX,18           ; Test for limit
        JBE     NOT_TRIVIAL     ; Not trivial case if <= 18
;**********************|
;  trivial operation   |
;**********************|
; If one element is too small to affect the sum, then the result is
; the larger element
; Assume that x > y
        MOV     ESI,bcd1        ; x value to ESI
        MOV     EBX,result
        CMP     LARGER,'x'      ; Test if x > y (assumed case)
        JNE     NOT_LARGER_X    ; Go is x is not larger
        JMP     MOVE_2Z         ; Exit routine for x > y
NOT_LARGER_X:
        MOV     ESI,bcd2        ; y entry value to ESI
        JMP     MOVE_2Z
;**********************|
;    offset smaller    |
;     significand      |
;**********************|
; DX holds the difference between exponents if this difference is 18
; or less. This value is used to offset the second significand
; ESI --> BCD12 encoding of smaller number
NOT_TRIVIAL:
        LEA     EDI,SIG_S       ; Buffer for smaller number
        AND     EDX,0FFFFH
        ADD     EDI,EDX         ; Add offset to destination
        ADD     ESI,3           ; Mantissa digits in BCD12 encoding
        MOV	    WORD PTR expDigs,9    ; 9 packed digits to move
		MOV     resi,ESI
		MOV     redi,EDI
        CALL    UNPACK_MOVE     ; Unpack and move digits routine
;**********************|
;    test for unequal  |
;        signs         |
;**********************|
; The sign of the number is the high nibble of the first BCD12 digit
        MOV     ESI,bcd1        ; Set pointers to entry values
        MOV     EDI,bcd2
; ESI --> x and EDI --> y
        MOV     AL,BYTE PTR [ESI]   ; x's first digit to AL
        MOV     AH,BYTE PTR [EDI]   ; y's first digit to AH
        AND     AL,11110000B    ; Clear x's low nibble
        AND     AH,11110000B    ; and y's low nibble
        CMP     AL,AH           ; Test for equal sign
        JE      SAME_SIGN
        JMP     DIFFERENT_SIGN
;************************************|
; CASE 1: x and y have the same sign |
;************************************|
SAME_SIGN:
        MOV     CASE_12,1       ; Case 1 processing switch
;**********************|
;   add significands   |
;**********************|
; At this point the significands of x and y are in the buffers SIG_L
; and SIG_S, ready for addition into SIG_R
        LEA     ESI,SIG_L       ; ESI --> to larger significand
        LEA     EDI,SIG_S       ; EDI --> to smaller significand
        LEA     EBX,SIG_R       ; EBX --> buffer for sum
        ADD     ESI,35          ; Point ESI to last digit in buffer
        ADD     EDI,35          ; also EDI
        ADD     EBX,36          ; and EBX
        MOV     ECX,36          ; 36 digits to add
        CLC                     ; Clear carry flag for first sum
ADD_36_DIGITS:
        MOV     AL,BYTE PTR [ESI] ; One digit to AL
        MOV     AH,BYTE PTR [EDI]
        ADC     AL,AH           ; Add other digit
        AAA                     ; Convert sum to unpacked BCD
        MOV     BYTE PTR [EBX],AL     ; Store digit in buffer
        DEC     ESI             ; Bump pointers to next digit
        DEC     EDI
        DEC     EBX
        LOOP    ADD_36_DIGITS   ; Loop for 36 digits
; At this point carry is set if sum overflows 36 digits
        JNC     NO_OVF          ; Go if no overflow
        MOV     BYTE PTR [EBX],01H    ; Enter 1 if overlfow
        INC     CS:BIAS_Z       ; Adjust exponent
NO_OVF:
; SIG_R buffer now holds sum of significands
        JMP     FORM_BCD        ; Construct BCD12 result
;************************************|
; CASE 2: x and y have different sign|
;************************************|
DIFFERENT_SIGN:
        MOV     CASE_12,2      ; Case 2 processing switch
;**********************|
;    discard x = y     |
;**********************|
; Signed addition of numbers with different signs and the same values
; result in zero
        CMP     LARGER,'e'      ; Encoding for x = y
        JNE     X_NOTEQ_Y       ; Continue if not equal
; Clear result buffer to encode zero and exit
OUTPUT_ZERO:
        MOV     ESI,result      ; Result buffer for BCD12 number
        MOV     ECX,12          ; 12 bytes in BCD12 format
ZERO_RESULT:
        MOV     BYTE PTR [ESI],00H       ; Place zero byte in buffer
        INC     ESI             ; Bump buffer pointer
        LOOP    ZERO_RESULT     ; Loop until buffer cleared
        JMP     SIGN_ADD_EXIT
;**********************|
; subtract significands|
;**********************|
; At this point the significands of x and y are in the buffers SIG_L
; and SIG_S. The number with the larger exponent is in SIG_L and the
; number with the smaller exponent in SIG_S
X_NOTEQ_Y:
        LEA     ESI,SIG_L       ; ESI --> to larger significand
        LEA     EDI,SIG_S       ; EDI --> to smaller significand
        LEA     EBX,SIG_R       ; EBX --> buffer for sum
        ADD     ESI,35          ; Point ESI to last digit in buffer
        ADD     EDI,35          ; also EDI
        ADD     EBX,36          ; and EBX
        MOV     ECX,36          ; 36 digits to subtract
        CLC                     ; Clear carry flag for first sum
SUB_36_DIGITS:
        MOV     AL,BYTE PTR [ESI]     ; x digit to AL
        SBB     AL,BYTE PTR [EDI]     ; Subtract y digit
        AAS                     ; Convert difference to unpacked BCD
        MOV     BYTE PTR [EBX],AL     ; Store digit in buffer
        DEC     ESI             ; Bump pointers to next digit
        DEC     EDI
        DEC     EBX
        LOOP    SUB_36_DIGITS   ; Loop for 36 digits
; At this point carry is set if there is a carry out of the last
; digit
        JNC     FORM_BCD        ; Go if no overflow
        MOV     BYTE PTR [EBX],01H    ; Enter 1 if overlfow
; SIG_R buffer now holds difference of significands
;*************************************|
;   Construct BCD12 number and exit   |
;*************************************|
FORM_BCD:
        CALL    ROUND_SIGF      ; Round off significand
        MOV     EDI,result      ; Set pointer to result buffer
                                ; for BCD12 output
        ADD     EDI,3           ; Point to significand area
;**********************|
;   store significand  |
;**********************|
; Significand of result is unpacked BCD form in the first 18 digits
; of the SIG_R buffer
; EDI --> significand field of the BCD12 result buffer
        MOV     ECX,9           ; Counter for 18 packed digits
        LEA     ESI,SIG_R       ; Pointer to unpacked digits
; Skip possible leftmost zero in buffer
        CMP     BYTE PTR [ESI],00H    ; Test for zero
        JNE     PACK_AND_MOVE   ; Go if no leading zero
        INC     ESI             ; Skip zero
; Other leading zeros in buffer require adjustment of exponent
LEADING_ZEROS:
        CMP     BYTE PTR [ESI],00H    ; Test for leading zero
        JNE     PACK_AND_MOVE   ; Go if no leading zero
        INC     ESI             ; Skip leading zero
        DEC     BIAS_Z          ; Adjust exponent
        LOOP    LEADING_ZEROS
; At this point all significand is zeros, therefore the number is
; zero
        JMP     OUTPUT_ZERO     ; Zero result exit
PACK_AND_MOVE:
        MOV     AH,BYTE PTR [ESI]     ; Get one digit
        INC     ESI             ; Bump pointer
        MOV     AL,BYTE PTR [ESI]     ; Get next digit
        INC     ESI             ; Bump pointer to next digit
        SHL     AH,4            ; Shift high digit to high nibble
        OR      AL,AH           ; Merge high nibble in AH with low
                                ; nibble in AH (pack digits)
        MOV     BYTE PTR [EDI],AL        ; Store packed digits
        INC     EDI
        LOOP    PACK_AND_MOVE
; All digits moved to the destination BCD12 buffer
        MOV     EDI,result      ; Set pointer to result buffer
                                ; sign area
; Test case switch to determine sign of result
        CMP     CASE_12,1       ; See if CASE 1
        JE      CASE1_SIGN      ; Go to sign routine for CASE 1
;**********************|
; CASE 2 sign routine  |
;**********************|
; In CASE 2 the sign of the result is the sign of the larger element
; The variable CS:LARGER holds 'x' if x > y
; Note that the possibily x = y is discarded previously
        CMP     LARGER,'x'      ; Test for x > y
        JE      CASE1_SIGN      ; Use sign of x for result
; At this point the sign of y must be used for the result
; This sign is obtained from the entry value of EDI
        MOV     ESI,bcd2        ; Set pointer the y element
        MOV     AL,BYTE PTR [ESI]       ; Get sign byte
        AND     AL,11110000B            ; Isolate sign of number
        MOV     BYTE PTR [EDI],AL       ; Store in result buffer
        JMP     EXPONENT_SIGN           ; Continue processing
;**********************|
; CASE 1 sign routine  |
;**********************|
; In CASE 1 both numbers have the same sign. The sign of the sum is
; the sign common to both elements
CASE1_SIGN:
        MOV     ESI,bcd1             ; Set pointer the x element
        MOV     AL,BYTE PTR [ESI]    ; Get sign byte
        AND     AL,11110000B         ; Isolate sign of number
        MOV     BYTE PTR [EDI],AL    ; Store in result buffer
;**********************|
; get sign of exponent |
;**********************|
EXPONENT_SIGN:
; Exponent of result is stored in BIAS_Z in bias 10000 form
; Values < 10000 indicate a negative exponent
        MOV     DX,WORD PTR BIAS_Z    ; Exponent of result to DX
        CMP     DX,10000        ; Compare with bias
        JAE     GET_POS_VALUE   ; Go if positive exponent
; Positive exponent is already encoded in result buffer
; Encode negative exponent (0001B) into BCD12 number
; EDI --> sign field
        MOV     AL,[EDI]        ; Get sign digit into AL
        MOV     AH,00000001B    ; Code for negative exponent
        OR      AL,AH           ; OR exponent sign into byte
        MOV     BYTE PTR [EDI],AL        ; Store in buffer
; Calculate absolute value of negative exponent
        MOV     AX,10000        ; Bias to AX
        SUB     AX,DX           ; Bias minus exponent
        XCHG    AX,DX           ; Move exponent to DX
        JMP     CALC_EXPONENT   ; Go to conversion routine
;**********************|
;  calculate value of  |
;  positive exponent   |
;**********************|
; DX = exponent in bias 10000 form
GET_POS_VALUE:
        MOV     AX,10000        ; Bias to AX
        SUB     DX,AX           ; Exponent minus bias
CALC_EXPONENT:
; DX = absolute value of exponent (in binary)
        MOV     AX,DX           ; For conversion routine
        PUSH    EDI             ; Save pointer for result
        MOV     redx,EDX        ; Register to variable
        CALL    BIN_TO_BCD      ; Conversion routine
        POP     EDI             ; Restore pointer
; DH and DL holds 4 packed BCD digits
        INC     EDI             ; Bump pointer to exponent filed
        MOV     BYTE PTR [EDI],DH     ; Store two digits
        INC     EDI
        MOV     BYTE PTR [EDI],DL     ; Store two other digits
;**********************************|
;           exit routine           |
;**********************************|
SIGN_ADD_EXIT:
        MOV     ESI,bcd1        ;Restore entry pointers
        MOV     EDI,bcd2
        MOV     EBX,result
}
}


//**************************************|
// signed floating point BCD subtraction|
//**************************************|
void SignSubBcd12(char bcd1[], char bcd2[], char result[])
{
// Subtraction of two signed BCD numbers stored in BCD12 floating
// point format
// Operation:
//           z = x - y
//           where x, y, and z are signed, floating point numbers
// On entry:
//          bcd1[] = Minuend (element x in z = x - y)
//          bcd2[] = Subtraend (element y in z = x - y)
//          result[] is 12-byte storage area for difference (element z)
// Note: the code assumes that the BCD12 numbers are in normalized
//       form, that is, that there are no leading zeros in the
//       significand
// On exit:
//       result[] holds difference (element z in z = x - y)
// Operation:
//   Processing is based on the algebraic priciple of changing the
//   sign of the subtraend and proceeding as in addition
//
// change subtraend sign
        REVERSE_SIGN(bcd2);   // Procedure to reverse the sign of
                              // a BCD12 number
// Call signed addition routine
        SignAddBcd12(bcd1, bcd2, result); 

// Restore sign of subtraend to entry value
        REVERSE_SIGN(bcd2);
        return;
}

//******************************************************************
//                  BCD multiplication function
//******************************************************************
void SignMulBcd12(char bcd1[], char bcd2[], char result[])
{
_asm
{
// Multiplication of two signed BCD numbers stored in BCD12 floating
// point format
// Operation:
//           z = x * y
//           where x, y, and z are signed, floating point numbers
// On entry:
//          bcd1[] = Muliplicand (element x in z = x * y)
//          bcd2[] = Multiplier (element y in z = x * y)
//          result[] = byte storage area for product in BCD12 format
// Note: the code assumes that the BCD12 numbers are in normalized
//       form, that is, that there are no leading zeros in the
//       significand
// On exit:
//       result[] = product (element z in z = x * y)
//*******************************************************************
// BCD signed multiplication algorithm
// A. If one of the factors is zero, the product is zero
//    If neither factor is zero then:
// B. If the factors have equal signs the product is positive
//    If the factors have unequal sings the product is negative
// C. The significand of the product is the product of the
//    significands of the factors
// D. The exponent of the product is the sum of the exponents of
//    the factors
//*******************************************************************
; Store entry variables and clear buffers
         MOV     ESI,bcd1
         MOV     EDI,bcd2
		 MOV     EBX,result
         MOV     rebx,EBX
		 MOV     resi,ESI
		 MOV     redi,EDI
; Store pointers
         MOV     ENTRY_SI,ESI
		 MOV     ENTRY_DI,EDI
		 MOV     ENTRY_BX,EBX

         CALL    CLEAR_SIG_BUFS ; Local procedure
;**********************|
;    test x or y = 0   |
;**********************|
; If x or y = 0 then the product is zero
        MOV     resi,ESI
        CALL    TEST_4_ZERO     ; Test routine
		CMP     ZERO_VAL,1
        JNE     NOT_ZERO_X      ; Go if no carry
; At this point x is zero
MOVE_00:
        MOV     ECX,12          ; Count for digits to move
MOVE_12_00:
        MOV     BYTE PTR[EBX],00; Store zero by BX pointer
        INC     EBX             ; Bump pointer
        LOOP    MOVE_12_00
		JMP     SIGN_MUL_EXIT
;**********************|
;    test for y = 0    |
;**********************|
NOT_ZERO_X:
        PUSH    ESI             ; Save x pointer
        MOV     resi,EDI        ; Passed to function
        CALL    TEST_4_ZERO     ; Test routine
        POP     ESI             ; Restore x pointer
		MOV     resi,ESI
		CMP     ZERO_VAL,1
        JNE     NOT_ZERO_XY     ; Continue if no carry
        JMP     MOVE_00         ; Exit routine if y = 0
;**********************|
;  neither x nor y = 0 |
;     add exponents    |
;**********************|
NOT_ZERO_XY:
        MOV     resi,ESI
        CALL    EXP_2_BIAS      ; Conversion procedure
; DX now holds binary exponent for x in bias 10,000 form
        MOV     WORD PTR BIAS_X,DX    ; Store in scratch pad area
; Obtain biased exponent for y
        XCHG    ESI,EDI         ; Save x and set y
		MOV     resi,ESI
        CALL    EXP_2_BIAS      ; Conversion procedure
        MOV     WORD PTR BIAS_Y,DX    ; Store in scratch pad area
        XCHG    ESI,EDI         ; Restore x and y pointers
;**********************|
;    add exponents     |
;**********************|
        MOV     AX,WORD PTR BIAS_X      ; Exponent of x to AX
        ADD     AX,DX           ; Add exponents of x and y
        SUB     AX,10000        ; Subtract bias from sum
                                ; to re-scale exponent
        INC     AX              ; Normalize siginifcand by
                                ; incrementing the exponent
        MOV     WORD PTR BIAS_Z,AX    ; Store sum in scratch pad area
;**********************|
;    unpack and move   |
;      significands    |
;**********************|
; Move 18 packed BCD digits from BCD12 encoding to SIG_L buffer
        MOV     ESI,bcd1        ; Set pointer to x
        ADD     ESI,3           ; Mantissa digits in BCD12 encoding
        LEA     EDI,SIG_L       ; Offset of CS buffer
        MOV     WORD PTR expDigs,9     ; 9 packed digits to move
		MOV     resi,ESI
		MOV     redi,EDI
        CALL    UNPACK_MOVE     ; Unpack and move digits routine
; Move significand of y to SIG_S buffer
        MOV     ESI,ENTRY_DI    ; Set pointer to y
        ADD     ESI,3           ; Mantissa digits in BCD12 encoding
        LEA     EDI,SIG_S       ; Offset of CS buffer
        MOV     WORD PTR expDigs,9     ; 9 packed digits to move
		MOV     resi,ESI
		MOV     redi,EDI
        CALL    UNPACK_MOVE     ; Unpack and move digits routine

;**********************|
; multiply significands|
;**********************|
;*******************************************************************
; Structures for multidigit unpacked BCD multiplication
; 1. PART_PROD is a 37-digit buffer to hold the product or the
;    multiplicand times each digit in the multiplier. The operation
;    can be visualized as follows:
;       COLUMN NUMBERS
;       3 3 3 3 3 3 3 3 2 2 ..... 1 1 1 1 1 1 1 1 1
;       7 6 5 4 3 2 1 0 9 8 ..... 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1
;                                 x x x x x x x x x x x x x x x x x x
;                                 y y y y y y y y y y y y y y y y y y
; ROWS --------------------------------------------------------------
; 1     0 0 0 0 0 0 0 0 0 0 ....C z z z z z z z z z z z z z z z z z z
; 2     0 0 0 0 0 0 0 0 0 ... C z z z z z z z z z z z z z z z z z z 0
; 3     0 0 0 0 0 0 0 0 ... C z z z z z z z z z z z z z z z z z z 0 0
; 4     0 0 0 0 0 0 0 ... C z z z z z z z z z z z z z z z z z z 0 0 0
; . . .
; 17    0 C z z z z ... z z z z z z 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
; 18    C z z z z ... z z z z z z 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
;       -------------------------------------------------------------
;       C p p p p ... p p p p p p p p p p p p p p p p p p p p p p p p
;
; In the actual routine the PART_PROD buffer contains a single row
; of digits. This product row is added to the previous contents of
; the SIG_R buffer before the next partial product is calculated
;*******************************************************************
; At this point the significands of x and y are in the buffers SIG_L
; and SIG_S
        LEA     EDI,SIG_S       ; EDI --> to multiplier
        ADD     EDI,17          ; Point ESI to last multiplier digit
        MOV     ROWS,1          ; ROWS is counter for 18 digit rows
;**********************|
; clear product buffer |
;**********************|
ONE_ROW_OP:
        LEA     ESI,SIG_L       ; ESI --> to multiplicand
        ADD     ESI,17          ; Point ESI to end of multiplicand
        LEA     EBX,PART_PROD   ; EBX --> buffer for partial products
        MOV     ECX,37          ; Bytes to clear
        PUSH    EBX             ; Save buffer start address
CLEAR_PP:
        MOV     BYTE PTR [EBX],0H     ; Store a zero digit
        INC     EBX             ; Bump buffer pointer
        LOOP    CLEAR_PP
; Offset of rightmost digit in PART_PROD is 37 minus the row
; count. Row count is in ROWS. The offset of the last digit in the
; first partial product is EBX + (37-1) = EBX + 36
        POP     EBX             ; EBX --> start of products buffer
        MOV     AX,37           ; Offset of last buffer digit
        SUB     AX,ROWS         ; Minus current row
		AND     EAX,0FFFFH      ; Clear high word of EAX
        ADD     EBX,EAX         ; Add offset to pointer
;**********************|
;    multiply row      |
;**********************|
        MOV     DL,[EDI]        ; Multiplier digit
        MOV     DH,0            ; Register for carry
        MOV     ECX,18          ; Counter for 18 digits
; At this point:
; ESI --> rightmost digit in the multiplicand
; DL = multiplier digit
; EBX --> last digit offset into partial products buffer
MUL_18_DIGITS:
        MOV     AH,0            ; Clear high digit for product
        MOV     AL,BYTE PTR [ESI]     ; Multiplicand digit to AL
        MUL     DL              ; AX = AL * DL
        ADD     AL,DH           ; Add in carry from last product
        AAM                     ; Adjust product to unpacked BCD
        MOV     BYTE PTR [EBX],AL     ; Store digit in buffer
        MOV     DH,AH           ; Carry into next digit
        DEC     EBX             ; Bump to next storage position
        DEC     ESI             ; and to next multiplicand digit
        LOOP    MUL_18_DIGITS
; 18 digits multiplied. Store possible carry
        MOV     BYTE PTR [EBX],DH     ; Digit to buffer
; At this point PART_PROD holds product of multiplicand times
; current multiplier digit
;**********************|
;      add row         |
;**********************|
; PART_PROD is now added to the accumulated sum stored in SIG_R
; Both buffers have 37 digits
        PUSH    EDI             ; Save pointer to multiplier
        LEA     ESI,PART_PROD   ; Pointer to first digit row
        LEA     EDI,SIG_R       ; Pointer to destination of sum
        ADD     ESI,36          ; Point ESI to last digit
        ADD     EDI,36          ; also EDI
        MOV     ECX,37          ; 37 digits to add
        CLC                     ; Clear carry for first addition
ADD_ROW:
        MOV     AL,BYTE PTR [ESI]     ; One digit to AL
        ADC     AL,BYTE PTR [EDI]     ; Add other digit
        AAA                     ; Convert sum to unpacked BCD
        MOV     BYTE PTR [EDI],AL     ; Store digit in same buffer
        DEC     ESI             ; Bump pointers to next digit
        DEC     EDI
        LOOP    ADD_ROW         ; Loop for 37 digits
; Since the buffers contain 37 digits there can be no carry out of
; the last digit
        POP     EDI             ; Restore multiplier pointer
;**********************|
;   multiply next row  |
;**********************|
        DEC     EDI             ; Index to next multiplier
        INC     ROWS            ; Increment row counter
        CMP     ROWS,19         ; Test for all rows multiplied
        JE      DONE_18         ; Go if at end of operation
        JMP     ONE_ROW_OP
; At this point the SIG_R buffer contains the product of the
; significands
;*************************************|
;   Construct BCD12 number and exit   |
;*************************************|
DONE_18:
        CALL    RESTORE_BCD     ; Local procedure used by multiply
                                ; and divide rouitnes
SIGN_MUL_EXIT:
}
}

//******************************************************************
//                  BCD division function
//******************************************************************
void SignDivBcd12(char bcd1[], char bcd2[], char result[])
{
_asm
{
//************************************|
// signed floating point BCD division |
//************************************|
// Division of two signed BCD numbers stored in BCD12 floating point
// format
// Operation:
//           z = x / y
//           where x, y, and z are signed, floating point numbers
// On entry:
//          bcd1[] = dividend (element x in z = x / y)
//          bcd2[] = divisor (element y in z = x / y)
//          result[] = 12-byte storage area for quotient
// Note: the code assumes that the BCD12 numbers are in normalized
//       form, that is, that there are no leading zeros in the
//       significand
// On exit:
//       result = quotient (element z in z = x / y)
//       if divisor = 0, then BCD in result has FFH for Ee byte
//*******************************************************************
// BCD signed division algorithm
// A. If the dividend is zero the quotient is zero
//    If the divisor is zero the operation is undefined
//    If neither element is zero then:
// B. If the factors have equal signs the quotient is positive
//    If the factors have unequal sings the quotient is negative
// C. The significand of the quotient is the quotient resulting from
//    dividing the significand of the dividend by the siginificand of
//    the divisor
// D. The exponent of the quotient is the difference between the
//    exponent of the dividend minus the exponent of the divisor
//*******************************************************************
; Store entry variables
         MOV     ESI,bcd1
         MOV     EDI,bcd2
		 MOV     EBX,result
         MOV     rebx,EBX
		 MOV     resi,ESI
		 MOV     redi,EDI
; Store pointers
         MOV     ENTRY_SI,ESI
		 MOV     ENTRY_DI,EDI
		 MOV     ENTRY_BX,EBX
; Clear buffers for significands
         CALL    CLEAR_SIG_BUFS ; Local procedure
;**********************|
;    clear division    |
;        buffers       |
;**********************|
        PUSH    EBX             ; Save pointer passed by caller
        LEA     EBX,DIVIDEND    ; First of 5 buffers
        MOV     rebx,EBX
        MOV     recx,19         ; bytes to clear
        CALL    CLEAR_BUF       ; Local procedure to clear buffer
        LEA     EBX,DIVISOR     ; Next buffer
        MOV     rebx,EBX
		MOV     recx,18
        CALL    CLEAR_BUF
        LEA     EBX,TRIAL_PROD  ; Next buffer
		MOV     rebx,EBX
		MOV     recx,19
        CALL    CLEAR_BUF
        LEA     EBX,NEW_DIVIDEND; Next buffer
		MOV     rebx,EBX
		MOV     recx,19
        CALL    CLEAR_BUF
        LEA     EBX,QUOTIENT    ; Last buffer
		MOV     rebx,EBX
		MOV     recx,37
        CALL    CLEAR_BUF
        POP     EBX             ; Restore caller's pointer

;**********************|
; test for dividend = 0|
;**********************|
; If the dividiend = 0 then the quotient = 0
; resi holds address of bcd1[]
        CALL    TEST_4_ZERO     ; Test routine
		CMP     ZERO_VAL,1
        JNE     NOT_ZERO_DID    ; Go if no carry
; At this point x is zero
        MOV     ECX,12          ; Count for digits to move
MOVE_12_QUOT:
        MOV     BYTE PTR[EBX],00; Store zero by BX pointer
        INC     EBX             ; Bump pointer
        LOOP    MOVE_12_QUOT
		JMP     SIGN_DIV_EXIT
;**********************|
; test for divisor = 0 |
;**********************|
; Division is undefined if the divisor = 0
; This is signaled to the caller with the carry flag set
NOT_ZERO_DID:
        PUSH    ESI             ; Save x pointer
        MOV     resi,EDI
        CALL    TEST_4_ZERO     ; Test routine
        POP     ESI             ; Restore x pointer
		MOV     resi,ESI        ; Restore variable
		CMP     ZERO_VAL,1
        JNE     NOT_ZERO_ELMS   ; Continue if no carry
;**********************************|
;    error exit from division      |
;**********************************|
; Sign byte at BCD12 offset 0 = FFH to signal error
        MOV     ESI,result
		MOV     BYTE PTR[ESI],0FFH
        JMP     SIGN_DIV_EXIT
;**********************|
;  neither x nor y = 0 |
;  subtract exponents  |
;**********************|
NOT_ZERO_ELMS:
		MOV     ESI,bcd1
        MOV     resi,ESI 
        CALL    EXP_2_BIAS      ; Conversion procedure
; DX now holds binary exponent for x in bias 10,000 form
        MOV     WORD PTR BIAS_X,DX   ; Store in scratch pad area
; Obtain biased exponent for y
        XCHG    ESI,EDI         ; Save x and set y
		MOV     resi,ESI
        CALL    EXP_2_BIAS      ; Conversion procedure
        MOV     WORD PTR BIAS_Y,DX    ; Store in scratch pad area
        XCHG    ESI,EDI         ; Restore x and y pointers
;**********************|
;  subtract exponents  |
;**********************|
        MOV     AX,WORD PTR BIAS_X    ; Exponent of x to AX
        SUB     AX,DX           ; Subtract exponent of x - y
        ADD     AX,10000        ; Add bias to difference
                                ; to re-scale exponent
        MOV     WORD PTR BIAS_Z,AX    ; Store sum in scratch pad area
;*******************************************************************
; Buffers used in the division of significands:
;        DIVIDEND     (-->EDI)        19 bytes
;        NEW_DIVIDEND                 19 bytes
;        DIVISOR      (-->ESI)        18 bytes
;        TRIAL_PROD                   19 bytes
;        QUOTIENT                     37 bytes
;*******************************************************************
;**********************|
;    unpack and move   |
;      significands    |
;**********************|
; Move 18 packed BCD digits from BCD12 encoding to DIVIDEND buffer
        MOV     ESI,bcd1        ; Set pointer to x
        ADD     ESI,3           ; Mantissa digits in BCD12 encoding
        LEA     EDI,DIVIDEND    ; Offset of CS buffer
        INC     EDI             ; Right-justify dividend in buffer
        MOV     WORD PTR expDigs,9     ; 9 packed digits to move
		MOV     resi,ESI
		MOV     redi,EDI
        CALL    UNPACK_MOVE     ; Unpack and move digits routine
; Move significand of y to DIVISOR buffer
        MOV     ESI,bcd2        ; Set pointer to y
        ADD     ESI,3           ; Mantissa digits in BCD12 encoding
        LEA     EDI,DIVISOR   ; Offset of CS buffer
        MOV     WORD PTR expDigs,9     ; 9 packed digits to move
		MOV     resi,ESI
		MOV     redi,EDI
        CALL    UNPACK_MOVE     ; Unpack and move digits routine
;**********************|
; init quotiend digit  |
;      counter         |
;**********************|
        MOV     DIGS,0          ; DIGS is counter for digits
                                ; in the quotient buffer
; Go to entry point for initial undersize dividend
        JMP     EXPU
;**********************|
; obtain trial divisor |
;**********************|
TEST_DIVIDE:
        CALL    TRIAL_QUOT      ; Local routine to find a trial
                                ; quotient digit
        MOV     DL,AL           ; Trial quotient digit to DL
TEST0:
        LEA     ESI,DIVISOR   ; Set pointer to divisor
        LEA     EBX,TRIAL_PROD        ; Pointer to trial product
        ADD     ESI,17          ; To last digit in buffer
        ADD     EBX,18          ; also this pointer
        MOV     DH,0            ; Clear carry register
        MOV     ECX,18          ; Set counter for 18 digits
;**********************|
;  multiply dividend   |
; times trial quotient |
;**********************|
TRI0:
        MOV     AH,0            ; Clear high byte
        MOV     AL,[ESI]        ; Get high order divisor digit
        MUL     DL              ; Divisor times trial quotient
        ADD     AL,DH           ; Add possible carry to product
        AAM                     ; Convert product to BCD
        MOV     DH,AH           ; Save carry in DH
        MOV     BYTE PTR [EBX],AL   ; Store digit as trial product
        DEC     EBX             ; Bump pointers
        DEC     ESI
        LOOP    TRI0            ; Continue for 18 digits
; At this point there can be a carry into the 19th digit
; Place possible carry or zero in buffer
        MOV     BYTE PTR [EBX],DH        ; Store extra digit
;**********************|
;   subtract product   |
;    from dividend     |
;**********************|
        PUSH    EDX             ; Save registers and pointers
        PUSH    ESI
        PUSH    EDI
        PUSH    EBX
        LEA     EBX,NEW_DIVIDEND; Set context pointers
        LEA     ESI,TRIAL_PROD
        LEA     EDI,DIVIDEND
        ADD     EBX,18          ; Index pointers to low order byte
        ADD     ESI,18
        ADD     EDI,18
        MOV     ECX,19          ; Counter for 19 digits
        CLC                     ; Clear carry flag
; Perform subtraction operation
REST0:
        MOV     AL,BYTE PTR [EDI]   ; Dividend digit to AL
        MOV     DL,BYTE PTR [ESI]   ; Trial product digit to DL
        SBB     AL,DL           ; Dividend minus trial product
        AAS                     ; Convert to BCD
        MOV     BYTE PTR [EBX],AL      ; Store result in buffer
        DEC     EBX             ; Bump buffer pointers
        DEC     ESI
        DEC     EDI
        LOOP    REST0           ; Continue for 19 bytes
        POP     EBX             ; Restore entry pointers
        POP     EDI
        POP     ESI
        POP     EDX
; If there is a carry into the buffer's 19th digit then the trial
; quotient is too large
        JNC     PUTQ            ; Go if no carry (quotient digit OK)
;**********************|
;   invalid quotient   |
;        digit         |
;**********************|
; At this point the trial quotient must be adjusted and retried
; If the trial digit goes to zero, then it is stored as a quotient
; digit and processing continues
        DEC     DL              ; Decrement quotient digit
        JZ      PUTQ            ; Store zero
        JMP     TEST0           ; Retry new trial quotiend digit
;**********************|
; valid quotient digit |
;**********************|
; Processing:
; 1. Store quotient digit in buffer (DIGIT_TO_QUOT procedure)
; 2. Make the trial dividend (in NEW_DIVIDEND) the dividend
PUTQ:
        MOV     redx,EDX
        CALL    DIGIT_TO_QUOT   ; Store digit in buffer
		CMP     BUF_FULL,1
        JE      SIG_DONE        ; Go if buffer full returned
; Move NEW_DIVIDEND to DIVIDEND multiplying by 10
        LEA     EDI,DIVIDEND  ; Pointer to dividend buffer
        LEA     ESI,NEW_DIVIDEND      ; Pointer to new dividend
        MOV     BYTE PTR [ESI],0H     ; Clear high byte
        INC     ESI
        MOV     resi,ESI
		MOV     redi,EDI
		MOV     recx,18
		CALL    TRANSFER_CS_CX  ; Local procedure
; New dividend is now in NEW_DIVIDEND buffer
; Chech size of dividend. Place zeros in buffer is the dividend
; is smaller than the divisor
EXPU:
        CALL    COMPARE_DI_DR   ; Compare divisor and dividend
        CMP     LARGER,1
        JE      EXPU1           ; Go if divisor > dividend
        JMP     TEST_DIVIDE     ; Continue processing
;**********************|
;   quotient = 0       |
;**********************|
EXPU1:
        MOV     DL,0            ; Digit to store
        MOV     redx,EDX
        CALL    DIGIT_TO_QUOT   ; Local procedure
		CMP     BUF_FULL,1
        JE      SIG_DONE        ; Go if buffer full returned
; Multiply dividend by 10 by offsetting into buffer
        LEA     EDI,DIVIDEND    ; DI points to dividend area
        MOV     ESI,EDI         ; and so does ESI
        INC     ESI             ; Skip first digit (0) in source
                                ; effectively multiplying by 10
        MOV     resi,ESI
		MOV     redi,EDI
		MOV     recx,18
        CALL    TRANSFER_CS_CX  ; Local procedure
        MOV     BYTE PTR [EDI],0      ; Add zero at end of buffer
        JMP     EXPU
;**********************|
;  move significand    |
;**********************|
; Significand digits from QUOTIENT buffer to SIG_R buffer
SIG_DONE:
        LEA     ESI,QUOTIENT    ; Source buffer
        LEA     EDI,SIG_R       ; Destination buffer
        INC     EDI             ; Skip first digit
        MOV     resi,ESI
		MOV     redi,EDI
		MOV     recx,36
        CALL    TRANSFER_CS_CX  ; Local procedure
        CALL    RESTORE_BCD     ; Local procedure used by multiply
                                ; and divide rouitnes
;**********************************|
;   normal exit from division      |
;**********************************|
SIGN_DIV_EXIT:
}
}

//*************************************************
//      Support functions for conversions
//*************************************************

void BIN_TO_BCD()
{
// Local buffer for BIN_TO_BCD
char BCD_BUF[5];

_asm
{
// BIN_TO_BCD
// Convert a 16-bit binary number to 5 BCD digits
// On entry:
//          redx holds binary number
// On exit:
//         5 packed BCD digits in the following registers:
//                CL = digit number 1 (high order digit)
//                DH = digits number 2 and 3
//                DL = digits number 4 and 5 (low order BCD digit)
//
// Operation:
// Binary in AX is first converted into 5 unpacked BCD digits and
// stored in a global buffer (BCD_BUF). Digits are then moved
// from this buffer and packed into DX and CL
//
//
// CPU compatibility chart:
//                       | 8086/8088       | YES |
//                       | 80286           | YES |
//                       | 80386           | YES |
//                       | 80486/PENTIUM   | YES |

        MOV     EDX,redx 	
	    MOV     AX,DX
;**********************|
;    clear buffer      |
;**********************|
        LEA     EDI,BCD_BUF     ; Set pointer to buffer
        PUSH    EDI             ; Save buffer start
        MOV     ECX,5           ; Five digits to clear
CLEAR_5BCD:
        MOV     BYTE PTR [EDI],0         ; Clear digit
        INC     EDI             ; Bump pointer
        LOOP    CLEAR_5BCD      ; Repeat for 5 digits
        DEC     EDI             ; Adjust buffer pointer to last
                                ; digit
        MOV     ECX,10          ; Decimal divisor to CX
GET_BIN:
        MOV     DX,0            ; Clear for word division
; Divide binary by 10
        DIV     CX              ; Perform division AX/CX
; Quotient is in AX and remainder in DL
        MOV     BYTE PTR [EDI],DL     ; Store digit in buffer
        DEC     EDI             ; Bump destination pointer
; Note: step 5 is performed automatically since the binary
; quotient is left in AX
        CMP     AX,0            ; Test for end of binary
        JNZ     GET_BIN         ; Continue if not end
; Last digit converted
;**********************|
;   BCD to registers   |
;      and pack        |
;**********************|
        POP     EDI             ; Restore pointer to start of
                                ; buffer
        ADD     EDI,4           ; Point to last digit
        MOV     AL,BYTE PTR [EDI]     ; BCD digit to AL
        MOV     BL,AL           ; To results register low nibble
        DEC     EDI             ; Point to fourth BCD digit
        MOV     AL,BYTE PTR [EDI]     ; BCD digit to AL
        MOV     CL,4            ; Shift counter
        SHL     AL,CL           ; Shift BCD to high nibble
        OR      BL,AL           ; OR into result register high
                                ; nibble
        DEC     EDI              ; Pointer to third BCD digit
        MOV     AL,BYTE PTR [EDI]     ; BCD digit to AL
        MOV     BH,AL           ; To results register low nibble
        DEC     EDI              ; Point to second BCD digit
        MOV     AL,BYTE PTR [EDI]     ; BCD digit to AL
        MOV     CL,4            ; Shift counter
        SHL     AL,CL           ; Shift BCD to high nibble
        OR      BH,AL           ; OR into result register high
                                ; nibble
; Move 4 low order BCD digits to DX
        MOV     DX,BX
; Move high order digit to CL
        DEC     EDI              ; Pointer to first BCD digit
        MOV     AL,BYTE PTR [EDI]     ; BCD digit to AL
        MOV     CL,AL           ; To results register low nibble
}
}


void BCD_INTO_REG()
{
char ASC_TEMP_BUF[5];
_asm
{
// Convert a 5-digit BCD number into a binary in DX
// On entry:
//         AL-BH-BL-CH-CL hold 5 BCD digits
//                        holding a value < 65536
// On exit:
//         DX = binary number (range 0 to 65535)
//
//
// CPU compatibility chart:
//                       | 8086/8088       | NO  |
//                       | 80286           | NO  |
//                       | 80386           | YES |
//                       | 80486/PENTIUM   | YES |
//
//

		CALL	RESTORE_RS
        LEA     ESI,ASC_TEMP_BUF ; Address of buffer
		MOV		reax,EAX
		MOV		rebx,EBX
		MOV		recx,ECX
		MOV     redx,EDX
        CALL    BCD_TO_ASCII    ; Convert BCD digits to ASCII
                                ; string
		MOV		reax,EAX
		MOV		rebx,EBX
		MOV		recx,ECX
		MOV     redx,EDX
					
        CALL    ASCII_TO_DX     ; Conversion routine
}
}

void BCD_TO_ASCII()
{
_asm
{
// Convert 5 unpacked BCD digits in machine registers into to 5
// ASCII digits
// On entry:
//         AL-BH-BL-CH-CL hold 5 unpacked BCD digits
//                        holding a value < 65536
//         ESI --> 5 byte buffer to hold ASCII string
// On exit:
//         ESI --> buffer holding ASCII digits
//
// CPU compatibility chart:
//                       | 8086/8088       | NO  |
//                       | 80286           | NO  |
//                       | 80386           | YES |
//                       | 80486/PENTIUM   | YES |
//

		CALL	RESTORE_RS
; Clear buffer for ASCII digits
        PUSH    ESI              ; Save start of buffer
        PUSH    ESI              ; Twice
        PUSH    ECX              ; Save BCD high digit
        MOV     ECX,5            ; Counter for 5 digits
CLEAR_BUF_5:
        MOV     BYTE PTR [ESI],' '    ; Set blank in buffer
        INC     ESI              ; Bump pointer
        LOOP    CLEAR_BUF_5
        POP     ECX              ; Restore BCD digit
;**********************|
; convert to ASCII and |
;    move to buffer    |
;**********************|
; Convert BCD digits to ASCII
        ADD     AL,30H          ; Add 30H to each digit
        ADD     BH,30H
        ADD     BL,30H
        ADD     CH,30H
        ADD     CL,30H
        POP     ESI              ; Restore buffer start
        MOV     BYTE PTR [ESI],AL     ; Digit to buffer
        INC     ESI              ; Bump buffer pointer
        MOV     BYTE PTR [ESI],BH     ; Digit to buffer
        INC     ESI              ; Bump buffer pointer
        MOV     BYTE PTR [ESI],BL     ; Digit to buffer
        INC     ESI              ; Bump buffer pointer
        MOV     BYTE PTR [ESI],CH     ; Digit to buffer
        INC     ESI              ; Bump buffer pointer
        MOV     BYTE PTR [ESI],CL     ; Digit to buffer
        POP     ESI              ; Restore start of buffer
}
}


void ASCII_TO_DX()
{
_asm
{
// Convert an ASCII number to binary and store in DX register
// On entry:
// ESI -->   start of 5-digit ASCII buffer holding decimal
//           string in the range 0 to 65,535
// On exit:
//           DX = binary number
//
// CPU compatibilty chart:
//                       | 8086/8088       | NO  |
//                       | 80286           | NO  |
//                       | 80386           | YES |
//                       | 80486/PENTIUM   | YES |
//

		CALL	RESTORE_RS      
;***********************|
;  initialize registers |
;***********************|
        MOV     DX,0            ; Clear accumulator register
        MOV     ECX,10          ; Load multiplier into CX
;***********************|
; test digit for range  |
;        0 to 9         |
;***********************|
DIGIT_TO_ACC16:
        MOV     AL,BYTE PTR [ESI]         ; Get ASCII digit
        CMP     AL,'0'          ; Test for lower limit
        JB      EXIT_ASC16      ; Exit if less than 0
        CMP     AL,'9'          ; Test for higher limit
        JA      EXIT_ASC16      ; Exit if larger than 9
;***********************|
; ASCII digit to binary |
;***********************|
        SUB     AL,30H          ; ASCII to decimal
;***********************|
;    accumulator x 10   |
;      + digit          |
;***********************|
        MOV     AH,0            ; Clear high byte in AX
        PUSH    AX              ; Save digit in stack
        MOV     AX,DX           ; Previous product to AX
        MUL     CX              ; Perform AX = AX * 10
        MOV     DX,AX           ; Move product to accumulator
        POP     AX              ; Restore decimal digit
        ADD     DX,AX           ; Add digit to accumulator
;***********************|
; pointer to next digit |
;***********************|
        INC     ESI              ; Bump pointer
        JMP     DIGIT_TO_ACC16  ; Continue
EXIT_ASC16:
}
}

void RESTORE_RS()
{
_asm
{
		MOV		EAX,reax
		MOV		EBX,rebx
		MOV		ECX,recx
		MOV		EDX,redx
}
}

//********************************************************
//         support functions for BCD arithmetic
//********************************************************
void REVERSE_SIGN(char aBcd[])
{
_asm
{
// Invert sign of a number encoded in BCD12 format
// On entry:
//         parameter aBcd[] is BCD operand
// On exit: 
//         parameter aBcd[] is sign reversed
//

	    MOV     EDI,aBcd
	    TEST    BYTE PTR [EDI],00010000B ; Bit 4 set if sign is -
        JNZ     SIGN_MINUS      ; Go if bit set
;**********************|
;    change + to -     |
;**********************|
        OR      BYTE PTR [EDI],00010000B ; Make negative
        JMP     EXIT_RS
;**********************|
;    change - to +     |
;**********************|
SIGN_MINUS:
        AND     BYTE PTR [EDI],00001111B ; Make postive
EXIT_RS:
}
}

void CLEAR_SIG_BUFS()
{
_asm
{
// Clear buffers for BCD significands SIG_L, SIG_S, and SIG_R
; Store entry pointers in scratchpad area and clear work buffers
;**********************|
; get entry pointers   |
;**********************|
        MOV     EDI,redi
        MOV     ESI,resi
        MOV     EBX,rebx
;**********************|
;  clear work buffers  |
;**********************|
        PUSH    ESI              ; Save x value entry pointer
        LEA     ESI,SIG_L        ; Pointer to buffer
        MOV     ECX,36           ; Clear buffer
CLEAR_B1:
        MOV     BYTE PTR[ESI],00H    ; Clear one byte in buffer
        INC     ESI                      ; Bump pointer
        LOOP    CLEAR_B1
; Next buffer
        LEA     ESI,SIG_S
		MOV     ECX,36
CLEAR_B2:
        MOV     BYTE PTR[ESI],00H    ; Clear one byte in buffer
        INC     ESI                      ; Bump pointer
        LOOP    CLEAR_B2       
; Next buffer
        LEA     ESI,SIG_R
		MOV     ECX,37
CLEAR_B3:
        MOV     BYTE PTR[ESI],00H    ; Clear one byte in buffer
        INC     ESI                      ; Bump pointer
        LOOP    CLEAR_B3  

        POP     ESI              ; Restore x pointer
}
}

void UNPACK_MOVE()
{
_asm
{
// Move the number of packed digits in CX from -> ESI
// -> EDI and unpack
// On entry:
//         expDigs = number of packed BCD digits to be moved
//         ESI --> packed BCD string to move
//         EDI --> destination of unpacked string
// On exit:
//         EDI = unpacked string of BCD digits
//
        MOV     ESI,resi
        MOV     EDI,redi

		MOV     CX,WORD PTR expDigs
; Move CX digits
MOVE_DIGS:
        MOV     AL,BYTE PTR [ESI]         ; Get one byte from source
; AL has two packed BCD digits
        MOV     AH,AL           ; Copy packed digits to AH
        AND     AL,00001111B    ; Clear high nibble to isolate low
                                ; BCD digit
        SHR     AH,4            ; Shift right to isolate high BCD
                                ; digit
; AH holds the unpacked high digit and AL holds the low digit
        MOV     BYTE PTR[EDI],AH      ; High BCD digit to buffer
        INC     EDI             ; Bump buffer pointer
        MOV     BYTE PTR[EDI],AL        ; Low BCD digit to buffer
        INC     EDI             ; Bump pointer again
        INC     ESI             ; Bump source pointer
        LOOP    MOVE_DIGS
		MOV     ESI,resi
		MOV     EDI,redi
}
}


void ROUND_SIGF()
{
_asm
{
// Rounding of significand digits stored in SIG_R buffer
// Rationale:
// SIG_R holds 37 digits in unpacked BCD format. The BCD12 format
// encodes 18 siginificand digits
// Rounding proceeds with the 19th digit in SIG_R, as follows:
// 1. If the 19th digit is < 5 then no rounding is required
// 2. If the 19th digit is > 5 then the first 18 digits are rounded
// 3. If the 19th digit = 5 then the digits to the right are examined.
//    If the digits to the right are all zeros (5000..00) then the
//    18th digit is incremented only if it is odd. This rounding
//    method is known as "round to nearest even".
// 4. If the result of incrementing the 18th digit is 10, the digit
//    to the left is also incremented, until no carry is generated.
//
	    LEA     ESI,SIG_R       ; Set pointer to buffer in CS
        ADD     ESI,18          ; Index to the 19th digit
        MOV     AL,[ESI]        ; 19th digit to AL
        CMP     AL,05H          ; Test for 5 or greater
        JB      EXIT_ROUNDING   ; No rounding necessary
; 19th digit is 5 or greater
        JE      MIDWAY_CASE     ; Go to "round to even" routine
; At this point the 19th digit is > 5
;**********************|
;  bump digit and test |
;**********************|
ROUND_OFF:
        DEC     ESI             ; Point to 18th digit
ROUND_EVEN:
        MOV     AL,[ESI]        ; Get BCD digit into AL
        ADD     AL,01H          ; Add 1 to digit
        AAA                     ; Adjust result
        JNC     EXIT_ROUNDING   ; Go if no carry generated
        JMP     ROUND_OFF
; Note that no testing for the end of buffer is required since the
; leftmost digit in SIG_R cannot be greater than 8
;**********************|
; test for midway case |
;**********************|
; ESI --> to the 19th digit
MIDWAY_CASE:
        PUSH    ESI             ; Save pointer to 19th digit
        MOV     ECX,17          ; Set counter for 17 digits
TEST_FOR_0H:
        INC     ESI             ; Bump pointer to next digit
        CMP     BYTE PTR [ESI],00H    ; Test for a zero digit
        JNE     NOT_MIDWAY      ; Exit if not midway case
        LOOP    TEST_FOR_0H     ; Continue testing
; At this point the undisplayed digits are in the pattern 5000..00
        POP     ESI             ; Restore pointer to 19th digit
        DEC     ESI             ; Point to 18th digit
        MOV     AL,[ESI]        ; Get 18th digit into AL
;**********************|
;  test for odd 18th   |
;        digit         |
;**********************|
; Binary encodings of odd numbers have the LOB set
        TEST    AL,00000001B    ; Test LOB to determine if odd
        JNZ     ROUND_EVEN      ; Go to rounding routine if odd
        JMP     EXIT_ROUNDING   ; Do not round if already even
;**********************|
;    not midway case   |
;**********************|
; The number is rounded if the undiplayed significand is not in the
; form 50000..00
NOT_MIDWAY:
        POP     ESI             ; Restore pointer
        JMP     ROUND_OFF       ; Go to rounding routine
;**********************|
;   exit procedure     |
;**********************|
EXIT_ROUNDING:
}
}

void TEST_4_ZERO()
{
_asm
{
// Test a number encoded in BCD12 format for a zero value
// Note: a number in BCD12 format is zero if all numbers in the
//       exponent and significand are zero
// On entry:
//         ESI --> Number stored in BCD12 format
// On exit:
//         Carry set if number is zero
//         Carry clear if number is not zero
//
	    MOV     ESI,resi
        PUSH    ESI             ; Save entry pointer
        INC     ESI             ; Point to first exponent digit
        MOV     ECX,11          ; Counter for 11 bytes to test
TEST_11_ZERO:
        CMP     BYTE PTR[ESI],0 ; Is the byte zero
        JNE     NOT_11_ZERO     ; Go if not zero
        INC     ESI             ; Bump pointer
        LOOP    TEST_11_ZERO    ; Test all digits
; At this point the number is zero
        POP     ESI             ; Restore entry pointer
        MOV     ZERO_VAL,1      ; Set zero return signal
        JMP     EXIT_T40
NOT_11_ZERO:
        POP     ESI             ; Restore entry pointer
        MOV     ZERO_VAL,0      ; Set non-zero return signal
EXIT_T40:
}
}

void EXP_2_BIAS()
{
_asm
{
// Convert exponent of BCD12 number into binary and add bias
// On entry:
//         resi --> Number encoded in BCD12 format
// On exit:
//         DX = binary value of exponent
//

        MOV     ESI,resi
	    PUSH    EBX              ; Save pointers
        PUSH    ESI
        PUSH    EDI
        INC     ESI              ; Pointer to exponent field
        INC     ESI              ; Second exponent digit
; Move BCD digits to AL-BH-BL-CH-CL for conversion routine
        MOV     AL,0            ; Highest digit is zero
        MOV     DL,BYTE PTR [ESI]        ; DL holds 2 packed digits
        MOV     DH,DL           ; Copy digits to DH
        AND     DL,00001111B    ; Erase high nibble in DL
        AND     DH,11110000B    ; and low in DH
        SHR     DH,4            ; Shift digit to low position
        DEC     ESI              ; Bump pointer to previous digit
        MOV     BL,BYTE PTR [ESI]        ; BL holds 2 packed digits
        MOV     BH,BL           ; Copy digits to BH
        AND     BL,00001111B    ; Erase high nibble in BL
        AND     BH,11110000B    ; and low in BH
        SHR     BH,4            ; Shift digit to low position
        MOV     CX,DX           ; Move digits to CX
// Save registers
		MOV		reax,EAX
		MOV		rebx,EBX
		MOV		recx,ECX
		MOV     redx,EDX
		CALL    BCD_INTO_REG    ; BCD to binary conversion
        POP     EDI              ; Restore entry pointers
        POP     ESI
        POP     EBX
; At this point DX hold binary value for exponent
; ESI --> sign byte
; Test for negative exponent. Add bias if positive
        TEST    BYTE PTR[ESI],01H        ; Test for negative exponent
        JNZ     EXP_IS_NEG
; At this point exponent is positive
        ADD     DX,10000        ; Add bias to exponent value
        JMP     EXIT_E2B
; Subtract exponent from bias
EXP_IS_NEG:
        MOV     AX,10000        ; Bias
        SUB     AX,DX           ; Value minus bias
        MOV     DX,AX           ; Result to DX
EXIT_E2B:
}
}

void RESTORE_BCD()
{
_asm
{
// Reconstruct BCD number after multiplication and division operation
// Note that the sign rules for significand and exponents are the
// same
// On entry:
//         BIAS_Z holds biased exponent of result
//         SIG_R buffer holds significand
// On exit:
//         Buffer at CS:ENTRY_BX contains results in BCD12 format
//
        CALL    ROUND_SIGF      ; Round off significand
  		MOV     EDI,ENTRY_BX    ; Set pointer to result buffer
                                ; for BCD12 output
        MOV     BYTE PTR [EDI],00H       ; Set default positive sign
        ADD     EDI,3            ; Point to significand area
;**********************|
;   store significand  |
;**********************|
; Significand of result is unpacked BCD form in the first 18 digits
; of the SIG_R buffer
; EDI --> significand field of the BCD12 result buffer
        MOV     ECX,9           ; Counter for 18 packed digits
        LEA     ESI,SIG_R       ; Pointer to unpacked digits
        INC     ESI             ; Skip leading zero
; Skip possible leftmost zero in buffer
        CMP     BYTE PTR [ESI],00H    ; Test for zero
        JNE     PACK_DIGITS     ; Go if no leading zero
        INC     ESI              ; Skip zero
        DEC     BIAS_Z          ; Adjust exponent
; In multiplication there can be only one leading zero in the buffer
; because the input significands are in normalized form
PACK_DIGITS:
        MOV     AH,BYTE PTR [ESI]     ; Get one digit
        INC     ESI             ; Bump pointer
        MOV     AL,BYTE PTR [ESI]     ; Get next digit
        INC     ESI             ; Bump pointer to next digit
        SHL     AH,4            ; Shift high digit to high nibble
        OR      AL,AH           ; Merge high nibble in AH with low
                                ; nibble in AH (pack digits)
        MOV     BYTE PTR [EDI],AL        ; Store packed digits
        INC     EDI
        LOOP    PACK_DIGITS
; All digits moved to the destination BCD12 buffer
        MOV     EBX,ENTRY_BX    ; Set pointer to result buffer
                                ; sign area
;**********************|
;   sign of number     |
;**********************|
; Rule: if the factors have equal signs the product is positive
;       If the factors have unequal sings the product is negative
; The sign of the number is the high nibble of the first BCD12 digit
        MOV     ESI,ENTRY_SI  ; Set pointers to entry values
        MOV     EDI,ENTRY_DI
; ESI --> x and EDI --> y
        MOV     AL,BYTE PTR [ESI]        ; x's first digit to AL
        MOV     AH,BYTE PTR [EDI]        ; y's first digit to AH
        AND     AL,11110000B    ; Clear x's low nibble
        AND     AH,11110000B    ; and y's low nibble
        CMP     AL,AH           ; Test for equal sign
        JE      EXPN_SIGN       ; Numbers have equal sign
                                ; This encoding already in buffer
; At this point the factors have unequal sign. Product is negative
        MOV     BYTE PTR [EBX],10H       ; Negative sign for number
;**********************|
; get sign of exponent |
;**********************|
EXPN_SIGN:
; Exponent of result is stored in BIAS_Z in bias 10000 form
; Values < 10000 indicate a negative exponent
        MOV     DX,BIAS_Z    ; Exponent of result to DX
        CMP     DX,10000        ; Compare with bias
        JAE     POS_VALUE       ; Go if positive exponent
; Positive exponent is already encoded in result buffer
; Encode negative exponent (0001B) into BCD12 number
; EDI --> sign field
        MOV     AL,BYTE PTR [EBX]        ; Get sign digit into AL
        MOV     AH,00000001B    ; Code for negative exponent
        OR      AL,AH           ; OR exponent sign into byte
        MOV     BYTE PTR [EBX],AL        ; Store in buffer
; Calculate absolute value of negative exponent
        MOV     AX,10000        ; Bias to AX
        SUB     AX,DX           ; Bias minus exponent
        XCHG    AX,DX           ; Move exponent to DX
        JMP     CALCULATE_VALUE ; Go to conversion routine
;**********************|
;  calculate value of  |
;  positive exponent   |
;**********************|
; DX = exponent in bias 10000 form
POS_VALUE:
        MOV     AX,10000        ; Bias to AX
        SUB     DX,AX           ; Exponent minus bias
CALCULATE_VALUE:
; DX = absolute value of exponent (in binary)
        MOV     AX,DX           ; For conversion routine
        PUSH    EBX             ; Save pointer for result
        MOV     redx,EDX        ; Register to variable
        CALL    BIN_TO_BCD      ; Conversion routine
        POP     EBX             ; Restore pointer
; DH and DL holds 4 packed BCD digits
        INC     EBX             ; Bump pointer to exponent filed
        MOV     BYTE PTR [EBX],DH        ; Store two digits
        INC     EBX
        MOV     BYTE PTR [EBX],DL        ; Store two other digits
}
}


void TRIAL_QUOT()
{
_asm
{
// Obtain larger possible trial quotient for division
// On entry:
//          nothing
// On exit:
//          AL = trial quotient
//
        LEA     EBX,DIVIDEND    ; Pointer to dividend buffer
        MOV     AL,BYTE PTR [EBX]  ; First dividend digit to AL
        INC     EBX             ; Pointer to next dividend digit
        MOV     CH,BYTE PTR [EBX]      ; Second dividend digit to CH
        MOV     DL,BYTE PTR DIVISOR   ; High order divisor digit to DL
; Convert dividend digits (in AL-CH) to binary by multiplying the
; first digit by 10 and adding the second digit
        MOV     CL,10           ; Multiplier
        MUL     CL              ; First digit (in AL) times 10
        ADD     AL,CH           ; Add in second digit (in CH)
; AL now holds the binary dividend
; Trial quotient is obtained by dividing by the high order divisor
; digit (in DL)
        MOV     AH,0            ; Clear high nibble
        DIV     DL              ; Dividend / divisor = quotient
; The quotient must be a single digit. Test for a trial quotient
; larger than 9 and adjust
        CMP     AL,9            ; Test for limit value
        JA      ADJUST_DIGIT    ; Go is larger than 9
; At this point there is a valid trial quotient digit in AL
        JMP     TRIAL_EXIT
ADJUST_DIGIT:
        MOV     AL,9            ; Use largest single digit
TRIAL_EXIT:
}
}

void TRANSFER_CS_CX()
{
_asm
{
// Transfer ECX bytes from buffer (by ESI) to another one
// (by EDI)
// On entry:
//         ESI --> source buffer for move
//         EDI --> destination buffer for move
//         ECX = number of bytes to move
// On exit:
//         EDI --> moved string

        MOV     ESI,resi
		MOV     EDI,redi
		MOV     ECX,recx
        PUSH    ESI             ; Save entry registers
        PUSH    EDI
        PUSH    ECX
MOVE_TO_CS:
        MOV     AL,BYTE PTR [ESI]  ; Get byte from source buffer
        MOV     BYTE PTR [EDI],AL  ; and place it in the destination
        INC     ESI             ; Bump buffer pointers
        INC     EDI
        LOOP    MOVE_TO_CS
        POP     ECX             ; Restore caller's registers
        POP     EDI
        POP     ESI
}
}


void DIGIT_TO_QUOT()
{
_asm
{
// Place digit (in DL) in QUOTIENT buffer (if not full)
// On entry:
//			redx = quotient digit to store
// On exit:
//          BUF_FULL = 1 if QUOTIENT buffer full
//
        MOV     EDX,redx 
        PUSH    EBX             ; Save entry registers
        PUSH    EAX
        LEA     EBX,QUOTIENT    ; Pointer to destination buffer
		MOV     AX,DIGS         ; DIGS is buffer offset
		AND     EAX,0FFFFH
        ADD     EBX,EAX
; Test for end of QUOTIENT buffer
        CMP     BYTE PTR [EBX],0FFH   ; Marker for end of buffer
        JE      QUOT_BUF_END    ; Go if at end
; Place digit in buffer and index pointer
        MOV     BYTE PTR [EBX],DL      ; Store digit
        INC     DIGS            ; Bump offset counter
;**********************|
;    normal exit       |
;**********************|
        POP     EAX             ; Restore registers
        POP     EBX
        MOV     BUF_FULL,0      ; Signal buffer not full
        JMP     DIG_QUO_EXIT
;**********************|
;  buffer full exit    |
;**********************|
QUOT_BUF_END:
        POP     EAX             ; Restore entry registers
        POP     EBX
        MOV     BUF_FULL,1      ; Signal buffer full to caller
DIG_QUO_EXIT:
}
}


void COMPARE_DI_DR()
{
_asm
{
// Compare dividend to divisor
// On entry:
//          nothing
// On exit:
//          LARGER = 1 if dividend =< divisor
//          LARGER = 0 if dividend > divisor
//
        LEA     EDI,DIVIDEND
        LEA     ESI,DIVISOR
; If first digit of divided is not zero then it must be larger than
; the divisor
        CMP     BYTE PTR [EDI],00; Test for zero digit
        JNE     LARGER_DIVIDEND ; Dividend is larger
        INC     EDI             ; Bump pointer to next digit
COMP_DI_DR:
        MOV     AL,BYTE PTR [ESI]   ; Divisor digit
        CMP     AL,0FFH         ; Test for end of divisor
        JE      LARGER_DIVIDEND ; Dividend is larger exit
        MOV     AH,BYTE PTR [EDI]      ; Get dividend digit to AH
; Compare dividend and divisor digits
        CMP     AH,AL           ; Operation is dividend - divisor
        JB      LARGER_DIVISOR  ; Go if divisor > dividend
        JA      LARGER_DIVIDEND ; or if divisor < dividend
; At this point divisor = dividend. Continue comparing digits
        INC     ESI             ; Bump pointers
        INC     EDI
        JMP     COMP_DI_DR
;**********************|
;  dividend > divisor  |
;**********************|
LARGER_DIVIDEND:
        MOV     LARGER,0        ; Dividend is larger 
        JMP     LAR_DIV_EXIT
;**********************|
;  divisor > dividend  |
; or divisor = dividend|
;**********************|
LARGER_DIVISOR:
        MOV     LARGER,1        ; Divisor is larger
LAR_DIV_EXIT:
}
}

void CLEAR_BUF()
{
_asm
{
// Clear bytes in buffer
// On entry:
//         rebx --> buffer area
//         recx = number of bytes to clear
// On exit:
//         Buffer bytes = 0
//         Buffer terminator = FFH
        MOV     EBX,rebx
		MOV     ECX,recx
		PUSH    EBX
		PUSH    ECX
CLEAR_BYCS:
        MOV     BYTE PTR [EBX],0H       ; Clear one byte
        INC     EBX                     ; Bump pointer
        LOOP    CLEAR_BYCS
; Place FFH marker at end of buffer
        POP     ECX
		POP     EBX
		MOV     BYTE PTR [EBX+ECX],0FFH
}
}
