//************************************************************
//              Support functions module
//************************************************************
// Reference:   SOFTWARE SOLUTIONS - Chapters 3/5
// This file:   support.cpp
// Header:      support.h
// Development: Visual C++ 6.0 Console Application
// Date:        January 2001
//
// Description:
//              Function module with support functions
// Note:
//              Some functions uses inline assembly
//***********************************************************
#include "stdafx.h"
#include <iostream.h>




void IdMathUnit(int *userCode)
{

// Local data
unsigned short CONTROL_87 = 0; // Storage for control word
unsigned short STATUS_87 = 0;  // Storage for status word

_asm
{
; Determine if there is a mathematical coprocessor installed and
; if it is an 8087, 80287, 80387, or the math unit of a 486 or
; Pentium
; On exit:
;         userCode = 0 if no coprocessor present
;                    1 if 8087
;                    2 if 80287
;                    3 if 80387
;                    4 if math unit of 486 or 486SX / 487 system
;                    5 if math unit of Pentium
;****************************************************************

        FNINIT          ; Initialize coprocessor (if present)
; Note that the no-wait form must be used to prevent a wait
; forever condition if a coprocessor is not present
        MOV     AX,5A5AH        ; Value to set in status word
        MOV     STATUS_87,AX
        FNSTSW  STATUS_87       ; Store status word
        MOV     AX,STATUS_87    ; Read status into AX
        CMP     AL,0            ; Test for no status bits
        JE      CHK_CONTROL     ; Go if 0
; At this point no FPU unit is detected in system
        MOV     AX,0           ; Return code for no coprocessor
        JMP     EXIT_FPU
; A secondary test is based on the FPU control word
CHK_CONTROL:
        FNSTCW  CONTROL_87      ; Store control word
        MOV     AX,CONTROL_87   ; Read into AX
        CMP     AL,0            ; Test for no status bits
        AND     AX,103FH        ; Bit mask is
                                ; AH = 0001 0000
                                ; AL = 0011 1111
        CMP     AX,003FH        ; Test for AL bits unchanged
        JE      SYS_80X87       ; 80x87 is present
; At this point no FPU unit is detected in system
        MOV     AX,0           ; Return code for no coprocessor
        JMP     EXIT_FPU
;**********************|
;  coprocessor present |
;**********************|
SYS_80X87:
; The first test for the coprocessor type can be based on the CPU
; type installed in the host
; determine the CPU type, as follows:
; AX =       1 if CPU is 8086 or 8088
;            2 if CPU is 80286
;            3 if CPU is 80386
;            4 if CPU is 486
;            5 if CPU is Pentium
;***********************|
;   test for 8086/8088  |
;***********************|
; Bits 12 to 15 in the flag register are always set in the 8086
; and 8088 CPU
        PUSHF                   ; Flag register to stack
        POP     AX              ; Store flags in AX
        AND     AX,0FFFH        ; Clear bits 12 to 15
        PUSH    AX              ; AX to stack
        POPF                    ; and to flags register
        PUSHF                   ; Flags to stack
        POP     AX              ; and to AX for reading
        AND     AX,0F000H       ; Preserve bits 12 to 15
        CMP     AX,0F000H       ; Test for bits set
        JNE     TEST_286        ; Go if bits not set
; At this point processor is a 8086 or 8088
        MOV     AX,1            ; Return code
        JMP     ID_EXIT         ; Exit
;***********************|
;     test for 80286    |
;***********************|
; Bits 12 to 15 in the flag register are always clear in the Intel
; 80286 CPU
TEST_286:
        PUSHF                   ; Flag register to stack
        POP     BX              ; Store flags in BX
        OR      BX,0F000H       ; Make sure bit field is set
        PUSH    BX              ; To stack
        POPF                    ; And to flag register
        PUSHF                   ; Flags to stack
        POP     AX              ; And to AX
        AND     AX,0F000H       ; Clear all other bits
        JNZ     TEST_386        ; Go if bits not clear
; At this point processor is an 80286
        MOV     AX,2            ; Return code
        JMP     ID_EXIT         ; Exit
;***********************|
;     test for 80386    |
;***********************|
; Bit 18 of the E flags register was introduced in the 486 CPU
; This bit cannot be set in the 80386
TEST_386:
        PUSHFD                  ; 32-bits E flags to stack
        POP     EAX             ; Flags to EAX
        OR      EAX,40000H      ; Make sure bit 18 is set
        PUSH    EAX             ; New flags to stack
        POPFD                   ; An to E flags register
        PUSHFD                  ; Back to stack
        POP     EAX             ; And to EAX
        AND     EAX,40000H      ; Clear all except bit 18
        JNZ     TEST_486        ; Go if bit 18 is clear
; At this point processor is a 80386
        MOV     AX,3            ; Return code
        JMP     ID_EXIT         ; Exit
;***********************|
;     test for 486      |
;***********************|
; Bit 21 (ID flag) of the E flags register cannot be set in the
; 486
TEST_486:
        PUSHFD                  ; 32-bits E flags to stack
        POP     EAX             ; Flags to EAX
        OR      EAX,200000H     ; Make sure bit 21 is set
        PUSH    EAX             ; New flags to stack
        POPFD                   ; An to E flags register
        PUSHFD                  ; Back to stack
        POP     EAX             ; And to EAX
        AND     EAX,200000H     ; Clear all except bit 21
        JNZ     IS_PENTIUM      ; Go if bit 21 is clear
; At this point processor is a 486
        MOV     AX,4            ; Return code
        JMP     ID_EXIT         ; Exit
;***********************|
;  processor is PENTIUM |
;***********************|
IS_PENTIUM:
        MOV     AX,5            ; Return code
ID_EXIT:
; At this point AL holds CPU ID code as follows:
;           AL = 1 if FPU is 8087
;           AL = 2 if FPU is 80287
;           AL = 3 if FPU is 80287 or 80387 since the 80287 can
;                be installed in an 80386 system
;           AL = 4 if FPU is 486 math unit
;           AL = 5 if FPU is Pentium math unit
;
; If return code is not 3 then there is no uncertainty regarding
; the FPU type
        CMP     AX,3            ; Test for undefined case
        JE      IS_386          ; Go if 386 CPU
; If not 386, return with coprocessor code in AL
        JMP     EXIT_FPU  
;**********************|
;    80287 or 80387    |
;        system        |
;**********************|
; The test for an 80287 system is based on the state of the
; condition code bits after FINIT. In the 80387 C0 to C3 bits
; are zero after FINIT, in the 80287 they contain the previous
; value
; Note: at this time the wait form of FPU opcodes can be used
;       since the presence of a coprocessor is certain
IS_386:
        FINIT           ; Make sure stack is EMPTY
        FXAM            ; C0 and C3 are set after FXAM if
                        ; the stack top is EMPTY
        FINIT           ; In an 80287 C0 and C3 will remain
                        ; set after the second FINIT
        FSTSW   STATUS_87    ; Status word to memory
        FWAIT                ; Synchronize CPU and FPU
        MOV     AX,STATUS_87 ; Status word to AX register
        AND     AH,01000001B ; Clear all bits except C3
                             ; and C0
        JNZ     TYPE_287     ; Go if bits not clear
;**********************|
;     80387 or 486     |
;**********************|
        MOV     AX,3         ; Return code for 80387
        JMP     EXIT_FPU
;**********************|
;        80287         |
;**********************|
TYPE_287: 
        MOV     AX,2         ; Return code for 80287
EXIT_FPU:
		AND     EAX,0FFFFH
        MOV     ESI,userCode
		MOV     [ESI],EAX
}
}

