//************************************************************
//           Math unit input and output functions
//************************************************************
// Reference:   SOFTWARE SOLUTIONS
// Workspace  : Chapter8\Test Un32_5
// Entry point: Test Un32_5
// Other files: Un32_5.lib
//              Un32_5.h
// Development: Visual C++ 6.0 Console Application
// Date:        January 12/2001
//
// Description:
//              Header file for conversion testing functions
//              in the Un32_5 module of math32.lib
//************************************************************
// Library functions
extern "C" TEN_TO_Y_BYFAC();//10**y by exponent factoring
extern "C" X_TO_Y_BYFAC();  // x**y by exponent factoring
extern "C" X_TO_Y_BYBP();   // x**y by binary powering
extern "C" X_TO_Y_BYLOG();  // x**y by logarithms
extern "C" X_TO_Y_MIXED();  // x**y by logs and binary powering
extern "C" LOG_10();        // base 10 logarithm
extern "C" LOG_E();         // base e logarithm
extern "C" ALOG_10();       // base 10 antilogarithm
extern "C" ALOG_E();        // base e antilogarithm

//**************************************************************
//                       Prototypes
//**************************************************************
unsigned long TenToYByFac(int);  // 10^y by exponent factoring
double XtoYByFac(double, int);   // x^y by exponent factoring
double XtoYByBP(double, double); // x^y by binary powers
double XtoYByLog(double, double);// x^y by loagarithms 
double XtoYMixed(double, double);// x^y by logs and binary power
double LogE(double);             // Natural logarithm
double Log10(double x);          // Common logarithm 
double AlogE(double x);          // Natural antilogarithm 
double Alog10(double x);         // Common antilogarithm
double Alog2(double x);          // Base 2 logarithm

//**************************************************************
//            C++ interface functions to Un32_5 module
//**************************************************************
unsigned long TenToYByFac(int y)
{
//Exact calculation of 10**y by exponent factorization 
// On entry:
//     parameter y is positive exponent of 10
// On exit:
//     returns 10**y

unsigned long power10 = 0;
_asm
{
	FILD    y
	CALL    TEN_TO_Y_BYFAC
	FISTP   power10
}
  return power10;  
}

double XToYByFac(double x, unsigned int y)
{
//Exact calculation of x**y by exponent factorization 
// On entry:
//     parameter x is base
//     parameter y is the positive integer exponent
// On exit:
//     returns x**y

double xtoy = 0;
_asm
{
	FLD     x
    FILD    y
	CALL    X_TO_Y_BYFAC
	FSTP    xtoy
}
  return xtoy;  
}

double XToYByBP(double x, unsigned int y)
{
//Exact calculation of x**y by binary powers 
// On entry:
//     parameter x is base
//     parameter y is the positive integer exponent
// On exit:
//     returns x**y

double xtoy = 0;
_asm
{
	FLD     x
    FILD    y
	CALL    X_TO_Y_BYBP
	FSTP    xtoy
}
  return xtoy;  
}

double XToYByLog(double x, double y)
{
// Approximate calculation of x**y by logarithms
// On entry:
//     parameter x is base
//     parameter y is the exponent
// On exit:
//     returns x**y

double xtoy = 0;
_asm
{
	FLD     x
    FLD     y
	CALL    X_TO_Y_BYLOG
	FSTP    xtoy
}
  return xtoy;  
}

double XToYMixed(double x, double y)
{
// Approximate calculation of x**y by logarithms
// and binary powering
// On entry:
//     parameter x is base
//     parameter y is the exponent
// On exit:
//     returns x**y

double xtoy = 0;
_asm
{
	FLD     x
    FLD     y
	CALL    X_TO_Y_MIXED
	FSTP    xtoy
}
  return xtoy;  
}


double LogE(double x)
{
// Logarithm base E
// On entry:
//     parameter x number
// On exit:
//     returns Log E of x
double logx;
_asm
{
	FLD     x
	CALL    LOG_E
	FSTP    logx
}
  return logx;  
}

double Log10(double x)
{
// Logarithm base 10
// On entry:
//     parameter x number
// On exit:
//     returns Log 10 of x
double logx;
_asm
{
	FLD     x
	CALL    LOG_10
	FSTP    logx
}
  return logx;  
}

double AlogE(double x)
{
// Antilogarithm base E
// On entry:
//     parameter x is logarithm base E
// On exit:
//     returns number whose log is x
double numx;
_asm
{
	FLD     x
	CALL    ALOG_E
	FSTP    numx
}
  return numx;  
}


double Alog10(double x)
{
// Antilogarithm base 10
// On entry:
//     parameter x is logarithm base 10
// On exit:
//     returns number whose log is x
double numx;
_asm
{
	FLD     x
	CALL    ALOG_10
	FSTP    numx
}
  return numx;  
}
