//************************************************************
//           Math unit trigonometric functions
//************************************************************
// Reference:   SOFTWARE SOLUTIONS
// Workspace  : Chapter8\Test Un32_6
// Entry point: Test Un32_6
// Other files: Un32_6.asm
//              Un32_6.h
// Development: Visual C++ 6.0 Console Application
// Date:        January 12/2001
//
// Description:
//              Header file for testing trigonometric 
//              functions in the Un32_6 module of
//              math32.lib
//************************************************************
// Library functions
extern "C" DEG_TO_RAD();    // degress to radian conversion
extern "C" RAD_TO_DEG();    // radians to degrees conversion
extern "C" TANGENT();       // tangent
extern "C" SINE();          // sine
extern "C" COSINE();        // cosine
extern "C" COTANGENT();     // cotangent
extern "C" SECANT();        // secant
extern "C" COSECANT();      // cosecant
extern "C" ARC_SINE();      // arcsine
extern "C" ARC_COSINE();    // arccosine
extern "C" ARC_TANGENT();   // arctangent
extern "C" SINH();          // hyperbolic sine
extern "C" COSH();          // hyperbolic cosine
extern "C" TANH();          // hyperbolic tangent
extern "C" ARC_SINH();      // hyperbolic arcsine
extern "C" ARC_COSH();      // hyperbolic arcosin
extern "C" ARC_TANH();      // hyperbolic arctangent
//**************************************************************
//                       Prototypes
//**************************************************************
double DegToRad(double);
double RadToDeg(double);
double Tangent(double);
double Sine(double);
double Cosine(double);
double CoTangent(double);
double Secant(double);
double CoSecant(double);
double ArcSine(double);
double ArcCosine(double);
double ArcTangent(double);
double Sinh(double);
double Cosh(double);
double Tanh(double);
double AscSinh(double);
double ArcCosh(double);
double ArcTanh(double);
//**************************************************************
//            C++ interface functions to Un32_6 module
//**************************************************************
double DegToRad(double degs)
// Convert degrees to radian measure
// On entry:
//         parameter degs = value in degrees
// On exit:
//         returns radians
{
double rads; 
_asm
{
    FLD     degs
	CALL    DEG_TO_RAD
    FSTP    rads
}
return rads;
}

double RadToDeg(double rads)
// Convert radians to degrees
// On entry:
//         parameter rads = value in radian measure
// On exit:
//         returns degrees
{
double degs; 
_asm
{
    FLD     rads
	CALL    RAD_TO_DEG
    FSTP    degs
}
return degs;
}

double Tangent(double angle)
// Compute tangent of an angle in radians using the 80387 or
// the math unit of the 486 and Pentium.
// (no restrictions on range or sign)
//
// On entry:
//       parameter angle is an angle in radian
//       (can be positive or negative)
// On exit:
//       returns tangent of angle
{
double result;
_asm
{
	FLD      angle
	CALL     TANGENT
	FSTP     result
}
return result;
}
 
double Sine(double angle)
// Compute sine of an angle in radians using the 80387 or
// the math unit of the 486 and Pentium.
// (no restrictions on range or sign)
//
// On entry:
//       parameter angle is an angle in radian
//       (can be positive or negative)
// On exit:
//       returns sine of angle
{
double result;
_asm
{
	FLD      angle
	CALL     SINE
	FSTP     result
}
return result;
}

double Cosine(double angle)
// Compute cosine of an angle in radians using the 80387 or
// the math unit of the 486 and Pentium.
// (no restrictions on range or sign)
//
// On entry:
//       parameter angle is an angle in radian
//       (can be positive or negative)
// On exit:
//       returns cosine of angle
{
double result;
_asm
{
	FLD      angle
	CALL     COSINE
	FSTP     result
}
return result;
}  


double CoTangent(double angle)
// Compute cotangent of an angle in radians using the 80387 or
// the math unit of the 486 and Pentium.
// (no restrictions on range or sign)
//
// On entry:
//       parameter angle is an angle in radian
//       (can be positive or negative)
// On exit:
//       returns cotangent of angle
{
double result;
_asm
{
	FLD      angle
	CALL     COTANGENT
	FSTP     result
}
return result;
}

double Secant(double angle)
// Compute secant of an angle in radians using the 80387 or
// the math unit of the 486 and Pentium.
// (no restrictions on range or sign)
//
// On entry:
//       parameter angle is an angle in radian
//       (can be positive or negative)
// On exit:
//       returns secant of angle
{
double result;
_asm
{
	FLD      angle
	CALL     SECANT
	FSTP     result
}
return result;
}

double CoSecant(double angle)
// Compute cosecant of an angle in radians using the 80387 or
// the math unit of the 486 and Pentium.
// (no restrictions on range or sign)
//
// On entry:
//       parameter angle is an angle in radian
//       (can be positive or negative)
// On exit:
//       returns cosecant of angle
{
double result;
_asm
{
	FLD      angle
	CALL     COSECANT
	FSTP     result
}
return result;
}

double ArcTangent(double trigFun)
// Computes the angle in radian corresponding to a tangent function
// using the 80387 or the math unit of the 486 and Pentium.
// (no restrictions on range or sign)
//
// On entry:
//       parameter trigFun is a tangent function
// On exit:
//       returns the angle in radian
{
double result;
_asm
{
	FLD      trigFun
	CALL     ARC_TANGENT
	FSTP     result
}
return result;
}
  
double ArcSine(double trigFun)
// Computes the angle in radian corresponding to a sine function
// using the 80387 or the math unit of the 486 and Pentium.
// (no restrictions on range or sign)
//
// On entry:
//       parameter trigFun is a sine function
// On exit:
//       returns the angle in radian
{
double result;
_asm
{
	FLD      trigFun
	CALL     ARC_SINE
	FSTP     result
}
return result;
}

double ArcCosine(double trigFun)
// Computes the angle in radian corresponding to a cosine function
// using the 80387 or the math unit of the 486 and Pentium.
// (no restrictions on range or sign)
//
// On entry:
//       parameter trigFun is a cosine function
// On exit:
//       returns the angle in radian
{
double result;
_asm
{
	FLD      trigFun
	CALL     ARC_COSINE
	FSTP     result
}
return result;
}

double Sinh(double angle)
// Compute hyperbolic sine of an angle in radians using the
// 80387 or the math unit of the 486 and Pentium.
// (no restrictions on range or sign)
//
// On entry:
//       parameter angle is an angle in radian
//       (can be positive or negative)
// On exit:
//       returns hyperbolic sine of angle
{
double result;
_asm
{
	FLD      angle
	CALL     SINH
	FSTP     result
}
return result;
}


double Cosh(double angle)
// Compute hyperbolic cosine of an angle in radians using the
// 80387 or the math unit of the 486 and Pentium.
// (no restrictions on range or sign)
//
// On entry:
//       parameter angle is an angle in radian
//       (can be positive or negative)
// On exit:
//       returns hyperbolic cosine of angle
{
double result;
_asm
{
	FLD      angle
	CALL     COSH
	FSTP     result
}
return result;
}


double Tanh(double angle)
// Compute hyperbolic tangent of an angle in radians using the
// 80387 or the math unit of the 486 and Pentium.
// (no restrictions on range or sign)
//
// On entry:
//       parameter angle is an angle in radian
//       (can be positive or negative)
// On exit:
//       returns hyperbolic tangent of angle
{
double result;
_asm
{
	FLD      angle
	CALL     TANH
	FSTP     result
}
return result;
}

double ArcSinh(double trigFun)
// Computes the angle in radian corresponding to a hyperbolic
// sine function, using the 80387 or the math unit of the 486
// and Pentium.
// (no restrictions on range or sign)
//
// On entry:
//       parameter trigFun is a hyperbolic sine function
// On exit:
//       returns the angle in radian
{
double result;
_asm
{
	FLD      trigFun
	CALL     ARC_SINH
	FSTP     result
}
return result;
}

double ArcCosh(double trigFun)
// Computes the angle in radian corresponding to a hyperbolic
// cosine function, using the 80387 or the math unit of the 486
// and Pentium.
// (no restrictions on range or sign)
//
// On entry:
//       parameter trigFun is a hyperbolic cosine function
// On exit:
//       returns the angle in radian
{
double result;
_asm
{
	FLD      trigFun
	CALL     ARC_COSH
	FSTP     result
}
return result;
}


double ArcTanh(double trigFun)
// Computes the angle in radian corresponding to a hyperbolic
// tangent function, using the 80387 or the math unit of the 486
//  and Pentium.
// (no restrictions on range or sign)
//
// On entry:
//       parameter trigFun is a hyperbolic sine function
// On exit:
//       returns the angle in radian
{
double result;
_asm
{
	FLD      trigFun
	CALL     ARC_TANH
	FSTP     result
}
return result;
}
