; Module name: BIOVIDEO.ASM                   Library:   SOLUTION.LIB
; Created:
; Update:
;
;********************************************************************
;********************************************************************
;                            BIOVIDEO.ASM
;********************************************************************
;********************************************************************
;
; Library module for video services using BIOS functions in INT 10H
; and DOS functions in INT 21H
;
;*******************|
;      PUBLICS      |
;*******************|
PUBLIC  BIO_CLEAR_SCR
PUBLIC  BIO_SET_MODE
PUBLIC  BIO_GET_MODE
PUBLIC  BIO_SET_CUR
PUBLIC  BIO_GET_CUR
PUBLIC  BIO_CUR_OFF
PUBLIC  BIO_SHOW_ATR
PUBLIC  BIO_SHOW_CHR
PUBLIC  BIO_SHOW_BLOK
PUBLIC  BIO_SCROLL_UP
PUBLIC  BIO_SCROLL_DN
PUBLIC  BIO_SHOW_MESS
PUBLIC  BIO_TTY
PUBLIC  DOS_STRING_OUT
PUBLIC  BIO_CX_VERT
;
;
P_CODE  SEGMENT PUBLIC
        ASSUME  CS:P_CODE
;
;*******************************************************************
BIO_CLEAR_SCR   PROC    FAR
; Clear the video display using BIOS Service number 6, INT 10H
; On entry:
;         nothing
; On exit:
;         carry clear
        PUSH    AX              ; Save entry registers
        PUSH    BX
        PUSH    CX
        PUSH    DX
        MOV     AH,6            ; Service request number
        MOV     AL,0            ; Code for blanking entire window
        MOV     BH,07           ; Normal attribute
        MOV     CX,0            ; Start at row # 0, column # 0
        MOV     DH,24           ; End at row # 24
        MOV     DL,79           ; Column # 79
        INT     10H             ; Transfer to BIOS
        POP     DX              ; Restore registers
        POP     CX
        POP     BX
        POP     AX
        CLC                     ; No error detection
        RET
BIO_CLEAR_SCR   ENDP
;*******************************************************************
;
BIO_SET_MODE    PROC     FAR
; Set video display mode using BIOS service number 0 of INT 10H
; On entry:
;         AL = number of vide mode to set
; On exit:
;         carry clear
;
        PUSH     AX             ; Save AX
        MOV      AH,0           ; BIOS service request number
        INT      10H
        POP      AX             ; Restore caller's register
        CLC                     ; No error detection
        RET
BIO_SET_MODE     ENDP
;*******************************************************************
;
BIO_GET_MODE     PROC      FAR
; Obtain current video mode using BIOS service number 15 of INT 10H
; On entry:
;         nothing
; On exit:
;         AH = character columns on screen (40 or 80)
;         AL = active video mode
;         BH = active video page
;         carry clear
;
        MOV     AH,15           ; BIOS service request number
        INT     10H
        CLC                     ; No error detection
        RET
BIO_GET_MODE    ENDP
;******************************************************************
;
BIO_SET_CUR     PROC    FAR
; Set system cursor to a row and column coordinate
; On entry:
;         DH = screen row (range 0 to 24)
;         DL = screen column (range 0 to 79)
; On exit:
;         carry clear if no error
;
        CALL    TEST_RANGE      ; Local error handler
        PUSH    AX              ; Save entry registers
        PUSH    BX
        PUSH    DX
        MOV     AH,2            ; Service request number
        MOV     BH,0            ; Assume display page No. 0
        INT     10H
        POP     DX              ; Restore registers
        POP     BX
        POP     AX
        CLC                     ; No error
        RET
BIO_SET_CUR     ENDP
;******************************************************************
;
BIO_GET_CUR     PROC    FAR
; Read coordinates of current cursor position and start and
; end lines of cursor size
; On entry:
;         BH = active page number (normally 0)
; On exit:
;         CH = starting line of cursor size
;         CL = ending line of cursor size
;         DH = row number of current cursor position
;         DL = column number of current cursor position
;         carry clear
;
        PUSH    AX              ; Save entry registers
        MOV     AH,3            ; Service request number
        INT     10H
        POP     AX              ; Restore AX
        CLC                     ; No error
        RET
BIO_GET_CUR     ENDP
;*******************************************************************
;
BIO_CUR_OFF     PROC    FAR
; Turn off cursor by setting the cursor address outside of the
; current display page
        PUSH    AX              ; Save entry registers
        PUSH    BX
        PUSH    DX
        MOV     DH,25           ; y coordinate out of page
        MOV     DL,0
        MOV     AH,2            ; Service request number
        MOV     BH,0            ; Assume display page No. 0
        INT     10H
        POP     DX              ; Restore registers
        POP     BX
        POP     AX
        CLC                     ; No error
BIO_CUR_OFF     ENDP
;*******************************************************************
;
BIO_SHOW_ATR    PROC    FAR
; Display character and attribute at current cursor position
; On entry:
;         AL = character
;         BL = attribute to use in displaying character
; On exit:
;         carry clear
;
        PUSH    AX              ; Save entry registers
        PUSH    BX
        PUSH    CX
        PUSH    DX
        MOV     AH,9            ; Service request number
        MOV     BH,0            ; Display page
        MOV     CX,1            ; Repetition factor
        INT     10H
        POP     DX              ; Restore caller's registers
        POP     CX
        POP     BX
        POP     AX
        CLC                     ; No error detection
        RET
BIO_SHOW_ATR    ENDP
;*******************************************************************
;
BIO_SHOW_CHR    PROC    FAR
; Display the character in AL at the current cursor position
; On entry:
;         AL = character
; On exit:
;         carry clear
;
        PUSH    AX              ; Save caller's registers
        PUSH    BX
        PUSH    CX
        MOV     AH,10           ; Service request number
        MOV     BH,0            ; Display page zero
        MOV     CX,1            ; One character
        INT     10H
        POP     CX              ; Restore caller's registers
        POP     BX
        POP     AX
        CLC                     ; No error detection
        RET
BIO_SHOW_CHR    ENDP
;*******************************************************************
;
BIO_SCROLL_UP   PROC    FAR
; Scroll up video display window using BIOS service number 6
; of INT 10H
; On entry:
;         AL = number of lines to scroll
;         CH = y coordinate of upper left corner of window
;         CL = x coordinate of upper left corner of window
;         DH = y coordinate of lower right corner of window
;         DL = x coordinate of lower right corner of window
;
; CH/CL ->|------------------------------------|
;         |                                    |
;         |    screen area to be scrolled      |
;         |                                    |
;         |------------------------------------|<- DH/DL
;
        MOV     AH,6            ; BIOS service request number
        INT     10H
        CLC                     ; No error detection
        RET
BIO_SCROLL_UP   ENDP
;****************************************************************
;
BIO_SCROLL_DN   PROC    FAR
; Scroll down video display window using BIOS service number 7
; of INT 10H
; On entry:
;         AL = number of lines to scroll
;         CH = y coordinate of upper left corner of window
;         CL = x coordinate of upper left corner of window
;         DH = y coordinate of lower right corner of window
;         DL = x coordinate of lower right corner of window
;
; CH/CL ->|------------------------------------|
;         |                                    |
;         |    screen area to be scrolled      |
;         |                                    |
;         |------------------------------------|<- DH/DL
;
        MOV     AH,7        ; BIOS service request number
        INT     10H
        CLC                 ; No error detection
        RET
BIO_SCROLL_DN   ENDP
;****************************************************************
;
BIO_SHOW_BLOK   PROC    FAR
; Display a preformatted block message using BIOS service number 9
; of INT 10H
; On entry:
;          SI --> start of block message
; Message format:
;       Offset:         Function:
;              0 ------ Display row for start of message
;              1 ------ Display column for start of message
;              2 ------ Attribute for block
; Embedded codes:
;              00H = End of message
;              FFH = End of row
; On exit:
;        carry clear
;
        MOV     DH,[SI]         ; Get row
        INC     SI              ; Point to start column byte
        MOV     DL,[SI]         ; Column
        MOV     CL,DL           ; Save start column in CL
        CALL    BIO_SET_CUR     ; Set cursor to message start
        INC     SI              ; Point to attribute byte
; Get attribute into BL
        MOV     BL,[SI]         ; SI --> attribute
        INC     SI
;
GET_AND_SHOW:
        MOV     AL,[SI]         ; Character
;*********************|
;   test for control  |
;      characters     |
;*********************|
        CMP     AL,00H          ; Terminator
        JE      SHOW_END
        CMP     AL,0FFH         ; End of line
        JE      LINE_END
; Display it
        CALL    BIO_SHOW_CHR
        INC     DL              ; Next column
;*********************|
;    bump cursor      |
;*********************|
BUMP_PTRS:
        CALL    BIO_SET_CUR
        INC     SI              ; Bump message pointer
        JMP     GET_AND_SHOW
;********************|
;    end of line     |
;********************|
LINE_END:
        INC     DH              ; Bump row pointer
        MOV     DL,CL           ; Reset column to start column
        JMP     BUMP_PTRS
SHOW_END:
        CLC                     ; No error detection
        RET
BIO_SHOW_BLOK   ENDP
;*******************************************************************
;
BIO_SHOW_MESS   PROC    FAR
; Display a string terminated in a NULL byte
; On entry:
;         SI --> message terminated in 00H
; On exit:
;         carry clear
;         SI points to message terminator
;         cursor bumped once for each character displayed
;
        PUSH    AX              ; Save caller's register
        PUSH    BX
SHOW_MS:
        MOV     AL,[SI]         ; Get character
        CMP     AL,0            ; Test for terminator
        JE      END_OF_MESS
        MOV     AH,14           ; BIOS service request number
                                ; for ASCII teletype write
        MOV     BX,0            ; Display page
        INT     10H             ; BIOS service request
        INC     SI              ; Bump pointer
        JMP     SHOW_MS
END_OF_MESS:
        POP     BX              ; Restore caller's register
        POP     AX
        CLC                     ; No error detection
        RET
BIO_SHOW_MESS   ENDP
;*******************************************************************
;
BIO_TTY         PROC    FAR
; Display character or control code at cursor position using BIOS
; teletype service number 14, INT 10H
; On entry:
;         AL = character or control code
;              Recognized control codes are:
;              07H = bell (beep speaker)
;              08H = backspace
;              0AH = line feed
;              0DH = carriage return
; On exit:
;         carry clear
;         cursor bumped to next position
;
        PUSH    AX              ; Save character
        PUSH    BX
        MOV     AH,14           ; BIOS service request number
                                ; for ASCII teletype write
        MOV     BX,0            ; Display page
        INT     10H             ; BIOS service request
        POP     BX              ; Restore caller's registers
        POP     AX
        CLC                     ; No error detection
        RET
BIO_TTY         ENDP
;*******************************************************************
;
DOS_STRING_OUT  PROC    FAR
; Display a string of characters terminated in a $ sign using
; DOS service number 9 of INT 21H
;
; On entry:
;         DX --> start of string
; On exit:
;         carry clear

        MOV     AH,9
        INT     21H
        CLC                     ; No error detection
        RET
DOS_STRING_OUT  ENDP
;*******************************************************************
;
BIO_CX_VERT     PROC    FAR
; Display CX number of characters and attributes vertically using
; a BIOS service
; On entry:
;         AL = character to display
;         AH = attribute (if AH = 0 then present attribute is used)
;         DH = screen row
;         DL = screen column
;         CX = number of characters to repeat
; On exit:
;        carry clear
;
        CALL    BIO_SET_CUR
        CMP     AH,0            ; Test attribute unchanged code
        JE      KEEP_ATTR
CHANGE_ATTR:
        CALL    BIO_SHOW_ATR
        INC     DH              ; Next screen row
        CALL    BIO_SET_CUR     ; Move cursor
        LOOP    CHANGE_ATTR     ; Continue
        RET
KEEP_ATTR:
        CALL    BIO_SHOW_CHR
        INC     DH              ; Next screen row
        CALL    BIO_SET_CUR     ; Move cursor
        LOOP    KEEP_ATTR
        RET
BIO_CX_VERT     ENDP
;
;*******************************************************************
;                         internal procedures
;*******************************************************************
TEST_RANGE      PROC    NEAR
; Procedure to test that the screen address passed in the DH and DL
; registers is a valid one
; On entry:
;         DH = screen row (legal range is 0 to 24)
;         DL = screen column (legal range is 0 to 79)
; On exit:
;        carry set if error. AX destroyed. Returns to original caller
;        nothing if no error
        CMP     DH,24           ; Test for valid input
        JG      BAD_CR          ; Invalid value
        CMP     DL,79           ; Maximum column value
        JG      BAD_CR          ; Invalid value
; Input is in legal range
        RET                     ; Return
BAD_CR:
        POP     AX              ; Adjust stack
        STC                     ; Error flag on
        RET                     ; Return to original caller
TEST_RANGE      ENDP
;
P_CODE  ENDS
        END
