; Module name: CONVERT.ASM                    Library:   SOLUTION.LIB
; Created:
; Update:
;
;********************************************************************
;********************************************************************
;                             CONVERT.ASM
;********************************************************************
;********************************************************************
;
; Library module for numeric conversions
;
;*******************|
;      PUBLICS      |
;*******************|
PUBLIC	BIN_TO_ASC
PUBLIC	ASC_TO_BIN
PUBLIC	BIN_TO_ASCHEX
PUBLIC	BIN_TO_ASCBIN
;
P_CODE	SEGMENT PUBLIC
	ASSUME	CS:P_CODE
;
;*******************************************************************
BIN_TO_ASC	PROC	FAR
;
; Convert a 16-bit binary in the DX register to an ASCII decimal
; number
; On entry:
;         DX = binary source
;         DI --> 5-byte output buffer
; On exit:
;         DI --> start of 5-byte buffer holding the ASCII decimal
;                number
;         carry clear
;
	PUSH	DI		; Save start of buffer
	CALL	CLEAR_5		; Clear user's buffer
; DI --> last byte in user's buffer
	MOV	CX,0		; Clear counter
BINA0:	PUSH	CX		; Save count
	MOV	AX,DX		; Add in numerator
	MOV	DX,0		; Clear top half
	MOV	CX,10		; Enter decimal divisor
	DIV	CX		; Perform division AX/CX
	XCHG	AX,DX		; Get quotient
	ADD	AL,30H		; Make digit ASCII
	MOV	[DI],AL		; Store digit in buffer
	DEC	DI		; Bump destination pointer
	POP	CX		; Restore counter
	INC	CX		; Count the digit
	CMP	DX,0		; Test for end of binary
	JNZ	BINA0		; Continue if not end
;
	POP	DI		; Restore pointer to start of
				; buffer
	CLC			; No error detection  
	RET
BIN_TO_ASC	ENDP
;*******************************************************************
;
ASC_TO_BIN	PROC	FAR
;
; Convert an ASCII number to binary
; On entry:
;         BX --> start of 5 digit ASCII buffer
; On exit:
;         DX = binary number (range 0 to 65535)
;         carry clear
;
	MOV	DX,0		; Clear binary output
ASC0:	MOV	AL,[BX]		; Get ASCII digit
	INC	BX		; Bump pointer to next digit
	SUB	AL,30H		; ASCII to decimal
	JL	EXASC		; Exit if invalid ASCII
	CMP	AL,9		; Test for highest value
	JG	EXASC		; Exit if larger than 9
	CBW			; Extend to AX
	PUSH	AX		; Save digit in stack
	MOV	AX,DX		; Move into output register
	MOV	CX,10		; Decimal multiplier
	MUL	CX		; Perform AX = AX * 10
	MOV	DX,AX		; Move product to output register
	POP	AX		; Restore decimal digit
	ADD	DX,AX		; Add in digit
	JMP	ASC0		; Continue
EXASC:	CLC			; No error detection
	RET
;
ASC_TO_BIN	ENDP
;******************************************************************
;
BIN_TO_ASCHEX	PROC	FAR
; Convert a 16-bit binary in the DX register to an ASCII
; hexadecimal number
; On entry:
;         DX = binary source
;         DI --> 5-byte output buffer
; On exit:
;        DI --> start of 5-byte buffer. The last character
;               in the buffer is the letter H
;        carry clear
;
	PUSH	DI		; Save user's buffer pointer
	PUSH	DI		; twice
	CALL	CLEAR_5
	MOV	BYTE PTR[DI],'H'     ; Place H in buffer
	POP	DI		; Restore buffer pointer
	MOV	CX,4		; Count 4 digits
HEX_CVR:
	PUSH	CX		; Save counter
	MOV	CL,4		; 4 digits
	ROL	DX,CL		; Rotate source left
	MOV	AL,DL		; Move digit into AL
	AND	AL,00001111B	; Clear high nibble
	DAA			; Adjust AL if A through F
	ADD	AL,11110000B	; and bump the carry
	ADC	AL,40H		; Convert hex to ASCII
	MOV	[DI],AL		; Store in buffer
	INC	DI		; Bump buffer pointer
	POP	CX		; Restore digit counter
	LOOP	HEX_CVR		; Continue with next digit
; Converted number is now in buffer
	POP	DI		; Restore caller's pointer
	CLC			; No error detection
	RET
BIN_TO_ASCHEX	ENDP
;******************************************************************
;
BIN_TO_ASCBIN	PROC	FAR
; Convert a 8-bit binary number in DL into 8 ASCII binary digits
; On entry:
;         DL = 8-bit binary number
;         DI --> 9-byte output buffer
; On exit:
;        DI --> start of 9-byte buffer. The last character
;               in the buffer is the letter B
;        carry clear
;
	PUSH	DI		; Save user's buffer pointer
	PUSH	DI		; twice
	MOV	CX,9		; Nine digits to clear
CLEAR_9:
	MOV	BYTE PTR [DI],20H	; Clear digit
	INC	DI		; Bump pointer
	LOOP	CLEAR_9		; Repeat for 9 digits
	DEC	DI		; Adjust buffer pointer to last
				; digit
	MOV	BYTE PTR[DI],'B'     ; Place B in buffer
	POP	DI		; Start of buffer
	MOV	CX,8		; Count 8 digits
; DL has binary value to convert to ASCII
BIN_CVR:
	TEST	DL,10000000B	; Test high-order bit of DL
	JNZ	SET_1		; Bit is set
	MOV	BYTE PTR[DI],'0'
	JMP	NEXT_BIN
SET_1:
	MOV	BYTE PTR[DI],'1'
NEXT_BIN:
	INC	DI		; Bump pointer
	SAL	DL,1		; Shift bits left once
	LOOP	BIN_CVR
; Converted number is now in buffer
	POP	DI		; Restore caller's pointer
	CLC			; No error detection
	RET
BIN_TO_ASCBIN	ENDP
;
;*******************************************************************
;                        local procedures
;*******************************************************************
CLEAR_5		PROC	NEAR
; On entry:
;         DS:DI --> start of a 5-byte user buffer
; On exit:
;         DS:DI --> last byte in the 5-byte buffer
	MOV	CX,5		; Five digits to clear
CLEAR_BUF:
	MOV	BYTE PTR [DI],20H	; Clear digit
	INC	DI		; Bump pointer
	LOOP	CLEAR_BUF	; Repeat for 5 digits
	DEC	DI		; Adjust buffer pointer to last
				; digit
	RET
CLEAR_5		ENDP
;*******************************************************************
;
P_CODE	ENDS
	END

				; digit
	RET
CLEAR_5		ENDP
;*****************