;*****************************************************************
;*****************************************************************
;                           UNIT_1.ASM
;*****************************************************************
;*****************************************************************
; Library of math routines from the book SOFTWARE SOLUTIONS by
; Julio Sanchez and Maria P. Canton
; Creation date:  January 13, 1994
; Updated:        January 25, 1994
;
; Libary name: MATHUNIT.LIB
;
;*****************************************************************
;                           publics
;*****************************************************************
PUBLIC BIN_TO_ASC10  ; .. 16-bit binary into 5-digit ASCII decimal
PUBLIC BIN_TO_ASC16  ; .. 16-bit binary into 4-digit ASCII hex
PUBLIC ASCII_TO_DX   ; .. 4-digit ASCII decimal to binary in DX
PUBLIC BINF_TO_DECF  ; .. 8-bit binary to 8-digit decimal fraction
PUBLIC DECODE_SINGLE ; .. Isolate elements of single precision
PUBLIC BIN_TO_BCD    ; .. 16-bit binary to 5 BCD digits
PUBLIC BCD_TO_ASCII  ; .. 5 unpacked BCD to 5 ASCII digits
PUBLIC BCD_INTO_REG  ; .. 5-digit BCD into a binary in DX
PUBLIC BCD12_TO_ASCSTR;.. BCD12 number in ASCII string
PUBLIC ASCII_TO_BCD12 ;.. ASCII decimal string into BCD12 format
PUBLIC ASCII_TO_EDX  ; .. ASCII to binary in EDX register
;
;****************************************************************
;                          code segment
;****************************************************************
;
CODE    SEGMENT BYTE
        ASSUME  CS:CODE
;
BIN_TO_ASC10    PROC    FAR
; Procedure to convert a 16-bit binary number into 5 ASCII decimal
; digits
; On entry:
;         AX = binary number
;         DI --> 5 byte ASCII buffer
; On exit:
;         DI --> start of ASCII buffer holding ASCII decimal
;
; CPU compatibilty chart:
;                       | 8086/8088       | YES |
;                       | 80286           | YES |
;                       | 80386           | YES |
;                       | 80486/PENTIUM   | YES |
;
;***********************|
;   clear ASCII buffer  |
;***********************|
        PUSH    DI              ; Save buffer start
        MOV     CX,5            ; Five digits to clear
CLEAR_5:
        MOV     BYTE PTR [DI],' '       ; Clear digit
        INC     DI              ; Bump pointer
        LOOP    CLEAR_5         ; Repeat for 5 digits
        DEC     DI              ; Adjust buffer pointer to last
                                ; digit
        MOV     CX,10           ; Decimal divisor to CX
;***********************|
;  obtain ASCII decimal |
;        digit          |
;***********************|
GET_ASC10:
        MOV     DX,0            ; Clear for word division
        DIV     CX              ; Perform division AX/CX
; Quotient is in AX and remainder in DL
; Convert decimal to ASCII decimal
        ADD     DL,30H          ; Add 30H to bring to ASCII range
;***********************|
;     store digit       |
;  bump buffer pointer  |
;***********************|
        MOV     BYTE PTR [DI],DL        ; Store digit in buffer
        DEC     DI              ; Buffer pointer to next digit
; Note: the binary quotient is left in AX by the DIV CX instruction
;***********************|
;   is quotient = 0 ?   |
;***********************|
        CMP     AX,0            ; Test for end of binary
        JNZ     GET_ASC10       ; Continue if not 0
;***********************|
;   end of conversion   |
;***********************|
        POP     DI              ; Restore pointer to start of
                                ; buffer
        RET
BIN_TO_ASC10     ENDP
;
;
;*****************************************************************
BIN_TO_ASC16    PROC    FAR
; Procedure to convert a 16-bit binary number into 4 ASCII
; hexadecimal digits
; On entry:
;         AX = binary number
;         DI --> 4 byte ASCII buffer
; On exit:
;         DI --> start of ASCII buffer holding ASCII decimal

; CPU compatibilty chart:
;                       | 8086/8088       | YES |
;                       | 80286           | YES |
;                       | 80386           | YES |
;                       | 80486/PENTIUM   | YES |
;
;***********************|
;   clear ASCII buffer  |
;***********************|
        PUSH    DI              ; Save buffer start
        MOV     CX,4            ; Five digits to clear
CLEAR_4:
        MOV     BYTE PTR [DI],' '       ; Clear digit
        INC     DI              ; Bump pointer
        LOOP    CLEAR_4         ; Repeat for 4 digits
        DEC     DI              ; Adjust buffer pointer to last
                                ; digit
        MOV     CX,16           ; Hexadecimal divisor to CX
;***********************|
;   obtain ASCII hex    |
;        digit          |
;***********************|
GET_ASC16:
        MOV     DX,0            ; Clear for word division
        DIV     CX              ; Perform division AX/CX
; Quotient is in AX and remainder in DL
; Convert decimal to ASCII decimal
        ADD     DL,30H          ; Add 30H to bring to ASCII range
;***********************|
;     store digit       |
;  bump buffer pointer  |
;***********************|
        MOV     BYTE PTR [DI],DL        ; Store digit in buffer
        DEC     DI              ; Buffer pointer to next digit
; Note: the binary quotient is left in AX by the DIV CX instruction
;***********************|
;   is quotient = 0 ?   |
;***********************|
        CMP     AX,0            ; Test for end of binary
        JNZ     GET_ASC16       ; Continue if not 0
;***********************|
;   end of conversion   |
;***********************|
        POP     DI              ; Restore pointer to start of
                                ; buffer
        RET
BIN_TO_ASC16     ENDP
;

;*****************************************************************
ASCII_TO_DX     PROC    FAR
; Convert an ASCII number to binary and store in DX register
; On entry:
;         SI --> start of 5-digit ASCII buffer holding decimal
;                string in the range 0 to 65,535
; On exit:
;         DX = binary number
;
; CPU compatibilty chart:
;                       | 8086/8088       | YES |
;                       | 80286           | YES |
;                       | 80386           | YES |
;                       | 80486/PENTIUM   | YES |
;
;***********************|
;  initialize registers |
;***********************|
        MOV     DX,0            ; Clear accumulator register
        MOV     CX,10           ; Load multiplier into CX
;***********************|
; test digit for range  |
;        0 to 9         |
;***********************|
DIGIT_TO_ACC16:
        MOV     AL,[SI]         ; Get ASCII digit
        CMP     AL,'0'          ; Test for lower limit
        JB      EXIT_ASC16      ; Exit if less than 0
        CMP     AL,'9'          ; Test for higher limit
        JA      EXIT_ASC16      ; Exit if larger than 9
;***********************|
; ASCII digit to binary |
;***********************|
        SUB     AL,30H          ; ASCII to decimal
;***********************|
;    accumulator x 10   |
;      + digit          |
;***********************|
        MOV     AH,0            ; Clear high byte in AX
        PUSH    AX              ; Save digit in stack
        MOV     AX,DX           ; Previous product to AX
        MUL     CX              ; Perform AX = AX * 10
        MOV     DX,AX           ; Move product to accumulator
        POP     AX              ; Restore decimal digit
        ADD     DX,AX           ; Add digit to accumulator
;***********************|
; pointer to next digit |
;***********************|
        INC     SI              ; Bump pointer
        JMP     DIGIT_TO_ACC16  ; Continue
EXIT_ASC16:
        RET
ASCII_TO_DX     ENDP

;******************************************************************
;
; BINF_TO_DECF
; Procedure to convert an 8-bit binary fraction into an 8-digit
; decimal fraction in packed BCD format
; On entry:
;         AL = binary fraction
;         DI --> pointer to 4-byte buffer to hold packed BCD
;                decimal fraction
; On exit:
;         DI --> buffer holding 8 packed BCD digits
;
; CPU compatibility chart:
;                       | 8086/8088       | YES |
;                       | 80286           | YES |
;                       | 80386           | YES |
;                       | 80486/PENTIUM   | YES |
;
; BCD conversion factors stored in code segment
BCD_FACTORS     DB      50H,00,00,00    ; .50000000
                DB      25H,00,00,00    ; .25000000
                DB      12H,50H,00,00   ; .12500000
                DB      06H,25H,00,00   ; .06250000
                DB      03H,12H,50H,00  ; .03125000
                DB      01H,56H,25H,00  ; .01562500
                DB      00H,78H,12H,50H ; .00781250
                DB      00H,39H,06H,25H ; .00390625
;
BINF_TO_DECF     PROC    FAR
        LEA     SI,CS:BCD_FACTORS       ; Pointer to conversion table
        MOV     AH,10000000B    ; Mask to test binary digits
        MOV     CX,8            ; Counter for 8 bits
;
ADD_FACTOR:
        TEST    AL,AH           ; Test binary fraction bit
        JZ      NO_FACTOR_ADD   ; Do not add factor if not set
;***********************|
; add conversion factor |
;***********************|
        PUSH    AX              ; Save binary fraction
        ADD     SI,3    ; Set pointer to last byte
        ADD     DI,3    ; for both numbers
        MOV     CX,4    ; Counter for 4 bytes to add
        CLC             ; Clear carry for first add operation
ADD_BCD:
        MOV     AL,BYTE PTR CS:[SI]     ; Get conversion factor
        MOV     AH,BYTE PTR [DI]        ; Get factor sum
        ADC     AL,AH           ; AL+AH+carry from previous add
        DAA                     ; Adjust to packed BCD format
        MOV     BYTE PTR [DI],AL        ; Store sum
        DEC     SI              ; Adjust all pointers
        DEC     DI
        LOOP    ADD_BCD         ; Loop 4 times
        INC     DI              ; Adjust pointers to buffer start
        INC     SI
        POP     AX              ; Restore binary fraction
;***********************|
;     bump pointers     |
;***********************|
NO_FACTOR_ADD:
        ADD     SI,4            ; To next factor
        SHR     AH,1            ; Shift test mask to next bit
        JNZ     ADD_FACTOR      ; Continue for 8 digits
; DI --> buffer holding sum of all conversion factors
; Note that sum cannot overflow 8 digits. Carry can be ignored
        RET
BINF_TO_DECF     ENDP
;***************************************************************

DECODE_SINGLE   PROC    FAR
; Procedure to separate the elements of a floating point number
; stored in memory in IEEE single format
; On entry:
;         SI --> LOB of stored IEEE floating point single
; On exit:
;         AH = Sign field
;                    0000 0001 = negative sign
;                    0000 0000 = positive sign
;         AL = 8-bit exponent in bias 127 form
;   BH-BL-CH = 24-bit significand
;              Implicit significand bit is set if exponent is
;              not zero
;
; CPU compatibility chart:
;                       | 8086/8088       | YES |
;                       | 80286           | YES |
;                       | 80386           | YES |
;                       | 80486/PENTIUM   | YES |
;
;***********************|
;       decode sign     |
;***********************|
        ADD     SI,3            ; Set pointer to last byte
        MOV     AH,[SI]         ; Get byte into AH
        AND     AH,10000000B    ; Mask off exponent bits
        MOV     CL,7            ; Counter to shift 7 bits
        SHR     AH,CL           ; Shift sign bit to right-most bit
;***********************|
;     decode exponent   |
;***********************|
        MOV     AL,[SI]         ; Get 7 high bits of exponent
        SHL     AL,1            ; Shift left one bit
        DEC     SI              ; Pointer to byte at offset 2
        MOV     BH,[SI]         ; Get byte into BH
        TEST    BH,10000000B    ; Test for exponent bit 8 set
        JZ      OK_EXPONENT     ; Exponent OK if bit not set
        OR      AL,00000001B    ; Set low exponent bit
;***********************|
;  decode significand   |
;***********************|
OK_EXPONENT:
; Test for a non-zero exponent so set implicit significand bit
        CMP     AL,0            ; Is exponent zero
        JE      IMPLICIT_0      ; Implicit bit is zero with a
                                ; zero exponent. Do not set
;***********************|
;     explicit 1 bit    |
;***********************|
        OR      BH,10000000B    ; Set BH high bit
IMPLICIT_0:
        DEC     SI              ; Point to byte at offset 1
        MOV     BL,[SI]         ; Load into BL
        DEC     SI              ; Point to byte at offset 0
        MOV     CH,[SI]         ; Load into CH
        RET
DECODE_SINGLE   ENDP
;***************************************************************
;
; BIN_TO_BCD
; Convert a 16-bit binary number to 5 BCD digits
; On entry:
;         AX = binary number
; On exit:
;         5 packed BCD digits in the following registers:
;                CL = digit number 1 (high order digit)
;                DH = digits number 2 and 3
;                DL = digits number 4 and 5 (low order BCD digit)
;
; Operation:
; Binary in AX is first converted into 5 unpacked BCD digits and
; stored in a temporary buffer (BCD_BUF). Digits are then moved
; from this buffer and packed into DX and CL
;
; Conversion buffer in code segment
BCD_BUF         DB      5 DUP (0)
;
; CPU compatibility chart:
;                       | 8086/8088       | YES |
;                       | 80286           | YES |
;                       | 80386           | YES |
;                       | 80486/PENTIUM   | YES |
;
BIN_TO_BCD      PROC    FAR
;**********************|
;    clear buffer      |
;**********************|
        LEA     DI,CS:BCD_BUF   ; Set pointer to buffer
        PUSH    DI              ; Save buffer start
        MOV     CX,5            ; Five digits to clear
CLEAR_5BCD:
        MOV     BYTE PTR CS:[DI],0         ; Clear digit
        INC     DI              ; Bump pointer
        LOOP    CLEAR_5BCD      ; Repeat for 5 digits
        DEC     DI              ; Adjust buffer pointer to last
                                ; digit
        MOV     CX,10           ; Decimal divisor to CX
GET_BIN:
        MOV     DX,0            ; Clear for word division
; Divide binary by 10
        DIV     CX              ; Perform division AX/CX
; Quotient is in AX and remainder in DL
        MOV     BYTE PTR CS:[DI],DL     ; Store digit in buffer
        DEC     DI              ; Bump destination pointer
; Note: step 5 is performed automatically since the binary
; quotient is left in AX
        CMP     AX,0            ; Test for end of binary
        JNZ     GET_BIN         ; Continue if not end
; Last digit converted
;**********************|
;   BCD to registers   |
;      and pack        |
;**********************|
        POP     DI              ; Restore pointer to start of
                                ; buffer
        ADD     DI,4            ; Point to last digit
        MOV     AL,BYTE PTR CS:[DI]     ; BCD digit to AL
        MOV     BL,AL           ; To results register low nibble
        DEC     DI              ; Point to fourth BCD digit
        MOV     AL,BYTE PTR CS:[DI]     ; BCD digit to AL
        MOV     CL,4            ; Shift counter
        SHL     AL,CL           ; Shift BCD to high nibble
        OR      BL,AL           ; OR into result register high
                                ; nibble
        DEC     DI              ; Pointer to third BCD digit
        MOV     AL,BYTE PTR CS:[DI]     ; BCD digit to AL
        MOV     BH,AL           ; To results register low nibble
        DEC     DI              ; Point to second BCD digit
        MOV     AL,BYTE PTR CS:[DI]     ; BCD digit to AL
        MOV     CL,4            ; Shift counter
        SHL     AL,CL           ; Shift BCD to high nibble
        OR      BH,AL           ; OR into result register high
                                ; nibble
; Move 4 low order BCD digits to DX
        MOV     DX,BX
; Move high order digit to CL
        DEC     DI              ; Pointer to first BCD digit
        MOV     AL,BYTE PTR CS:[DI]     ; BCD digit to AL
        MOV     CL,AL           ; To results register low nibble
        RET
BIN_TO_BCD      ENDP
;****************************************************************
;
BCD_TO_ASCII    PROC    FAR
; Convert 5 unpacked BCD digits in machine registers into to 5
; ASCII digits
; On entry:
;         AL-BH-BL-CH-CL hold 5 unpacked BCD digits
;                        holding a value < 65536
;         SI --> 5 byte buffer to hold ASCII string
; On exit:
;         SI --> buffer holding ASCII digits
;
; CPU compatibility chart:
;                       | 8086/8088       | YES |
;                       | 80286           | YES |
;                       | 80386           | YES |
;                       | 80486/PENTIUM   | YES |
;
; Clear buffer for ASCII digits
        PUSH    SI              ; Save start of buffer
        PUSH    SI              ; Twice
        PUSH    CX              ; Save BCD high digit
        MOV     CX,5            ; Counter for 5 digits
CLEAR_BUF_5:
        MOV     BYTE PTR [SI],' '    ; Set blank in buffer
        INC     SI              ; Bump pointer
        LOOP    CLEAR_BUF_5
        POP     CX              ; Restore BCD digit
;**********************|
; convert to ASCII and |
;    move to buffer    |
;**********************|
; Convert BCD digits to ASCII
        ADD     AL,30H          ; Add 30H to each digit
        ADD     BH,30H
        ADD     BL,30H
        ADD     CH,30H
        ADD     CL,30H
        POP     SI              ; Restore buffer start
        MOV     BYTE PTR [SI],AL     ; Digit to buffer
        INC     SI              ; Bump buffer pointer
        MOV     BYTE PTR [SI],BH     ; Digit to buffer
        INC     SI              ; Bump buffer pointer
        MOV     BYTE PTR [SI],BL     ; Digit to buffer
        INC     SI              ; Bump buffer pointer
        MOV     BYTE PTR [SI],CH     ; Digit to buffer
        INC     SI              ; Bump buffer pointer
        MOV     BYTE PTR [SI],CL     ; Digit to buffer
        POP     SI              ; Restore start of buffer
        RET
BCD_TO_ASCII    ENDP
;*******************************************************************
;
; BCD_INTO_REG
; Convert a 5-digit BCD number into a binary in DX
; On entry:
;         AL-BH-BL-CH-CL hold 5 BCD digits
;                        holding a value < 65536
; On exit:
;         DX = binary number (range 0 to 65535)
;
; Code segment temporary data
ASC_TEMP_BUF    DB      5 DUP (20H)
                DB      0               ; Terminator
;
; CPU compatibility chart:
;                       | 8086/8088       | YES |
;                       | 80286           | YES |
;                       | 80386           | YES |
;                       | 80486/PENTIUM   | YES |
;
;
BCD_INTO_REG    PROC    FAR
; Set segment for conversion procedure
        PUSH    DS              ; Save current DS
        PUSH    CS              ; Push CS
        POP     DS              ; and pop into DS
        ASSUME  DS:CODE         ; Change assembler assumption
        LEA     DI,ASC_TEMP_BUF ; Address of buffer off DS
        CALL    FAR PTR BCD_TO_ASCII  ; Convert BCD digits to ASCII
                                ; string
        CALL    FAR PTR ASCII_TO_DX     ; Conversion routine
        POP     DS              ; Restore DS
        RET
BCD_INTO_REG    ENDP
;*******************************************************************

; BCD12_TO_ASCSTR
; Convert a number in BCD12 format into an ASCII string
; On entry:
;         SI --> 12-byte BCD12 buffer holding number
;         DI --> 31-byte ASCII buffer to hold string
; On exit:
;         DI holds ASCII string
;
; Operation:
;         Numbers with exponent < 10 are converted to decimal
;         format. If exponent is > 10 the number is converted to
;         exponential format
;
; BCD12 format:
;          S s e e e e m . m m m m m m m m m m m m m m m m m m
;                        |
;                        |____ implicit decimal point
;
;    S = sign of number (1 BCD digit)
;    s = sign of exponent (1 BCD digit)
;    e = exponent (4 BCD digits)
;    m = normalized significand (18 BCD digits)
;        (in positive and negative numbers the first significand
;         digit must be non-zero)
;    . = implicit decimal point between the first and second
;        significand digits
;
; CPU compatibility chart:
;                       | 8086/8088       | YES |
;                       | 80286           | YES |
;                       | 80386           | YES |
;                       | 80486/PENTIUM   | YES |
;
;**********************|
;   code segment data  |
;**********************|
OFFSET_ASCBUF   DW      0       ; Entry offset of ASCII buffer
;
BCD12_TO_ASCSTR  PROC    FAR
;**********************|
; save buffer addresses|
;**********************|
        MOV     WORD PTR CS:OFFSET_ASCBUF,DI
; Save pointer registers
        PUSH    SI
        PUSH    DI
; Clear results buffer
        PUSH    DI              ; Save pointer
        MOV     CX,31           ; Bytes to clear
CLEAR_23:
        MOV     BYTE PTR [DI],0 ; Clear to zero
        INC     DI              ; Bump pointer
        LOOP    CLEAR_23
        POP     DI              ; Restore pointer to start of buffer
;**********************|
;  get sign of number  |
;**********************|
        MOV     BYTE PTR [DI],' '       ; In case of positive
        MOV     AL,[SI]         ; First byte of BCD12 number
        TEST    AL,00010000B    ; Bit 4 is negative number
        JZ      NNOT_NEG        ; Number is not negative
; Negative number. Set sign
        MOV     BYTE PTR [DI],'-'       ; Sign to results buffer
NNOT_NEG:
        INC     DI              ; Bump pointer
;**********************|
; get value of exponent|
;**********************|
        INC     SI              ; Pointer to BCD12 exponent field
        INC     SI              ; Point to low byte of BCD exponent
; Move BCD digits to AL-BH-BL-CH-CL for conversion routine
        MOV     AL,0            ; Highest digit is zero
        MOV     DL,BYTE PTR [SI]        ; DL holds 2 packed digits
        MOV     DH,DL           ; Copy digits to DH
        AND     DL,00001111B    ; Erase high nibble in DL
        AND     DH,11110000B    ; and low in DH
        MOV     CX,4            ; For a count of 4
        SHR     DH,CL           ; Shift digit to low position
        DEC     SI              ; Bump pointer to previous digit
        MOV     BL,BYTE PTR [SI]        ; BL holds 2 packed digits
        MOV     BH,BL           ; Copy digits to BH
        AND     BL,00001111B    ; Erase high nibble in BL
        AND     BH,11110000B    ; and low in BH
        MOV     CX,4            ; For a count of 4
        SHR     BH,CL           ; Shift digit to low position
        MOV     CX,DX           ; Move digits to CX
        PUSH    SI              ; Save source
        PUSH    DI              ; and destination pointers
        CALL    FAR PTR BCD_INTO_REG    ; BCD to binary conversion
        POP     DI              ; Restore pointers
        POP     SI
; DX holds binary value of exponent
; Numbers with exponent = or < 10 are stored in conventional form
; Number with exponent > 10 are stored in exponential form
        CMP     DX,10           ; Test for limit
        JA      EXPO_FORM       ; Exponential form routine
        JMP     NORMAL_FORM     ; Conventional form if 10 or less
;**********************|
;   exponential form   |
;**********************|
; Buffer is formatted as follows for exponential form:
;       -1.2345 ...89 E-23
;       |||          ||_________ offset 21
;       |||          |__________ offset 20
;       |||_____________________ offset 2
;       ||______________________ offset 1
;       |_______________________ offset 0 (- sign or blank)
;**********************|
;  initialize buffer   |
;**********************|
EXPO_FORM:
        MOV     DI,CS:OFFSET_ASCBUF     ; Start of ASCII buffer
        MOV     BYTE PTR [DI+2],'.'     ; Set period in buffer
        MOV     BYTE PTR [DI+20],' '    ; Blank space separator
        MOV     BYTE PTR [DI+21],'E'    ; Letter E
; Get and encode exponent sign
        ADD     DI,22           ; Set pointer to sign position
        DEC     SI              ; Set pointer to first BCD digit
        MOV     AL,BYTE PTR [SI]        ; Get BCD sign digit
        TEST    AL,00000001B    ; Low bit is negative exponent
        JZ      POS_EXP_EXP     ; Exponent is positive
        MOV     BYTE PTR [DI],'-'       ; Place negative sign
        INC     DI
POS_EXP_EXP:
        INC     SI              ; BCD pointer to exponent area
; DI --> position for first exponent digit
        MOV     BX,CS:OFFSET_ASCBUF     ; BX --> start of buffer
        ADD     BX,3            ; Buffer offset 3 to 6 is used
                                ; as a temporary storage
        MOV     AL,[SI]         ; Get 2 packed BCD digits
        MOV     AH,AL           ; Copy digit in AH
        AND     AL,00001111B    ; Unpack low digit
        AND     AH,11110000B    ; and high digit
        MOV     CL,4            ; For a count of 4
        SHR     AH,CL           ; Move high digit to low nibble
; Convert BCD digits to ASCII
        ADD     AL,30H          ; BCD + 30H = ASCII
        ADD     AH,30H
        MOV     BYTE PTR [BX],AH        ; Store ASCII
        INC     BX              ; Bump pointer to temporary area
        MOV     BYTE PTR [BX],AL
        INC     BX              ; Bump pointer again
        INC     SI              ; BCD pointer to next digit pair
; Get next two BCD digits and store
        MOV     AL,[SI]         ; Get 2 packed BCD digits
        MOV     AH,AL           ; Copy digit in AH
        AND     AL,00001111B    ; Unpack low digit
        AND     AH,11110000B    ; and high digit
        MOV     CL,4            ; For a count of 4
        SHR     AH,CL           ; Move high digit to low nibble
; Convert BCD digits to ASCII
        ADD     AL,30H          ; BCD + 30H = ASCII
        ADD     AH,30H
        MOV     BYTE PTR [BX],AH        ; Store ASCII
        INC     BX              ; Bump pointer to temporary area
        MOV     BYTE PTR [BX],AL
        SUB     BX,3            ; BX --> first digit in temporary
                                ; area
; Move digits to final position skipping leading zeros
        MOV     CX,4            ; For a count of 4
SKIP_LEAD0:
        MOV     AL,BYTE PTR [BX]        ; Digit in temporary area
        CMP     AL,'0'          ; Test for leading zero
        JNE     NOT_LEAD00      ; Go if not a leading zero
        INC     BX              ; Bump pointer
        LOOP    SKIP_LEAD0      ; Test 4 digits
; Execution drops if all exponent digits are 0
        MOV     BYTE PTR [DI],'0'       ; Place one zero in buffer
        INC     DI              ; Bump buffer pointer
        JMP     EXPONENT_RDY
NOT_LEAD00:
        MOV     AL,BYTE PTR [BX]        ; Get digit
        MOV     BYTE PTR [DI],AL        ; and store it
        INC     BX              ; Bump pointers
        INC     DI
        LOOP    NOT_LEAD00      ; Continue for remaining digits
;
EXPONENT_RDY:
        MOV     BYTE PTR [DI],0FFH      ; Special code for clean up
                                        ; routine
; Exponent is in buffer. Move significand using existing routine
        SUB     SI,2            ; Set SI to start of BCD buffer
        MOV     DI,CS:OFFSET_ASCBUF     ; Start of ASCII buffer
        INC     DI              ; Skip sign position
        MOV     DX,0            ; Normalized significand
        JMP     POS_EX_FORM     ; Handle by positive exponent
;**********************|
;     normal form      |
;**********************|
NORMAL_FORM:
; DX holds binary value of exponent, which must be 10 or less
; Possible storage forms:
; If sign of exponent is negative:
;       0.1234
;       0.00001234
; If sign of exponent is positive
;       12.3456
;       12345.67
;       1.2345 (exponent = 0)
;
        DEC     SI              ; Set pointer to first BCD digit
        MOV     AL,BYTE PTR [SI]        ; Get BCD sign digit
        TEST    AL,00000001B    ; Low bit is negative exponent
        JZ      POS_EX_FORM     ; Exponent is positive
        JMP     NEG_EX_FORM     ; Go to negative exponent routine
;**********************|
; exponent is positive |
;**********************|
; DX holds the number of digits to move to the right the decimal
; point from the normalized position
; DI --> ASCII buffer
POS_EX_FORM:
        PUSH    DI              ; Save start of ASCII buffer
        ADD     DI,DX           ; Add exponent to start of buffer
        INC     DI              ; Add one for normalized format
        MOV     BYTE PTR [DI],'.'       ; Place point in buffer
        POP     DI              ; Restore pointer to start position
MOVE_SIG:
        ADD     SI,3            ; Set pointer to BCD significand
; SI --> first significand digit in packed BCD format
;**********************|
;   unpack and move    |
;     BCD digits       |
;**********************|
EXP_FORM_ENTRY:
        MOV     DX,9            ; Counter for 9 packed BCD digits
UNPACK_2:
        MOV     AL,[SI]         ; Get 2 packed BCD digits
        MOV     AH,AL           ; Copy digit in AH
        AND     AL,00001111B    ; Unpack low digit
        AND     AH,11110000B    ; and high digit
        MOV     CL,4            ; For a count of 4
        SHR     AH,CL           ; Move high digit to low nibble
; Convert BCD digits to ASCII
        ADD     AL,30H          ; BCD + 30H = ASCII
        ADD     AH,30H
; Place two digits in buffer
        CMP     BYTE PTR [DI],'.'       ; Test for period in buffer
        JNE     NO_PERIOD1      ; OK if no period
        INC     DI              ; Bump pointer if on period
NO_PERIOD1:
        MOV     BYTE PTR [DI],AH        ; Store high digit
        INC     DI              ; Bump pointer to next digit
        CMP     BYTE PTR [DI],'.'       ; Test for period in buffer
        JNE     NO_PERIOD2      ; OK if no period
        INC     DI              ; Bump pointer if on period
NO_PERIOD2:
        MOV     BYTE PTR [DI],AL        ; Store low digit
        INC     DI
        INC     SI              ; Bump pointer to next BCD
        DEC     DX              ; Adjust digits counter
        JNZ     UNPACK_2
;**********************|
;       clean up       |
;**********************|
; Clear trailing zeros from buffer
; Example buffer digits:  1.23450000000000000(00)
;                         1234.00000000000000(00)
; Buffer after clean-up:  1.2345(00)
;                         1234(00)
;
        MOV     DI,CS:OFFSET_ASCBUF     ; Entry offset from storage
        ADD     DI,30           ; Point to last digit
; Replace up to 27 trailing zeros with NULL bytes
        MOV     CX,29           ; Counter for maximum digits
REPLACE_27:
        CMP     BYTE PTR [DI],0FFH      ; Special code for
                                        ; exponential form
        JNE     NT_EXP_FORM     ; Go if not
        MOV     BYTE PTR [DI],00H       ; Clear special code
        JMP     EXIT_REP27      ; Do not fix trailing zeros
NT_EXP_FORM:
        CMP     BYTE PTR [DI],'0'       ; Test for trailing 0
        JNE     TEST_27         ; Test for other values
        MOV     BYTE PTR [DI],0 ; Replace zero with NULL
        JMP     NEXT_27
TEST_27:
        CMP     BYTE PTR [DI],0 ; Test for NULL at end of buffer
        JNE     EXIT_REP27      ; Exit if not 0 or NULL
NEXT_27:
        DEC     DI              ; Adjust pointer
        LOOP    REPLACE_27      ; Continue for 27 digits
EXIT_REP27:
; If the last character found is a period, then number is in the
; form 1234.0000..00.
        CMP     BYTE PTR [DI],'.'       ; Test for period
        JNE     NO_PERIOD_FIX   ; Go if not
        MOV     BYTE PTR [DI],0 ; Replace period with NULL
NO_PERIOD_FIX:
        JMP     EXIT_ASCSTR
;**********************|
; exponent is negative |
;**********************|
; Display forms for negative exponent
;       0.1234 (exponent = -1)
;       0.00001234 (exponent = -5)
; Operation:
;    The decimal point is first placed in the buffer
;    DX number of zeros are placed in the buffer, skipping the
;    decimal point
NEG_EX_FORM:
        INC     DI              ; Leave blank for leading zero
        MOV     BYTE PTR [DI],'.'       ; Place point in buffer
        DEC     DI
PLACE_0S:
        CMP     BYTE PTR [DI],'.'       ; Test for period in buffer
        JNE     NO_PERIOD3      ; No period, proceed
        INC     DI              ; Bump buffer pointer to skip period
NO_PERIOD3:
        MOV     BYTE PTR [DI],'0'       ; Place 0 digit in buffer
        INC     DI              ; Bump pointer to next digit
        DEC     DX              ; Decrement counter for zeros
        JNZ     PLACE_0S        ; Continue if not zero
        JMP     MOVE_SIG        ; Use existing code to move
                                ; significand
;**********************|
;    end of routine    |
;**********************|
EXIT_ASCSTR:
        POP     DI              ; Restore pointer registers
        POP     SI
        RET
BCD12_TO_ASCSTR   ENDP
;*******************************************************************

; ASCII_TO_BCD12
; Conversion of an ASCII decimal number into BCD12 format
; On entry:
;       DS:SI --> an ASCII decimal string, terminated in a NULL
;                 byte or control code
;             Note: the caller's buffer may contain the $ sign
;                   and commas as formatting symbols. The string
;                   terminator (00H) may be preceded by a CR
;                   (0DH) or LF (0AH) control codes. If the
;                   characters E or e are in the string, the
;                   procedure assumes that the string is in
;                   ASCII exponential form. The code assumes that
;                   exponential strings are normalized.
;
;       DS:DI --> 12-byte memory area to holds the number encoded
;                 in BCD12 format
;
; Sample input strings:
;        123.223344
;        -0000.000002345
;        1.273E-22
;        -1.44556677 E230
;
;       DS:DI --> 12-byte memory area to holds the number encoded
;                 in BCD12 format
;
; On exit:
;       DS:DI --> the caller's buffer with the input number
;                 reformatted in BCD12 format, as follows:
;
;          S s e e e e m . m m m m m m m m m m m m m m m m m m
;                        |
;                        |____ implicit decimal point
;
;    S = sign of number (1 BCD digit)
;    s = sign of exponent (1 BCD digit)
;    e = exponent (4 BCD digits)
;    m = normalized significand (18 BCD digits)
;    . = implicit decimal point between the first and second
;        significand digits
;
; CPU compatibility chart:
;                       | 8086/8088       | YES |
;                       | 80286           | YES |
;                       | 80386           | YES |
;                       | 80486/PENTIUM   | YES |
;
;**********************|
;   code segment data  |
;**********************|
OFFSET_ASC      DW      0       ; Entry offset of ASCII buffer
OFFSET_BCD      DW      0       ; Entry offset of BCD buffer
;
ASCII_TO_BCD12   PROC    FAR
;**********************|
; save buffer addresses|
;**********************|
        MOV     WORD PTR CS:OFFSET_ASC,SI
        MOV     WORD PTR CS:OFFSET_BCD,DI
;**********************|
; save input registers |
;**********************|
; Pointer registers are saved. All others are destroyed
        PUSH    SI
        PUSH    DI
;**********************|
;   clear buffer for   |
;     BCD12 number     |
;**********************|
; DI --> 12-byte buffer to hold BCD12 number
NOT_EXP:
        PUSH    DI              ; Save output pointer
        MOV     CX,12           ; Bytes to clear
CLEAR_12:
        MOV     BYTE PTR [DI],0 ;
        INC     DI              ; Bump buffer pointer
        LOOP    CLEAR_12
        POP     DI              ; Restore pointer to start
        MOV     SI,CS:OFFSET_ASC        ; Source pointer to start
                                        ; of ASCII
;**********************|
;  get sign of number  |
;**********************|
; Find first significant character in the caller's string
; Skip leading blanks, $ and # sign, and other invalid symbols
 GET_NB:
        MOV     AL,BYTE PTR [SI]        ; Get character
        CMP     AL,'+'          ; Test for + sign
        JAE     FIRST_NB        ; Go if equal or larger
        INC     SI
        JMP     GET_NB
; first non-blank is a symbol or number larger than the + sign
FIRST_NB:
        CMP     AL,'+'          ; Test for +
        JE      SIGN_IS_PLUS
        CMP     AL,'-'          ; Test for -
        JE      SIGN_IS_MINUS
; First non-blank character is not a sign. Assume a digit
; Positive number is the default setting in BCD buffer
        JMP     TEST_4_EXP      ; Test for exponential form
;**********************|
;  number is negative  |
;**********************|
SIGN_IS_MINUS:
        MOV     BYTE PTR [DI],00010000B ; Set first BCD to 1
;**********************|
;  number is positive  |
;**********************|
; Encoding for positive number (0000B) is already in buffer
SIGN_IS_PLUS:
        INC     SI              ; Bump ASCII digits pointer
;**********************|
;   test for E or e    |
;**********************|
; The caller's string may be in ASCII exponential form
; Any value less than 20H terminates the ASCII string
TEST_4_EXP:
        MOV     WORD PTR CS:OFFSET_ASC,SI
                                ; Save pointer to first digit
TEST_AGAIN:
        CMP     BYTE PTR [SI],'E'
        JE      STRING_ISEX             ; String in FP format
        CMP     BYTE PTR [SI],'e'
        JE      STRING_ISEX
        CMP     BYTE PTR [SI],' '       ; Less than 20H is end
                                        ; of string
        JB      SKIP_ZERO               ; Not exponential form
        INC     SI                      ; Bump digits pointer
        JMP     TEST_AGAIN
;**********************|
;   input is in ASCII  |
;   exponential form   |
;**********************|
STRING_ISEX:
; Sample exponential form ASCII strings:
;            1.2345 E12
;           -1.2345E-12
;           +1.2345 e+12
; SI --> letter e or E
; First test for negative exponent
        INC     SI              ; Bump pointer past letter E
        PUSH    SI              ; Save pointer to possible exponent
                                ; sign
; SI --> string of exponent digits
; Skip sign, if there is one
FIRST_0PLUS:
        CMP     BYTE PTR [SI],'0'       ; First valid ASCII number
        JAE     OK_SI_PTR       ; SI pointer is OK if 0 or >
        INC     SI              ; Bump pointer if not
        JMP     FIRST_0PLUS     ; Continue
OK_SI_PTR:
        CALL    FAR PTR ASCII_TO_DX     ; Convert to binary in DX
; DX now holds binary value of exponent
        POP     SI              ; Restore point to possible sign
        CMP     BYTE PTR [SI],'-'       ; Negative exponent
        JNE     NON_NEG_EXP     ; Go if not negative
        JMP     NEG_EXPONENT    ; Routine for negative exponent
; Positive exponent code is already in place
NON_NEG_EXP:
        JMP     CONV_EXPONENT
;**********************|
; skip leading zeros   |
;**********************|
SKIP_ZERO:
; Leading zeros are not significant. skip them
; First reset pointer to start of string
        MOV     SI,CS:OFFSET_ASC        ; Pointer saved in memory
SKIP_1ZERO:
        CMP     BYTE PTR [SI],'0'       ; Test for 0
        JNE     NOT_LEAD0
        INC     SI              ; Bump to next digit
        JMP     SKIP_1ZERO
;**********************|
;  find decimal point  |
;    (if present)      |
;**********************|
;  DX is used to count the non-zero digits preceding the decimal
;  point. This value determines the number's exponent
NOT_LEAD0:
        MOV     DX,0            ; Clear integer digit counter
; Caller's string must end in a NULL byte (00H) or other control
; code < 20H
; Test for an integer zero input in the form 0000
        CMP     BYTE PTR [SI],' '       ; Smaller than 20H is
                                ; end of input buffer
        JAE     GET_POINT       ; Continue if not end
        JMP     NUM_ISZERO      ; Number is zero
GET_POINT:
        MOV     AL,BYTE PTR [SI]; Caller's ASCII digit
        CMP     AL,'.'          ; Test for decimal point
        JE      POINT_FOUND
        CMP     AL,','          ; Test for formatting comma
                                ; in the caller's string
        JE      SKIP_COMMA      ; Ignore comma
        CMP     AL,20H          ; String terminator
        JNB     LOOK_4_PT       ; Continue searching if not < 20H
        JMP     NO_POINT        ; No decimal point in string
LOOK_4_PT:
        INC     DX              ; Count one integer digit
SKIP_COMMA:
        INC     SI              ; Bump integer digits counter
        JMP     GET_POINT
;**********************|
; caller's string is a |
;   decimal fractions  |
;**********************|
; If DX = 0 then the caller's string < 1 and the EXPONENT IS
;           NEGATIVE or the number is zero
; If DX = 1 then the caller's string is already normalized
;         and the EXPONENT = +0
; If DX > 1 then the caller's string is a number =< 10 and
;         the EXPONENT IS POSITIVE
POINT_FOUND:
        CMP     DX,1            ; Test for normalized input
        JNE     TEST_4_DX0
;**********************|
;       DX = 1         |
;**********************|
; ASCII input is in the form: d.dddd...ddd
; Number is between 1 and 10. Sign of exponent = 0000B
; 4 exponent digits = 0000H (these values are already in place)
        JMP     EXP_MOVED
TEST_4_DX0:
        CMP     DX,0            ; Test for negative exponent
        JE      DX_ISZERO       ; Go to DX = 0 processing
; If DX > 1 adjust and store
        DEC     DX
        JMP     CONV_EXPONENT
;**********************|
;       DX = 0         |
;**********************|
; Possible cases:
; non-zero -->   000.002345
; non-zero -->      .2345
;     zero -->     0.00
;     zero -->      .0
; Count decimal digits to normalized position, that is, following
; the first non-zero digit
; DX is used as counter
; At this point:
;                dddd.ddd...
;                    |____ SI points to decimal point
DX_ISZERO:
        INC     SI              ; Skip decimal point
        INC     DX              ; Count one digit
;**********************|
;     normalize        |
;**********************|
NORMALIZE:
        CMP     BYTE PTR [SI],'0'       ; Test for zero
        JE      ZERO_DIGIT              ; Go if zero digit
        JB      NUM_ISZERO              ; exit if less than zero
; If not equal or less than zero it must be larger
        JMP     NEG_EXPONENT    ; Normalized position found
ZERO_DIGIT:
        INC     SI              ; Bump pointer
        INC     DX              ; and digits counter
        JMP     NORMALIZE
;**********************|
;   ASCII number = 0   |
;**********************|
NUM_ISZERO:
; Make sure zero encoding is in BCD buffer
ZERO_EXIT:
        MOV     CX,12           ; Bytes to clear
CLEAR_12_D:
        MOV     BYTE PTR [DI],0 ;
        INC     DI              ; Bump buffer pointer
        LOOP    CLEAR_12_D
        JMP     A2FP_EXIT       ; exit conversion
;**********************|
;  negative exponent   |
;**********************|
; DX holds binary value of negative exponent
; Convert to BCD and store
NEG_EXPONENT:
        MOV     DI,CS:OFFSET_BCD        ; Saved in CS data
; DI --> start of BCD buffer
        MOV     AL,[DI]         ; Get BCD sign digit
        MOV     AH,00000001B    ; Code for minus exponent
        OR      AL,AH           ; OR into sign digit
        MOV     BYTE PTR [DI],AL        ; Store exponent sign
        JMP     CONV_EXPONENT
;**********************|
;  ASCII is an integer |
;**********************|
NO_POINT:
; Exponent of integer is positive. This encoding is already in place
; String has no decimal point. DX = exponent plus 1
        DEC     DX
;**********************|
;  convert and store   |
;      exponent        |
;**********************|
CONV_EXPONENT:
        MOV     AX,DX           ; Move binary to source register
        CALL    FAR PTR BIN_TO_BCD       ; Binary to BCD conversion
        MOV     DI,CS:OFFSET_BCD        ; Saved in CS data
; DI --> start of BCD buffer
        INC     DI              ; Bump pointer
; DI --> first BCD exponent digit
        MOV     BYTE PTR [DI],DH        ; Store two high BCD digits
        INC     DI              ; Bump pointer again
        MOV     BYTE PTR [DI],DL        ; Store two low BCD digits
;**********************|
;  convert and store   |
;     significand      |
;**********************|
; Set pointer to first non-zero and non-sign digit in ASCII number
EXP_MOVED:
        MOV     SI,CS:OFFSET_ASC        ; Pointer to start of number
LOOK_4_1:
        MOV     AL, BYTE PTR [SI]       ; Get ASCII digit
        CMP     AL,'1'          ; Test for non-zero and non-control
        JAE     FIRST_NON0      ; Go if 1 or greater
        INC     SI              ; Bump pointer
        CMP     AL,20H          ; Test for control or NULL
        JAE     LOOK_4_1        ; Continue if not control or NULL
        JMP     ZERO_EXIT       ; Force zero and exit if no valid
                                ; significand digits
FIRST_NON0:
        MOV     CX,18           ; Counter for 18 digits maximum
        MOV     DI,CS:OFFSET_BCD        ; Set pointer
; DI --> start of BCD buffer
        ADD     DI,3            ; Offset of significand area
; DI --> BCD significand area in buffer
; SI --> first non-zero digits in the ASCII string
; ASCII string must have at least one valid digit at this point
; Sample significand strings:
;                1.23456(0DH)
;                1.23456(00H)
;                -1.2345 E-12
;                123.45E345
; Operation:
;    The conversion routine takes the ASCII digits from the buffer,
;    converts them to BCD, and stores the digits in packed format.
;    The DH register is used to signal the high-order BCD digit
        MOV     DH,1            ; Code for next digit is high nibble
ASCII_MOVE:
        MOV     AL,BYTE PTR [SI]        ; ASCII digit
; Test for invalid values
        CMP     AL,'0'          ; Lower limit
        JAE     VALID_1         ; Digit is 0 or larger
; If digit is less than 0, test for period, comma, and other symbols
; and skip. Exit routine if control code or NULL
        CMP     AL,20H          ; Control codes are < 20H
        JAE     NEXT_ASCII      ; Skip if space or symbol
        JMP     A2FP_EXIT       ; Exit with first non ASCII digit
VALID_1:
        CMP     AL,'9'          ; Upper limit of digit
        JBE     VALID_2         ; Go if OK
        JMP     A2FP_EXIT       ; Exit with first non ASCII digit
                                ; greater than 9
VALID_2:
; ASCII digit in AL is ready for storing in packed BCD format
        SUB     AL,30H          ; ASCII to BCD
        CMP     DH,1            ; Test high digit indicator
        JE      HIGH_BCD        ; Go if high
; Low order BCD digit
        MOV     AH,BYTE PTR [DI]        ; Get previous high digit
        OR      AH,AL           ; OR low digit into AH
        MOV     BYTE PTR [DI],AH        ; Store low and high
        INC     DI              ; Bump destination pointer
        MOV     DH,1            ; Set switch for next digit is high
        JMP     NEXT_ASCII      ; Get next ASCII digit
; Routine to store high-order ASCII digit
; Binary value (in AL) must be shifted left 4 position
HIGH_BCD:
        PUSH    CX              ; Save digits counter
        MOV     CL,4            ; For a count of 4
        SHL     AL,CL           ; Shift AL digits
        POP     CX              ; Restore counter
        MOV     BYTE PTR [DI],AL        ; Store digit in buffer
        MOV     DH,0            ; Clear high digit switch
                                ; Next digit is low
NEXT_ASCII:
        INC     SI              ; Bump ASCII digits pointer
        LOOP    ASCII_MOVE      ; Continue moving digits
;**********************|
;  restore and exit    |
;**********************|
A2FP_EXIT:
        POP     DI              ; Restore caller's data registers
        POP     SI
        RET
ASCII_TO_BCD12   ENDP

;*****************************************************************
.386
ASCII_TO_EDX     PROC    FAR
; Convert an ASCII number to binary and store in EDX register
; On entry:
;         SI --> start of 10-digit ASCII buffer holding decimal
;                string in the range 0 to 4,294,967,295
; On exit:
;         EDX = binary number
;
; CPU compatibilty chart:
;                       | 8086/8088       | NO  |
;                       | 80286           | NO  |
;                       | 80386           | YES |
;                       | 80486/PENTIUM   | YES |
;
;***********************|
;  initialize registers |
;***********************|
        MOV     EDX,0           ; Clear accumulator register
        MOV     ECX,10          ; Load multiplier into ECX
;***********************|
; test digit for range  |
;        0 to 9         |
;***********************|
DIGIT_TO_ACC32:
        MOV     AL,[SI]         ; Get ASCII digit
        CMP     AL,'0'          ; Test for lower limit
        JB      EXIT_ASC32      ; Exit if less than 0
        CMP     AL,'9'          ; Test for higher limit
        JA      EXIT_ASC32      ; Exit if larger than 9
;***********************|
; ASCII digit to binary |
;***********************|
        SUB     AL,30H          ; ASCII to decimal
;***********************|
;    accumulator x 10   |
;      + digit          |
;***********************|
        MOV     AH,0            ; Clear high byte in AX
        PUSH    AX              ; Save digit in stack
        MOV     EAX,EDX         ; Previous product to EAX
        MUL     ECX             ; Perform EAX = EAX * 10
        MOV     EDX,EAX         ; Move product to accumulator
        MOV     EAX,0           ; Clear 32 bits
        POP     AX              ; Restore decimal digit
        ADD     EDX,EAX         ; Add digit to accumulator
;***********************|
; pointer to next digit |
;***********************|
        INC     SI              ; Bump pointer
        JMP     DIGIT_TO_ACC32  ; Continue
EXIT_ASC32:
        RET
ASCII_TO_EDX    ENDP
;
CODE    ENDS
        END

r
        JMP     DIGIT_TO_ACC32  ; Continue
EXIT_ASC32:
        