;*****************************************************************
;*****************************************************************
;                           UNIT_4.ASM
;*****************************************************************
;*****************************************************************
; Library of math routines from the book SOFTWARE SOLUTIONS by
; Julio Sanchez and Maria P. Canton
; Creation date:  January 13, 1994
; Last update:    February 12, 1994
;
; Libary name: MATHUNIT.LIB
;
;*****************************************************************
;                           publics
;*****************************************************************
; MATHEMATICAL ROUTINES IN THIS MODULE
PUBLIC ASCII_TO_EXP ;.... ASCII decimal string to exponential form
PUBLIC FPU_OUTPUT   ;.... ST(0) register to ASCII decimal string
PUBLIC FPU_INPUT    ;..... ASCII exponential string to ST(0)
;
;*****************************************************************
;                          externals
;*****************************************************************
; Routines in other modules of MATHUNIT.LIB
EXTRN   NUM_AT_ST0:     FAR
;
;****************************************************************
;                          code segment
;****************************************************************
;
CODE    SEGMENT BYTE
        ASSUME  CS:CODE
;
;***************************|
;   code segment data for   |
; procedures in this module |
;***************************|
PACKED_EXP      DT      ?       ; 18-digit BCD exponent
PACKED_SIG      DT      ?       ; 18-digit BCD significand
DECIMAL_EXP     DW      ?       ; Storage for decimal exponent
SIGN_OF_EXP     DB      ?       ; 1 = negative
                                ; 0 = positive
TEN_TO_17       DT      00100000000000000000
TEN             DW      10
NAN_CODE        DD      0FFFFFFFFH      ; Not a number encoding
INF_NEGATIVE    DD      0FF800000H      ; Negative infinity
INF_POSITIVE    DD      07F800000H      ; Positive infinity
; Storage for entry pointers
ENTRY_SI        DW      0       ; Start of caller's input string
ENTRY_DI        DW      0       ; Start of caller's output
                                ; buffer
INPUT_NUM       DT      0       ; Storage for input in FPU_OUTPUT
;*******************|
;       buffers     |
;*******************|
TEMP_REAL       DT      ?
                DB      0
WORK_BUF        DB      18 DUP (?)
                DB      0
; Intermediate buffer for 5 exponent ASCII digits
EXP_BUFFER      DB      '     ',00H
; Storage area for the caller's 80X87 environment
STATE_80X87     DB      94 DUP (?)
;
;*******************|
;   control words   |
;*******************|
USERS_CW        DW      ?       ; Storage for caller's control
                                ; word
; Control words to force rounding to + or - infinity and reduce
; precision to 24 bits to avoid error during the extraction of
; the exponent
ROUND_DOWN      DW      14FFH
ROUND_UP        DW      18FFH
;
;*******************|
;   text messages   |
;*******************|
; Text for clearing 27 bytes in the caller's buffer
FPD_MES         DB      '  .                  E     '
POS_ZERO        DB      ' 0.00000000000000000 E+0   '
NEG_ZERO        DB      '-0.00000000000000000 E+0   '
EMPTY_MES       DB      ' EMPTY                     '
;
STATUS_WORD     DW      ?
;
                DB      0
;****************************************************************
;
;********************************|
; INPUT ASCII STRING INTO 80x87  |
;********************************|
FPU_INPUT       PROC    FAR
; On entry:
;     DS:DX points to the caller's buffer which holds an ASCII
;     decimal number formatted in floating-point decimal notation:
;
;                 sm.mmmmmmmmmmmmmmmmm ESeeee
; where
; s = sign of number (blank = +)
; m = up to 18 significand digits. Extra digits are ignored
; . = decimal point following the first significand digit.
;     If no decimal point in string, it is assumed
; E = explicit letter E (or e) to signal start of exponent
; S = + or <196> sign of exponent. If no sign, positive is assumed
; e = up to four exponent digits. The numerical value of the
;     exponent cannot exceed +/- 4932
; One or more spaces can be used to separate the last digit of
; the significand and the start of the exponent
;
; Examples of input:
;                    1.781252345E-1
;                    -3.14163397  E+0
;                    1.2233445566778899 e1387
;
; On exit:
;        a. If carry flag clear
;           Input number is loaded into NDP stack top register
;           Previous values in the stack are pushed down one
;           register
;           AL = 0 (no error)
;
;        b. If carry flag set
;           AL holds error code
;           AL = 1 if no E or e symbol in the caller's string
;                NAN encoding is loaded into ST(0)
;           AL = 2 if exponent exceeds valid range
;                Overflow - positive infinity is loaded into ST(0)
;                Underflow - negative infinity is loaded into ST(0)
;           AL = 3 if there is an invalid character in caller's
;                string. NAN encoding is loaded into ST(0)
;
;        AX is destroyed. All other registers are preserved.
;        The caller's NDP environment is also preserved, except
;        the stack top register.
;********************************************************************
; CPU compatibility chart:
;                       | 8086/8088/8087  | YES |
;                       | 80286/80287     | YES |
;                       | 80386/80387     | YES |
;                       | 486 DX          | YES |
;                       | PENTIUM         | YES|
;********************************************************************
;
;*******************|
; save NDP caller's |
;  environment and  |
;     registers     |
;*******************|
        FSAVE   CS:STATE_80X87  ; Environment area in scratchapad 
        FWAIT
        PUSH    CX              ; Save registers used by routine
        PUSH    DX
        PUSH    SI
        PUSH    DI
;*******************|
; clear scratchapad |
;      buffers      |
;*******************|
        LEA     DI,CS:PACKED_SIG     ; Scratchpad buffer
        MOV     CX,10                ; Bytes in buffer
CLEAR_10:
        MOV     BYTE PTR CS:[DI],0   ; Clear with zeros
        INC     DI                   ; Bump pointer
        LOOP    CLEAR_10
; Clear work buffer
        LEA     DI,CS:WORK_BUF
        MOV     CX,18
CLEAR_18:
        MOV     BYTE PTR CS:[DI],0   ; Clear byte
        INC     DI
        LOOP    CLEAR_18
;*******************|
;  move digits into |
;     WORK_BUF      |
;*******************|
; DS:DX --> caller's buffer with ASCII decimal number in
; exponential notation
        MOV     SI,DX           ; Copy source to SI
        LEA     DI,CS:WORK_BUF  ; Destination
        MOV     CX,18           ; Maximum digits to move
MOVE_18:
        MOV     AL,[SI]         ; Get caller's character
        INC     SI              ; Bump source pointer
        CMP     AL,'.'          ; Do not move decimal point
        JE      MOVE_18
        CMP     AL,'+'          ; or + and - signs
        JE      MOVE_18
        CMP     AL,'-'
        JE      MOVE_18
        CMP     AL,' '          ; Skip spaces
        JE      MOVE_18
; Test for E or e field terminators
        CMP     AL,'E'          ; Start of exponent
        JE      END_MOVE_18     ; End move if E
        CMP     AL,'e'          ; Lowercase e also allowed
        JE      END_MOVE_18     ; End move if e
; Test for invalid digit, larger than ASCII 9 or smaller than
; ASCII 0
        CMP     AL,'9'          ; High limit
        JA      BAD_ASCII       ; Exit if larger than 9
        CMP     AL,'0'          ; Low limit
        JB      BAD_ASCII
; Value is in legal range
OK_ASCII:
        MOV     CS:[DI],AL      ; Place ASCII in buffer
        INC     DI
        LOOP    MOVE_18
        JMP     END_MOVE_18
;*******************|
;      illegal      |
;     character     |
;*******************|
BAD_ASCII:
        MOV     AL,3            ; Error code for illegal character
                                ; in caller's string
        JMP     INPUT_ERROR
;*******************|
;  sign of number   |
;*******************|
END_MOVE_18:
; BCD significand is to be formatted as follows:
;            s0|dd|dd|dd|dd|dd|dd|dd|dd|dd
; where
; s = sign of significand. 1000B if number is negative
;                          0000B if number is positive
; d = 18 significand digits in packed BCD format
; 0 = required zero
;
; Set up buffer pointers
        MOV     SI,DX                   ; Caller's input buffer
        LEA     DI,CS:PACKED_SIG        ; Output buffer
        ADD     DI,9                    ; Point to sign byte
        MOV     AL,[SI]                 ; Get sign of number
        CMP     AL,'-'                  ; Test for negative
        JNE     POS_NUMBER              ; Go if not negative
        MOV     BYTE PTR CS:[DI],80H    ; Set sign byte
POS_NUMBER:
        DEC     DI              ; Point to byte number 8
;*******************|
; convert ASCII to  |
;     packed BCD    |
;*******************|
        LEA     SI,CS:WORK_BUF  ; Significand's digits
CONVERT_8:
        MOV     AH,CS:[SI]      ; Get ASCII digit
        CMP     AH,0            ; All characters read ?
        JE      FIND_E          ; Exit if zero
                                ; Nothing to store at this point
        INC     SI              ; Bump pointer to next digit
        SUB     AH,30H          ; Convert to binary
        MOV     CL,4            ; Set up shift counter
        SHL     AH,CL           ; Shift bits to high nibble of AL
; High-order BCD digit now in AH
        MOV     AL,CS:[SI]      ; Get ASCII digit
        CMP     AL,0            ; All characters read ?
        JNE     OK_DIGIT        ; Continue if not
; There is a valid digit in AH which must be stored
        MOV     CS:[DI],AH      ; Store it
        JMP     FIND_E          ; and exit routine
OK_DIGIT:
        SUB     AL,30H          ; Convert to binary
        OR      AL,AH           ; Pack BCD digits in AL
        MOV     CS:[DI],AL      ; Place in significand's buffer
        INC     SI              ; Point to next character
        DEC     DI              ; Bump pointers
        JMP     CONVERT_8
;*******************|
;  find start of    |
;   exponent        |
;*******************|
FIND_E:
        MOV     SI,DX           ; Caller's input buffer
        MOV     CX,22           ; Maximum character count
                                ; preceding the E symbol
SEARCH_4_E:
        MOV     AL,[SI]         ; Get character
        CMP     AL,'E'
        JE      VALID_EXP       ; Valid symbol
        CMP     AL,'e'
        JE      VALID_EXP       ; Also valid
        INC     SI              ; Bump buffer pointer
        LOOP    SEARCH_4_E
; No E or e in caller's input string
        MOV     AL,1            ; Error code number 1
;*******************|
;   input ERROR     |
;*******************|
INPUT_ERROR:
        FRSTOR  CS:STATE_80X87  ; Restore environment
        FWAIT
        FLD     CS:NAN_CODE     ; Load Not-a-Number code
ERROR_RETURN:
        STC                     ; Error
        JMP     RESTORE_EXIT
;*******************|
; convert exponent  |
;*******************|
VALID_EXP:
        INC     SI              ; Point to exponent sign or
                                ; to first exponent digit
;*******************|
;  sign of exponent |
;*******************|
        MOV     CS:SIGN_OF_EXP,0      ; Assume positive
        MOV     AL,[SI]         ; Get sign of exponent
        CMP     AL,'-'          ; Test for negative
        JNE     POS_EXPON       ; Go if not negative
        MOV     CS:SIGN_OF_EXP,1      ; Code for negative exponent
        INC     SI
        JMP     READ_EXP
POS_EXPON:
        CMP     AL,'+'          ; Test for explicit positive
        JNE     READ_EXP        ; Pointer is OK
        INC     SI              ; Skip + sign
;*******************|
; convert digits to |
;      binary       |
;*******************|
READ_EXP:
        CALL    ASC_2_BIN       ; Convert to binary (in DX)
        MOV     CS:DECIMAL_EXP,DX
;*******************|
; test for exponent |
;   out of range    |
;*******************|
; Absolute value of exponent cannot exceed the limit of the NDP
; long real format.
        CMP     DX,4933         ; Test for limit
        JB      EXPONENT_OK     ; Go if smaller than 4933
;*******************|
;      encode       |
;  + or - infinity  |
;*******************|
; Restore caller's environment
        FRSTOR  CS:STATE_80X87  ; Restore environment
        FWAIT
        MOV     AL,2            ; Exit code for this error
; Exponent is out of range. Determine if underflow or overflow
        CMP     CS:SIGN_OF_EXP,1      ; Test for negative
        JE      UNDERFLOW       ; Encode negative infinity
; Overflow. Encode positive infinity
        FLD     CS:INF_POSITIVE ; Load code for positive infinity
        JMP     ERROR_RETURN
UNDERFLOW:
        FLD     CS:INF_NEGATIVE ; Load code for negative infinity
        JMP     ERROR_RETURN
        MOV     AL,2            ; Error code number 2
        JMP     INPUT_ERROR
;*******************|
;      load BCD     |
;    significand    |
;*******************|
EXPONENT_OK:
        FBLD    CS:TEN_TO_17    ;  10^17   |  EMPTY  |
        FBLD    CS:PACKED_SIG   ;   sig    |  10^17  |  EMPTY |
        FDIV    ST,ST(1)        ; sig/10^17|  10^17  |  EMPTY |
        FXCH    ST(1)           ;  10^17   |   SIG   |  EMPTY |
        FSTP    ST(0)           ;   SIG    |  EMPTY  |
;*******************|
; select conversion |
;     option        |
;*******************|
; Option 1: if exponent = 0, then significand remains unchanged
; Option 2: if exponent is negative (SIGN_OF_EXP = 1), then divide
;           significand by 10^exponent
; Option 3: if exponent is positive (SIGN_OF_EXP = 0), then
;           multiply significand by 10^exponent
;
; DX holds binary exponent
        CMP     DX,0            ; Test option no. 1
        JNE     TEST_EXP_POS    ; Test for positive if not zero
        JMP     EXPONENT_ZERO
TEST_EXP_POS:
        CMP     CS:SIGN_OF_EXP,0      ; Exponent positive ?
        JNE     EXPONENT_NEG
        JMP     EXPONENT_POS    ; Exponent must be positive
;*******************|
; negative exponent |
;     (option 2)    |
;*******************|
EXPONENT_NEG:
        CALL    TEN_TO_EXP      ;  sig    |  EMPTY  |  EMPTY  |
        FDIVP   ST(1),ST        ; number  |  EMPTY  |
        JMP     EXIT_INPUT_87
;*******************|
; positive exponent |
;    (option 3)     |
;*******************|
EXPONENT_POS:
        CALL    TEN_TO_EXP      ;  sig    |  EMPTY  |  EMPTY  |
        FMULP   ST(1),ST        ; number  |  EMPTY  |
        JMP     EXIT_INPUT_87
;*******************|
;  exponent = 0     |
;   (option 1)      |
;*******************|
EXPONENT_ZERO:
        JMP     EXIT_INPUT_87
;*******************|
;   restore NDP     |
;   environment     |
;*******************|
EXIT_INPUT_87:
        FSTP    CS:TEMP_REAL    ; Save numbers in memory
        FRSTOR  CS:STATE_80X87  ; Restore original NDP environment
        FWAIT
        FLD     CS:TEMP_REAL    ; Load number into NDP stack
        CLC                     ; No error
        MOV     AL,0            ; Error code zero
RESTORE_EXIT:
        POP     DI              ; Restore caller's registers
        POP     SI
        POP     DX
        POP     CX
        RET
FPU_INPUT       ENDP
;*****************************************************************
;*********************************|
;  convert ST(0) to ASCII string  |
;*********************************|
FPU_OUTPUT      PROC    FAR
; Operation:
;          Converts the value in the top register of the FPU
;          stack into an ASCII decimal number in scientific
;          notation
;
; On entry:
;         DS:DX points to an unformatted caller's buffer area
;         with minimum space for 25 bytes of storage
;
;         ST(0) holds value to be converted
;
; On exit:
;         a. Carry flag clear
;         ST(0) holds value to converted
;         The caller's buffer will contain the ASCII decimal
;         representation of the number at the FPU stack top
;         register, formatted in scientific notation, as follows:
;
;                 sm.mmmmmmmmmmmmmmmmm ESeeee
; where
; s = sign of number (blank = +)
; m = 18 significand digits
; . = decimal point following the first significand digit
; E = explicit letter E to signal start of exponent
; S = + or - sign of exponent
; e = up to four exponent digits
;
; Examples of output:
;                    1.78125234500000000 E-12
;                   -3.14163397000000000 E+0
;                    1.22334455667788998 E+1388
;
;       b. Carry flag set (invalid value in top of stack)
;
; Other number types return the following 12-character text srings
; in the caller's buffer
;                        + denormal
;                        - denormal
;                        + NAN
;                        - NAN
;                        + infinity
;                        - infinity
;                        unsupported
;                         EMPTY
;
;      Signed zeros are reported as follows:
;                    0.00000000000000000 E+0 (positive zero)
;                   -0.00000000000000000 E+0 (negative zero)
;
;       AX is destroyed. All other registers and the caller's
;       FPU environment are preserved
;********************************************************************
; CPU compatibility chart:
;                       | 8086/8088/8087  | YES |
;                       | 80286/80287     | YES |
;                       | 80386/80387     | YES |
;                       | 486 DX          | YES |
;                       | PENTIUM         | YES|
;********************************************************************
;
;***************************|
;      save caller's ST(0)  |
;***************************|
        FSTP    CS:INPUT_NUM    ; Save in extended precision
        FWAIT
        FLD     CS:INPUT_NUM    ; Reload ST(0)
;***************************|
;    test for EMPTY ST(0)   |
;***************************|
; The condition of an empty ST(0) must be tested before the FPU
; stack is used
        PUSH    AX              ; Save accumulator
        FXAM                    ; Examine condition codes
        FWAIT                   ; For 8087 compatibility
        FSTSW   CS:STATUS_WORD  ; Store status word in memory
        FWAIT
;***************************|
;   store condition codes   |
;          in BX            |
;***************************|
;  AH bits: 7  6  5  4  3  2  1  0
;              3           2  1  0  condition code bits
;              |___________|__|__|_____ Condition code settings:
        MOV     AX,CS:STATUS_WORD       ; Status word to AX
; Clear all condition codes escept C0 and C3
        AND     AH,01000001B            ; Mask to clear bits
; If C0 and C3 are set ST(0) is tagged EMPTY
        CMP     AH,01000001B
        JNE     ST_NOT_EMPTY        ; Go if both bits not set
;***************************|
;       ST(0) is EMPTY      |
;***************************|
; AX is already in stack
        PUSH    CX              ; Save caller's environment
        PUSH    DX
        PUSH    SI
        LEA     SI,CS:EMPTY_MES ; Error message string
        CALL    FILL_CALLERS    ; Text message to caller's buffer
        POP     SI              ; Restore context
        POP     DX
        POP     CX
        POP     AX              ; Restore accumulator
        STC                     ; Set carry to signal invalid
        RET
;*******************|
; save caller's FPU |
;  environment and  |
;     registers     |
;*******************|
ST_NOT_EMPTY:
        POP     AX              ; Restore stack
        FSAVE   CS:STATE_80X87  ; Save context
        FWAIT
        FLD     CS:INPUT_NUM    ; Restore stack top value
        PUSH    CX              ; Save registers used by routine
        PUSH    DX
        PUSH    SI
        PUSH    DI
;***************************|
;  get number type at ST(0) |
;***************************|
; Code uses the procedure named NUM_AT_ST0 in the UNIT_2 module
; of MATHLIB to identify the number at ST(0)
; DS:DX --> caller's buffer for number string or error message
        PUSH    DX              ; Save caller's buffer
        XCHG    DX,DI           ; Transfer pointer to DI
        CALL    NUM_AT_ST0      ; Library procedure
        POP     DX              ; Restore caller's pointer
;*******************|
; test for invalid  |
;     values        |
;*******************|
; Code is returned in AL. Caller's buffer contains text string
        CMP     AL,1            ; Test for + or - normal
        JA      NOT_1OR2        ; Go if not
        JMP     NORMAL_NUMBER   ; Go to processing routine
; Test for zeros
NOT_1OR2:
        CMP     AL,4            ; Is it positive zero ?
        JNE     TRY_NEG_ZERO    ; If not keep testing
        JMP     IS_POS_ZERO     ; Processing for positive zero
TRY_NEG_ZERO:
        CMP     AL,5            ; Is it negative zero ?
        JNE     INVALID_EXIT    ; Go if not
        JMP     IS_NEG_ZERO     ; Processing for negative zero
;***************************|
;  exit for invalid types   |
;***************************|
INVALID_EXIT:
        FRSTOR  CS:STATE_80X87  ; Restore original FPU environment
        FWAIT
; Restore registers
        POP     DI
        POP     SI
        POP     DX
        POP     CX
        STC                             ; Set carry flag
        RET                             ; to flag invalid type
;***************************|
;  exit for zero types      |
;***************************|
; Zero can be signed positive or negative. ANSI/IEEE 764 requires
; preserving the sign of zero and infinity
IS_POS_ZERO:
        LEA     SI,CS:POS_ZERO          ; Error message string
        JMP     EITHER_ZERO
IS_NEG_ZERO:
        LEA     SI,CS:NEG_ZERO          ; Error message string
EITHER_ZERO:
        CALL    FILL_CALLERS
        MOV     AL,0                    ; No error
        FRSTOR  CS:STATE_80X87  ; Restore original FPU environment
        FWAIT
; Restore registers
        POP     DI
        POP     SI
        POP     DX
        POP     CX
        CLC                             ; No error
        RET
;*******************|
;  extract exponent |
;*******************|
; The binary exponent must be converted into a decimal exponent
NORMAL_NUMBER:
        FSTCW  CS:USERS_CW      ; Save caller's control word
        FWAIT
        FLDCW   CS:ROUND_UP     ; Force rounding up with reduced
                                ; precision to avoid spurious
                                ; rounding of results
        FWAIT
;                               |  ST(0)  |  ST(1)  |  ST(2)  |
        FLD     ST(0)           ;    #    |    #    |  EMPTY  |
        FABS                    ;   |#|   |    #    |  EMPTY  |
        FLD1                    ;    1    |   |#|   |   #     |
        FXCH    ST(1)           ;   |#|   |    1    |   #     |
; Make |#| = x
        FYL2X                   ; Log2-x  |    #    |
        FLDL2T                  ; Log2-10 | Log2-x  |   #     |
        FXCH    ST(1)           ; Log2-x  | Log2-10 |   #     |
        FDIV    ST,ST(1)        ;  exp    | Log2-10 |   #     |
        FXCH    ST(1)           ; Log2-10 |  exp    |   #     |
        FSTP    ST(0)           ;  exp    |   #     |  EMPTY  |
; Change control word to round down
        FLDCW   CS:ROUND_DOWN   ; Force rounding control
        FWAIT
        FRNDINT                 ;  EXP    |   #     |  EMPTY  |
TEST_OVER:
        FLD     ST(0)           ;  EXP    |  EXP    |   #     |
; Change again to normal rounding control
        FBSTP   CS:PACKED_EXP   ;  EXP    |   #     |  EMPTY  |
        FISTP   CS:DECIMAL_EXP  ;   #     |  EMPTY  |
;
        FLDCW   CS:USERS_CW     ; Restore original control word
;*********************|
; extract significand |
;*********************|
; The validity and required precision of the significand field
; is achieved by testing the 18th significand bit of the BCD
; result. If this bit is not a BCD digit, the significand is
; divided by 10. If this bit is 0, the significand is multiplied
; by 10. Conversion is repeated until the 18-th digit is valid
; and not zero
;                               |  ST(0)  |  ST(1) |  ST(2)  |
;                               |   #     |  EMPTY |  EMPTY  |
        FLD     ST(0)           ;   #     |    #   |  EMPTY  |
; Round and store significand, just in case number it is integer
MANTISSA:
        FRNDINT
        FBSTP   CS:PACKED_SIG   ;   #     |    #   |  EMPTY  |
; Test 18-th BCD for invalid or not zero
        FWAIT
        LEA     SI,CS:PACKED_SIG
        ADD     SI,8            ; To most significand BCD byte
        MOV     AL,CS:[SI]
        CMP     AL,0FFH         ; Integer is too large to test
        JE      TOO_LARGE
        AND     AL,11110000B    ; Mask to clear bits 0 to 3
        JNZ     END_MANTISSA
; The required precision is not yet achieved
; Multiply by 10 and repeat
        FILD    CS:TEN          ; Load multiplier 
        FMULP   ST(1),ST        ; Multiply by 10 
        FLD     ST(0)           ; Adjust stack 
        JMP     MANTISSA        ; Repeat conversion
; The significand contains invalid BCD digits
; Divide by 10 and repeat
TOO_LARGE:
        FILD    CS:TEN          ; Load divisor
        FDIVP   ST(1),ST        ; Divide by 10 
        FLD     ST(0)           ; Adjust stack
        JMP     MANTISSA        ; Repeat conversion 
; The significand is valid and to sufficient precision
;*******************|
;    restore FPU    |
;    environment    |
;*******************|
END_MANTISSA:
        FRSTOR  CS:STATE_80X87  ; Restore original FPU environment
        FWAIT
;*******************|
; format caller's   |
;      buffer       |
;*******************|
        LEA     SI,CS:FPD_MES           ; Blank message
        CALL    FILL_CALLERS
;*******************|
;   build decimal   |
;     number        |
;*******************|
; The decimal number is built from the data stored in the buffers
; PACKED_SIG and PACKED_EXP. The output number is formatted in
; exponential form and stored in the caller's buffer
        LEA     SI,CS:PACKED_SIG; Significand source
        ADD     SI,9            ; To sign byte of BCD
        MOV     DI,DX           ; User's buffer area
        MOV     AL,CS:[SI]      ; Get sign byte from BCD
        CMP     AL,0            ; Zero is positive
        JE      POS_SIGNIF      ; Significand is positive
        MOV     BYTE PTR [DI],'-'    ; Set negative sign
POS_SIGNIF:
        INC     DI              ; Bump decimal buffer pointer
        DEC     SI              ; and BCD significand pointer
; Move digit preceding the significand's decimal point
        MOV     AL,CS:[SI]      ; Get packed BCD digits into AL
        CALL    NIBBS_2_DEC     ; Decimal digits in AH and AL
        MOV     [DI],AH         ; Set first decimal digit
        INC     DI              ; Skip decimal point
        INC     DI
        MOV     [DI],AL         ; Set second digit
        INC     DI              ; Bump pointer
; Move next eight BCD digit pairs as a block
        DEC     SI              ; Point to next BCD pair
        MOV     CX,8            ; Digits to move
MOVE_8_DIGITS:
        MOV     AL,CS:[SI]      ; Get packed BCD digits into AL
        CALL    NIBBS_2_DEC     ; Decimal digits in AH and AL
        MOV     [DI],AH         ; Set first digit
        INC     DI              ; Bump decimal digits pointer
        MOV     [DI],AL         ; Place second digit
        INC     DI              ; Bump pointer
        DEC     SI              ; Source pointer to next BCD pair
        LOOP    MOVE_8_DIGITS
        ADD     DI,2            ; Bump to exponent sign position
;*******************|
;  sign of exponent |
;*******************|
        LEA     SI,CS:PACKED_EXP; Exponent BCD storage
        ADD     SI,9            ; To sign byte of BCD
        MOV     BYTE PTR [DI],'+'       ; Assume positive exponent
        MOV     AL,CS:[SI]      ; Get sign byte from BCD
        CMP     AL,0            ; Zero is positive
        JE      POS_EXPONENT    ; Significand is positive
        MOV     BYTE PTR [DI],'-'    ; Set negative sign
POS_EXPONENT:
        INC     DI              ; Bump decimal buffer pointer
;*******************|
; 4 exponent digits |
;*******************|
        LEA     SI,CS:PACKED_EXP; Reset pointer
        INC     SI              ; to first BCD digit pair of
                                ; exponent
; Move 4 exponent digits, skipping leading zeros
; BL = 0 if first non-zero digit not yet found
        MOV     BL,0            ; Switch for first non-zero
        MOV     CX,2            ; Counter for 2 packed BCDs
MOVE_2_DIGITS:
        MOV     AL,CS:[SI]      ; Get packed BCD digits into AL
        CALL    NIBBS_2_DEC     ; Decimal digits in AH and AL
        CMP     BL,0            ; Execute leading zero test ?
        JNE     NO_LEAD_TEST1   ; Switch OFF, skip test
; Test for leading zero
        CMP     AH,'0'          ; Zero ASCII digit ?
        JE      SKIP_LEAD0_1    ; Skip leading zero
; Valid no zero digit, reset lead zero test switch
        MOV     BL,0FFH         ; BL not zero
NO_LEAD_TEST1:
        MOV     [DI],AH         ; Set first decimal digit
        INC     DI              ; Bump buffer pointer
SKIP_LEAD0_1:
; Second test for leading zeros
        CMP     BL,0            ; Execute test ?
        JNE     NO_LEAD_TEST2   ; Switch OFF, skip test
        CMP     AL,'0'          ; Zero ASCII digit ?
        JE      SKIP_LEAD0_2
NO_LEAD_TEST2:
        MOV     [DI],AL         ; Set second digit
; Valid no zero digit, reset lead zero test switch
        MOV     BL,0FFH         ; BL not zero
        INC     DI              ; Bump buffer pointer
SKIP_LEAD0_2:
        DEC     SI              ; To next BCD digit pair
        LOOP    MOVE_2_DIGITS

;*******************|
;  test for all 0s  |
;*******************|
; If exponent is all zeros, the leading zero switch will be set
        CMP     BL,0            ; Is exponent all zeros?
        JNE     EXIT_OUTPUT     ; No, exit
; Set a zero digit in the exponent
        MOV     BYTE PTR [DI],'0'
;*******************|
; restore and exit  |
;*******************|
EXIT_OUTPUT:
        POP     DI
        POP     SI
        POP     DX
        POP     CX
        CLC                     ; No error exit
        RET
FPU_OUTPUT      ENDP
;*****************************************************************
;
;*******************************|
;  ASCII to exponential form    |
;*******************************|
ASCII_TO_EXP    PROC    FAR
; Check for an input string not in exponential form and convert
; On entry:
;        DS:SI --> an ASCII decimal string, terminated in a NULL
;                  byte, which holds the caller's ASCII decimal
;                  input
;             Note: The caller's buffer may contain the $ sign
;                   and commas as formatting symbols. The string
;                   terminator (00H) may be preceded by a CR
;                   (0DH) or LF (0AH) control codes. If the
;                   characters E or e are in the string, the
;                   procedure assumes that the string is already
;                   in exponential notation
;
;        DS:DI --> a 27-byte unformatted output buffer which will
;                  be filled with the output string formatted in
;                  exponential form
;
; On exit:
;        DS:DI --> the caller's output buffer with the input number
;                  reformatted in exponential form, as follows:
;
;                 sm.mmmmmmmmmmmmmmmmm ESeeee
; where
; s = sign of number (blank = +)
; m = 18 significand digits
; . = decimal point following the first significand digit
; E = letter E signals the start of the exponent field
; S = + or - sign of exponent
; e = up to four exponent digits
;                All registers are preserved
;********************************************************************
; CPU compatibility chart:
;                       | 8086/8088/8087  | YES |
;                       | 80286/80287     | YES |
;                       | 80386/80387     | YES |
;                       | 486 DX          | YES |
;                       | PENTIUM         | YES|
;********************************************************************
;
;*******************|
; save entry data   |
;*******************|
; Entry pointers are saved in scratchpad area
        MOV     CS:ENTRY_SI,SI
        MOV     CS:ENTRY_DI,DI
; Caller's registers are saved in stack
        PUSH    AX
        PUSH    CX
        PUSH    DX
        PUSH    SI
        PUSH    DI
;*******************|
;  test for E or e  |
;*******************|
; The caller's string may already be in floating-point format
TEST_4_FPF:
        CMP     BYTE PTR [SI],'E'
        JE      STRING_ISFP             ; String in FP format
        CMP     BYTE PTR [SI],'e'
        JE      STRING_ISFP
        CMP     BYTE PTR [SI],' '       ; Less than 20H is
                                           ; end
                                           ; of string
        JB      NOT_FPF         ; Not floating-point
        INC     SI
        JMP     TEST_4_FPF
;*******************|
;  move FP input    |
; to output buffer  |
;*******************|
STRING_ISFP:
        MOV     SI,CS:ENTRY_SI  ; Restore input string pointer
; Clear 27 bytes in output buffer
        PUSH    DI              ; Save output pointer
        MOV     CX,27           ; Set up digits counter
PREP_27:
        MOV     BYTE PTR [DI],' '    ; Place blank in buffer
        INC     DI
        LOOP    PREP_27
        POP     DI              ; Restore output buffer pointer
; Move up to 27 digits from input to output buffers. No checks
; are made for an invalid input string
        MOV     CX,27           ; Digits counter
MOVE_INPUT:
        MOV     AL,[SI]         ; Get character
        CMP     AL,' '          ; Less than 20H ends input string
        JB      END_ISTRING     ; End move if terminator
        MOV     [DI],AL         ; Move character to output buffer
        INC     SI              ; Bump pointers
        INC     DI
        LOOP    MOVE_INPUT
END_ISTRING:
        JMP     A2FP_EXIT
;*******************|
; clear output area |
;*******************|
NOT_FPF:
        PUSH    DI              ; Save output pointer
        LEA     SI,CS:POS_ZERO  ; Zero is default value
        MOV     CX,27           ; Byte to clear
CLEAR_27:
        MOV     AL,CS:[SI]      ; Get message character
        MOV     [DI],AL         ; Place in buffer area
        INC     DI              ; Bump pointers
        INC     SI
        LOOP    CLEAR_27
; Caller's buffer is now formatted. Restore pointer
        MOV     SI,CS:ENTRY_SI  ; Start of caller's buffer
        POP     DI              ; Output buffer start
;*******************|
;    get sign       |
;*******************|
; Find first significant character in the caller's string
; Skip leading blanks, $ and # sign, and other invalid symbols
 GET_NB:
        MOV     AL,[SI]         ; Get character
        CMP     AL,'+'          ; Test for + sign
        JAE     FIRST_NB        ; Go if equal or larger
        INC     SI
        JMP     GET_NB
; First nonblank is a symbol or number larger than the + sign
FIRST_NB:
        CMP     AL,'+'          ; Test for +
        JE      SIGN_IS_PLUS
        CMP     AL,'-'          ; Test for -
        JE      SIGN_IS_MINUS
; First nonblank character is not a sign. Assume a digit and
; force a positive significand
        JMP     IMPLICIT_POS
SIGN_IS_MINUS:
        MOV     BYTE PTR [DI],'-'    ; Set minus sign in
                                        ; caller's buffer
; Positive numbers are formatted with a leading blank
SIGN_IS_PLUS:
        INC     SI              ; Bump source pointer
IMPLICIT_POS:
        INC     DI              ; Bump destination pointers
;*******************|
;    skip leading   |
;       zeros       |
;*******************|
SKIP_ZERO:
; Leading zeros are not significant. Skip them
        CMP     BYTE PTR [SI],'0'    ; Test for 0
        JNE     NOT_LEAD0
        INC     SI              ; Bump to next digit
        JMP     SKIP_ZERO
;*******************|
; find decimal point|
;*******************|
;  Clear counter for digits preceding the decimal point
NOT_LEAD0:
;
        MOV     DX,0            ; DX is integer digit counter
        PUSH    SI              ; Save pointer to caller's
                                ; leading digit
; Caller's string must end in a NULL byte (00H), but it can have
; other embedded control characters, for example, 0DH, or 20H
; Test for an integer zero input
        CMP     BYTE PTR [SI],' '    ; Smaller than 20H is
                                ; end of input buffer
        JAE     GET_POINT       ; Continue if not end
        JMP     NUM_ISZERO      ; Number is zero
GET_POINT:
        MOV     AL,[SI]         ; Get caller's characters
        CMP     AL,'.'          ; Test for decimal point
        JE      POINT_FOUND
        CMP     AL,','          ; Test for formatting comma
                                ; in the caller's string
        JE      SKIP_COMMA      ; Ignore comma
        CMP     AL,0DH          ; String terminator
        JA      LOOK_4_PT       ; Continue searching
        JMP     NO_POINT        ; No decimal point in string
LOOK_4_PT:
        INC     DX              ; Count one integer digit
SKIP_COMMA:
        INC     SI              ; Bump integer digits counter
        JMP     GET_POINT
;*******************|
;   string is a     |
;  decimal fraction |
;*******************|
; If DX = 0, then the caller's string < 1 and the EXPONENT IS
;            NEGATIVE or the number is zero
; If DX = 1, then the caller's string is already normalized
;            and the EXPONENT = +0
; If DX > 1, then the caller's string is a positive fraction and
;            the EXPONENT IS POSITIVE
;
POINT_FOUND:
        CMP     DX,1            ; Test for normalized input
        JNE     TEST_4_DX0
;*******************|
;     DX = 1        |
;*******************|
; Number is between 1 and 10. Set exponent sign and value
        MOV     SI,CS:ENTRY_DI  ; SI --> start of output buffer
        ADD     SI,22           ;    --> first character after E
; DS:SI --> position for exponent sign in output buffer
        MOV     BYTE PTR [SI],'+'    ; Sign must be +
        INC     SI
        MOV     BYTE PTR [SI],'0'    ; Set zero exponent
; Move exponent digits from conversion buffer into output buffer
        JMP     EXP_MOVED
TEST_4_DX0:
        CMP     DX,0            ; Test for negative exponent
        JE      DX_ISZERO       ; Go to DX = 0 processing
; If not zero, treat number as integer
        JMP     POS_FRACTION
;*******************|
;     DX = 0        |
;*******************|
; Possible cases:
; nonzero -->   000.002345
; nonzero -->      .2345
;    zero -->     0.00
;    zero -->      .0
; Count decimal digits to normalized position, that is, following
; the first nonzero digit
; DX is used as counter
; SI --> decimal point
DX_ISZERO:
        INC     SI              ; Skip decimal point
        INC     DX              ; Count one digit
NORM_SIGNF:
        CMP     BYTE PTR [SI],'0'    ; Test for zero
        JE      ZERO_DIGIT
        JB      NUM_ISZERO
; If not equal or less, it must be larger
        JMP     OK_SIGNF        ; Normalized position found
ZERO_DIGIT:
        INC     SI              ; Bump pointer
        INC     DX              ; and digits counter
        JMP     NORM_SIGNF
;*******************|
;  input string = 0 |
;*******************|
NUM_ISZERO:
; Zero is the default floating-point value in the buffer,
; but exponent may have been changed
        MOV     SI,CS:ENTRY_DI  ; SI --> start of output buffer
        ADD     SI,22           ;    --> first character after E
; DS:SI --> position for exponent sign in output buffer
        MOV     BYTE PTR [SI],'+'    ; Sign must be +
        INC     SI
        MOV     BYTE PTR [SI],'0'    ; Set zero exponent
        POP     SI                   ; Restore stack
        JMP     A2FP_EXIT       ; Exit conversion
;
OK_SIGNF:
; Test for an invalid digit indicating a zero input
        POP     AX              ; Discard old pointer in stack
        PUSH    SI              ; Save pointer to first nonzero
                                ; digit
        MOV     SI,CS:ENTRY_DI  ; SI --> start of output buffer
        ADD     SI,22           ;    --> first character after E
; DS:SI --> position for exponent sign in output buffer
        MOV     BYTE PTR [SI],'-'    ; Sign must be -
        JMP     CONV_EXPONENT
;
;***********************|
;   string is integer   |
; or positive fractions |
;***********************|
NO_POINT:
POS_FRACTION:
; Exponent of integer is positive or number is a positve fraction
        MOV     SI,CS:ENTRY_DI  ; SI --> start of output buffer
        ADD     SI,22           ;    --> first character after E
; DS:SI --> position for exponent sign in output buffer
        MOV     BYTE PTR [SI],'+'    ; Sign must be +
        INC     SI
; String has no decimal point. DX = exponent plus 1
        DEC     DX
CONV_EXPONENT:
        LEA     DI,CS:EXP_BUFFER       ; Buffer for conversion
        CALL    BIN_TO_ASC      ; Binary in DX to a 5-digit
                                ; ASCII in EXP_BUFFER
        MOV     SI,CS:ENTRY_DI  ; SI --> start of output buffer
        ADD     SI,23           ;    --> first character after
                                ; exponent sign
; Caller's string pointer in stack
; DI --> first nonzero ASCII digit in conversion buffer
; Move exponent digits from conversion buffer into output buffer
TO_OUTPUT_BUF:
        MOV     AL,CS:[DI]      ; Get character
        CMP     AL,0            ; Test for end of buffer
        JE      EXP_MOVED       ; Exponent digits have been moved
        MOV     [SI],AL         ; Place digit in output buffer
        INC     SI              ; Bump pointers
        INC     DI
        JMP     TO_OUTPUT_BUF
;***************************|
; reset pointer to string   |
;***************************|
EXP_MOVED:
        POP     SI              ; Pointer to first nonzero digit
                                ; in the caller's input
;*******************|
;  move up to 18    |
;     digits        |
;*******************|
        MOV     DI,CS:ENTRY_DI  ; Start of output buffer
        INC     DI              ; Skip sign of significand
; Move one digit preceding the decimal point. Caller's string
; will have at least one digit at this point
        MOV     AL,[SI]         ; Get digit
; Test for invalid values
        CMP     AL,'0'          ; Lower limit
        JAE     VALID_1         ; Go if OK
        PUSH    SI              ; Adjust stack
        JMP     NUM_ISZERO      ; Force zero if invalid digit
VALID_1:
        CMP     AL,'9'          ; Upper limit of digit
        JBE     VALID_2         ; Go if OK
        PUSH    SI              ; Adjust stack
        JMP     NUM_ISZERO      ; Force zero if invalid digit
VALID_2:
        MOV     [DI],AL         ; Move to buffer
        INC     DI              ; Bump buffer pointers
        INC     SI
        INC     DI              ; Skip decimal point in buffer
        MOV     CX,17           ; Load digits counter
UPTO_17:
        MOV     AL,[SI]         ; Get digit
        CMP     AL,'.'          ; Test for period symbol
                                ; in caller's string
        JE      SKIP_SYMBOL     ; Skip it
        CMP     AL,','          ; and for commas
        JE      SKIP_SYMBOL
        CMP     AL,' '          ; Test for terminator symbol
        JB      A2FP_EXIT       ; Go if less than 20H
        JE      SKIP_SYMBOL     ; Skip if space
; Test for invalid values
        CMP     AL,'0'          ; Lower limit
        JAE     VALID_3         ; Go if OK
        PUSH    SI              ; Adjust stack
        JMP     NUM_ISZERO      ; Force zero if invalid digit
VALID_3:
        CMP     AL,'9'          ; Upper limit of digit
        JBE     VALID_4         ; Go if OK
        PUSH    SI              ; Adjust stack
        JMP     NUM_ISZERO      ; Force zero if invalid digit
VALID_4:
        MOV     [DI],AL         ; Move digit into buffer
        INC     DI              ; Bump pointers
        INC     SI
        LOOP    UPTO_17
        JMP     A2FP_EXIT
; Skip symbol without decrementing character counter
SKIP_SYMBOL:
        INC     SI
        JMP     UPTO_17
;*******************|
; restore and exit  |
;*******************|
A2FP_EXIT:
        POP     DI              ; Restore caller's data registers
        POP     SI
        POP     DX
        POP     BX
        POP     AX
        RET
;
ASCII_TO_EXP    ENDP
;
;****************************************************************
;                      local support procedures
;****************************************************************
;***************************|
;    code segment constants |
;***************************|
; Tables of constants holding 40 values, 10 for each processing
; stage
TABLE_1000      DT      0, 1.0E1000, 1.0E2000, 1.0E3000, 1.0E4000
TABLE_100       DT      0, 1.0E100, 1.0E200, 1.0E300, 1.0E400
                DT      1.0E500, 1.0E600, 1.0E700, 1.0E800
                DT      1.0E900
TABLE_10        DT      0, 1.0E10, 1.0E20, 1.0E30, 1.0E40
                DT      1.0E50, 1.0E60, 1.0E70, 1.0E80
                DT      1.0E90
TABLE_1         DT      0, 1.0E1, 1.0E2, 1.0E3, 1.0E4
                DT      1.0E5, 1.0E6, 1.0E7, 1.0E8, 1.0E9
;
TEN_TO_EXP       PROC    NEAR
; Exact calculation of 10**y by exponent factorization
; On entry:
;         CS:DECIMAL_EXP = exponent
;
; On exit:
;         ST(0) = 10**y
;
;*****************************|
; move data to work variable  |
;*****************************|
        MOV     CX,CS:DECIMAL_EXP   ; Exponent to loop register
;*****************|
;     STAGE 3     |
;********************************|
;    calculate (10**F3)**I3      |
;********************************|
; Test for y > 1000
; First load 1 for the case that exp < 1000
        FLD1                    ;    1    | ------- | ------- |
        CMP     CX,1000         ; Is exp > 1000
        JNA     STAGE_2V3       ; Go if not above than 100
        FSTP    ST(0)           ; ------- | ------- | ------- |
; At this point CX is > 1000
;**************************|
;   calculate element I3   |
;      INT (y/1000)        |
;**************************|
; The division by 1000 is performed in 2 steps in order to avoid
; a divide overflow
        MOV     DX,CX           ; Save original exponent in DX
        MOV     AX,CX           ; Exponent to AX
        MOV     BX,100          ; First divisor
        DIV     BL              ; Exponent / 100
        MOV     AH,0            ; Clear remainder of byte division
        MOV     BX,10           ; Second divisor
        DIV     BL              ; First dividend / 10
        MOV     AH,0            ; Clear remainder of byte division
        MOV     CX,AX           ; Copy divisor to CX
; Calculate R3 = y MOD (I3 * F3)
; At this point:
;               DX = y
;               CX = INT(y/1000) = I3
        PUSH    DX              ; Save y in stack
        MOV     AX,1000         ; F3 to AX
        MUL     CX              ; AX = I3 * F3
        POP     DX              ; Restore y
        SUB     DX,AX           ; DX = y MOD (I3 * F3)
; CX = INT(y/1000)
        MOV     AX,CX           ; AX = y MOD 1000 (not 0)
        MOV     CL,10           ; Multiplier to CL
        MUL     CL              ; AX x 10 = offset in table
        MOV     SI,AX           ; Offset to index register
        FLD     CS:TABLE_1000[SI]
; Test for exponent exact multiple of 1000 (no remainder)
        MOV     CX,DX           ; R3 to CX
        CMP     CX,0            ; In this case CX = 0
        JNE     STAGE_2V3       ; Go if not
;*************************|
;  quick exit if REM = 0  |
;*************************|
; Return if exponent is exact multiple of 1000
        RET                     ; 10**y   | ------- |
;
;*****************|
;     STAGE 2     |
;********************************|
;    calculate (10**F2)**I2      |
;********************************|
; At this point:                |  ST(0)  |  ST(1)  | ST(2)  |
;                               |   S3    | ------- |
;                               |  or 1   |
STAGE_2V3:
; Test for EXP > 100
; First load 1 for the case that exp < 100
        FLD1                    ;    1    |   S3    | ------- |
        CMP     CX,100          ; Is exp > 100
        JNA     STAGE_1V3       ; Go if not above than 100
        FSTP    ST(0)           ;    S3   | ------- | ------- |
; At this point CX is > 100
;**************************|
;   calculate element I2   |
;      INT (y/100)         |
;**************************|
; CX = INT(y/1000)
        MOV     AX,CX           ; Move S3 to AX
        MOV     DX,CX           ; Update DX
        MOV     BX,100          ; Divisor
        DIV     BL              ; Exponent / 100
        MOV     AH,0            ; Clear reminder
        MOV     CX,AX           ; Copy divisor to CX
; Calculate R2 = R3 MOD (I2 * F2)
; At this point:
;               DX = R3
;               CX = INT(y/100) = I2
        PUSH    DX              ; Save y in stack
        MOV     AX,100          ; F2 to AX
        MUL     CL              ; AX = I2 * F2
        POP     DX              ; Restore y
        SUB     DX,AX           ; DX = R3 MOD (I2 * F2)
; CX = INT(y/100)
        MOV     AX,CX           ; AX = REM y MOD 25 (not 0)
        MOV     CL,10           ; Multiplier to CL
        MUL     CL              ; AX x 10 = offset in table
        MOV     SI,AX           ; Offset to index register
        FLD     CS:TABLE_100[SI]
; AX holds R2
        MOV     CX,DX           ; Copy S2 to CX
; Test for exponent exact multiple of 100 (no remainder)
        CMP     CX,0            ; In this case CX = 0
        JNE     STAGE_1V3       ; Go if not
;*************************|
;  quick exit if REM = 0  |
;*************************|
; Return if exponent is exact multiple of 100
;                               |    S3    |    S2   | -------- |
        FMUL    ST,ST(1)
        FSTP    ST(1)           ; 10**y   | ------- | -------- |
        RET                     ; Exponent is exact multiple of 100
;
;*****************|
;     STAGE 1     |
;********************************|
;    calculate (10**F1)**I1      |
;********************************|
; At this point:                |  ST(0)  |  ST(1)  |  ST(2)  |
;                               |   S2    |   S3    | ------- |
;                               |  or 1   |   or 1
; CX = INT(y/10)
; Test for EXP > 10
; First load 1 for the case that exp < 10 
STAGE_1V3:
        FLD1                    ;    1    |    S2   |    S3   |
        CMP     CX,10           ; Is exp > 25 
        JNA     STAGE_0V3       ; Go if not above 10
        FSTP    ST(0)           ;   S2    |    S3   | ------- |
; At this point CX is > 10
;**************************|
;   calculate element I1   |
;      INT (y/10)          |
;**************************|
; CX = INT(y/100)
        MOV     AX,CX           ; New exponent to AX
        MOV     DX,CX           ; Update DX
        MOV     BX,10           ; Divisor
        DIV     BL              ; Exponent / 25 
        MOV     AH,0            ; Clear reminder
        MOV     CX,AX           ; Copy divisor to CX
; Calculate R1 = R2 MOD (I1 * F1)
; At this point:
;               DX = R2
;               CX = INT(y/10) = I1
        PUSH    DX              ; Save y in stack
        MOV     AX,10           ; F1 to AX
        MUL     CL              ; AX = I1 * F1
        POP     DX              ; Restore y
        SUB     DX,AX           ; DX = R2 MOD (I1 * F1)
; CX = INT(y/10)
        MOV     AX,CX           ; AX = REM y MOD 25 (not 0)
        MOV     CL,10           ; Multiplier to CL
        MUL     CL              ; AX x 10 = offset in table
        MOV     SI,AX           ; Offset to index register
        FLD     CS:TABLE_10[SI]
; DX holds remainder
        MOV     CX,DX           ; R1 to CX
; Test for exponent exact multiple of 10
        CMP     CX,0            ; In this case CX = 0
        JNE     STAGE_0V3       ; Go if not
;*************************|
;  quick exit if REM = 0  |
;*************************|
; Return if exponent is exact multiple of 10
;                               |    S1   |    S2   |    S3   |
        FMUL    ST,ST(1)
        FSTP    ST(1)           ;    PP   |    S3   | ------- |
        FMUL    ST,ST(1)
        FSTP    ST(1)           ; 10**y   | ------- | -------- |
        RET
;
;*****************|
;     STAGE 0     |
;********************************|
;    calculate (10**F0)**I0      |
;********************************|
; At this point:                |  ST(0)  |  ST(1)  |  ST(2)  |
;                               |   S1    |    S2   |   S3    |
;                               |  or 1   |   or 1  |   or 1  |
; CX = INT(y/10)
STAGE_0V3:
        MOV     AX,CX           ; AX = REM y MOD 25 (not 0)
        MOV     CL,10           ; Multiplier to CL
        MUL     CL              ; AX x 10 = offset in table
        MOV     SI,AX           ; Offset to index register
        FLD     CS:TABLE_1[SI]
;                               |    P0    |    P1    |    P2    |.. 
        FMUL    ST,ST(1)
        FSTP    ST(1)           ;    PP    |    P2    |    P3    |
        FMUL    ST,ST(1)
        FSTP    ST(1)           ;    PP    |    P3    | -------- |
        FMUL    ST,ST(1)
        FSTP    ST(1)           ;  10**y   | -------- | -------- |
        RET
;
TEN_TO_EXP       ENDP
;
;*****************************************************************
;
ASC_2_BIN       PROC    NEAR
; Convert an ASCII number to binary
; On entry:
;               SI --> start ASCII buffer
; On exit:
;               DX = binary number (range 0 to 65535)
;
        MOV     DX,0            ; Clear binary output
ASC0:   MOV     AL,[SI]         ; Get ASCII digit
        INC     SI              ; Bump pointer to next digit
        SUB     AL,30H          ; ASCII to decimal
        JL      EXASC           ; Exit if invalid ASCII
        CMP     AL,9            ; Test for highest value
        JG      EXASC           ; Exit if larger than 9
        CBW                     ; Extend to AX
        PUSH    AX              ; Save digit in stack
        MOV     AX,DX           ; Move into output register
        MOV     CX,10           ; Decimal multiplier
        MUL     CX              ; Perform AX = AX * 10
        MOV     DX,AX           ; Move product to output register
        POP     AX              ; Restore decimal digit
        ADD     DX,AX           ; Add in digit
        JMP     ASC0            ; Continue
EXASC:  RET
ASC_2_BIN       ENDP
;****************************************************************
;
BIN_TO_ASC      PROC    NEAR
; Convert a 16-bit binary in the DX register to an ASCII decimal
; number
; On entry:
;               DX = binary source
;               CS:DI --> 5-byte output buffer
;
; Clear the buffer with blanks
        PUSH    DI              ; Save buffer start
        MOV     CX,5            ; Five digits to clear
        MOV     AL,' '          ; Blank character
CLEAR_5:
        MOV     CS:[DI],AL      ; Clear digit
        INC     DI              ; Bump pointer
        LOOP    CLEAR_5         ; Repeat for 5 digits
        DEC     DI              ; Adjust buffer pointer to last
                                ; digit
;
        MOV     CX,0            ; Clear counter
BINA0:  PUSH    CX              ; Save count
        MOV     AX,DX           ; Add in numerator
        MOV     DX,0            ; Clear top half
        MOV     CX,10           ; Enter decimal divisor
        DIV     CX              ; Perform division AX/CX
        XCHG    AX,DX           ; Get quotient
        ADD     AL,30H          ; Make digit ASCII
        MOV     CS:[DI],AL      ; Store digit in buffer
        DEC     DI              ; Bump destination pointer
        POP     CX              ; Restore counter
        INC     CX              ; Count the digit
        CMP     DX,0            ; Test for end of binary
        JNZ     BINA0           ; Continue if not end
;
        POP     DI              ; Restore pointer to start of
                                ; buffer
; Skip leading blanks in buffer
BUF_BLANK:
        MOV     AL,CS:[DI]      ; Get buffer digit
        CMP     AL,' '          ; Test for ASCII blank
        JNE     NOT_BLANK       ; Exit if not leading blank
        INC     DI
        JMP     BUF_BLANK
; DI --> first nonzero digit in buffer
NOT_BLANK:
        RET
BIN_TO_ASC      ENDP
;****************************************************************
;
NIBBS_2_DEC     PROC    NEAR
; Convert binary value in low and high nibble of AL into two
; ASCII decimal digits in AH and AL
; On entry:
;       AL holds two packed BCD digits in the range 0 to 9
;
; On exit:
;       AH = high-order ASCII digit (from AL bits xxxx ????)
;       AL = low-order ASCII digit (from AL bits ???? xxxx)
;
        PUSH    CX              ; Preserve entry register
        MOV     AH,AL           ; Copy AL in AH
        MOV     CL,4
        SHR     AH,CL           ; Isolate high nibble
        ADD     AH,30H          ; Convert to ASCII
;
        AND     AL,0FH          ; Isolate low nibble
        ADD     AL,30H          ; Convert to ASCII
        POP     CX
        RET
;
NIBBS_2_DEC     ENDP
;****************************************************************
;
FILL_CALLERS    PROC    NEAR
; Fill caller's buffer by DS:DX with 27-byte message by CS:SI
        MOV     DI,DX           ; Set up destination pointer
        MOV     CX,27           ; Characters to clear
TRANSFER_26:
        MOV     AL,CS:[SI]      ; Character from source
        MOV     [DI],AL         ; Into buffer area
        INC     DI              ; Bump pointers
        INC     SI
        LOOP    TRANSFER_26
        RET
FILL_CALLERS    ENDP
;
CODE    ENDS
        END

 pointers
        INC     SI
        LOOP    TRANSFER_26
        RET
FILL_