;*****************************************************************
;*****************************************************************
;                           UNIT_7.ASM
;*****************************************************************
;*****************************************************************
; Library of math routines from the book SOFTWARE SOLUTIONS by
; Julio Sanchez and Maria P. Canton
; Creation date:  March 19, 1994
; Updated:        March 22, 1994
;
; Libary name: MATHUNIT.LIB
;
;*****************************************************************
;                           publics
;*****************************************************************
; MATHEMATICAL ROUTINES IN THIS MODULE
; Calculator-type primtives
PUBLIC  FACTORIAL       ; .. find x!
PUBLIC  ORDER           ; .. organize x > y
PUBLIC  MAX             ; .. find larger or two
PUBLIC  MIN             ; .. find smaller of two
PUBLIC  MODULUS         ; .. remainder of x/y
PUBLIC  PITHAGORAS      ; .. solve triangle using Pithagoras
PUBLIC  SOLVE_FOR_A     ; .. solve triangle using tangent
PUBLIC  INT_FRAC        ; .. separate integer and fractional parts
;
; Solution of a quadratic equation
PUBLIC  QUAD_FORM_REAL  ; .. solve equation for real roots
PUBLIC  QUAD_FORMULA    ; .. solve equation for all roots
;
; Complex arithmetic
PUBLIC  ADD_CMPLX       ; .. add complex numbers
PUBLIC  SUB_CMPLX       ; .. subtract idem
PUBLIC  MUL_CMPLX       ; .. multiply idem
PUBLIC  DIV_CMPLX       ; .. divide idem
PUBLIC  ABS_CMPLX       ; .. find absolute value of complex number
PUBLIC  POLAR_TO_CART   ; .. convert polar to Cartesian coordinates
PUBLIC  CART_TO_POLAR   ; .. convert Cartesian to polar coordinates
;
; Financial calculations
PUBLIC  PAYMENTS        ; .. periodic payments on a loan
PUBLIC  PRESENT_VALUE   ; .. present value of an investment
PUBLIC  FUTURE_VALUE    ; .. future value of an investment
PUBLIC  SL_DEPRECIATION ; .. straight-line depreciation
PUBLIC  COMP_PERIODS    ; .. number of compounding periods
;
;*****************************************************************
;                           externals
;*****************************************************************
; In other modules of the MATHUNIT library
EXTRN   DEG_TO_RAD:     FAR
EXTRN   TANGENT:        FAR
EXTRN   COSINE:         FAR
EXTRN   SINE:           FAR
EXTRN   ARC_TANGENT:    FAR
EXTRN   RAD_TO_DEG:     FAR
EXTRN   X_TO_Y_BYBP:    FAR
EXTRN   LOG_E:          FAR
;
;****************************************************************
;                          code segment
;****************************************************************
;
CODE    SEGMENT BYTE
        ASSUME  CS:CODE
;*****************************************************************
;              calculator-type arithmetic procedures
;*****************************************************************
; calculate factorial
; order FPU data
; find maximum
; find minimum
; find modulus
; solve triangle using Pithagoras
; solve triangle in terms of side and adjacent angle
; find integer and fractional parts
;*****************************************************************
;***************************|
;    code segment data      |
; for calculator arithmetic |
;***************************|
FAC_ARGUMENT    DW      0       ; Storage for factorial argument
ROUND_DOWN      DW      177FH   ; Control word for round down
                                ; operations
CONTROL_WW      DW      0       ; Storage for caller's control word
;
;
FACTORIAL       PROC    FAR
; Calculation of the factorial function
; On entry:
;         ST(0) = x (x is an integer)
; On exit:
;         ST(0) = x!
;
        FRNDINT                 ; Make sure argument is integer
        FABS                    ; and positive
        FIST    CS:FAC_ARGUMENT         ; Store argument in memory
        MOV     CX,CS:FAC_ARGUMENT      ; Load into counter register
        CMP     CX,0            ; Test for 0!
        JE      FACTORIAL_ZERO  ; Special case is 0!
        DEC     CX              ; Number of factors is 1 less
        JNZ     FAC_OK
; At this point operation is 0! or 1!
FACTORIAL_ZERO:
        FSTP    ST(0)           ; Clear stack
        FLD1                    ;    1     | -------- |
        RET
;                               ;    x    | --------- | -------- |
FAC_OK:
        FLD1                    ;    1    |     x     | -------- |
        FXCH                    ;    x    |     1     | -------- |
        FLD     ST(0)           ;    x    |     x     |    1     |
DO_FACTORS:
        FSUB    ST,ST(2)        ;  x - 1  |  pfact    |    1     |
        FXCH                    ;  pfact  |   x - 1   |    1     |
        FMUL    ST,ST(1)        ;  pfact  |   x - 1   |    1     |
        FXCH                    ;  x - 1  |   pfact   |    1     |
        LOOP    DO_FACTORS      ; Repeat
        FSTP    ST(0)           ;   x!    |     1     | -------- |
        FSTP    ST(1)           ;   x!    | --------- |
        RET
FACTORIAL       ENDP
;*****************************************************************
;
STATUS_WW       DW      0       ; Storage for status word
;
ORDER   PROC    FAR
; Find largest of two values, x and y
;
; On entry:
;          ST(0) = x
;          ST(1) = y
; On exit:
;          ST(0) = the larger of x and y
;          ST(1) = the smaller of x and y
;
        FCOM                    ;    x     |     y    | --------- |
        FSTSW   CS:STATUS_WW    ; Store status word
        FWAIT
        MOV     AH,BYTE PTR CS:STATUS_WW + 1
        SAHF                    ; AH to 80x86 flags register
        JA      STACK_OK        ; Go if x > y
; At this point ST(0) =< ST(1)
        FXCH                    ;   y      |     x    | --------- |
STACK_OK:
        RET
ORDER   ENDP
;*****************************************************************
;
MAX     PROC    FAR
; Find largest of two values, x and y
;
; On entry:
;          ST(0) = x
;          ST(1) = y
; On exit:
;          ST(0) = the larger of x and y
;          ST(1) = empty
;
        FCOM                    ;    x     |     y    | --------- |
        FSTSW   CS:STATUS_WW    ; Store status word
        FWAIT
        MOV     AH,BYTE PTR CS:STATUS_WW + 1
        SAHF                    ; AH to 80x86 flags register
        JA      MAX_IS_X        ; Go if x > y
; At this point ST(0) =< ST(1)
        FXCH                    ;   y      |     x    | --------- |
MAX_IS_X:
        FSTP    ST(1)           ; Scrap ST(1)
        RET
MAX     ENDP
;******************************************************************
;
MIN     PROC    FAR
; Find smallest of two values, x and y
;
; On entry:
;          ST(0) = x
;          ST(1) = y
; On exit:
;          ST(0) = the smaller of x and y
;          ST(1) = empty
;
        FCOM                    ;    x     |     y    | --------- |
        FSTSW   CS:STATUS_WW    ; Store status word
        FWAIT
        MOV     AH,BYTE PTR CS:STATUS_WW + 1
        SAHF                    ; AH to 80x86 flags register
        JB      MIN_IS_X        ; Go if x > y
; At this point ST(0) =< ST(1)
        FXCH                    ;   y      |     x    | --------- |
MIN_IS_X:
        FSTP    ST(1)           ; Scrap ST(1)
        RET
MIN     ENDP
;*****************************************************************
;
MODULUS         PROC    FAR
; Find x (MOD y), which is the remainder of x/y
;
; On entry:
;         ST(0) = x (argument)
;         ST(1) = y (modulus)
;
; On exit:
;        ST(0) = x (MOD y)

;                               ;  ST(0)    |   ST(1)   |   ST(2) |
                                ;    x      |    y      | ------- |
REPEAT_MOD:
        FPREM                   ; REM(x/y)  |    y      | ------- |
        FSTSW   CS:STATUS_WW    ; Store status bits in memory
        FWAIT                   ; Delay until end of store
        MOV     AH,BYTE PTR CS:STATUS_WW+1
                        ; Move status bits into AH
        AND     AH,00000100B    ; Mask off all except C2
        JNZ     REPEAT_MOD      ; Repeat if C2 not clear
        FSTP    ST(1)           ; x (MOD y) | -------- |
        RET
MODULUS         ENDP
;*****************************************************************
PITHAGORAS      PROC    FAR
; Solve the Pithagorean equation: c = SQR (x**2 + y**2)
;
; On entry:
;         ST(0) = x
;         ST(1) = y
;
; On exit:
;         ST(0) = SQR (x**2 + y**2)
;
;                               ;  ST(0)    |   ST(1)   |   ST(2) |
                                ;    x      |    y      | ------- |
        FMUL    ST(0),ST        ;   x**2    |    y      | ------- |
        FXCH                    ;    y      |   x**2    | ------- |
        FMUL    ST(0),ST        ;   y**2    |   x**2    | ------- |
        FADDP   ST(1),ST        ;y**2 + x**y| --------- |
        FSQRT                   ;    c      | --------- |
        RET
PITHAGORAS      ENDP
;******************************************************************
;
SOLVE_FOR_A     PROC    FAR
; Solve a right triangle in terms of one side and the adjacent
; angle (A) using the formula
;
;                     a = b * Tan A
;
; On entry:
;         ST(0) = angle A (in degrees)
;         ST(1) = side b
;
; On exit:
;         ST(0) = side a
;
;                               ;  ST(0)   |  ST(1)   |   ST(2) |
                                ;    A     |     b    | ------- |
; Convert degrees to radians
        CALL    DEG_TO_RAD      ;  A (rads)|     b    | ------- |
        CALL    TANGENT         ; Tan A    |     b    | ------- |
        FMULP   ST(1),ST        ; b * Tan A| -------- |
        RET
SOLVE_FOR_A     ENDP
;*****************************************************************
;
INT_FRAC        PROC    FAR
; Separate integer and fractional parts a number in ST(0)
;
; On entry:
;         ST(0) = n
;
; On exit:
;         ST(0) = integer part of n
;         ST(1) = fractional part of n
;**********************|
;   set to round down  |
;**********************|
        FSTCW   CS:CONTROL_WW   ; Store control word in memory
        FLDCW   CS:ROUND_DOWN   ; Install new control word
; At this point the FPU is set to round down
;                               |   ST(0)   |   ST(1)  |   ST(2)
                                ;     n     | -------- |
        FLD     ST(0)           ;     n     |     n    | -------- |
        FRNDINT                 ;     i     |     n    | -------- |
        FLDCW   CS:CONTROL_WW   ; Restore original control word
        FSUB    ST(1),ST        ;     i           f    | -------- |
        RET
INT_FRAC        ENDP
;
;*****************************************************************
;            solving quadratic equations by formula
;*****************************************************************
;****************************|
;     code segment data      |
;   for quadratic formula    |
;****************************|
CONST_A         DT      0       ; Storage for constant a
CONST_B         DT      0       ; Storage for constant b
CONST_C         DT      0       ; Storage for constant c
DISCRIMINANT    DT      0       ; Storage for discriminant
;
QUAD_FORM_REAL  PROC    FAR
; Solution of a quadratic equation in the form:
;
;                     ax**2 + bx + c = 0
;
; by means of the formula
;
;                   -b + SQR [(b**2 - 4ac)]
;            x(1) = -------------------------
;                            2a
;
;                   -b - [SQR (b**2 - 4ac)]
;            x(2) = -------------------------
;                            2a
;
; where x(1) and x(2) are two possible roots and a, b, and c are
; constants
;
; On entry:
;                 ST(0) = c
;                 ST(1) = b
;                 ST(2) = a
; On exit:
;                 ST(0) = x(1)
;                 ST(1) = x(2)
;
;                               |  ST(0)   |   ST(1)  |   ST(2)
                                ;    c     |     b    |     a    |
;***************************|
;      store constants      |
;***************************|
        FSTP    CS:CONST_C      ;    b     |    a     | -------- |
        FSTP    CS:CONST_B      ;    a     | -------- |
        FSTP    CS:CONST_A      ; -------- |
;****************************|
;  evaluate the discriminant |
;      SQR (b**2 - 4ac)      |
;****************************|
        FLD     CS:CONST_C      ;    c     | -------- |
        FLD     CS:CONST_A      ;    a     |     c    | -------- |
        FMULP   ST(1),ST        ;    ac    | -------- |
        FLD1                    ;    1     |    ac    | -------- |
        FADD    ST(0),ST        ;    2     |    ac    | -------- |
        FADD    ST(0),ST        ;    4     |    ac    | -------- |
        FMULP   ST(1),ST        ;   4ac    | -------- |
        FLD     CS:CONST_B      ;    b     |    4ac   | -------- |
        FMUL    ST(0),ST        ;  b**2    |    4ac   | -------- |
        FXCH                    ;   4ac    |  b**2    | -------- |
        FSUB                    ; b**2-4ac | -------- |
        FSQRT                   ;  dscr    | -------- |
; Store discriminant
        FSTP    CS:DISCRIMINANT ; -------- |
;****************************|
;  evaluate the root x(1)    |
;****************************|
        FLD     CS:CONST_A      ;    a     | --------|
        FLD1                    ;    1     |    a    | -------- |
        FADD    ST(0),ST        ;    2     |    a    | -------- |
        FMULP   ST(1),ST        ;   2a     | ------- |
        FLD     CS:CONST_B      ;    b     |    2a   | -------- |
        FCHS                    ;   -b     |    2a   | -------- |
        FLD     CS:DISCRIMINANT ;  dscr    |   -b    |    2a    |
        FADDP   ST(1),ST        ; -b + dscr|    2a   | -------- |
        FDIVR                   ;   x(1)   | ------- |
; ST(0) holds root x(1)
;****************************|
;  evaluate the root x(2)    |
;****************************|
        FLD     CS:CONST_A      ;    a     |   x(1)  |
        FLD1                    ;    1     |    a    |    x(1)  |
        FADD    ST(0),ST        ;    2     |    a    |    x(1)  |
        FMULP   ST(1),ST        ;   2a     |   x(1)  | -------- |
        FLD     CS:CONST_B      ;    b     |    2a   |    x(1)  |
        FCHS                    ;   -b     |    2a   |    x(1)  |
        FLD     CS:DISCRIMINANT ;  dscr    |   -b    |    2a    |
        FSUBP   ST(1),ST        ; -b - dscr|    2a   |    x(1)  |
        FDIVR                   ;   x(2)   |   x(1)  | -------- |
        FXCH                    ;   x(1)   |   x(2)  | -------- |
        RET
QUAD_FORM_REAL  ENDP
;*****************************************************************
;
QUAD_FORMULA    PROC    FAR
; Solution of a quadratic equation in the form:
;
;                     ax**2 + bx + c = 0
;
; in real or complex numbers, by means of the formula
;
;                   -b + SQR [(b**2 - 4ac)]
;            x(1) = -------------------------
;                            2a
;
;                   -b - [SQR (b**2 - 4ac)]
;            x(2) = -------------------------
;                            2a
; where x(1) and x(2) are two possible roots and a, b, and c are
; constants
;
; Solution sets:
;          x(1) and x(2) are the real roots
;          x(r) +/- x(i) is the complex root
;
; On entry:
;                 ST(0) = c
;                 ST(1) = b
;                 ST(2) = a
; On exit:
;        carry flag clear if roots are real numbers, then
;                 ST(0) = x(1)
;                 ST(1) = x(2)
;        carry flag set if roots are complex, then
;        solution is:
;                 ST(0) = -b/2a = real part = x(r)
;                 ST(1) = +/- i * SQR(b**2 - 4ac)/2a = imaginary
;                         part = +/- x(i)
;
;                               |  ST(0)   |   ST(1)  |   ST(2)
                                ;    c     |     b    |     a    |
;***************************|
;      store constants      |
;***************************|
        FSTP    CS:CONST_C      ;    b     |    a     | -------- |
        FSTP    CS:CONST_B      ;    a     | -------- |
        FSTP    CS:CONST_A      ; -------- |
;****************************|
;  evaluate the discriminant |
;      SQR (b**2 - 4ac)      |
;****************************|
        FLD     CS:CONST_C      ;    c     | -------- |
        FLD     CS:CONST_A      ;    a     |     c    | -------- |
        FMULP   ST(1),ST        ;    ac    | -------- |
        FLD1                    ;    1     |    ac    | -------- |
        FADD    ST(0),ST        ;    2     |    ac    | -------- |
        FADD    ST(0),ST        ;    4     |    ac    | -------- |
        FMULP   ST(1),ST        ;   4ac    | -------- |
        FLD     CS:CONST_B      ;    b     |    4ac   | -------- |
        FMUL    ST(0),ST        ;  b**2    |    4ac   | -------- |
        FXCH                    ;   4ac    |  b**2    | -------- |
        FSUB                    ; b**2-4ac | -------- |
;***************************|
;test for real or imaginary |
;***************************|
; If discriminant is positive or zero the roots (or root) are real
; If discriminant is negative the roots are complex
        FTST                    ;  dscr    | -------- | --------- |
        FSTSW   CS:STATUS_WW    ; Store status word
        FWAIT
        MOV     AH,BYTE PTR CS:STATUS_WW + 1
        AND     AH,01000101B    ; Keep C3, C2, and C0
        CMP     AH,00000001B    ; Value for negative and not zero
        JNE     REAL_ROOTS      ; Go if discriminant is not negative
;*********************************************|
;         processing for complex roots        |
;*********************************************|
; At this point the discriminant is negative, therefore the roots
; of the equation are complex
        FCHS                    ; Change discriminant to positive
        FSQRT                   ;  dscr    | -------- |
; Store discriminant
        FSTP    CS:DISCRIMINANT ; -------- |
;****************************|
;   evaluate the real part   |
;    x(r) of complex root    |
;****************************|
; Calculate -b/2a
        FLD     CS:CONST_A      ;    a     | --------|
        FLD1                    ;    1     |    a    | -------- |
        FADD    ST(0),ST        ;    2     |    a    | -------- |
        FMULP   ST(1),ST        ;   2a     | ------- |
        FLD     CS:CONST_B      ;    b     |    2a   | -------- |
        FCHS                    ;   -b     |    2a   | -------- |
        FDIV    ST,ST(1)        ;   x(r)   |    2a   | -------- |
; ST(0) holds real part of complex root
;****************************|
;  evaluate imaginary part   |
;  +/- x(i) of complex root  |
;****************************|
; Calculate discriminant/2a
;                               ;   x(r)   |    2a   | -------- |
        FXCH                    ;    2a    |   x(r)  | -------- |
        FLD     CS:DISCRIMINANT ;  dscr    |   2a    |    x(r)  |
        FDIVR                   ;   x(i)   |   x(r)  | -------- |
        FXCH                    ;   x(r)   |   x(i)  | -------- |
        STC                     ; Carry means roots are complex
        RET
;
;*********************************************|
;           processing for real roots         |
;*********************************************|
REAL_ROOTS:
        FSQRT                   ;  dscr    | -------- |
; Store discriminant
        FSTP    CS:DISCRIMINANT ; -------- |
;****************************|
;  evaluate the root x(1)    |
;****************************|
        FLD     CS:CONST_A      ;    a     | --------|
        FLD1                    ;    1     |    a    | -------- |
        FADD    ST(0),ST        ;    2     |    a    | -------- |
        FMULP   ST(1),ST        ;   2a     | ------- |
        FLD     CS:CONST_B      ;    b     |    2a   | -------- |
        FCHS                    ;   -b     |    2a   | -------- |
        FLD     CS:DISCRIMINANT ;  dscr    |   -b    |    2a    |
        FADDP   ST(1),ST        ; -b + dscr|    2a   | -------- |
        FDIVR                   ;   x(1)   | ------- |
; ST(0) holds root x(1)
;****************************|
;  evaluate the root x(2)    |
;****************************|
        FLD     CS:CONST_A      ;    a     |   x(1)  |
        FLD1                    ;    1     |    a    |    x(1)  |
        FADD    ST(0),ST        ;    2     |    a    |    x(1)  |
        FMULP   ST(1),ST        ;   2a     |   x(1)  | -------- |
        FLD     CS:CONST_B      ;    b     |    2a   |    x(1)  |
        FCHS                    ;   -b     |    2a   |    x(1)  |
        FLD     CS:DISCRIMINANT ;  dscr    |   -b    |    2a    |
        FSUBP   ST(1),ST        ; -b - dscr|    2a   |    x(1)  |
        FDIVR                   ;   x(2)   |   x(1)  | -------- |
        FXCH                    ;   x(1)   |   x(2)  | -------- |
        CLC                     ; No carry means roots are real
        RET
QUAD_FORMULA    ENDP
;
;****************************************************************
;                  complex arithmtic procedures
;*****************************************************************
;****************************|
;     code segment data      |
;   for complex arithmetic   |
;****************************|
PART_A          DT      0       ; Storage for a
PART_B          DT      0       ; Storage for bi
PART_C          DT      0       ; Storage for c
PART_D          DT      0       ; Storage for di
;
ADD_CMPLX       PROC    FAR
; Addition of two complex numbers (a + bi and c + di) by means of
; the formula:
;
;       (a + bi) + (c + di) = (a + c) + i(b + d)
;
; On entry:
;         ST(3) = a
;         ST(2) = bi
;         ST(1) = c
;         ST(0) = di
; On exit:
;         ST(0) = (a + c) = real part
;         ST(1) = (b + d) = imaginary part
;
;                               | ST(0) | ST(1) | ST(2) | ST(3) |
;                               |   a   |   bi  |   c   |  di   |
;***************************|
;      store inputs         |
;***************************|
        FSTP    CS:PART_D       ;   bi  |   c   |   di  | ----- |
        FSTP    CS:PART_C       ;   c   |   di  | ----- | ----- |
        FSTP    CS:PART_B       ;   di  | ----- | ----- | ----- |
        FSTP    CS:PART_A       ; ----- | ----- | ----- | ----- |
;***************************|
;    calculate (a + c)      |
;***************************|
        FLD     CS:PART_A       ;  a    | ----- |
        FLD     CS:PART_C       ;  c    |   a   | ----- |
        FADD                    ; a + c | ----- |
;***************************|
;    calculate i(b + d)     |
;***************************|
        FLD     CS:PART_B       ;  b    | a + c | ----- |
        FLD     CS:PART_D       ;  d    |   b   | a + c | ----- |
        FADD                    ; b + d | a + c | ----- |
        FXCH                    ; a + c | b + d | ----- |
                                ;   |       |___ imaginary part
                                ;   |___________ real part
        RET
ADD_CMPLX       ENDP
;*****************************************************************
SUB_CMPLX       PROC    FAR
; Subtraction of two complex numbers (a + bi and c + di) by means
; of the formula:
;
;       (a + bi) - (c + di) = (a - c) + i(b - d)
;
; On entry:
;         ST(3) = a
;         ST(2) = bi
;         ST(1) = c
;         ST(0) = di
; On exit:
;         ST(0) = (a - c) = real part
;         ST(1) = (b - d) = imaginary part
;
;                               | ST(0) | ST(1) | ST(2) | ST(3) |
;                               |   a   |   bi  |   c   |  di   |
;***************************|
;      store inputs         |
;***************************|
        FSTP    CS:PART_D       ;   bi  |   c   |   di  | ----- |
        FSTP    CS:PART_C       ;   c   |   di  | ----- | ----- |
        FSTP    CS:PART_B       ;   di  | ----- | ----- | ----- |
        FSTP    CS:PART_A       ; ----- | ----- | ----- | ----- |
;***************************|
;    calculate (a - c)      |
;***************************|
        FLD     CS:PART_A       ;  a    | ----- |
        FLD     CS:PART_C       ;  c    |   a   | ----- |
        FSUBR                   ; a - c | ----- |
;***************************|
;    calculate i(b - d)     |
;***************************|
        FLD     CS:PART_B       ;   b   | a - c | ----- |
        FLD     CS:PART_D       ;   d   |   b   | a - c | ----- |
        FSUBR                   ; b - d | a - c | ----- |
        FXCH                    ; a - c | b - d | ----- |
                                ;   |       |___ imaginary part
                                ;   |___________ real part
        RET
SUB_CMPLX       ENDP
;*****************************************************************
;
MUL_CMPLX       PROC    FAR
; Multiplication of two complex numbers (a + bi and c + di) by
; means of the formula:
;
;       (a + bi)(c + di) = (ac - bd) + i(ad + bc)
;
; On entry:
;         ST(3) = a
;         ST(2) = bi
;         ST(1) = c
;         ST(0) = di
; On exit:
;         ST(0) = (ac - bd) = real part
;         ST(1) = (ad + bc) = imaginary part
;
;                               | ST(0) | ST(1) | ST(2) | ST(3) |
;                               |   a   |   bi  |   c   |  di   |
;***************************|
;      store inputs         |
;***************************|
        FSTP    CS:PART_D       ;   bi  |   c   |   di  | ----- |
        FSTP    CS:PART_C       ;   c   |   di  | ----- | ----- |
        FSTP    CS:PART_B       ;   di  | ----- | ----- | ----- |
        FSTP    CS:PART_A       ; ----- | ----- | ----- | ----- |
;***************************|
;    calculate (ac - bd)    |
;***************************|
;                               ; ----- |
        FLD     CS:PART_A       ;   a   | ----- |
        FLD     CS:PART_C       ;   c   |   a   | ----- |
        FMULP   ST(1),ST        ;  ac   | ----- |
        FLD     CS:PART_B       ;   b   |   ac  | ------|
        FLD     CS:PART_D       ;   d   |   b   |  ac   |
        FMULP   ST(1),ST        ;  bd   |   ac  | ----- |
        FSUB                    ; ac-bd | ----- |
;***************************|
;    calculate i(ad + bc)   |
;***************************|
;                               ; ac-bd |
        FLD     CS:PART_A       ;   a   | ac-bd | ----- |
        FLD     CS:PART_D       ;   a   |   d   | ac-bd |
        FMULP   ST(1),ST        ;  ad   | ac-bd | ----- |
        FLD     CS:PART_B       ;   b   |   ad  | ac-bd | ----- |
        FLD     CS:PART_C       ;   c   |   b   |  ad   | ac-bd |
        FMULP   ST(1),ST        ;  bc   |   ad  | ac-bd | ----- |
        FADD                    ; ad+bc | ac-bd | ----- |
        FXCH                    ; ac-bd | ad+bc | ----- |
                                ;   |       |___ imaginary part
                                ;   |___________ real part
        RET
MUL_CMPLX       ENDP
;*****************************************************************
DIV_CMPLX       PROC    FAR
; Division of two complex numbers (a + bi and c + di) by means of
; the formula:
;
;       (a + bi)            (ac + bd) + i(bc - ad)
;       --------     =      ----------------------
;       (c + di)                 c**2 + d**2
;
; On entry:
;         ST(3) = a
;         ST(2) = bi
;         ST(1) = c
;         ST(0) = di
; On exit:
;         ST(0) = (ac + bd)/(c**2 + d**2) = real part
;         ST(1) = (bc - ad)/(c**2 + d**2) = imaginary part
;
;                               | ST(0) | ST(1) | ST(2) | ST(3) |
;                               |   a   |   bi  |   c   |  di   |
;***************************|
;      store inputs         |
;***************************|
        FSTP    CS:PART_D       ;   bi  |   c   |   di  | ----- |
        FSTP    CS:PART_C       ;   c   |   di  | ----- | ----- |
        FSTP    CS:PART_B       ;   di  | ----- | ----- | ----- |
        FSTP    CS:PART_A       ; ----- | ----- | ----- | ----- |
;***************************|
;    calculate (ac + bd)    |
;***************************|
;                               ; ----- |
        FLD     CS:PART_A       ;   a   | ----- |
        FLD     CS:PART_C       ;   c   |   a   | ----- |
        FMULP   ST(1),ST        ;  ac   | ----- |
        FLD     CS:PART_B       ;   b   |   ac  | ----- |
        FLD     CS:PART_D       ;   d   |   b   |  ac   |
        FMULP   ST(1),ST        ;  bd   |   ac  | ----- |
        FADD                    ; ac+bd | ----- |
;***************************|
;    calculate i(bc - ad)   |
;***************************|
;                               ; ac+bd |
        FLD     CS:PART_B       ;   b   | ac+bd | ----- |
        FLD     CS:PART_C       ;   c   |   b   | ac+bd | ----- |
        FMULP   ST(1),ST        ;  bc   | ac+bd | ----- |
        FLD     CS:PART_A       ;   a   |   bc  | ac+bd | ----- |
        FLD     CS:PART_D       ;   d   |   a   |  bc   | ac+bd |
        FMULP   ST(1),ST        ;  ad   |   bc  | ac+bd | ----- |
        FSUB                    ; bc-ad | ac+bd | ----- |
;***************************|
;    calculate c**2 + d**2  |
;***************************|
;                               | bc-ad | ac+bd | ----- |
        FLD     CS:PART_C       ;   c   | bc-ad | ac+bd | ----- |
        FMUL    ST(0),ST        ;  c**2 | bc-ad | ac+bd | ----- |
        FLD     CS:PART_D       ;   d   |  c**2 | bc-ad | ac+bd |
        FMUL    ST(0),ST        ;  d**2 | c**2  | bc-ad | ac+bd |
        FADD                    ; c2+d2 | bc-ad | ac+bd | ----- |
;***************************|
; calculate imaginary part  |
; (bc - ad)/(c**2 + d**2)   |
;***************************|
;                               | ST(0) | ST(1) | ST(2) | ST(3) |
;                               | c2+d2 | bc-ad | ac+bd | ----- |
        FXCH                    ; bc-ad | c2+d2 | ac+bd | ----- |
        FLD     ST(1)           ; c2+d2 | bc-ad | c2+d2 | ac+bd |
        FDIV                    ; bc-ad |
                                ;/c2*d2 | c2+d2 | ac+bd | ----- |
;***************************|
;   calculate real part     |
; (ac + bd)/(c**2 + d**2)   |
;***************************|
        FXCH    ST(2)           ; ac+bd | c2+d2 | img p | ----- |
        FXCH    ST(1)           ; c2+d2 | ac+bd | img p | ----- |
        FDIV                    ; ac+bd | bc-ad |
                                ;/c2+d2 |/c2+d2 | ----- |
                                ;   |       |___ imaginary part
                                ;   |___________ real part
        RET
DIV_CMPLX       ENDP
;*****************************************************************
ABS_CMPLX       PROC    FAR
; Find the absolute value of a complex number using the formula:
;
;           |a + bi| = SQR (a**2 + b**2)
;
; On entry:
;         ST(0) = a
;         ST(1) = bi
; On exit:
;         ST(0) = |a + bi|
;                               | ST(0) | ST(1) | ST(2) | ST(3) |
;                               |   a   |   b   |
        FMUL    ST(0),ST        ;  a**2 |   b   | ----- |
        FXCH                    ;   b   | a**2  | ----- |
        FMUL    ST(0),ST        ;  b**2 | a**2  | ----- |
        FADD                    ; a2+b2 | ----- |
        FSQRT                   ;  abs  |
                                ; value |
        RET
ABS_CMPLX       ENDP
;******************************************************************
;           procedures for coordinate system conversions
;******************************************************************
;
POLAR_TO_CART   PROC    FAR
; Procedure for converting a coordinate pair in polar form to a
; coordinate pair in Cartesina (rectangular) form by means of the
; formulas:
;
;            x = r cos A
;            y = r sin A
;
; Where r is the directed distance from the pole, A is the
; directed angle from the polar axis, expressed in degrees,
; and x, y are the Cartesian coordinate pair

; On entry:
;          ST(0) = A
;          ST(1) = r
; On exit:
;          ST(0) = x
;          ST(1) = y
;
;                               ;  ST(0)   |  ST(1)   |   ST(2) |
                                ;    A     |     r    | ------- |
; Convert degrees to radians
        CALL    DEG_TO_RAD      ;  A (rads)|     r    | ------- |
;***************************|
;    solve x = r cos A      |
;***************************|
        FXCH                    ;    r     |     A    | ------- |
        FLD     ST(1)           ;    A     |     r    |    A    |
        CALL    COSINE          ; cos A    |     r    |    A    |
        FMUL    ST,ST(1)        ;    x     |     r    |    A    |
        FXCH    ST(2)           ;    A     |     r    |    x    |
;***************************|
;    solve y = r sin A      |
;***************************|
        CALL    SINE            ;  sin A   |     r    |    x    |
        FMULP   ST(1),ST        ;    y     |     x    | ------- |
        FXCH                    ;    x     |     y    | ------- |
        RET
POLAR_TO_CART   ENDP
;*****************************************************************
CART_TO_POLAR   PROC    FAR
; Procedure for converting a coordinate pair in Cartesian form to
; a coordinate pair in Polar form, by means of the formulas:
;
;                      y
;            tan A = ----
;                      x
;
;            r = SQR (x**2 + y**2)
;
; Where x and y are the Cartesian coordinates, r is the directed
; distance from the pole and A is the directed angle from the
; polar axis, expressed in degrees
;
; On entry:
;          ST(0) = x
;          ST(1) = y
; On exit:
;          ST(0) = A
;          ST(1) = r

; Make copies of input values
;                               | ST(0) | ST(1) | ST(2) | ST(3) |
                                ;   x   |   y   | ----- |
        FLD     ST(0)           ;   x   |   x   |   y   |
        FLD     ST(2)           ;   y   |   x   |   x   |   y   |
        FXCH                    ;   x   |   y   |   x   |   y   |
;***************************|
;    find tan A = y/x       |
;***************************|
;                               | ST(0) | ST(1) | ST(2) | ST(3) |
;                               |   x   |   y   |   x   |   y   |
        FDIV                    ;  y/x  |   x   |   y   | ----- |
        CALL    ARC_TANGENT     ;A(rads)|   x   |   y   | ----- |
        CALL    RAD_TO_DEG      ;   A   |   x   |   y   | ----- |
;***************************|
; find r = SQR(x**2 + y**2) |
;***************************|
;                               | ST(0) | ST(1) | ST(2) | ST(3) |
;                               |   A   |   x   |   y   | ----- |
        FXCH    ST(2)           ;   y   |   x   |   A   | ----- |
        CALL    PITHAGORAS      ;   r   |   A   | ----- |
        FXCH                    ;   A   |   r   | ----- |
        RET
CART_TO_POLAR   ENDP
;
;*****************************************************************
;*****************************************************************
;                      financial procedures
;*****************************************************************
;***************************|
;     code segement data    |
;  for financial procedures |
;***************************|
PRINCIPAL       DT      0       ; Storage for principal
INTEREST        DT      0       ; Storage for interest rate
TERM            DT      0       ; Storage for term of loan
PAYMENT         DT      0       ; Storage for periodic payment
;

PAYMENTS        PROC    FAR
; Financial procedure to calculate the ammount of each payment on
; a loan using the formula:
;
;                     i
;      P =  p * ---------------
;               1-(i + 1)**-n
; where
;     P = ammount of periodic payment
;     p = principal
;     i = periodic interest rate (in fractional form)
;     n = term of loan (number of payment periods)
;
; On entry:
;          ST(0) = n (term)
;          ST(1) = i (interest rate)
;          ST(2) = p (principal)
; On exit:
;          ST(0) = P (ammount of periodic payment)
;
; Note: interest rate (i) and term (n) must be in the same units
;       of time, for example: to calculate monthly payments on a
;       3-year loan of $4000.00, at 0.14% yearly interest rate
;                   p = 4000
;                   i = 0.14/12 = 0.0116666...
;                   n = 3*12 = 36
;
;***************************|
;   store input data        |
;***************************|
;                               | ST(0) | ST(1) | ST(2) | ST(3) |
;                               |   n   |   i   |   p   | ----- |
        FSTP    CS:TERM         ;   i   |   p   | ----- |
        FSTP    CS:INTEREST     ;   p   | ----- |
        FSTP    CS:PRINCIPAL    ; ----- |
;***************************|
; calculate 1-(i + 1)**-n   |
;***************************|
        FLD     CS:TERM         ;   n   | ----- |
        FLD     CS:INTEREST     ;   i   |   n   | ----- |
        FLD1                    ;   1   |   i   |   n   | ----- |
        FADD                    ; i + 1 |   n   | ----- |
        FXCH                    ;   n   |  i+1  | ----- |
        CALL    X_TO_Y_BYBP     ;(i+1)^n| ----- |
; x**-n = 1/x**n
        FLD1                    ;   1   |(i+1)^n| ----- |
        FDIVR                   ;  1/   |
                                ;(i+1)^n| ----- |
        FLD1                    ;  1    |1/(i+..| ----- |
        FSUBR                   ;1-(1/..| ----- |
;***************************|
; calculate:                |
;     p * i/1-(i + 1)**-n   |
;***************************|
; den = 1-(i + 1)**-n
;                               ; den   | ----- |
        FLD1                    ;  1    |  den  | ----- |
        FSUBR                   ; 1-den | ----- |
        FLD     CS:INTEREST     ;   i   | 1-den | ----- |
        FDIVR                   ;i/1-den| ----- |
        FLD     CS:PRINCIPAL    ;  p    |i/1-den| ----- |
        FMULP   ST(1),ST        ;  P    | ----- |
        RET
PAYMENTS        ENDP
;*****************************************************************
;
PRESENT_VALUE   PROC    FAR
; Financial procedure to calculate the present value of an
; investment using the formula:
;
;                 1-(i + 1)**-n
;      PV =  pm * ---------------
;                       i
;
; where:
;     PV = present value of an investment
;     pm = periodic payment
;     i = periodic interest rate (in fractional form)
;     n = term of loan (number of payment periods)
;
; On entry:
;          ST(0) = n (term)
;          ST(1) = i (interest rate)
;          ST(2) = pm (periodic payment)
; On exit:
;          ST(0) = PV (present value of investment)
;
; Note: interest rate (i) and term (n) must be in the same units
;       of time (see note in the PAYMENTS procedure)
;
;***************************|
;   store input data        |
;***************************|
;                               | ST(0) | ST(1) | ST(2) | ST(3) |
;                               |   n   |   i   |   pm  | ----- |
        FSTP    CS:TERM         ;   i   |   pm  | ----- |
        FSTP    CS:INTEREST     ;   pm  | ----- |
        FSTP    CS:PAYMENT      ; ----- |
;***************************|
; calculate 1-(i + 1)**-n   |
;***************************|
        FLD     CS:TERM         ;   n   | ----- |
        FLD     CS:INTEREST     ;   i   |   n   | ----- |
        FLD1                    ;   1   |   i   |   n   | ----- |
        FADD                    ; i + 1 |   n   | ----- |
        FXCH                    ;   n   |  i+1  | ----- |
        CALL    X_TO_Y_BYBP     ;(i+1)^n| ----- |
; x**-n = 1/x**n
        FLD1                    ;   1   |(i+1)^n| ----- |
        FDIVR                   ;  1/   |
                                ;(i+1)^n| ----- |
        FLD1                    ;  1    |1/(i+..| ----- |
        FSUBR                   ;1-(1/..| ----- |
;***************************|
; calculate:                |
;           1-(i + 1)**-n   |
;   pm *  ---------------   |
;                i          |
;***************************|
; num = 1-(i + 1)**-n
;                               ; num   | ----- |
        FLD     CS:INTEREST     ;   i   |  num  | ----- |
        FDIV                    ; num/i | ----- |
        FLD     CS:PAYMENT      ;  pm   | i/num | ----- |
        FMULP   ST(1),ST        ;  PV   | ----- |
        RET
PRESENT_VALUE   ENDP
;*****************************************************************
;
FUTURE_VALUE    PROC    FAR
; Financial procedure to calculate the future value of an
; investment using the formula:
;
;                 (i + 1)**n - 1
;      FV =  pm * ---------------
;                       i
;
; where:
;     PV = present value of an investment
;     pm = periodic payment
;     i = periodic interest rate (in fractional form)
;     n = term of loan (number of payment periods)
;
; On entry:
;          ST(0) = n (term)
;          ST(1) = i (interest rate)
;          ST(2) = pm (periodic payment)
; On exit:
;          ST(0) = FV (future value of investment)
;
; Note: interest rate (i) and term (n) must be in the same units
;       of time (see note in the PAYMENTS procedure)
;
;***************************|
;   store input data        |
;***************************|
;                               | ST(0) | ST(1) | ST(2) | ST(3) |
;                               |   n   |   i   |   pm  | ----- |
        FSTP    CS:TERM         ;   i   |   pm  | ----- |
        FSTP    CS:INTEREST     ;   pm  | ----- |
        FSTP    CS:PAYMENT      ; ----- |
;***************************|
; calculate (i + 1)**n - 1  |
;***************************|
        FLD     CS:TERM         ;   n   | ----- |
        FLD     CS:INTEREST     ;   i   |   n   | ----- |
        FLD1                    ;   1   |   i   |   n   | ----- |
        FADD                    ; i + 1 |   n   | ----- |
        FXCH                    ;   n   |  i+1  | ----- |
        CALL    X_TO_Y_BYBP     ;(i+1)^n| ----- |
        FLD1                    ;   1   |(i+1)^n| ----- |
        FSUB                    ;(i+1)^n|
                                ;  -1   | ----- |
;***************************|
; calculate:                |
;          (i + 1)**n - 1   |
;   pm *  ---------------   |
;                i          |
;***************************|
; num = (i + 1)**n - 1
;                               ; num   | ----- |
        FLD     CS:INTEREST     ;   i   |  num  | ----- |
        FDIV                    ; num/i | ----- |
        FLD     CS:PAYMENT      ;  pm   | i/num | ----- |
        FMULP   ST(1),ST        ;  PV   | ----- |
        RET
FUTURE_VALUE    ENDP
;*****************************************************************
;
SL_DEPRECIATION PROC    FAR
; Financial procedure to calculate the straight-line depreciation
; of an asset during 1 period, using the formula:
;
;              (c - s)
;      SLD =  --------
;                 ul
; where:
;     SLD = straight-line depreciation allowance
;       c = initial cost of the asset
;       s = salvage value of the asset at the end of the
;           depreciation period
;      ul = useful life of the asset
;
; On entry:
;          ST(0) = ul (useful life)
;          ST(1) = s (salvage value)
;          ST(2) = c (initial cost)
; On exit:
;          ST(0) = SLD allowance per period
;
; Note: the depreciation period (SLD) and the useful life (ul) of
;       asset must be in the same units of time
;
;***************************|
;   calculate (c-s)/ul      |
;***************************|
;                               | ST(0) | ST(1) | ST(2) | ST(3) |
;                               |   ul  |   s   |   c   | ----- |
        FXCH    ST(2)           ;   c   |   s   |   ul  | ----- |
        FSUB    ST,ST(1)        ;  c-s  |   s   |   ul  | ----- |
        FSTP    ST(1)           ;  c-s  |   ul  | ----- |
        FDIV    ST,ST(1)        ; c-s/ul|   ul  | ----- |
        FSTP    ST(1)           ;  SLD  | ----- |
        RET
SL_DEPRECIATION ENDP
;*****************************************************************
;
COMP_PERIODS    PROC    FAR
; Financial procedure to calculate the number of compounding
; periods necessary for an investment to grow to a predetermined
; value, by means of the formula:
;
;           ln(fv/sv)
;      CP = ---------
;           ln(1 + i)
;
; where:
;      CP = number of compounding periods
;      fv = final value
;      sv = starting value
;       i = interest rate
;      ln = log base e (natural logarithm)
; On entry:
;          ST(0) = i (interest rate)
;          ST(1) = fv (final value of investment)
;          ST(2) = sv (starting value of investment)
; On exit:
;          ST(0) = CP (number of compounding periods for growth)
;
; Note: the number of compounding periods (CP) will be in the same
;       units of time as the interest rate (i). For example: to
;       calculate compounding periods (in months) for an investment
;       at a yearly interest of 14 percent
;                 i = 0.14/12 = 0.0116666...
;
;***************************|
;   calculate ln(1 + i)     |
;***************************|
;                               | ST(0) | ST(1) | ST(2) | ST(3) |
;                               |   i   |   fv  |   sv  | ----- |
        FLD1                    ;   1   |   i   |   fv  |   sv  |
        FADD                    ; 1 + i |   fv  |   sv  | ----- |
        CALL    LOG_E           ;ln(1+i)|   fv  |   sv  | ----- |
;***************************|
;   calculate:
;      ln(fv/sv)/ln(1 + i)  |
;***************************|
        FXCH    ST(2)           ;  sv   |   fv  |ln(1+i)| ----- |
        FDIV                    ; fv/sv |ln(1+i)| ----- |
        CALL    LOG_E           ;ln(fv/ |
                                ;  sv)  |ln(1+i)| ----- |
        FDIVR                   ;  CP   | ----- |
        RET
COMP_PERIODS    ENDP

CODE    ENDS
        END

 ----- |
        FDIVR                   ;  CP   | ----- |
    