;*****************************************************************
;*****************************************************************
;                           UN32_1.ASM
;*****************************************************************
;*****************************************************************
; Library of math routines from the book SOFTWARE SOLUTIONS 
; by Julio Sanchez and Maria P. Canton
; Creation date:  January 9, 2001
; Updated: March 31/2001
; Libary name: MATH32.LIB
; Reference: Chapter 1
; Module name: UN32_1.ASM
; Description:
;         Procedures for number system conversions.
;
;*****************************************************************
;                      About the MATH32 library
;*****************************************************************  
; Environent:
;     All modules in this library were assembled with MASM Version
;     6.12 and used in console applications, under Visual C++ 6.0.
;     32-bit code cannot be used in DOS.
; Procedure names:
;     Procedures referenced by C++ code have names that start with
;     the _ symbol. Local support procedure names do not use the
;     _ symbol as the first character.
;      
;*****************************************************************
;                           publics
;*****************************************************************
PUBLIC _BIN_TO_ASC10  ; .. 16-bit binary into 5-digit ASCII decimal
PUBLIC _BIN_TO_ASC16  ; .. 16-bit binary into 4-digit ASCII hex
PUBLIC _ASCII_TO_DX   ; .. 4-digit ASCII decimal to binary in DX
PUBLIC _BINF_TO_DECF  ; .. 8-bit binary to 8-digit decimal fraction
PUBLIC _DECODE_SINGLE ; .. Isolate elements of single precision
PUBLIC _BIN_TO_BCD    ; .. 16-bit binary to 5 BCD digits
PUBLIC _BCD_TO_ASCII  ; .. 5 unpacked BCD to 5 ASCII digits
PUBLIC _BCD_INTO_REG  ; .. 5-digit BCD into a binary in DX
PUBLIC _BCD12_TO_ASCSTR;.. BCD12 number in ASCII string
PUBLIC _ASCII_TO_BCD12 ;.. ASCII decimal string into BCD12 format
PUBLIC _ASCII_TO_EDX  ; .. ASCII to binary in EDX register
;
PUBLIC _BCD20_TO_ASCSTR;.. BCD20 number in ASCII string
PUBLIC _ASCII_TO_BCD20 ;.. ASCII decimal string into BCD20 format
;
;****************************************************************
;              definitions for 32-bit flat model
;****************************************************************

             .486
		.MODEL flat
		.DATA
;***************************************
;       data for this module
;***************************************
; BCD conversion factors
BCD_FACTORS     DB      50H,00,00,00    ; .50000000
                DB      25H,00,00,00    ; .25000000
                DB      12H,50H,00,00   ; .12500000
                DB      06H,25H,00,00   ; .06250000
                DB      03H,12H,50H,00  ; .03125000
                DB      01H,56H,25H,00  ; .01562500
                DB      00H,78H,12H,50H ; .00781250
                DB      00H,39H,06H,25H ; .00390625

; Local dataa
ASC_TEMP_BUF    DB      5 DUP (20H)
                DB      0               ; Terminator

OFFSET_ASC      DD      0       ; Entry offset of ASCII buffer
OFFSET_BCD      DD      0       ; Entry offset of BCD buffer

OFFSET_ASCBUF   DD      0       ; Entry offset of ASCII buffer
BCD_BUF         DB      5 DUP (0)

;*****************************************************************
;                          code
;*****************************************************************
         .CODE      

_BIN_TO_ASC10    PROC
; Procedure to convert a 16-bit binary number into 5 ASCII decimal
; digits
; On entry:
;         AX = binary number
;         EDI --> 5 byte ASCII buffer
; On exit:
;         Buffer holds ASCII decimal digits
;
;***********************|
;   clear ASCII buffer  |
;***********************|
        MOV     ECX,5           ; Five digits to clear
CLEAR_5:
        MOV     BYTE PTR [EDI],20H       ; Clear digit
        INC     EDI             ; Bump pointer
        LOOP    CLEAR_5         ; Repeat for 5 digits
        DEC     EDI             ; Adjust buffer pointer to last
                                ; digit
        MOV     ECX,10          ; Decimal divisor to CX
;***********************|
;  obtain ASCII decimal |
;        digit          |
;***********************|
GET_ASC10:
        MOV     DX,0            ; Clear for word division
        DIV     CX              ; Perform division AX/CX
; Quotient is in AX and remainder in DL
; Convert decimal to ASCII decimal
        ADD     DL,30H          ; Add 30H to bring to ASCII range
;***********************|
;     store digit       |
;  bump buffer pointer  |
;***********************|
        MOV     BYTE PTR [EDI],DL        ; Store digit in buffer
        DEC     EDI             ; Buffer pointer to next digit
; Note: the binary quotient is left in AX by the DIV CX instruction
;***********************|
;   is quotient = 0 ?   |
;***********************|
        CMP     AX,0            ; Test for end of binary
        JNZ     GET_ASC10       ; Continue if not 0
;***********************|
;   end of conversion   |
;***********************|
        CLD 
        RET
_BIN_TO_ASC10     ENDP
;
;
_BIN_TO_ASC16    PROC   USES edi esi ebx ebp
; Procedure to convert a 16-bit binary number into 4 ASCII
; hexadecimal digits
; On entry:
;         AX = binary number
;         EDI --> 4 byte ASCII buffer
; On exit:
;         Buffer holds ASCII decimal digits

; CPU compatibilty chart:
;                       | 80386           | YES |
;                       | 80486/PENTIUM   | YES |
;
;***********************|
;   clear ASCII buffer  |
;***********************|
        MOV     ECX,4            ; Five digits to clear
CLEAR_4:
        MOV     BYTE PTR [EDI],20H      ; Clear digit
        INC     EDI             ; Bump pointer
        LOOP    CLEAR_4         ; Repeat for 4 digits
        DEC     EDI             ; Adjust buffer pointer to last
                                ; digit
        MOV     ECX,16          ; Hexadecimal divisor to CX
;***********************|
;   obtain ASCII hex    |
;        digit          |
;***********************|
GET_ASC16:
        MOV     DX,0            ; Clear for word division
        DIV     CX              ; Perform division AX/CX
; Quotient is in AX and remainder in DL
; Convert decimal to ASCII decimal
; Test for digit range 0 to 9
        CMP     DL,9
        JA      IS_LETTER       ; Digit is Hex letter
        ADD     DL,30H          ; Add 30H to bring to ASCII range
		JMP     STORE_DIGIT
IS_LETTER:
        ADD     DL,55           ; Convert to ASCII letter
;***********************|
;     store digit       |
;  bump buffer pointer  |
;***********************|
STORE_DIGIT:
        MOV     BYTE PTR [EDI],DL       ; Store digit in buffer
        DEC     EDI             ; Buffer pointer to next digit
; Note: the binary quotient is left in AX by the DIV CX instruction
;***********************|
;   is quotient = 0 ?   |
;***********************|
        CMP     AX,0            ; Test for end of binary
        JNZ     GET_ASC16       ; Continue if not 0
;***********************|
;   end of conversion   |
;***********************|
        CLD  
        RET
_BIN_TO_ASC16     ENDP



_ASCII_TO_EDX     PROC
; Convert an ASCII number to binary and store in EDX register
; On entry:
;         ESI --> start of 10-digit ASCII buffer holding decimal
;                 string in the range 0 to 4,294,967,295
; On exit:
;         EDX = binary number
;
;***********************|
;  initialize registers |
;***********************|
        MOV     EAX,0           ; Clear adder
        MOV     EDX,0           ; Clear accumulator register
        MOV     ECX,10          ; Load multiplier into ECX
;***********************|
; test digit for range  |
;        0 to 9         |
;***********************|
DIGIT_TO_ACC32:
        MOV     AL,[ESI]        ; Get ASCII digit
        CMP     AL,'0'          ; Test for lower limit
        JB      EXIT_ASC32      ; Exit if less than 0
        CMP     AL,'9'          ; Test for higher limit
        JA      EXIT_ASC32      ; Exit if larger than 9
;***********************|
; ASCII digit to binary |
;***********************|
        SUB     AL,30H          ; ASCII to decimal
;***********************|
;    accumulator x 10   |
;      + digit          |
;***********************|
        MOV     AH,0            ; Clear high byte in AX
        PUSH    EAX             ; Save digit in stack
        MOV     EAX,EDX         ; Previous product to EAX
        MUL     ECX             ; Perform EAX = EAX * 10
        MOV     EDX,EAX         ; Move product to accumulator
        MOV     EAX,0           ; Clear 32 bits
        POP     EAX             ; Restore decimal digit
        ADD     EDX,EAX         ; Add digit to accumulator
;***********************|
; pointer to next digit |
;***********************|
        INC     ESI             ; Bump pointer
        JMP     DIGIT_TO_ACC32  ; Continue
EXIT_ASC32:
        CLD
		RET
_ASCII_TO_EDX    ENDP

; BINF_TO_DECF
; Procedure to convert an 8-bit binary fraction into an 8-digit
; decimal fraction in packed BCD format
; On entry:
;         AL = binary fraction
;         EDI -> pointer to 4-byte buffer to hold packed BCD
;                decimal fraction
; On exit:
;         EDI -> buffer holding 8 packed BCD digits
;
; CPU compatibility chart:
;                       | 8086/8088       | NO  |
;                       | 80286           | NO  |
;                       | 80386           | NO  |
;                       | 80486/PENTIUM   | YES |
;

;
_BINF_TO_DECF     PROC USES esi edi ebx ebp
        LEA     ESI,BCD_FACTORS       ; Pointer to conversion table
        MOV     AH,10000000B    ; Mask to test binary digits
        MOV     ECX,8           ; Counter for 8 bits
;
ADD_FACTOR:
        TEST    AL,AH           ; Test binary fraction bit
        JZ      NO_FACTOR_ADD   ; Do not add factor if not set
;***********************|
; add conversion factor |
;***********************|
        PUSH    AX              ; Save binary fraction
        ADD     ESI,3   ; Set pointer to last byte
        ADD     EDI,3   ; for both numbers
        MOV     ECX,4   ; Counter for 4 bytes to add
        CLC             ; Clear carry for first add operation
ADD_BCD:
        MOV     AL,BYTE PTR [ESI]     ; Get conversion factor
        MOV     AH,BYTE PTR [EDI]        ; Get factor sum
        ADC     AL,AH           ; AL+AH+carry from previous add
        DAA                     ; Adjust to packed BCD format
        MOV     BYTE PTR [EDI],AL        ; Store sum
        DEC     ESI             ; Adjust all pointers
        DEC     EDI
        LOOP    ADD_BCD         ; Loop 4 times
        INC     EDI             ; Adjust pointers to buffer start
        INC     ESI
        POP     AX              ; Restore binary fraction
;***********************|
;     bump pointers     |
;***********************|
NO_FACTOR_ADD:
        ADD     ESI,4           ; To next factor
        SHR     AH,1            ; Shift test mask to next bit
        JNZ     ADD_FACTOR      ; Continue for 8 digits
; EDI --> buffer holding sum of all conversion factors
; Note that sum cannot overflow 8 digits. Carry can be ignored
        CLD
        RET
_BINF_TO_DECF     ENDP


_BCD_INTO_REG    PROC
        LEA     ESI,ASC_TEMP_BUF ; Address of buffer off DS
        CALL    _BCD_TO_ASCII    ; Convert BCD digits to ASCII
                                 ; string
        CALL    _ASCII_TO_DX     ; Conversion routine
        RET
_BCD_INTO_REG    ENDP

_BCD_TO_ASCII    PROC
; Convert 5 unpacked BCD digits in machine registers into to 5
; ASCII digits
; On entry:
;         AL-BH-BL-CH-CL hold 5 unpacked BCD digits
;                        holding a value < 65536
;         ESI --> 5 byte buffer to hold ASCII string
; On exit:
;         ESI --> buffer holding ASCII digits
;
;
; Clear buffer for ASCII digits
        PUSH    ESI              ; Save start of buffer
        PUSH    ESI              ; Twice
        PUSH    ECX              ; Save BCD high digit
        MOV     ECX,5            ; Counter for 5 digits
CLEAR_BUF_5:
        MOV     BYTE PTR [ESI],' '    ; Set blank in buffer
        INC     ESI              ; Bump pointer
        LOOP    CLEAR_BUF_5
        POP     ECX              ; Restore BCD digit
;**********************|
; convert to ASCII and |
;    move to buffer    |
;**********************|
; Convert BCD digits to ASCII
        ADD     AL,30H          ; Add 30H to each digit
        ADD     BH,30H
        ADD     BL,30H
        ADD     CH,30H
        ADD     CL,30H
        POP     ESI              ; Restore buffer start
        MOV     [ESI],AL     ; Digit to buffer
        INC     ESI              ; Bump buffer pointer
        MOV     [ESI],BH     ; Digit to buffer
        INC     ESI              ; Bump buffer pointer
        MOV     [ESI],BL     ; Digit to buffer
        INC     ESI              ; Bump buffer pointer
        MOV     [ESI],CH     ; Digit to buffer
        INC     ESI              ; Bump buffer pointer
        MOV     BYTE PTR [ESI],CL     ; Digit to buffer
        POP     ESI              ; Restore start of buffer
        RET
_BCD_TO_ASCII    ENDP


_ASCII_TO_DX     PROC
; Convert an ASCII number to binary and store in DX register
; On entry:
;         ESI --> start of 5-digit ASCII buffer holding decimal
;                string in the range 0 to 65,535
; On exit:
;         DX = binary number
;
;
;***********************|
;  initialize registers |
;***********************|
        MOV     DX,0            ; Clear accumulator register
        MOV     ECX,10           ; Load multiplier into CX
;***********************|
; test digit for range  |
;        0 to 9         |
;***********************|
DIGIT_TO_ACC16:
        MOV     AL,[ESI]         ; Get ASCII digit
        CMP     AL,'0'          ; Test for lower limit
        JB      EXIT_ASC16      ; Exit if less than 0
        CMP     AL,'9'          ; Test for higher limit
        JA      EXIT_ASC16      ; Exit if larger than 9
;***********************|
; ASCII digit to binary |
;***********************|
        SUB     AL,30H          ; ASCII to decimal
;***********************|
;    accumulator x 10   |
;      + digit          |
;***********************|
        MOV     AH,0            ; Clear high byte in AX
        PUSH    AX              ; Save digit in stack
        MOV     AX,DX           ; Previous product to AX
        MUL     CX              ; Perform AX = AX * 10
        MOV     DX,AX           ; Move product to accumulator
        POP     AX              ; Restore decimal digit
        ADD     DX,AX           ; Add digit to accumulator
;***********************|
; pointer to next digit |
;***********************|
        INC     ESI              ; Bump pointer
        JMP     DIGIT_TO_ACC16  ; Continue
EXIT_ASC16:
        RET
_ASCII_TO_DX     ENDP



_DECODE_SINGLE   PROC USES edi esi ebx ebp
; Procedure to separate the elements of a floating point number
; stored in memory in IEEE single format
; On entry:
;         ESI --> LOB of stored IEEE floating point single
; On exit:
;         AH = Sign field
;                    0000 0001 = negative sign
;                    0000 0000 = positive sign
;         AL = 8-bit exponent in bias 127 form
;   BH-BL-CH = 24-bit significand
;              Implicit significand bit is set if exponent is
;              not zero
;
; CPU compatibility chart:
;                       | 8086/8088       | NO  |
;                       | 80286           | NO  |
;                       | 80386           | YES |
;                       | 80486/PENTIUM   | YES |
;
;***********************|
;       decode sign     |
;***********************|
        ADD     ESI,3           ; Set pointer to last byte
        MOV     AH,[ESI]        ; Get byte into AH
        AND     AH,10000000B    ; Mask off exponent bits
        SHR     AH,7            ; Shift sign bit to right-most bit
;***********************|
;     decode exponent   |
;***********************|
        MOV     AL,[ESI]        ; Get 7 high bits of exponent
        SHL     AL,1            ; Shift left one bit
        DEC     SI              ; Pointer to byte at offset 2
        MOV     BH,[ESI]        ; Get byte into BH
        TEST    BH,10000000B    ; Test for exponent bit 8 set
        JZ      OK_EXPONENT     ; Exponent OK if bit not set
        OR      AL,00000001B    ; Set low exponent bit
;***********************|
;  decode significand   |
;***********************|
OK_EXPONENT:
; Test for a non-zero exponent so set implicit significand bit
        CMP     AL,0            ; Is exponent zero
        JE      IMPLICIT_0      ; Implicit bit is zero with a
                                ; zero exponent. Do not set
;***********************|
;     explicit 1 bit    |
;***********************|
        OR      BH,10000000B    ; Set BH high bit
IMPLICIT_0:
        DEC     ESI             ; Point to byte at offset 1
        MOV     BL,[ESI]        ; Load into BL
        DEC     ESI             ; Point to byte at offset 0
        MOV     CH,[ESI]        ; Load into CH
        CLD
		RET
_DECODE_SINGLE   ENDP


; BIN_TO_BCD
; Convert a 16-bit binary number to 5 BCD digits
; On entry:
;         AX = binary number
; On exit:
;         5 packed BCD digits in the following registers:
;                CL = digit number 1 (high order digit)
;                DH = digits number 2 and 3
;                DL = digits number 4 and 5 (low order BCD digit)
;
; Operation:
; Binary in AX is first converted into 5 unpacked BCD digits and
; stored in a temporary buffer (BCD_BUF). Digits are then moved
; from this buffer and packed into DX and CL
;
;
_BIN_TO_BCD      PROC 
;**********************|
;    clear buffer      |
;**********************|
        LEA     EDI,BCD_BUF     ; Set pointer to buffer
        PUSH    EDI             ; Save buffer start
        MOV     ECX,5           ; Five digits to clear
CLEAR_5BCD:
        MOV     BYTE PTR [EDI],0         ; Clear digit
        INC     EDI             ; Bump pointer
        LOOP    CLEAR_5BCD      ; Repeat for 5 digits
        DEC     EDI             ; Adjust buffer pointer to last
                                ; digit
        MOV     ECX,10          ; Decimal divisor to CX
GET_BIN:
        MOV     DX,0            ; Clear for word division
; Divide binary by 10
        DIV     CX              ; Perform division AX/CX
; Quotient is in AX and remainder in DL
        MOV     [EDI],DL     ; Store digit in buffer
        DEC     EDI             ; Bump destination pointer
; Note: step 5 is performed automatically since the binary
; quotient is left in AX
        CMP     AX,0            ; Test for end of binary
        JNZ     GET_BIN         ; Continue if not end
; Last digit converted
;**********************|
;   BCD to registers   |
;      and pack        |
;**********************|
        POP     EDI             ; Restore pointer to start of
                                ; buffer
        ADD     EDI,4           ; Point to last digit
        MOV     AL,BYTE PTR [EDI]     ; BCD digit to AL
        MOV     BL,AL           ; To results register low nibble
        DEC     EDI             ; Point to fourth BCD digit
        MOV     AL,[EDI]        ; BCD digit to AL
        MOV     CL,4            ; Shift counter
        SHL     AL,CL           ; Shift BCD to high nibble
        OR      BL,AL           ; OR into result register high
                                ; nibble
        DEC     EDI             ; Pointer to third BCD digit
        MOV     AL,[EDI]        ; BCD digit to AL
        MOV     BH,AL           ; To results register low nibble
        DEC     EDI             ; Point to second BCD digit
        MOV     AL,[EDI]        ; BCD digit to AL
        MOV     CL,4            ; Shift counter
        SHL     AL,CL           ; Shift BCD to high nibble
        OR      BH,AL           ; OR into result register high
                                ; nibble
; Move 4 low order BCD digits to DX
        MOV     DX,BX
; Move high order digit to CL
        DEC     EDI             ; Pointer to first BCD digit
        MOV     AL,[EDI]        ; BCD digit to AL
        MOV     CL,AL           ; To results register low nibble
        RET
_BIN_TO_BCD      ENDP


_BCD12_TO_ASCSTR  PROC
; Convert a number in BCD12 format into an ASCII string
; On entry:
;         ESI --> 12-byte BCD12 buffer holding number
;         EDI --> 31-byte ASCII buffer to hold string
; On exit:
;         EDI holds ASCII string
;
; Operation:
;         Numbers with exponent < 10 are converted to decimal
;         format. If exponent is > 10 the number is converted to
;         exponential format
;
; BCD12 format:
;          S s e e e e m . m m m m m m m m m m m m m m m m m m
;                        |
;                        |____ implicit decimal point
;
;    S = sign of number (1 BCD digit)
;    s = sign of exponent (1 BCD digit)
;    e = exponent (4 BCD digits)
;    m = normalized significand (18 BCD digits)
;        (in positive and negative numbers the first significand
;         digit must be non-zero)
;    . = implicit decimal point between the first and second
;        significand digits
;
;
;**********************|
; save buffer addresses|
;**********************|
        MOV     OFFSET_ASCBUF,EDI
; Clear results buffer
        PUSH    EDI             ; Save pointer
        MOV     ECX,31          ; Bytes to clear
CLEAR_31:
        MOV     BYTE PTR [EDI],20H ; Clear
        INC     EDI             ; Bump pointer
        LOOP    CLEAR_31
        POP     EDI             ; Restore pointer to start of buffer
;**********************|
;  get sign of number  |
;**********************|
        MOV     BYTE PTR [EDI],' '       ; In case of positive
        MOV     AL,[ESI]        ; First byte of BCD12 number
        TEST    AL,00010000B    ; Bit 4 is negative number
        JZ      NNOT_NEG        ; Number is not negative
; Negative number. Set sign
        MOV     BYTE PTR [EDI],'-'       ; Sign to results buffer
NNOT_NEG:
        INC     EDI             ; Bump pointer
;**********************|
; get value of exponent|
;**********************|
        INC     ESI             ; Pointer to BCD12 exponent field
        INC     ESI             ; Point to low byte of BCD exponent
; Move BCD digits to AL-BH-BL-CH-CL for conversion routine
        MOV     AL,0            ; Highest digit is zero
        MOV     DL,BYTE PTR [ESI]        ; DL holds 2 packed digits
        MOV     DH,DL           ; Copy digits to DH
        AND     DL,00001111B    ; Erase high nibble in DL
        AND     DH,11110000B    ; and low in DH
        SHR     DH,4            ; Shift digit to low position
        DEC     ESI             ; Bump pointer to previous digit
        MOV     BL,BYTE PTR [ESI]        ; BL holds 2 packed digits
        MOV     BH,BL           ; Copy digits to BH
        AND     BL,00001111B    ; Erase high nibble in BL
        AND     BH,11110000B    ; and low in BH
        SHR     BH,4            ; Shift digit to low position
        MOV     CX,DX           ; Move digits to CX
        PUSH    ESI             ; Save source
        PUSH    EDI             ; and destination pointers
        CALL    _BCD_INTO_REG   ; BCD to binary conversion
        POP     EDI             ; Restore pointers
        POP     ESI
; DX holds binary value of exponent
; Numbers with exponent = or < 10 are stored in conventional form
; Number with exponent > 10 are stored in exponential form
        CMP     DX,10           ; Test for limit
        JA      EXPO_FORM       ; Exponential form routine
        JMP     NORMAL_FORM     ; Conventional form if 10 or less
;**********************|
;   exponential form   |
;**********************|
; Buffer is formatted as follows for exponential form:
;       -1.2345 ...89 E-23
;       |||          ||_________ offset 21
;       |||          |__________ offset 20
;       |||_____________________ offset 2
;       ||______________________ offset 1
;       |_______________________ offset 0 (- sign or blank)
;**********************|
;  initialize buffer   |
;**********************|
EXPO_FORM:
        MOV     EDI,OFFSET_ASCBUF       ; Start of ASCII buffer
        MOV     BYTE PTR [EDI+2],'.'    ; Set period in buffer
        MOV     BYTE PTR [EDI+20],' '   ; Blank space separator
        MOV     BYTE PTR [EDI+21],'E'   ; Letter E
; Get and encode exponent sign
        ADD     EDI,22           ; Set pointer to sign position
        DEC     ESI              ; Set pointer to first BCD digit
        MOV     AL,BYTE PTR [ESI]        ; Get BCD sign digit
        TEST    AL,00000001B    ; Low bit is negative exponent
        JZ      POS_EXP_EXP     ; Exponent is positive
        MOV     BYTE PTR [EDI],'-'      ; Place negative sign
        INC     EDI
POS_EXP_EXP:
        INC     ESI              ; BCD pointer to exponent area
; EDI --> position for first exponent digit
        MOV     EBX,OFFSET_ASCBUF     ; EBX --> start of buffer
        ADD     EBX,3           ; Buffer offset 3 to 6 is used
                                ; as a temporary storage
        MOV     AL,[ESI]        ; Get 2 packed BCD digits
        MOV     AH,AL           ; Copy digit in AH
        AND     AL,00001111B    ; Unpack low digit
        AND     AH,11110000B    ; and high digit
        SHR     AH,4            ; Move high digit to low nibble
; Convert BCD digits to ASCII
        ADD     AL,30H          ; BCD + 30H = ASCII
        ADD     AH,30H
        MOV     BYTE PTR [EBX],AH        ; Store ASCII
        INC     EBX             ; Bump pointer to temporary area
        MOV     BYTE PTR [EBX],AL
        INC     EBX             ; Bump pointer again
        INC     ESI             ; BCD pointer to next digit pair
; Get next two BCD digits and store
        MOV     AL,[ESI]        ; Get 2 packed BCD digits
        MOV     AH,AL           ; Copy digit in AH
        AND     AL,00001111B    ; Unpack low digit
        AND     AH,11110000B    ; and high digit
        SHR     AH,4            ; Move high digit to low nibble
; Convert BCD digits to ASCII
        ADD     AL,30H          ; BCD + 30H = ASCII
        ADD     AH,30H
        MOV     BYTE PTR [EBX],AH        ; Store ASCII
        INC     EBX             ; Bump pointer to temporary area
        MOV     BYTE PTR [EBX],AL
        SUB     EBX,3           ; BX --> first digit in temporary
                                ; area
; Move digits to final position skipping leading zeros
        MOV     ECX,4           ; For a count of 4
SKIP_LEAD0:
        MOV     AL,BYTE PTR [EBX]        ; Digit in temporary area
        CMP     AL,'0'          ; Test for leading zero
        JNE     NOT_LEAD00      ; Go if not a leading zero
        INC     EBX             ; Bump pointer
        LOOP    SKIP_LEAD0      ; Test 4 digits
; Execution drops if all exponent digits are 0
        MOV     BYTE PTR [EDI],'0'       ; Place one zero in buffer
        INC     EDI             ; Bump buffer pointer
        JMP     EXPONENT_RDY
NOT_LEAD00:
        MOV     AL,BYTE PTR [EBX]        ; Get digit
        MOV     BYTE PTR [EDI],AL        ; and store it
        INC     EBX             ; Bump pointers
        INC     EDI
        LOOP    NOT_LEAD00      ; Continue for remaining digits
;
EXPONENT_RDY:
        MOV     BYTE PTR [EDI],0FFH      ; Special code for clean up
                                        ; routine
; Exponent is in buffer. Move significand using existing routine
        SUB     ESI,2            ; Set SI to start of BCD buffer
        MOV     EDI,OFFSET_ASCBUF     ; Start of ASCII buffer
        INC     EDI              ; Skip sign position
        MOV     DX,0            ; Normalized significand
        JMP     POS_EX_FORM     ; Handle by positive exponent
;**********************|
;     normal form      |
;**********************|
NORMAL_FORM:
; DX holds binary value of exponent, which must be 10 or less
; Possible storage forms:
; If sign of exponent is negative:
;       0.1234
;       0.00001234
; If sign of exponent is positive
;       12.3456
;       12345.67
;       1.2345 (exponent = 0)
;
        DEC     ESI             ; Set pointer to first BCD digit
        MOV     AL,BYTE PTR [ESI]        ; Get BCD sign digit
        TEST    AL,00000001B    ; Low bit is negative exponent
        JZ      POS_EX_FORM     ; Exponent is positive
        JMP     NEG_EX_FORM     ; Go to negative exponent routine
;**********************|
; exponent is positive |
;**********************|
; DX holds the number of digits to move to the right the decimal
; point from the normalized position
; DI --> ASCII buffer
POS_EX_FORM:
        PUSH    EDI             ; Save start of ASCII buffer
		AND     EDX,0FFFFH      ; Clear high word
        ADD     EDI,EDX         ; Add exponent to start of buffer
        INC     EDI             ; Add one for normalized format
        MOV     BYTE PTR [EDI],'.'       ; Place point in buffer
        POP     EDI             ; Restore pointer to start position
MOVE_SIG:
        ADD     ESI,3            ; Set pointer to BCD significand
; ESI --> first significand digit in packed BCD format
;**********************|
;   unpack and move    |
;     BCD digits       |
;**********************|
EXP_FORM_ENTRY:
        MOV     DX,9            ; Counter for 9 packed BCD digits
UNPACK_2:
        MOV     AL,[ESI]         ; Get 2 packed BCD digits
        MOV     AH,AL           ; Copy digit in AH
        AND     AL,00001111B    ; Unpack low digit
        AND     AH,11110000B    ; and high digit
        SHR     AH,4            ; Move high digit to low nibble
; Convert BCD digits to ASCII
        ADD     AL,30H          ; BCD + 30H = ASCII
        ADD     AH,30H
; Place two digits in buffer
        CMP     BYTE PTR [EDI],'.'       ; Test for period in buffer
        JNE     NO_PERIOD1      ; OK if no period
        INC     EDI              ; Bump pointer if on period
NO_PERIOD1:
        MOV     BYTE PTR [EDI],AH        ; Store high digit
        INC     EDI              ; Bump pointer to next digit
        CMP     BYTE PTR [EDI],'.'       ; Test for period in buffer
        JNE     NO_PERIOD2      ; OK if no period
        INC     EDI              ; Bump pointer if on period
NO_PERIOD2:
        MOV     BYTE PTR [EDI],AL        ; Store low digit
        INC     EDI
        INC     ESI              ; Bump pointer to next BCD
        DEC     DX              ; Adjust digits counter
        JNZ     UNPACK_2
;**********************|
;       clean up       |
;**********************|
; Clear trailing zeros from buffer
; Example buffer digits:  1.23450000000000000(00)
;                         1234.00000000000000(00)
; Buffer after clean-up:  1.2345(00)
;                         1234(00)
;
        MOV     EDI,OFFSET_ASCBUF     ; Entry offset from storage
        ADD     EDI,30          ; Point to last digit
; Replace up to 27 trailing zeros with NULL bytes
        MOV     ECX,29          ; Counter for maximum digits
REPLACE_27:
        CMP     BYTE PTR [EDI],0FFH      ; Special code for
                                        ; exponential form
        JNE     NT_EXP_FORM     ; Go if not
        MOV     BYTE PTR [EDI],00H       ; Clear special code
        JMP     EXIT_REP27      ; Do not fix trailing zeros
NT_EXP_FORM:
        CMP     BYTE PTR [EDI],'0'       ; Test for trailing 0
        JNE     TEST_27         ; Test for other values
        MOV     BYTE PTR [EDI],0; Replace zero with NULL
        JMP     NEXT_27
TEST_27:
        CMP     BYTE PTR [EDI],0; Test for NULL at end of buffer
        JNE     EXIT_REP27      ; Exit if not 0 or NULL
NEXT_27:
        DEC     EDI             ; Adjust pointer
        LOOP    REPLACE_27      ; Continue for 27 digits
EXIT_REP27:
; If the last character found is a period, then number is in the
; form 1234.0000..00.
        CMP     BYTE PTR [EDI],'.'       ; Test for period
        JNE     NO_PERIOD_FIX   ; Go if not
        MOV     BYTE PTR [EDI],0; Replace period with NULL
NO_PERIOD_FIX:
        JMP     EXIT_ASCSTR
;**********************|
; exponent is negative |
;**********************|
; Display forms for negative exponent
;       0.1234 (exponent = -1)
;       0.00001234 (exponent = -5)
; Operation:
;    The decimal point is first placed in the buffer
;    DX number of zeros are placed in the buffer, skipping the
;    decimal point
NEG_EX_FORM:
        INC     EDI              ; Leave blank for leading zero
        MOV     BYTE PTR [EDI],'.'       ; Place point in buffer
        DEC     EDI
PLACE_0S:
        CMP     BYTE PTR [EDI],'.'       ; Test for period in buffer
        JNE     NO_PERIOD3      ; No period, proceed
        INC     EDI              ; Bump buffer pointer to skip period
NO_PERIOD3:
        MOV     BYTE PTR [EDI],'0'       ; Place 0 digit in buffer
        INC     EDI             ; Bump pointer to next digit
        DEC     DX              ; Decrement counter for zeros
        JNZ     PLACE_0S        ; Continue if not zero
        JMP     MOVE_SIG        ; Use existing code to move
                                ; significand
;**********************|
;    end of routine    |
;**********************|
EXIT_ASCSTR:
        MOV     EDI,OFFSET_ASCBUF
        CLD 
        RET
_BCD12_TO_ASCSTR   ENDP


_ASCII_TO_BCD12   PROC USES esi edi ebx ebp
; Conversion of an ASCII decimal number into BCD12 format
; On entry:
;         ESI --> an ASCII decimal string, terminated in a NULL
;                 byte or control code
;             Note: the caller's buffer may contain the $ sign
;                   and commas as formatting symbols. The string
;                   terminator (00H) may be preceded by a CR
;                   (0DH) or LF (0AH) control codes. If the
;                   characters E or e are in the string, the
;                   procedure assumes that the string is in
;                   ASCII exponential form. The code assumes that
;                   exponential strings are normalized.
;
;         EDI --> 12-byte memory area to holds the number encoded
;                 in BCD12 format
;
; Sample input strings:
;        123.223344
;        -0000.000002345
;        1.273E-22
;        -1.44556677 E230
;
;         EDI --> 12-byte memory area to holds the number encoded
;                 in BCD12 format
;
; On exit:
;         EDI --> the caller's buffer with the input number
;                 reformatted in BCD12 format, as follows:
;
;          S s e e e e m . m m m m m m m m m m m m m m m m m m
;                        |
;                        |____ implicit decimal point
;
;    S = sign of number (1 BCD digit)
;    s = sign of exponent (1 BCD digit)
;    e = exponent (4 BCD digits)
;    m = normalized significand (18 BCD digits)
;    . = implicit decimal point between the first and second
;        significand digits
;
;**********************|
; save buffer addresses|
;**********************|
        MOV     OFFSET_ASC,ESI
        MOV     OFFSET_BCD,EDI
;**********************|
;   clear buffer for   |
;     BCD12 number     |
;**********************|
; EDI --> 12-byte buffer to hold BCD12 number
NOT_EXP:
        PUSH    EDI              ; Save output pointer
        MOV     ECX,12           ; Bytes to clear
CLEAR_12:
        MOV     BYTE PTR [EDI],0 ;
        INC     EDI              ; Bump buffer pointer
        LOOP    CLEAR_12
        POP     EDI              ; Restore pointer to start
        MOV     ESI,OFFSET_ASC        ; Source pointer to start
                                      ; of ASCII
;**********************|
;  get sign of number  |
;**********************|
; Find first significant character in the caller's string
; Skip leading blanks, $ and # sign, and other invalid symbols
 GET_NB:
        MOV     AL,BYTE PTR [ESI]        ; Get character
        CMP     AL,'+'          ; Test for + sign
        JAE     FIRST_NB        ; Go if equal or larger
        INC     ESI
        JMP     GET_NB
; first non-blank is a symbol or number larger than the + sign
FIRST_NB:
        CMP     AL,'+'          ; Test for +
        JE      SIGN_IS_PLUS
        CMP     AL,'-'          ; Test for -
        JE      SIGN_IS_MINUS
; First non-blank character is not a sign. Assume a digit
; Positive number is the default setting in BCD buffer
        JMP     TEST_4_EXP      ; Test for exponential form
;**********************|
;  number is negative  |
;**********************|
SIGN_IS_MINUS:
        MOV     BYTE PTR [EDI],00010000B ; Set first BCD to 1
;**********************|
;  number is positive  |
;**********************|
; Encoding for positive number (0000B) is already in buffer
SIGN_IS_PLUS:
        INC     ESI              ; Bump ASCII digits pointer
;**********************|
;   test for E or e    |
;**********************|
; The caller's string may be in ASCII exponential form
; Any value less than 20H terminates the ASCII string
TEST_4_EXP:
        MOV     OFFSET_ASC,ESI
                                ; Save pointer to first digit
TEST_AGAIN:
        CMP     BYTE PTR [ESI],'E'
        JE      STRING_ISEX             ; String in FP format
        CMP     BYTE PTR [ESI],'e'
        JE      STRING_ISEX
        CMP     BYTE PTR [ESI],' '      ; Less than 20H is end
                                        ; of string
        JB      SKIP_ZERO               ; Not exponential form
        INC     ESI                     ; Bump digits pointer
        JMP     TEST_AGAIN
;**********************|
;   input is in ASCII  |
;   exponential form   |
;**********************|
STRING_ISEX:
; Sample exponential form ASCII strings:
;            1.2345 E12
;           -1.2345E-12
;           +1.2345 e+12
; SI --> letter e or E
; First test for negative exponent
        INC     ESI             ; Bump pointer past letter E
        PUSH    ESI             ; Save pointer to possible exponent
                                ; sign
; SI --> string of exponent digits
; Skip sign, if there is one
FIRST_0PLUS:
        CMP     BYTE PTR [ESI],'0'       ; First valid ASCII number
        JAE     OK_SI_PTR       ; SI pointer is OK if 0 or >
        INC     ESI             ; Bump pointer if not
        JMP     FIRST_0PLUS     ; Continue
OK_SI_PTR:
        CALL    _ASCII_TO_DX    ; Convert to binary in DX
; DX now holds binary value of exponent
        POP     ESI             ; Restore point to possible sign
        CMP     BYTE PTR [ESI],'-'       ; Negative exponent
        JNE     NON_NEG_EXP     ; Go if not negative
        JMP     NEG_EXPONENT    ; Routine for negative exponent
; Positive exponent code is already in place
NON_NEG_EXP:
        JMP     CONV_EXPONENT
;**********************|
; skip leading zeros   |
;**********************|
SKIP_ZERO:
; Leading zeros are not significant. skip them
; First reset pointer to start of string
        MOV     ESI,OFFSET_ASC        ; Pointer saved in memory
SKIP_1ZERO:
        CMP     BYTE PTR [ESI],'0'       ; Test for 0
        JNE     NOT_LEAD0
        INC     ESI              ; Bump to next digit
        JMP     SKIP_1ZERO
;**********************|
;  find decimal point  |
;    (if present)      |
;**********************|
;  DX is used to count the non-zero digits preceding the decimal
;  point. This value determines the number's exponent
NOT_LEAD0:
        MOV     DX,0            ; Clear integer digit counter
; Caller's string must end in a NULL byte (00H) or other control
; code < 20H
; Test for an integer zero input in the form 0000
        CMP     BYTE PTR [ESI],' '       ; Smaller than 20H is
                                ; end of input buffer
        JAE     GET_POINT       ; Continue if not end
        JMP     NUM_ISZERO      ; Number is zero
GET_POINT:
        MOV     AL,BYTE PTR [ESI]; Caller's ASCII digit
        CMP     AL,'.'          ; Test for decimal point
        JE      POINT_FOUND
        CMP     AL,','          ; Test for formatting comma
                                ; in the caller's string
        JE      SKIP_COMMA      ; Ignore comma
        CMP     AL,20H          ; String terminator
        JNB     LOOK_4_PT       ; Continue searching if not < 20H
        JMP     NO_POINT        ; No decimal point in string
LOOK_4_PT:
        INC     DX              ; Count one integer digit
SKIP_COMMA:
        INC     ESI             ; Bump integer digits counter
        JMP     GET_POINT
;**********************|
; caller's string is a |
;   decimal fractions  |
;**********************|
; If DX = 0 then the caller's string < 1 and the EXPONENT IS
;           NEGATIVE or the number is zero
; If DX = 1 then the caller's string is already normalized
;         and the EXPONENT = +0
; If DX > 1 then the caller's string is a number =< 10 and
;         the EXPONENT IS POSITIVE
POINT_FOUND:
        CMP     DX,1            ; Test for normalized input
        JNE     TEST_4_DX0
;**********************|
;       DX = 1         |
;**********************|
; ASCII input is in the form: d.dddd...ddd
; Number is between 1 and 10. Sign of exponent = 0000B
; 4 exponent digits = 0000H (these values are already in place)
        JMP     EXP_MOVED
TEST_4_DX0:
        CMP     DX,0            ; Test for negative exponent
        JE      DX_ISZERO       ; Go to DX = 0 processing
; If DX > 1 adjust and store
        DEC     DX
        JMP     CONV_EXPONENT
;**********************|
;       DX = 0         |
;**********************|
; Possible cases:
; non-zero -->   000.002345
; non-zero -->      .2345
;     zero -->     0.00
;     zero -->      .0
; Count decimal digits to normalized position, that is, following
; the first non-zero digit
; DX is used as counter
; At this point:
;                dddd.ddd...
;                    |____ SI points to decimal point
DX_ISZERO:
        INC     ESI              ; Skip decimal point
        INC     DX               ; Count one digit
;**********************|
;     normalize        |
;**********************|
NORMALIZE:
        CMP     BYTE PTR [ESI],'0'      ; Test for zero
        JE      ZERO_DIGIT              ; Go if zero digit
        JB      NUM_ISZERO              ; exit if less than zero
; If not equal or less than zero it must be larger
        JMP     NEG_EXPONENT    ; Normalized position found
ZERO_DIGIT:
        INC     ESI             ; Bump pointer
        INC     DX              ; and digits counter
        JMP     NORMALIZE
;**********************|
;   ASCII number = 0   |
;**********************|
NUM_ISZERO:
; Make sure zero encoding is in BCD buffer
ZERO_EXIT:
        MOV     ECX,12           ; Bytes to clear
CLEAR_12_D:
        MOV     BYTE PTR [EDI],0 ;
        INC     EDI              ; Bump buffer pointer
        LOOP    CLEAR_12_D
        JMP     A2FP_EXIT       ; exit conversion
;**********************|
;  negative exponent   |
;**********************|
; DX holds binary value of negative exponent
; Convert to BCD and store
NEG_EXPONENT:
        MOV     EDI,OFFSET_BCD        ; Saved in CS data
; DI --> start of BCD buffer
        MOV     AL,[EDI]         ; Get BCD sign digit
        MOV     AH,00000001B    ; Code for minus exponent
        OR      AL,AH           ; OR into sign digit
        MOV     BYTE PTR [EDI],AL        ; Store exponent sign
        JMP     CONV_EXPONENT
;**********************|
;  ASCII is an integer |
;**********************|
NO_POINT:
; Exponent of integer is positive. This encoding is already in place
; String has no decimal point. DX = exponent plus 1
        DEC     DX
;**********************|
;  convert and store   |
;      exponent        |
;**********************|
CONV_EXPONENT:
        MOV     AX,DX            ; Move binary to source register
        CALL    _BIN_TO_BCD      ; Binary to BCD conversion
        MOV     EDI,OFFSET_BCD        ; Saved in CS data
; EDI --> start of BCD buffer
        INC     EDI              ; Bump pointer
; EDI --> first BCD exponent digit
        MOV     BYTE PTR [EDI],DH        ; Store two high BCD digits
        INC     EDI              ; Bump pointer again
        MOV     BYTE PTR [EDI],DL        ; Store two low BCD digits
;**********************|
;  convert and store   |
;     significand      |
;**********************|
; Set pointer to first non-zero and non-sign digit in ASCII number
EXP_MOVED:
        MOV     ESI,OFFSET_ASC        ; Pointer to start of number
LOOK_4_1:
        MOV     AL, BYTE PTR [ESI]       ; Get ASCII digit
        CMP     AL,'1'          ; Test for non-zero and non-control
        JAE     FIRST_NON0      ; Go if 1 or greater
        INC     ESI             ; Bump pointer
        CMP     AL,20H          ; Test for control or NULL
        JAE     LOOK_4_1        ; Continue if not control or NULL
        JMP     ZERO_EXIT       ; Force zero and exit if no valid
                                ; significand digits
FIRST_NON0:
        MOV     ECX,18          ; Counter for 18 digits maximum
        MOV     EDI,OFFSET_BCD        ; Set pointer
; EDI --> start of BCD buffer
        ADD     EDI,3            ; Offset of significand area
; EDI --> BCD significand area in buffer
; ESI --> first non-zero digits in the ASCII string
; ASCII string must have at least one valid digit at this point
; Sample significand strings:
;                1.23456(0DH)
;                1.23456(00H)
;                -1.2345 E-12
;                123.45E345
; Operation:
;    The conversion routine takes the ASCII digits from the buffer,
;    converts them to BCD, and stores the digits in packed format.
;    The DH register is used to signal the high-order BCD digit
        MOV     DH,1            ; Code for next digit is high nibble
ASCII_MOVE:
        MOV     AL,BYTE PTR [ESI]        ; ASCII digit
; Test for invalid values
        CMP     AL,'0'          ; Lower limit
        JAE     VALID_1         ; Digit is 0 or larger
; If digit is less than 0, test for period, comma, and other symbols
; and skip. Exit routine if control code or NULL
        CMP     AL,20H          ; Control codes are < 20H
        JAE     NEXT_ASCII      ; Skip if space or symbol
        JMP     A2FP_EXIT       ; Exit with first non ASCII digit
VALID_1:
        CMP     AL,'9'          ; Upper limit of digit
        JBE     VALID_2         ; Go if OK
        JMP     A2FP_EXIT       ; Exit with first non ASCII digit
                                ; greater than 9
VALID_2:
; ASCII digit in AL is ready for storing in packed BCD format
        SUB     AL,30H          ; ASCII to BCD
        CMP     DH,1            ; Test high digit indicator
        JE      HIGH_BCD        ; Go if high
; Low order BCD digit
        MOV     AH,BYTE PTR [EDI]        ; Get previous high digit
        OR      AH,AL           ; OR low digit into AH
        MOV     BYTE PTR [EDI],AH        ; Store low and high
        INC     EDI              ; Bump destination pointer
        MOV     DH,1            ; Set switch for next digit is high
        JMP     NEXT_ASCII      ; Get next ASCII digit
; Routine to store high-order ASCII digit
; Binary value (in AL) must be shifted left 4 position
HIGH_BCD:
        SHL     AL,4             ; Shift AL digits
        MOV     BYTE PTR [EDI],AL        ; Store digit in buffer
        MOV     DH,0            ; Clear high digit switch
                                ; Next digit is low
NEXT_ASCII:
        INC     ESI              ; Bump ASCII digits pointer
        LOOP    ASCII_MOVE      ; Continue moving digits
;**********************|
;  restore and exit    |
;**********************|
A2FP_EXIT:
        CLD
        RET
_ASCII_TO_BCD12   ENDP

;****************************************************************
;                 BCD20 conversions
;****************************************************************
_ASCII_TO_BCD20   PROC USES esi edi ebx ebp
; Conversion of an ASCII decimal number into BCD12 format
; On entry:
;         ESI --> an ASCII decimal string, terminated in a NULL
;                 byte or control code
;             Note: the caller's buffer may contain the $ sign
;                   and commas as formatting symbols. The string
;                   terminator (00H) may be preceded by a CR
;                   (0DH) or LF (0AH) control codes. If the
;                   characters E or e are in the string, the
;                   procedure assumes that the string is in
;                   ASCII exponential form. The code assumes that
;                   exponential strings are normalized.
;
;         EDI --> 20-byte memory area to holds the number encoded
;                 in BCD20 format
;
; Sample input strings:
;        123.223344
;        -0000.000002345
;        1.273E-22
;        -1.44556677 E230
;
;         EDI --> 20-byte memory area to holds the number encoded
;                 in BCD20 format
;
; On exit:
;         EDI --> the caller's buffer with the input number
;                 reformatted in BCD12 format, as follows:
;
;          S s e e e e m . m m m m m m m m m m m m m m m m m m
;                        |
;                        |____ implicit decimal point
;
;    S = sign of number (1 BCD digit)
;    s = sign of exponent (1 BCD digit)
;    e = exponent (4 BCD digits)
;    m = normalized significand (18 BCD digits)
;    . = implicit decimal point between the first and second
;        significand digits
;
;**********************|
; save buffer addresses|
;**********************|
        MOV     OFFSET_ASC,ESI
        MOV     OFFSET_BCD,EDI
;**********************|
;   clear buffer for   |
;     BCD12 number     |
;**********************|
; EDI --> 20-byte buffer to hold BCD20 number
NOT_EXP2:
        PUSH    EDI              ; Save output pointer
        MOV     ECX,20           ; Bytes to clear
CLEAR_20:
        MOV     BYTE PTR [EDI],0 ;
        INC     EDI              ; Bump buffer pointer
        LOOP    CLEAR_20
        POP     EDI              ; Restore pointer to start
        MOV     ESI,OFFSET_ASC        ; Source pointer to start
                                      ; of ASCII
;**********************|
;  get sign of number  |
;**********************|
; Find first significant character in the caller's string
; Skip leading blanks, $ and # sign, and other invalid symbols
 GET_NB2:
        MOV     AL,BYTE PTR [ESI]        ; Get character
        CMP     AL,'+'          ; Test for + sign
        JAE     FIRST_NB2        ; Go if equal or larger
        INC     ESI
        JMP     GET_NB2
; first non-blank is a symbol or number larger than the + sign
FIRST_NB2:
        CMP     AL,'+'          ; Test for +
        JE      SIGN_IS_PLUS2
        CMP     AL,'-'          ; Test for -
        JE      SIGN_IS_MINUS2
; First non-blank character is not a sign. Assume a digit
; Positive number is the default setting in BCD buffer
        JMP     TEST_4_EXP2      ; Test for exponential form
;**********************|
;  number is negative  |
;**********************|
SIGN_IS_MINUS2:
        MOV     BYTE PTR [EDI],00010000B ; Set first BCD to 1
;**********************|
;  number is positive  |
;**********************|
; Encoding for positive number (0000B) is already in buffer
SIGN_IS_PLUS2:
        INC     ESI              ; Bump ASCII digits pointer
;**********************|
;   test for E or e    |
;**********************|
; The caller's string may be in ASCII exponential form
; Any value less than 20H terminates the ASCII string
TEST_4_EXP2:
        MOV     OFFSET_ASC,ESI
                                ; Save pointer to first digit
TEST_AGAIN2:
        CMP     BYTE PTR [ESI],'E'
        JE      STRING_ISEX2             ; String in FP format
        CMP     BYTE PTR [ESI],'e'
        JE      STRING_ISEX2
        CMP     BYTE PTR [ESI],' '      ; Less than 20H is end
                                        ; of string
        JB      SKIP_ZERO2               ; Not exponential form
        INC     ESI                     ; Bump digits pointer
        JMP     TEST_AGAIN2
;**********************|
;   input is in ASCII  |
;   exponential form   |
;**********************|
STRING_ISEX2:
; Sample exponential form ASCII strings:
;            1.2345 E12
;           -1.2345E-12
;           +1.2345 e+12
; SI --> letter e or E
; First test for negative exponent
        INC     ESI             ; Bump pointer past letter E
        PUSH    ESI             ; Save pointer to possible exponent
                                ; sign
; SI --> string of exponent digits
; Skip sign, if there is one
FIRST_0PLUS2:
        CMP     BYTE PTR [ESI],'0'       ; First valid ASCII number
        JAE     OK_SI_PTR2       ; SI pointer is OK if 0 or >
        INC     ESI             ; Bump pointer if not
        JMP     FIRST_0PLUS2     ; Continue
OK_SI_PTR2:
        CALL    _ASCII_TO_DX    ; Convert to binary in DX
; DX now holds binary value of exponent
        POP     ESI             ; Restore point to possible sign
        CMP     BYTE PTR [ESI],'-'       ; Negative exponent
        JNE     NON_NEG_EXP2     ; Go if not negative
        JMP     NEG_EXPONENT2    ; Routine for negative exponent
; Positive exponent code is already in place
NON_NEG_EXP2:
        JMP     CONV_EXPONENT2
;**********************|
; skip leading zeros   |
;**********************|
SKIP_ZERO2:
; Leading zeros are not significant. skip them
; First reset pointer to start of string
        MOV     ESI,OFFSET_ASC        ; Pointer saved in memory
SKIP_1ZERO2:
        CMP     BYTE PTR [ESI],'0'       ; Test for 0
        JNE     NOT_LEAD02
        INC     ESI              ; Bump to next digit
        JMP     SKIP_1ZERO2
;**********************|
;  find decimal point  |
;    (if present)      |
;**********************|
;  DX is used to count the non-zero digits preceding the decimal
;  point. This value determines the number's exponent
NOT_LEAD02:
        MOV     DX,0            ; Clear integer digit counter
; Caller's string must end in a NULL byte (00H) or other control
; code < 20H
; Test for an integer zero input in the form 0000
        CMP     BYTE PTR [ESI],' '       ; Smaller than 20H is
                                ; end of input buffer
        JAE     GET_POINT2       ; Continue if not end
        JMP     NUM_ISZERO2      ; Number is zero
GET_POINT2:
        MOV     AL,BYTE PTR [ESI]; Caller's ASCII digit
        CMP     AL,'.'          ; Test for decimal point
        JE      POINT_FOUND2
        CMP     AL,','          ; Test for formatting comma
                                ; in the caller's string
        JE      SKIP_COMMA2      ; Ignore comma
        CMP     AL,20H          ; String terminator
        JNB     LOOK_4_PT2       ; Continue searching if not < 20H
        JMP     NO_POINT2        ; No decimal point in string
LOOK_4_PT2:
        INC     DX              ; Count one integer digit
SKIP_COMMA2:
        INC     ESI             ; Bump integer digits counter
        JMP     GET_POINT2
;**********************|
; caller's string is a |
;   decimal fractions  |
;**********************|
; If DX = 0 then the caller's string < 1 and the EXPONENT IS
;           NEGATIVE or the number is zero
; If DX = 1 then the caller's string is already normalized
;         and the EXPONENT = +0
; If DX > 1 then the caller's string is a number =< 10 and
;         the EXPONENT IS POSITIVE
POINT_FOUND2:
        CMP     DX,1            ; Test for normalized input
        JNE     TEST_4_DX02
;**********************|
;       DX = 1         |
;**********************|
; ASCII input is in the form: d.dddd...ddd
; Number is between 1 and 10. Sign of exponent = 0000B
; 4 exponent digits = 0000H (these values are already in place)
        JMP     EXP_MOVED2
TEST_4_DX02:
        CMP     DX,0            ; Test for negative exponent
        JE      DX_ISZERO2       ; Go to DX = 0 processing
; If DX > 1 adjust and store
        DEC     DX
        JMP     CONV_EXPONENT2
;**********************|
;       DX = 0         |
;**********************|
; Possible cases:
; non-zero -->   000.002345
; non-zero -->      .2345
;     zero -->     0.00
;     zero -->      .0
; Count decimal digits to normalized position, that is, following
; the first non-zero digit
; DX is used as counter
; At this point:
;                dddd.ddd...
;                    |____ SI points to decimal point
DX_ISZERO2:
        INC     ESI              ; Skip decimal point
        INC     DX               ; Count one digit
;**********************|
;     normalize        |
;**********************|
NORMALIZE2:
        CMP     BYTE PTR [ESI],'0'      ; Test for zero
        JE      ZERO_DIGIT2              ; Go if zero digit
        JB      NUM_ISZERO2              ; exit if less than zero
; If not equal or less than zero it must be larger
        JMP     NEG_EXPONENT2    ; Normalized position found
ZERO_DIGIT2:
        INC     ESI             ; Bump pointer
        INC     DX              ; and digits counter
        JMP     NORMALIZE2
;**********************|
;   ASCII number = 0   |
;**********************|
NUM_ISZERO2:
; Make sure zero encoding is in BCD buffer
ZERO_EXIT2:
        MOV     ECX,20           ; Bytes to clear
CLEAR_20_D2:
        MOV     BYTE PTR [EDI],0 ;
        INC     EDI              ; Bump buffer pointer
        LOOP    CLEAR_20_D2
        JMP     A2FP_EXIT2       ; exit conversion
;**********************|
;  negative exponent   |
;**********************|
; DX holds binary value of negative exponent
; Convert to BCD and store
NEG_EXPONENT2:
        MOV     EDI,OFFSET_BCD        ; Saved in CS data
; DI --> start of BCD buffer
        MOV     AL,[EDI]         ; Get BCD sign digit
        MOV     AH,00000001B    ; Code for minus exponent
        OR      AL,AH           ; OR into sign digit
        MOV     BYTE PTR [EDI],AL        ; Store exponent sign
        JMP     CONV_EXPONENT2
;**********************|
;  ASCII is an integer |
;**********************|
NO_POINT2:
; Exponent of integer is positive. This encoding is already in place
; String has no decimal point. DX = exponent plus 1
        DEC     DX
;**********************|
;  convert and store   |
;      exponent        |
;**********************|
CONV_EXPONENT2:
        MOV     AX,DX            ; Move binary to source register
        CALL    _BIN_TO_BCD      ; Binary to BCD conversion
        MOV     EDI,OFFSET_BCD        ; Saved in CS data
; EDI --> start of BCD buffer
        INC     EDI              ; Bump pointer
; EDI --> first BCD exponent digit
        MOV     BYTE PTR [EDI],DH        ; Store two high BCD digits
        INC     EDI              ; Bump pointer again
        MOV     BYTE PTR [EDI],DL        ; Store two low BCD digits
;**********************|
;  convert and store   |
;     significand      |
;**********************|
; Set pointer to first non-zero and non-sign digit in ASCII number
EXP_MOVED2:
        MOV     ESI,OFFSET_ASC        ; Pointer to start of number
LOOK_4_12:
        MOV     AL, BYTE PTR [ESI]       ; Get ASCII digit
        CMP     AL,'1'          ; Test for non-zero and non-control
        JAE     FIRST_NON02      ; Go if 1 or greater
        INC     ESI             ; Bump pointer
        CMP     AL,20H          ; Test for control or NULL
        JAE     LOOK_4_12        ; Continue if not control or NULL
        JMP     ZERO_EXIT2       ; Force zero and exit if no valid
                                ; significand digits
FIRST_NON02:
        MOV     ECX,34          ; Counter for 34 digits maximum
        MOV     EDI,OFFSET_BCD        ; Set pointer
; EDI --> start of BCD buffer
        ADD     EDI,3            ; Offset of significand area
; EDI --> BCD significand area in buffer
; ESI --> first non-zero digits in the ASCII string
; ASCII string must have at least one valid digit at this point
; Sample significand strings:
;                1.23456(0DH)
;                1.23456(00H)
;                -1.2345 E-12
;                123.45E345
; Operation:
;    The conversion routine takes the ASCII digits from the buffer,
;    converts them to BCD, and stores the digits in packed format.
;    The DH register is used to signal the high-order BCD digit
        MOV     DH,1            ; Code for next digit is high nibble
ASCII_MOVE2:
        MOV     AL,BYTE PTR [ESI]        ; ASCII digit
; Test for invalid values
        CMP     AL,'0'          ; Lower limit
        JAE     VALID_12         ; Digit is 0 or larger
; If digit is less than 0, test for period, comma, and other symbols
; and skip. Exit routine if control code or NULL
        CMP     AL,20H          ; Control codes are < 20H
        JAE     NEXT_ASCII2      ; Skip if space or symbol
        JMP     A2FP_EXIT2       ; Exit with first non ASCII digit
VALID_12:
        CMP     AL,'9'          ; Upper limit of digit
        JBE     VALID_22         ; Go if OK
        JMP     A2FP_EXIT2       ; Exit with first non ASCII digit
                                ; greater than 9
VALID_22:
; ASCII digit in AL is ready for storing in packed BCD format
        SUB     AL,30H          ; ASCII to BCD
        CMP     DH,1            ; Test high digit indicator
        JE      HIGH_BCD2        ; Go if high
; Low order BCD digit
        MOV     AH,BYTE PTR [EDI]        ; Get previous high digit
        OR      AH,AL           ; OR low digit into AH
        MOV     BYTE PTR [EDI],AH        ; Store low and high
        INC     EDI              ; Bump destination pointer
        MOV     DH,1            ; Set switch for next digit is high
        JMP     NEXT_ASCII2      ; Get next ASCII digit
; Routine to store high-order ASCII digit
; Binary value (in AL) must be shifted left 4 position
HIGH_BCD2:
        SHL     AL,4             ; Shift AL digits
        MOV     BYTE PTR [EDI],AL        ; Store digit in buffer
        MOV     DH,0            ; Clear high digit switch
                                ; Next digit is low
NEXT_ASCII2:
        INC     ESI              ; Bump ASCII digits pointer
        LOOP    ASCII_MOVE2      ; Continue moving digits
;**********************|
;  restore and exit    |
;**********************|
A2FP_EXIT2:
        CLD
        RET
_ASCII_TO_BCD20   ENDP

;**********************************************
;             BCD20_TO_ASCSTR
;**********************************************
_BCD20_TO_ASCSTR  PROC
; Convert a number in BCD20 format into an ASCII string
; On entry:
;         ESI --> 20-byte BCD12 buffer holding number
;         EDI --> 51-byte ASCII buffer to hold string
; On exit:
;         EDI holds ASCII string
;
; Operation:
;         Numbers with exponent < 10 are converted to decimal
;         format. If exponent is > 10 the number is converted to
;         exponential format
;
; BCD12 format:
;          S s e e e e m . m m m m m m m m m m m m m m m m m m
;                        |
;                        |____ implicit decimal point
;
;    S = sign of number (1 BCD digit)
;    s = sign of exponent (1 BCD digit)
;    e = exponent (4 BCD digits)
;    m = normalized significand (18 BCD digits)
;        (in positive and negative numbers the first significand
;         digit must be non-zero)
;    . = implicit decimal point between the first and second
;        significand digits
;
;
;**********************|
; save buffer addresses|
;**********************|
        MOV     OFFSET_ASCBUF,EDI
; Clear results buffer
        PUSH    EDI             ; Save pointer
        MOV     ECX,51          ; Bytes to clear
CLEAR_51:
        MOV     BYTE PTR [EDI],20H ; Clear to zero
        INC     EDI             ; Bump pointer
        LOOP    CLEAR_51
        POP     EDI             ; Restore pointer to start of buffer
;**********************|
;  get sign of number  |
;**********************|
        MOV     BYTE PTR [EDI],' '       ; In case of positive
        MOV     AL,[ESI]        ; First byte of BCD12 number
        TEST    AL,00010000B    ; Bit 4 is negative number
        JZ      NNOT_NEG2        ; Number is not negative
; Negative number. Set sign
        MOV     BYTE PTR [EDI],'-'       ; Sign to results buffer
NNOT_NEG2:
        INC     EDI             ; Bump pointer
;**********************|
; get value of exponent|
;**********************|
        INC     ESI             ; Pointer to BCD20 exponent field
        INC     ESI             ; Point to low byte of BCD exponent
; Move BCD digits to AL-BH-BL-CH-CL for conversion routine
        MOV     AL,0            ; Highest digit is zero
        MOV     DL,BYTE PTR [ESI]        ; DL holds 2 packed digits
        MOV     DH,DL           ; Copy digits to DH
        AND     DL,00001111B    ; Erase high nibble in DL
        AND     DH,11110000B    ; and low in DH
        SHR     DH,4            ; Shift digit to low position
        DEC     ESI             ; Bump pointer to previous digit
        MOV     BL,BYTE PTR [ESI]        ; BL holds 2 packed digits
        MOV     BH,BL           ; Copy digits to BH
        AND     BL,00001111B    ; Erase high nibble in BL
        AND     BH,11110000B    ; and low in BH
        SHR     BH,4            ; Shift digit to low position
        MOV     CX,DX           ; Move digits to CX
        PUSH    ESI             ; Save source
        PUSH    EDI             ; and destination pointers
        CALL    _BCD_INTO_REG   ; BCD to binary conversion
        POP     EDI             ; Restore pointers
        POP     ESI
; DX holds binary value of exponent
; Numbers with exponent = or < 10 are stored in conventional form
; Number with exponent > 10 are stored in exponential form
        CMP     DX,10           ; Test for limit
        JA      EXPO_FORM2      ; Exponential form routine
        JMP     NORMAL_FORM2    ; Conventional form if 10 or less
;**********************|
;   exponential form   |
;**********************|
; Buffer is formatted as follows for exponential form:
;       -1.2345 ...89 E-23
;       |||          ||_________ offset 37
;       |||          |__________ offset 36
;       |||_____________________ offset 2
;       ||______________________ offset 1
;       |_______________________ offset 0 (- sign or blank)
;**********************|
;  initialize buffer   |
;**********************|
EXPO_FORM2:
        MOV     EDI,OFFSET_ASCBUF       ; Start of ASCII buffer
        MOV     BYTE PTR [EDI+2],'.'    ; Set period in buffer
        MOV     BYTE PTR [EDI+36],' '   ; Blank space separator
        MOV     BYTE PTR [EDI+37],'E'   ; Letter E
; Get and encode exponent sign
        ADD     EDI,38           ; Set pointer to sign position
        DEC     ESI              ; Set pointer to first BCD digit
        MOV     AL,BYTE PTR [ESI]        ; Get BCD sign digit
        TEST    AL,00000001B    ; Low bit is negative exponent
        JZ      POS_EXP_EXP2    ; Exponent is positive
        MOV     BYTE PTR [EDI],'-'      ; Place negative sign
        INC     EDI
POS_EXP_EXP2:
        INC     ESI              ; BCD pointer to exponent area
; EDI --> position for first exponent digit
        MOV     EBX,OFFSET_ASCBUF     ; EBX --> start of buffer
        ADD     EBX,3           ; Buffer offset 3 to 6 is used
                                ; as a temporary storage
        MOV     AL,[ESI]        ; Get 2 packed BCD digits
        MOV     AH,AL           ; Copy digit in AH
        AND     AL,00001111B    ; Unpack low digit
        AND     AH,11110000B    ; and high digit
        SHR     AH,4            ; Move high digit to low nibble
; Convert BCD digits to ASCII
        ADD     AL,30H          ; BCD + 30H = ASCII
        ADD     AH,30H
        MOV     BYTE PTR [EBX],AH        ; Store ASCII
        INC     EBX             ; Bump pointer to temporary area
        MOV     BYTE PTR [EBX],AL
        INC     EBX             ; Bump pointer again
        INC     ESI             ; BCD pointer to next digit pair
; Get next two BCD digits and store
        MOV     AL,[ESI]        ; Get 2 packed BCD digits
        MOV     AH,AL           ; Copy digit in AH
        AND     AL,00001111B    ; Unpack low digit
        AND     AH,11110000B    ; and high digit
        SHR     AH,4            ; Move high digit to low nibble
; Convert BCD digits to ASCII
        ADD     AL,30H          ; BCD + 30H = ASCII
        ADD     AH,30H
        MOV     BYTE PTR [EBX],AH        ; Store ASCII
        INC     EBX             ; Bump pointer to temporary area
        MOV     BYTE PTR [EBX],AL
        SUB     EBX,3           ; BX --> first digit in temporary
                                ; area
; Move digits to final position skipping leading zeros
        MOV     ECX,4           ; For a count of 4
SKIP_LEAD02:
        MOV     AL,BYTE PTR [EBX]        ; Digit in temporary area
        CMP     AL,'0'          ; Test for leading zero
        JNE     NOT_LEAD002      ; Go if not a leading zero
        INC     EBX             ; Bump pointer
        LOOP    SKIP_LEAD02      ; Test 4 digits
; Execution drops if all exponent digits are 0
        MOV     BYTE PTR [EDI],'0'       ; Place one zero in buffer
        INC     EDI             ; Bump buffer pointer
        JMP     EXPONENT_RDY2
NOT_LEAD002:
        MOV     AL,BYTE PTR [EBX]        ; Get digit
        MOV     BYTE PTR [EDI],AL        ; and store it
        INC     EBX             ; Bump pointers
        INC     EDI
        LOOP    NOT_LEAD002      ; Continue for remaining digits
;
EXPONENT_RDY2:
        MOV     BYTE PTR [EDI],0FFH   ; Special code for clean up
                                      ; routine
; Exponent is in buffer. Move significand using existing routine
        SUB     ESI,2            ; Set SI to start of BCD buffer
        MOV     EDI,OFFSET_ASCBUF; Start of ASCII buffer
        INC     EDI              ; Skip sign position
        MOV     DX,0             ; Normalized significand
        JMP     POS_EX_FORM2     ; Handle by positive exponent
;**********************|
;     normal form      |
;**********************|
NORMAL_FORM2:
; DX holds binary value of exponent, which must be 10 or less
; Possible storage forms:
; If sign of exponent is negative:
;       0.1234
;       0.00001234
; If sign of exponent is positive
;       12.3456
;       12345.67
;       1.2345 (exponent = 0)
;
        DEC     ESI             ; Set pointer to first BCD digit
        MOV     AL,BYTE PTR [ESI]        ; Get BCD sign digit
        TEST    AL,00000001B    ; Low bit is negative exponent
        JZ      POS_EX_FORM2     ; Exponent is positive
        JMP     NEG_EX_FORM2     ; Go to negative exponent routine
;**********************|
; exponent is positive |
;**********************|
; DX holds the number of digits to move to the right the decimal
; point from the normalized position
; DI --> ASCII buffer
POS_EX_FORM2:
        PUSH    EDI             ; Save start of ASCII buffer
		AND     EDX,0FFFFH      ; Clear high word
        ADD     EDI,EDX         ; Add exponent to start of buffer
        INC     EDI             ; Add one for normalized format
        MOV     BYTE PTR [EDI],'.'       ; Place point in buffer
        POP     EDI             ; Restore pointer to start position
MOVE_SIG2:
        ADD     ESI,3            ; Set pointer to BCD significand
; ESI --> first significand digit in packed BCD format
;**********************|
;   unpack and move    |
;     BCD digits       |
;**********************|
        MOV     DX,17           ; Counter for 17 packed BCD digits
UNPACK_22:
        MOV     AL,[ESI]        ; Get 2 packed BCD digits
        MOV     AH,AL           ; Copy digit in AH
        AND     AL,00001111B    ; Unpack low digit
        AND     AH,11110000B    ; and high digit
        SHR     AH,4            ; Move high digit to low nibble
; Convert BCD digits to ASCII
        ADD     AL,30H          ; BCD + 30H = ASCII
        ADD     AH,30H
; Place two digits in buffer
        CMP     BYTE PTR [EDI],'.'       ; Test for period in buffer
        JNE     NO_PERIOD12      ; OK if no period
        INC     EDI              ; Bump pointer if on period
NO_PERIOD12:
        MOV     BYTE PTR [EDI],AH        ; Store high digit
        INC     EDI              ; Bump pointer to next digit
        CMP     BYTE PTR [EDI],'.'       ; Test for period in buffer
        JNE     NO_PERIOD22      ; OK if no period
        INC     EDI              ; Bump pointer if on period
NO_PERIOD22:
        MOV     BYTE PTR [EDI],AL        ; Store low digit
        INC     EDI
        INC     ESI              ; Bump pointer to next BCD
        DEC     DX              ; Adjust digits counter
        JNZ     UNPACK_22
;**********************|
;       clean up       |
;**********************|
; Clear trailing zeros from buffer
; Example buffer digits:  1.23450000000000000(00)
;                         1234.00000000000000(00)
; Buffer after clean-up:  1.2345(00)
;                         1234(00)
;
        MOV     EDI,OFFSET_ASCBUF     ; Entry offset from storage
        ADD     EDI,50          ; Point to last digit
; Replace up to 27 trailing zeros with NULL bytes
        MOV     ECX,49          ; Counter for maximum digits
REPLACE_272:
        CMP     BYTE PTR [EDI],0FFH      ; Special code for
                                        ; exponential form
        JNE     NT_EXP_FORM2     ; Go if not
        MOV     BYTE PTR [EDI],00H       ; Clear special code
        JMP     EXIT_REP272      ; Do not fix trailing zeros
NT_EXP_FORM2:
        CMP     BYTE PTR [EDI],'0'       ; Test for trailing 0
        JNE     TEST_272        ; Test for other values
        MOV     BYTE PTR [EDI],0; Replace zero with NULL
        JMP     NEXT_272
TEST_272:
        CMP     BYTE PTR [EDI],0; Test for NULL at end of buffer
        JNE     EXIT_REP272      ; Exit if not 0 or NULL
NEXT_272:
        DEC     EDI             ; Adjust pointer
        LOOP    REPLACE_272      ; Continue for digits
EXIT_REP272:
; If the last character found is a period, then number is in the
; form 1234.0000..00.
        CMP     BYTE PTR [EDI],'.'       ; Test for period
        JNE     NO_PERIOD_FIX2   ; Go if not
        MOV     BYTE PTR [EDI],0; Replace period with NULL
NO_PERIOD_FIX2:
        JMP     EXIT_ASCSTR2
;**********************|
; exponent is negative |
;**********************|
; Display forms for negative exponent
;       0.1234 (exponent = -1)
;       0.00001234 (exponent = -5)
; Operation:
;    The decimal point is first placed in the buffer
;    DX number of zeros are placed in the buffer, skipping the
;    decimal point
NEG_EX_FORM2:
        INC     EDI              ; Leave blank for leading zero
        MOV     BYTE PTR [EDI],'.'       ; Place point in buffer
        DEC     EDI
PLACE_0S2:
        CMP     BYTE PTR [EDI],'.'       ; Test for period in buffer
        JNE     NO_PERIOD32      ; No period, proceed
        INC     EDI              ; Bump buffer pointer to skip period
NO_PERIOD32:
        MOV     BYTE PTR [EDI],'0'       ; Place 0 digit in buffer
        INC     EDI             ; Bump pointer to next digit
        DEC     DX              ; Decrement counter for zeros
        JNZ     PLACE_0S2        ; Continue if not zero
        JMP     MOVE_SIG2        ; Use existing code to move
                                ; significand
;**********************|
;    end of routine    |
;**********************|
EXIT_ASCSTR2:
        MOV     EDI,OFFSET_ASCBUF
        CLD 
        RET
_BCD20_TO_ASCSTR   ENDP

;
        end 

        